#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#ifndef GLIBC
#include <sys/socket.h>
#endif
#include <netinet/in_systm.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#ifdef GLIBC
#include <socketbits.h>
#endif
#include <sys/un.h>
#ifndef GLIBC
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <resolv.h>
#endif
#include <ctype.h>
#include <netdb.h>

#include "defines.h"
#include "config.h"
#include "autowin.h"
#include "log.h"
#include "prototypes.h"

extern int debug;
extern int doSuggest;
extern int ignoreConflicts;
extern int autoDeps;
extern int groupSelectType;
extern int returnReason;
extern int ignoreLocal;
extern int installer;
extern int ftpinstall;
extern int newtRunning;
extern int progmountedcdrom;
extern int progmountedfloppy;
extern int nocompsfile;
extern int norpmdir;
extern int nohdlist;
extern int ejectok;

extern int controlsocket;


DESCRIPTOR_DATA   *descriptor_list;
DESCRIPTOR_DATA   *d_next;
DESCRIPTOR_DATA *mydesc=NULL;
int num_connections=0;
int server_down=0;
fd_set in_set;
fd_set out_set;
fd_set exc_set;

int autoWelcome(short int mydesc) {
	char s0[256];
	sprintf(s0,"HELLO:TurboPkg/%s/%s/%s\n",VERSION,__DATE__,__TIME__);
	write_to_descriptor(mydesc,s0,strlen(s0));
	return SUCCESS;
}



int autoError(short int mydesc,int autoErrCode, int flags) {
	char s0[256];
	switch(autoErrCode) {
	case AUTO_ERR_SUPPORTFILES:
		sprintf(s0,"ERROR:%s\n",_("Error loading support files."));
		write_to_descriptor(mydesc,s0,strlen(s0));
		break;
	default:
		sprintf(s0,"ERROR:%s\n",_("Unknown error."));
		write_to_descriptor(mydesc,s0,strlen(s0));
		break;
	}
	if (flags & AUTO_ERRFLAG_FATAL) {
		return FAILURE;
	}
	return SUCCESS;
}

int autoSuccess(short int mydesc) {
	char *s0=strdup(_("SUCCESS:\n"));
	write_to_descriptor(mydesc,s0,strlen(s0));
	return SUCCESS;
}

int autoErrorMsg(short int mydesc, char *errmsg, int flags) {
	char s0[256];
	
	sprintf(s0,"%s:%s\n",(flags & AUTO_ERRFLAG_FATAL) ? "FATALERROR" : "ERROR",errmsg);
	write_to_descriptor(mydesc,s0,strlen(s0));
	if (flags & AUTO_ERRFLAG_FATAL) {
		return FAILURE;
	}
	return SUCCESS;
}

int autoMessage(short int mydesc, char *msg) {
	char s0[4096];
	if (FD_ISSET(mydesc,&out_set)) {
		sprintf(s0,"%s\n",msg);
		write_to_descriptor(mydesc,s0,strlen(s0));
	}
	return SUCCESS;
}

int autoStatus(short int mydesc, char *id, char *msg) {
	char s0[256];
	
	sprintf(s0,"STATUS:%s:%s\n",id,msg);
	write_to_descriptor(mydesc,s0,strlen(s0));
	return SUCCESS;
}

int autoStatusEnd(short int mydesc) {
	char s0[256];
	
	sprintf(s0,"STATUSEND:\n");
	write_to_descriptor(mydesc,s0,strlen(s0));
	return SUCCESS;
}

int init_socket(int port,int ptype, char *unixsockname)
{
    static struct sockaddr_in sa_zero;
    static struct sockaddr_un su_zero;
    struct sockaddr_in sa;
    struct sockaddr_un su;
    int x = 1;
    int fd;
    if ( ( fd = socket( ptype, SOCK_STREAM, 0 ) ) < 0 )
    {
	logMessage( "Init_socket: socket" );
	exit( 0 );
    }

    if ( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR,
    (char *) &x, sizeof(x) ) < 0 )
    {
	logMessage( "Init_socket: SO_REUSEADDR" );
	close(fd);
	exit( 0 );
    }

    if (ptype == AF_UNIX) {
	    su=su_zero;
	    su.sun_family=ptype;
	    strcpy(su.sun_path,unixsockname);
    } else {
	    sa		    = sa_zero;
	    sa.sin_family   = ptype;
	    sa.sin_port	    = htons( port );
    }

    if (ptype == AF_UNIX) {
	    if (bind(fd,(struct sockaddr *)&su,sizeof(su)) < 0) {
		    logMessage("Init unix socket: bind");
		    close(fd);
		    exit(1);
	    }
    } else {
	    if ( bind( fd, (struct sockaddr *) &sa, sizeof(sa) ) < 0 )
	    {
		    logMessage("Init socket: bind" );
		    close(fd);
		    exit(1);
	    }
    }


    if ( listen( fd, 3 ) < 0 )
    {
	logMessage("Init socket: listen");
	close(fd);
	exit(0);
    }
    logMessage("AUTOCONTROL_SRV: INIT_SOCKET - Port %d",port);
    return fd;
}

DESCRIPTOR_DATA *descriptor_free;

DESCRIPTOR_DATA *new_descriptor(void)
{
    static DESCRIPTOR_DATA d_zero;
    DESCRIPTOR_DATA *d;

    if (descriptor_free == NULL)
        d = (DESCRIPTOR_DATA *)malloc(sizeof(DESCRIPTOR_DATA));
    else
    {
        d = descriptor_free;
        descriptor_free = descriptor_free->next;
    }

    *d = d_zero;
    VALIDATE(d);
    return d;
}


void init_descriptor( int control, int ptype )
{
    char buf[MAX_STRING_LENGTH];
    DESCRIPTOR_DATA *dnew;
    struct sockaddr_in sock;
    struct sockaddr_un usock;
    struct hostent *from;
    int desc;
    int size;
    char tempstring[256];
    char cmd1[256];
    char *myhostname=(char *)malloc(100);

    if (ptype == AF_UNIX) {
	    size=sizeof(usock);
    } else {
	    size = sizeof(sock);
    }
    if (ptype == AF_UNIX) {
	    getsockname( control, (struct sockaddr *) &usock, &size );
	    if ( ( desc = accept( control, (struct sockaddr *) &usock, &size) ) < 0 )
	    {
		    logMessage( "New_descriptor: accept" );
		    return;
	    }
    } else {
	    getsockname( control, (struct sockaddr *) &sock, &size );
	    if ( ( desc = accept( control, (struct sockaddr *) &sock, &size) ) < 0 )
	    {
		    logMessage( "New_descriptor: accept" );
		    return;
	    }
    }

#if !defined(FNDELAY)
#define FNDELAY O_NDELAY
#endif

    if ( fcntl( desc, F_SETFL, FNDELAY ) == -1 )
    {
	logMessage( "New_descriptor: fcntl: FNDELAY" );
	return;
    }

    /*
     * Cons a new descriptor.
     */
    dnew = new_descriptor();

    dnew->descriptor	= desc;
    dnew->connected	= 1;
    dnew->showstr_head	= NULL;
    dnew->showstr_point = NULL;
    dnew->outsize	= 2000;
    dnew->outbuf	= (char *)malloc( dnew->outsize );
    dnew->command_mode  = CMD_ACCEPT;
    dnew->access_flag   = ACCESS_ACCEPTED;

    if (ptype == AF_UNIX) {
	    dnew->host = str_dup("(unix socket)");
    } else {
	    size = sizeof(sock);
	    if ( getpeername( desc, (struct sockaddr *) &sock, &size ) < 0 )
	    {
		    logMessage( "New_descriptor: getpeername" );
		    dnew->host = str_dup( "(unknown)" );
	    }
	    else
	    {
		    /*
		     * Would be nice to use inet_ntoa here but it takes a struct arg,
		     * which ain't very compatible between gcc and system libraries.
		     */
		    int addr;
		    
		    addr = ntohl( sock.sin_addr.s_addr );
		    sprintf( buf, "%d.%d.%d.%d",
			     ( addr >> 24 ) & 0xFF, ( addr >> 16 ) & 0xFF,
			     ( addr >>  8 ) & 0xFF, ( addr       ) & 0xFF
			    );
		    from = gethostbyaddr( (char *) &sock.sin_addr,
					  sizeof(sock.sin_addr), ptype );
		    dnew->host = str_dup( from ? from->h_name : buf );	
		    sprintf(tempstring,"[SYSTEM]: CONNECT from %s (%s)",dnew->host,buf);
		    wlog(tempstring);
	    }
    }
	
    /*
     * Init descriptor data.
     */
    dnew->next			= descriptor_list;
    descriptor_list		= dnew;

    /*
     * Send the greeting.
     */
    {
	    /*
	strcpy(greeting,"Welcome to TurboPkg - type @QUIT to leave.\r\n");
	write_to_descriptor(dnew->descriptor,greeting,strlen(greeting));
	sprintf(tempstring,"\r\n[%s]: (arrives)\r\n",dnew->host);
	*/
	    logMessage("AUTOCONTROL_SRV: connect: %s",dnew->host);
	    autoWelcome(dnew->descriptor);
	    num_connections++;
	    if (num_connections > AUTOCONTROL_MAX_CONNECTIONS) {
		    sprintf(cmd1,"ERROR:MaxConnReached/%d.\n",AUTOCONTROL_MAX_CONNECTIONS);
		    write_to_descriptor(dnew->descriptor,cmd1,strlen(cmd1));
		    dnew->outtop = 0;
		    close_socket( dnew );
	    }
	    if (ptype == AF_INET) {
		    gethostname(myhostname,100);
		    if ((strcmp(myhostname,dnew->host)) && (strcmp("localhost",dnew->host))) {
			    sprintf(cmd1,"ERROR:NotFromLocalHost/%d.\n",AUTOCONTROL_MAX_CONNECTIONS);
			    write_to_descriptor(dnew->descriptor,cmd1,strlen(cmd1));
			    dnew->outtop = 0;
			    close_socket( dnew );
		    }
	    }
	    controlsocket=dnew->descriptor;
	    mydesc=dnew;
	    /*
	for ( d = descriptor_list; d != NULL; d = d->next )
	{

	    if ((d != dnew)&&(d->access_flag==ACCESS_ACCEPTED))
	    {
		write_to_descriptor(d->descriptor,tempstring,strlen(tempstring));
	    }
	}
	*/

    }

    return;
}

bool read_from_descriptor( DESCRIPTOR_DATA *d )
{
    int iStart;

    /* Hold horses if pending command already. */
    if ( d->incomm[0] != '\0' )
	return TRUE;

    /* Check for overflow. */
    iStart = strlen(d->inbuf);
    if ( iStart >= (int)(sizeof(d->inbuf) - 10) )
    {
	write_to_descriptor( d->descriptor,
	    "\n*** PUT A LID ON IT!!! ***\n", 0 );
	return FALSE;
    }

    /* Snarf input. */
    for ( ; ; )
    {
	int nRead;

	nRead = read( d->descriptor, d->inbuf + iStart,
	    sizeof(d->inbuf) - 10 - iStart );
	if ( nRead > 0 )
	{
	    iStart += nRead;
	    if ( d->inbuf[iStart-1] == '\n' || d->inbuf[iStart-1] == '\r' )
		break;
	}
	else if ( nRead == 0 )
	{
	    return FALSE;
	}
	else if ( errno == EWOULDBLOCK )
	    break;
	else
	{
	    logMessage( "Read_from_descriptor" );
	    return FALSE;
	}
    }

    d->inbuf[iStart] = '\0';
    return TRUE;
}

void read_from_buffer( DESCRIPTOR_DATA *d )
{
    int i, j, k;

    /*
     * Hold horses if pending command already.
     */
    if ( d->incomm[0] != '\0' )
	return;

    /*
     * Look for at least one new line.
     */
    for ( i = 0; d->inbuf[i] != '\n' && d->inbuf[i] != '\r'; i++ )
    {
	if ( d->inbuf[i] == '\0' )
	    return;
    }

    /*
     * Canonical input processing.
     */
    for ( i = 0, k = 0; d->inbuf[i] != '\n' && d->inbuf[i] != '\r'; i++ )
    {
	if ( k >= MAX_INPUT_LENGTH - 2 )
	{
	    write_to_descriptor( d->descriptor, "Line too long.\n\r", 0 );

	    /* skip the rest of the line */
	    for ( ; d->inbuf[i] != '\0'; i++ )
	    {
		if ( d->inbuf[i] == '\n' || d->inbuf[i] == '\r' )
		    break;
	    }
	    d->inbuf[i]   = '\n';
	    d->inbuf[i+1] = '\0';
	    break;
	}

	if ( d->inbuf[i] == '\b' && k > 0 )
	    --k;
	else if ( isascii(d->inbuf[i]) && isprint(d->inbuf[i]) )
	    d->incomm[k++] = d->inbuf[i];
    }

    /*
     * Finish off the line.
     */
    if ( k == 0 )
	d->incomm[k++] = ' ';
    d->incomm[k] = '\0';

    /*
     * Shift the input buffer.
     */
    while ( d->inbuf[i] == '\n' || d->inbuf[i] == '\r' )
	i++;
    for ( j = 0; ( d->inbuf[j] = d->inbuf[i+j] ) != '\0'; j++ )
	;
    return;
}

bool write_to_descriptor( int desc, char *txt, int length )
{
    int iStart;
    int nWrite;
    int nBlock;

    if ( length <= 0 )
	length = strlen(txt);

    for ( iStart = 0; iStart < length; iStart += nWrite )
    {
	nBlock = UMIN( length - iStart, 4096 );
	if ( ( nWrite = write( desc, txt + iStart, nBlock ) ) < 0 )
	    {
		    /*    logMessage( "Write_to_descriptor" ); */
		    return FALSE;
	    }
    } 

    return TRUE;
}

void close_socket( DESCRIPTOR_DATA *dclose )
{
    if ( dclose == descriptor_list )
    {
	descriptor_list = descriptor_list->next;
    }
    else
    {
	DESCRIPTOR_DATA *d;

	for ( d = descriptor_list; d && d->next != dclose; d = d->next )
	    ;
	if ( d != NULL )
	    d->next = dclose->next;
    }

    close( dclose->descriptor );
    free_descriptor(dclose);
    num_connections--;
    if (num_connections == 0) {
	    server_down=1;
	    logMessage("AUTOCONTROL_SRV: Connections == 0, shutting down.");
    }
}

void free_descriptor(DESCRIPTOR_DATA *d)
{
    if (!IS_VALID(d))
        return;

    INVALIDATE(d);
    d->next = descriptor_free;
    descriptor_free = d;
}



int autoMainHandler(int argc, char **argv) {
	int clear_to_send=0;
	static struct timeval null_time;
	struct timeval last_time;
	int control;
	time_t			current_time;
	struct tm               current_time_struct;
	int ptype=AF_UNIX;

/*	unlink(AUTOCONTROL_UNIXSOCK); */
	control=init_socket(AUTOCONTROL_PORT,ptype,AUTOCONTROL_UNIXSOCK);
	/* Main loop */
	while (!server_down) {
		/*
		fd_set in_set;
		fd_set out_set;
		fd_set exc_set;
		*/
		DESCRIPTOR_DATA *d;
		int maxdesc;
		
		/*
		 * Poll all active descriptors.
		 */
		clear_to_send=0;
		FD_ZERO( &in_set  );
		FD_ZERO( &out_set );
		FD_ZERO( &exc_set );
		FD_SET( control, &in_set );
		maxdesc	= control;
		for ( d = descriptor_list; d; d = d->next )
		{
			maxdesc = UMAX( maxdesc, d->descriptor );
			FD_SET( d->descriptor, &in_set  );
			FD_SET( d->descriptor, &out_set );
			FD_SET( d->descriptor, &exc_set );
		}
		
		if ( select( maxdesc+1, &in_set, &out_set, &exc_set, &null_time ) < 0 )
		{
			logMessage( "Main_Loop: select: poll" );
			exit( 1 );
		}
		
		/*
		 * New connection?
		 */
		if ( FD_ISSET( control, &in_set ) ) {
			init_descriptor( control,ptype );
		}
		
		/*
		 * Kick out the freaky folks.
		 */
		for ( d = descriptor_list; d != NULL; d = d->next )
		{
			if ( FD_ISSET( d->descriptor, &exc_set ) )
			{
				FD_CLR( d->descriptor, &in_set  );
				FD_CLR( d->descriptor, &out_set );
				d->outtop	= 0;
				close_socket( d );
			}
		}
		
		/*
		 * Process input.
		 */
		for ( d = descriptor_list; d != NULL; d = d_next )
		{
			d_next	= d->next;
			d->fcommand	= FALSE;
			
			if ( FD_ISSET( d->descriptor, &in_set ) )
			{
				if ( !read_from_descriptor( d ) )
				{
					FD_CLR( d->descriptor, &out_set );
					d->outtop	= 0;
					close_socket( d );
					continue;
				}
			}
/* input stuff is here */
			read_from_buffer( d );
			if ( d->incomm[0] != '\0' )
			{
				autoParseCommand(d);
				d->incomm[0]='\0';
			}
		}
		{
			struct timeval now_time;
			long secDelta;
			long usecDelta;
			
			gettimeofday( &now_time, NULL );
			usecDelta	= ((int) last_time.tv_usec) - ((int) now_time.tv_usec)
					  + 1000000 / PULSE_PER_SECOND;
			secDelta	= ((int) last_time.tv_sec ) - ((int) now_time.tv_sec );
			while ( usecDelta < 0 )
			{
				usecDelta += 1000000;
				secDelta  -= 1;
			}
			
			while ( usecDelta >= 1000000 )
			{
				usecDelta -= 1000000;
				secDelta  += 1;
			}
			
			if ( secDelta > 0 || ( secDelta == 0 && usecDelta > 0 ) )
			{
				struct timeval stall_time;
				
				stall_time.tv_usec = usecDelta;
				stall_time.tv_sec  = secDelta;
				if ( select( 0, NULL, NULL, NULL, &stall_time ) < 0 )
				{
					logMessage( "Game_loop: select: stall" );
					exit( 1 );
				}
			}
		}
		
		gettimeofday( &last_time, NULL );
		current_time = (time_t) last_time.tv_sec;
		current_time_struct = *localtime( &current_time);
		
	}
	unlink(AUTOCONTROL_UNIXSOCK);
	return SUCCESS;
}
