#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <ctype.h>

#include "defines.h"
#include "config.h"
#include "autowin.h"
#include "log.h"
#include "prototypes.h"

extern int debug;
extern int doSuggest;
extern int ignoreConflicts;
extern int autoDeps;
extern int groupSelectType;
extern int returnReason;
extern int ignoreLocal;
extern int installer;
extern int ftpinstall;
extern int dodowngrades;
extern int newtRunning;
extern int progmountedcdrom;
extern int progmountedfloppy;
extern int nocompsfile;
extern int norpmdir;
extern int nohdlist;
extern int ejectok;

extern char **removemarked;

extern int controlsocket;
extern int server_down;
extern fd_set in_set;
extern fd_set out_set;
extern fd_set exc_set;

extern tlInstall *theInstall;
extern ftpConfig ftpconf;

int autoParseCommand(DESCRIPTOR_DATA *d) {
	char s0[4096];
	char s2[4096];
	char s3[4096];
	char *s1=NULL;
	char *s4=NULL;
	char *s5=NULL;
	int rc;
	tlInstall tmp;
	int xx,yy;
	static int zz;
	struct packageInfo *pptr=NULL;
	char **queryresults=NULL;
	char **qreq=NULL;
	char **qprov=NULL;
	char **sptr=NULL;
	char xc;
	int xa;
	int rcount;

	/* DOC: @REWELCOME (resend welcome strings) */
	if (strstr(d->incomm,"@REWELCOME")) {
		autoWelcome(d->descriptor);
	}
	/* DOC: @SHUTDOWN (shuts down server) */
	if (strstr(d->incomm,"@SHUTDOWN")) {
		logMessage("AUTOCONTROL_SRV: Received Shutdown command.");
	        strcpy(s0,"SHUTDOWN:NormalTermination\n");
		write_to_descriptor(d->descriptor,s0,strlen(s0));
		server_down=1;
	}
	/* DOC: @FINDFILES:basepath (finds support filenames based on basepath) */
	if (strstr(d->incomm,"@FINDFILES")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			tmp.compsfile=NULL;
			tmp.rpmdir=NULL;
			tmp.hdlist=NULL;
			rc=getLocationsByBasePath(&tmp,pathExpandHomeDir(s1));
			if (rc == FAILURE) {
				autoErrMsg(d->descriptor,_("CannotFindByBasePath"),0);
			} else {				
				if (FD_ISSET(d->descriptor,&out_set)) {
					sprintf(s0,"PATHS:%s:%s:%s\n",tmp.compsfile,tmp.rpmdir,tmp.hdlist);
					write_to_descriptor(d->descriptor,s0,strlen(s0));
				}
			}
		}
	}
	/* DOC: @INITIALIZE:inifilename (initializes data structures from inifilename) */
	if (strstr(d->incomm,"@INITIALIZE")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			logMessage("AUTOCONTROL_SRV: Initialize Data command, param '%s'.",s1);
			theInstall=initInstall(s1);
			if (!theInstall) {
				free(ftpconf.mirrors);
				ftpconf.mirrors=NULL;
				free(ftpconf.umirrors);
				ftpconf.umirrors=NULL;
				if (newtRunning) {
					newtFinished();
					newtRunning=0;
				}
				
				unmountCDROM();
				autoErrMsg(d->descriptor,_("PackageInit/Fatal"),1);
				server_down=1;
			}
			if (!server_down) {
				if ((!installer) && (!ignoreLocal) && (!norpmdir)) {
					getCurrentlyInstalled(theInstall);
				}
				autoSuccess(d->descriptor);
			}
		}
	}
	/* DOC: @SYSINFO (gets system information) */
	if (strstr(d->incomm,"@SYSINFO")) {
		s1=checkVersions();
		if (FD_ISSET(d->descriptor,&out_set)) {
			sprintf(s0,"SYSINFO:%s\n",s1);
			write_to_descriptor(d->descriptor,s0,strlen(s0));
			logMessage(s0);
		}		
	}
	/* DOC: @SUPPORTFILESTATUS (tells which support files are marked 'missing') */
	if (strstr(d->incomm,"@SUPPORTFILESTATUS")) {
		if (FD_ISSET(d->descriptor,&out_set)) {
			sprintf(s0,"SUPPFILESTAT:%d:%d:%d\n",nocompsfile,norpmdir,nohdlist);
			write_to_descriptor(d->descriptor,s0,strlen(s0));
			logMessage(s0);
		}
	}
	/* DOC: @BEGINPACKAGEMANAGEMENT (sends initial package data over socket) */
	if (strstr(d->incomm,"@BEGINPACKAGEMANAGEMENT")) {
		theInstall->selectedPackageSet=prepareCustom(NULL,theInstall->masterComponentList,theInstall->selectedPackageSet,&theInstall->masterPackageSet,theInstall->forcepackages,theInstall->installedPackages,theInstall->ipkglist);
	}
	/* DOC: @GETSETSIZE (returns size in MB and number of pkgs selected) */
	if (strstr(d->incomm,"@GETSETSIZE")) {
		xx=getPkgSetSize2(theInstall->selectedPackageSet,theInstall->installedPackages);
		yy=getNumPackages(theInstall->selectedPackageSet,removemarked);
		/*
		for (zz=0;zz < theInstall->selectedPackageSet->numPackages;zz++) {
			if (theInstall->selectedPackageSet->packages[zz]->selected == 1) {
				yy++;
			}
		}
		*/
		if (FD_ISSET(d->descriptor,&out_set)) {
			sprintf(s0,"SETSIZE:%d:%d\n",xx,yy);
			write_to_descriptor(d->descriptor,s0,strlen(s0));
			logMessage(s0);
		}	
	}
	/* DOC: @CHECKFREESPACE (checks if enough disk space is available for the install) */
	if (strstr(d->incomm,"@CHECKFREESPACE")) {
		xx=getPkgSetSize2(theInstall->selectedPackageSet,
				  theInstall->installedPackages);
		yy=checkFreeSpace((xx * 1024),&xa);
		sprintf(s0,"FREESPACE:%d:%d:%d\n",yy,xx,(xa / 1024));
		autoMessage(d->descriptor,s0);
	}
	/* DOC: @SETPACKAGESTATE:pkgname:state (sets the state (select/noselect) of pkgname) */
	if (strstr(d->incomm,"@SETPACKAGESTATE")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			pptr=getPackageByName(theInstall->selectedPackageSet,strtok(s1,":"));
			xx=atoi(strtok(NULL,":"));
			pptr->selected=xx;
			logMessage("AUTOCONTROL: Set '%s' to '%d'",pptr->name,pptr->selected);
			autoSuccess(d->descriptor);
		}
	}
	/* DOC: @MULTISETPACKAGESTATE:numpackages:state:pkg1:pkg2: ... :pkgN (sets multiple pkgs) */
	if (strstr(d->incomm,"@MULTISETPACKAGESTATE")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			if (s1[strlen(s1) - 1] == '\n') {
				s1[strlen(s1) - 1]='\0';
			}
			yy=atoi(strtok(s1,":"));
			xx=atoi(strtok(NULL,":"));
			for (zz=0;zz < yy;zz++) {
				s4=strtok(NULL,":");
				if (s4) {
					pptr=getPackageByName(theInstall->selectedPackageSet,s4);
					if (pptr) {
						pptr->selected=xx;
						logMessage("AUTOCONTROL: Set '%s' to '%d'",pptr->name,pptr->selected);
					}
				}
			}
			autoSuccess(d->descriptor);
		}
	}
	/* DOC: @QUERYPACKAGE:packagename (query packagename for information) */
	if (strstr(d->incomm,"@QUERYPACKAGE")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			if (s1[strlen(s1) - 1] == '\n') {
				s1[strlen(s1) - 1] = '\0';
			}
			queryresults=getRPMInfo(theInstall->selectedPackageSet,s1);
			if (!queryresults) {
				autoErrMsg(d->descriptor,_("NoResult"),0);
			} else {
				for (zz=0;zz < strlen(queryresults[0]);zz++) {
					if (queryresults[0][zz] == '\n') {
						queryresults[0][zz] = '';
					}
				}
				for (zz=0;zz < strlen(queryresults[1]);zz++) {
					if (queryresults[1][zz] == '\n') {
						queryresults[1][zz] = '';
					}
				}
				autoMessage(d->descriptor,queryresults[0]);
				autoMessage(d->descriptor,queryresults[1]);
				autoSuccess(d->descriptor);
			}
		}
	}
	/* DOC: @QUERYDEPS:packagename (query dependencies for packagename) */
	if (strstr(d->incomm,"@QUERYDEPS")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			if (s1[strlen(s1) - 1] == '\n') {
				s1[strlen(s1) - 1] = '\0';
			}
			qreq=getRequires(&theInstall->masterPackageSet,s1);
			qprov=getProvides(&theInstall->masterPackageSet,s1);
			if ((!qreq) || (!qprov)) {
				autoErrMsg(d->descriptor,_("NoResult"),0);
			} else {
				strcpy(s2,"");
				strcpy(s3,"");
				sptr=qreq;
				while (*sptr) {
					logMessage("REQUIRES: '%s'",*sptr);
					strcat(s2,*sptr);
					strcat(s2,"");
					sptr++;
				}
				if (s2[strlen(s2) - 1] == '') {
					s2[strlen(s2) - 1] = '\0';
				}
				sptr=qprov;
				while (*sptr) {
					logMessage("PROVIDES: '%s'",*sptr);
					strcat(s3,*sptr);
					strcat(s3,"");
					sptr++;
				}
				if (s3[strlen(s3) - 1] == '') {
					s3[strlen(s3) - 1] = '\0';
				}
				sprintf(s0,"REQS:%s\n",s2);
				logMessage(s0);
				autoMessage(d->descriptor,s0);
				sprintf(s0,"PROVS:%s\n",s3);
				logMessage(s0);
				autoMessage(d->descriptor,s0);
				autoSuccess(d->descriptor);
			}
		}
	}
	/* DOC: @ISPACKAGEINSTALLED:packagename [returns pkgname:character [+-=R]] (is packagename installed?) */
	if (strstr(d->incomm,"@ISPACKAGEINSTALLED")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			if (s1[strlen(s1) - 1] == '\n') {
				s1[strlen(s1) - 1] = '\0';
			}
			pptr=getPackageByName(theInstall->selectedPackageSet,s1);
			if (!pptr) {
				autoErrMsg(d->descriptor,"CannotFindPackage",1);
				return FAILURE;
			}
			xc=isPackageInstalled(theInstall->installedPackages,
					      theInstall->ipkglist,
					      &theInstall->masterPackageSet,
					      pptr,
					      removemarked);
			sprintf(s0,"PACKAGEINSTALLED:%s:%c\n",s1,xc);
			autoMessage(d->descriptor,s0);
		}
	}
	/* DOC: @ISPACKAGESELECTED:packagename [returns pkgname:int [0,1]] (is packagename selected?) */
	if (strstr(d->incomm,"@ISPACKAGESELECTED")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			if (s1[strlen(s1) - 1] == '\n') {
				s1[strlen(s1) - 1] = '\0';
			}
			pptr=getPackageByName(theInstall->selectedPackageSet,s1);
			if (!pptr) {
				autoErrMsg(d->descriptor,"CannotFindPackage",1);
				return FAILURE;
			}
			xx=pptr->selected;
			sprintf(s0,"PACKAGESELECTED:%s:%d\n",s1,xx);
			autoMessage(d->descriptor,s0);
		}
	}
	/* DOC: @REMOVEPACKAGE:packagename (toggles the remove flag) */
	if (strstr(d->incomm,"@REMOVEPACKAGE")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			if (s1[strlen(s1) - 1] == '\n') {
				s1[strlen(s1) - 1] = '\0';
			}
			if (!memberofStrArray(removemarked,s1)) {
				addToStrArray(removemarked,s1);
				if (memberofStrArray(removemarked,s1)) {
					autoSuccess(d->descriptor);
				} else {
					autoErrMsg(d->descriptor,"CouldNotRemove",0);
				}
			} else {
				removemarked=removeFromStrArray(removemarked,s1,MAX_PACKAGES);
				if (memberofStrArray(removemarked,s1)) {
					autoErrMsg(d->descriptor,"CouldNotUnRemove",0);
				} else {
					autoSuccess(d->descriptor);
				}
			}
		}
	}
	/* DOC: @CLEARSELECTION (unselects all packages) */
	if (strstr(d->incomm,"@CLEARSELECTION")) {
		for (zz=0;zz < theInstall->selectedPackageSet->numPackages;zz++) {
			theInstall->selectedPackageSet->packages[zz]->selected=0;
		}
		autoSuccess(d->descriptor);
	}
	/* DOC: @ALLUPGRADES:dodowngrades (selects all upgrades/downgrades) */
	if (strstr(d->incomm,"@ALLUPGRADES")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			if (s1[strlen(s1) - 1] == '\n') {
				s1[strlen(s1) - 1] = '\0';
			}
			dodowngrades=atoi(s1);
			for (zz=0;zz < theInstall->selectedPackageSet->numPackages;zz++) {
				xc=isPackageInstalled(theInstall->installedPackages,
						      theInstall->ipkglist,
						      &theInstall->masterPackageSet,
						      theInstall->selectedPackageSet->packages[zz],
						      removemarked);
				if ((xc == ' ') || (xc == '+')) {
					theInstall->selectedPackageSet->packages[zz]->selected=1;
				} else if ((xc == '-') && (dodowngrades)) {
					theInstall->selectedPackageSet->packages[zz]->selected=1;
				}
			}
			autoSuccess(d->descriptor);
		}
	}
	/* DOC: @EXTRACTPICTURE:packagename:basefn (Extracts picture from package to basefn) */
	if (strstr(d->incomm,"@EXTRACTPICTURE")) {
		s1=index(d->incomm,':');
		if (!s1) {
			autoErrMsg(d->descriptor,_("NoParameter"),0);
		} else {
			s1++;
			s4=strtok(s1,":");
			s5=strtok(NULL,":");
			xx=extractPictureFromPackage(theInstall,s4,s5);
			switch(xx) {
			case 0:
				sprintf(s0,"PICTURE:0\n");
				autoMessage(d->descriptor,s0);
				break;
			case 1:
				sprintf(s0,"PICTURE:%s.gif\n",s5);
				autoMessage(d->descriptor,s0);
				break;
			case 2:
				sprintf(s0,"PICTURE:%s.xpm\n",s5);
				autoMessage(d->descriptor,s0);
				break;
			default:
				sprintf(s0,"PICTURE:0\n");
				autoMessage(d->descriptor,s0);
				break;
			}
		}
	}
	/* DOC: @VERIFYDEPS (Verifies package dependencies) */
	if (strstr(d->incomm,"@VERIFYDEPS")) {
		xx=verifyDependencies(theInstall->selectedPackageSet,
				      theInstall->installedPackages,
				      removemarked);
		if (xx == SUCCESS) {
			autoSuccess(d->descriptor);
		} else {
			autoErrMsg(d->descriptor,"DepsFailedVerify",0);
		}
	}
	/* DOC: @REMOVECOUNT (Returns number of packages marked R) */
	if (strstr(d->incomm,"@REMOVECOUNT")) {
		rcount=0;
		sptr=removemarked;
		while(*sptr) {
			rcount++;
			sptr++;
		}
		sprintf(s0,"REMOVECOUNT:%d\n",rcount);
		autoMessage(d->descriptor,s0);
	}
	/* DOC: @DOREMOVES (Removes packages marked for removal */
	if (strstr(d->incomm,"@DOREMOVES")) {
		rc=removeSelectedPackages(removemarked);
		if (rc == SUCCESS) {
			autoSuccess(d->descriptor);
		} else {
			autoErrMsg(d->descriptor,"CannotUnInstall",0);
		}
	}
	/* DOC: @STARTINSTALL (Starts the package installation/upgrade process */
	if (strstr(d->incomm,"@STARTINSTALL")) {
		rc=doInstall(theInstall,&ftpconf);
		if (rc == FAILURE) {
			autoErrMsg(d->descriptor,"InstallFailed",1);
		} else {
			autoSuccess(d->descriptor);
		}
	}
	return SUCCESS;
}

int autoLoadConfig(char *configfn, char **compsfilepath, char **rpmpath, char **hdlistpath, ftpConfig *ftpconf, int *isftp, int *nocompsfile, int *norpmdir, int *nohdlist) {
	FILE *fp;
	char s0[256];

	logMessage("AUTOCONTROL_SRV: Loading Environment Config from '%s'",configfn);
	fp=fopen(configfn,"r");
	if (!fp) return FAILURE;

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	*compsfilepath=strdup(s0);
	
	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	*rpmpath=strdup(s0);
	
	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	*hdlistpath=strdup(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	ftpconf->ftphost=strdup(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	ftpconf->ftpdir=strdup(s0);	

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	ftpconf->username=strdup(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	ftpconf->password=strdup(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	ftpconf->proxy=strdup(s0);
	if (!strlen(ftpconf->proxy)) {
		ftpconf->proxy=NULL;
	}
	
	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	ftpconf->port=atoi(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	*isftp=atoi(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	*nocompsfile=atoi(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	*norpmdir=atoi(s0);

	fgets(s0,256,fp);
	s0[strlen(s0)-1]='\0';
	*nohdlist=atoi(s0);

	fclose(fp);
	
	return SUCCESS;
}

