/*
#### Copyright (c) 1997 Pacific HiTech, Inc. 
####              All Rights Reserved.
#### File:        autowin.h
#### Author:      koji@pht.com
#### Date:        2/17/98
#### Description: 
####       collection of useful newt window related functions.
*/


#ifndef AUTOWIN_H
#define AUTOWIN_H

#include <errno.h>
#include "includes.h"

typedef struct componentGroup_struct *componentGroup;
typedef struct selectionList_struct  *selectionList;

// data structure to hold selection list information.
typedef struct selectionList_struct {
    char *title;
    void *data;
    selectionList next;
    selectionList last;
} selectionList_struct_type;


// list box component.  it pretends to be a component group.
typedef struct selectListbox_struct {
    int height, width;
    int def;
    selectionList list;
    newtComponent *comp;
    bool multi;
} selectListbox_struct_type;


/*   types of dialogs.   */
enum cg_orientation { vertical, horizontal };
enum cg_gravity { topLeft, center, bottomRight };
enum open_method { show, ask, ask_keep, layout };



// simplified interface for opening a standard dialogs.
void msgWindow(char *title, char *buttontext, char *message, ...);
bool yesnoWindow(char *title, bool expectYes, 
    char *btext1, char *btext2, char *message, ...);
void errWindow(bool addErrorStr, char * message, ...);
void statusWindow(char *title, char *message, ...);
char *entryWindow(char *title, int entry_len, char *def_text,
    char *btext1, char *btext2, char *message, ...);
void openCenteredWindow(int width, int height, char *title);
void releaseForm(newtComponent *form);
int putRootText(int left, int top, char *message, ...);
void eraseRootText(int left, int top, int len);

// generic interface for arbitrary dialogs.
newtComponent askUser(enum open_method method, char * title, 
    newtComponent focus, componentGroup group);
newtComponent properTextbox(char * message, ...);
componentGroup compGroup(
    enum cg_orientation orientation, enum cg_gravity gravity, ...);
componentGroup groupMargine(int x_margine, int y_margine, 
    componentGroup group);

// listbox handling.
componentGroup selectionListbox(newtComponent *listboxComp, 
   bool multi, int height, int def, selectionList list);
selectionList addToSelectionList(selectionList *list, ...);

#endif /* AUTOWIN_H */
