/*
#### Copyright (c) 1998 Pacific HiTech, Inc.
####              All Rights Reserved
####              Distributable under the terms of the GPL.
#### File:        ftp.c
#### Author:      Scott Stone <sstone@turbolinux.com>
#### Version:     1.0J
#### Date:        January 28, 1998
#### Description:
# Source code for TurboLinux
#### ChangeLog:
# (none)
*/

#include "defines.h"
#include "config.h"
#include "autowin.h"
#include "log.h"
#include "prototypes.h"
#include <alloca.h>
#include <sys/types.h>
#include <netinet/in_systm.h>
#include <time.h>
#ifndef GLIBC
#include <linux/if.h>
#include <linux/route.h>
#endif
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>
#include <netinet/in.h>
#ifdef GLIBC
#include <socketbits.h>
#else
#include <netinet/ip.h>
#include <resolv.h>
#endif
#include <arpa/inet.h> 

#if ! HAVE_HERRNO
extern int h_errno;
#endif

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef GLIBC
#include "ftp-glibc.h"
#endif
#include "inet_aton.h"		/* for systems too stupid to provide this */
#ifdef INSTALLER
#ifdef GLIBC
#include "dns.h"
#endif
#endif

#define TIMEOUT_SECS 60
#define BUFFER_SIZE 4096

#ifndef IPPORT_FTP
# define IPPORT_FTP 21
#endif

#include "ftp.h"

extern int debug;
extern int ftpinstall;
extern int autocontrol;
extern int controlsocket;
extern fd_set in_set;
extern fd_set out_set;
extern DESCRIPTOR_DATA *mydesc;


newtComponent getScale,getLabel;

char *netdevices_installer[] = { "eth0", "tr0", "plip0", "plip1", "plip2", NULL };
char *netdevices[] = { "ppp0", "eth0", "tr0", "plip0", "plip1", "plip2", NULL };

static int ftpCheckResponse(int sock, char ** str);
static int ftpCommand(int sock, char * command, ...);
static int ftpReadData(int sock, int out, int filesize, int displaystatus, char *filename, newtComponent *scale);
static int getHostAddress(const char * host, struct in_addr * address);

char *tryPaths[] = {
	"TurboLinux/base",
	"TurboUpgrade/base",
	"RPMS/",
	ARCH_PATH,
	"/",
	NULL,
};

static int ftpCheckResponse(int sock, char ** str) {
    static char buf[BUFFER_SIZE + 1];
    int bufLength = 0; 
    fd_set emptySet, readSet;
    char * chptr, * start;
    struct timeval timeout;
    int bytesRead, rc = 0;
    int doesContinue = 1;
    char errorCode[4];
 
    errorCode[0] = '\0';
    
    do {
	FD_ZERO(&emptySet);
	FD_ZERO(&readSet);
	FD_SET(sock, &readSet);

	timeout.tv_sec = TIMEOUT_SECS;
	timeout.tv_usec = 0;
    
	rc = select(sock + 1, &readSet, &emptySet, &emptySet, &timeout);
	if (rc < 1) {
	    if (rc==0) 
		return FTPERR_BAD_SERVER_RESPONSE;
	    else
		rc = FTPERR_UNKNOWN;
	} else
	    rc = 0;

	bytesRead = read(sock, buf + bufLength, sizeof(buf) - bufLength - 1);

	bufLength += bytesRead;

	buf[bufLength] = '\0';

	/* divide the response into lines, checking each one to see if 
	   we are finished or need to continue */

	start = chptr = buf;

	do {
	    while (*chptr != '\n' && *chptr) chptr++;

	    if (*chptr == '\n') {
		*chptr = '\0';
		if (*(chptr - 1) == '\r') *(chptr - 1) = '\0';
		if (str) *str = start;

		if (errorCode[0]) {
		    if (!strncmp(start, errorCode, 3) && start[3] == ' ')
			doesContinue = 0;
		} else {
		    strncpy(errorCode, start, 3);
		    errorCode[3] = '\0';
		    if (start[3] != '-') {
			doesContinue = 0;
		    } 
		}

		start = chptr + 1;
		chptr++;
	    } else {
		chptr++;
	    }
	} while (*chptr);

	if (doesContinue && chptr > start) {
	    memcpy(buf, start, chptr - start - 1);
	    bufLength = chptr - start - 1;
	} else {
	    bufLength = 0;
	}
    } while (doesContinue && !rc);

    if (*errorCode == '4' || *errorCode == '5') {
	if (!strncmp(errorCode, "550", 3)) {
	    return FTPERR_FILE_NOT_FOUND;
	}

	return FTPERR_BAD_SERVER_RESPONSE;
    }

    if (rc) return rc;

    return 0;
}

int ftpCommand(int sock, char * command, ...) {
    va_list ap;
    int len;
    char * s;
    char * buf;
    int rc;

    va_start(ap, command);
    len = strlen(command) + 2;
    s = va_arg(ap, char *);
    while (s) {
	len += strlen(s) + 1;
	s = va_arg(ap, char *);
    }
    va_end(ap);

    buf = alloca(len + 1);

    va_start(ap, command);
    strcpy(buf, command);
    strcat(buf, " ");
    s = va_arg(ap, char *);
    while (s) {
	strcat(buf, s);
	strcat(buf, " ");
	s = va_arg(ap, char *);
    }
    va_end(ap);

    buf[len - 2] = '\r';
    buf[len - 1] = '\n';
    buf[len] = '\0';
     
    if (write(sock, buf, len) != len) {
        return FTPERR_SERVER_IO_ERROR;
    }

    if ((rc = ftpCheckResponse(sock, NULL)))
	return rc;

    return 0;
}

static int getHostAddress(const char * host, struct in_addr * address) {
#ifndef INSTALLER
	struct hostent * hostinfo;
#endif
#if (defined INSTALLER) && (defined GLIBC)
	struct sockaddr_in mysockaddr;
#endif

    if (isdigit(host[0])) {
      if (!inet_aton(host, address)) {
	  return FTPERR_BAD_HOST_ADDR;
      }
    } else {
#if (defined INSTALLER) && (defined GLIBC)
#warning Using mygethostbyname()
	    if (mygethostbyname(host,&mysockaddr.sin_addr)) {
		    return FTPERR_BAD_HOSTNAME;
	    }
	    memcpy(address,&mysockaddr.sin_addr,sizeof(*address));
#else
	    hostinfo = gethostbyname(host);
	    if (!hostinfo) {
		    errno = h_errno;
		    return FTPERR_BAD_HOSTNAME;
	    }
	    
	    memcpy(address, hostinfo->h_addr_list[0], hostinfo->h_length);
#endif
    }
    
    return 0;
}

int ftpOpen(char * host, char * name, char * password, char * proxy,
	    int port) {
    static int sock;
    /*static char * lastHost = NULL;*/
    struct in_addr serverAddress;
    struct sockaddr_in destPort;
    struct passwd * pw;
    char * buf;
    int rc;

    if (port < 0) port = IPPORT_FTP;

    if (!name)
	name = "anonymous";

    if (!password) {
	if (getuid()) {
	    pw = getpwuid(getuid());
	    password = alloca(strlen(pw->pw_name) + 2);
	    strcpy(password, pw->pw_name);
	    strcat(password, "@");
	} else {
	    password = "tl-install@";
	}
    }

    if (proxy) {
	buf = alloca(strlen(name) + strlen(host) + 5);
	sprintf(buf, "%s@%s", name, host);
	name = buf;
	host = proxy;
    }

    logMessage("FTP: Retrieving host address for '%s'...",host);
    if ((rc = getHostAddress(host, &serverAddress))) return rc;

    logMessage("FTP: Creating socket...");
    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (sock < 0) {
	    logMessage("FTP: Socket creation failed: %s",strerror(errno));
	    return FTPERR_FAILED_CONNECT;
    }

    destPort.sin_family = AF_INET;
    destPort.sin_port = htons(port);
    destPort.sin_addr = serverAddress;

    logMessage("FTP: Connecting...");
    if (connect(sock, (struct sockaddr *) &destPort, sizeof(destPort))) {
	close(sock);
	logMessage("FTP: Cannot connect socket!");
        return FTPERR_FAILED_CONNECT;
    }

    /* ftpCheckResponse() assumes the socket is nonblocking */
    if (fcntl(sock, F_SETFL, O_NONBLOCK)) {
	close(sock);
	logMessage("FTP: Cannot set socket to non-blocking mode");
        return FTPERR_FAILED_CONNECT;
    }

    if ((rc = ftpCheckResponse(sock, NULL))) {
	    logMessage("FTP: Bad response from ftpCheckResponse");
	    return rc;     
    }

    logMessage("FTP: Logging in, setting type to 'I'");
    if ((rc = ftpCommand(sock, "USER", name, NULL))) {
	close(sock);
	return rc;
    }

    if ((rc = ftpCommand(sock, "PASS", password, NULL))) {
	close(sock);
	return rc;
    }

    if ((rc = ftpCommand(sock, "TYPE", "I", NULL))) {
	close(sock);
	return rc;
    }

    return sock;
}

int ftpReadData(int sock, int out, int filesize, int displaystatus, char *filename, newtComponent *scale) {
    char buf[BUFFER_SIZE];
    fd_set emptySet, readSet;
    struct timeval timeout;
    int bytesRead,rc;
    int bytestotal=0;
    char s0[256];
    double done=0.0;

    while (1) {
	FD_ZERO(&emptySet);
	FD_ZERO(&readSet);
	FD_SET(sock, &readSet);

	timeout.tv_sec = TIMEOUT_SECS;
	timeout.tv_usec = 0;
    
	rc = select(sock + 1, &readSet, &emptySet, &emptySet, &timeout);
	if (rc == 0) {
	    close(sock);
	    return FTPERR_SERVER_TIMEOUT;
	} else if (rc < 0) {
	    close(sock);
	    return FTPERR_UNKNOWN;
	}

	bytesRead = read(sock, buf, sizeof(buf));
	if (bytesRead == 0) {
	    close(sock);
	    return 0;
	}
	bytestotal += bytesRead;
	if (displaystatus == 2) {
		if (filesize == 0) {
			sprintf(s0,"Unknown Progress - Server does not support SIZE\n");
		} else {			
			done=((double)bytestotal/(double)filesize) * 100.0;
			sprintf(s0,"%d of %d recvd, %.2f percent done\n",bytestotal,filesize,done);
		}
		write(1,s0,strlen(s0));
	} else if (displaystatus == 1) {
		done=((double)bytestotal/(double)filesize) * 100.0;
		if (autocontrol) {
#ifdef AUTOCONTROL
			if (filesize == 0) {
				sprintf(s0,"FTP:%s:0:0\n",filename);
			} else {
				sprintf(s0,"FTP:%s:%d:%d\n",filename,filesize,(int)done);
			}
			if (debug) {
				logMessage(s0);
			}
			autoMessage(controlsocket,s0);
			if ( FD_ISSET( mydesc->descriptor, &in_set ) )
			{
				if ( !read_from_descriptor( mydesc ) )
				{
					FD_CLR( mydesc->descriptor, &out_set );
					mydesc->outtop	= 0;
					close_socket( mydesc );
					continue;
				}
			}
			read_from_buffer( mydesc );
			if ( mydesc->incomm[0] != '\0' )
			{
				if (strstr(mydesc->incomm,"@CANCEL")) {
					autoMessage(mydesc->descriptor,"CANCELLED:");
					mydesc->incomm[0]='\0';
					close(sock);
					return FTPERR_USER_CANCELLED;
				}
				mydesc->incomm[0]='\0';
			}
#endif
		} else {
			newtScaleSet(*scale,(int)done);
			newtRefresh();
		}
	}

	if (write(out, buf, bytesRead) != bytesRead) {
	    close(sock);
	    return FTPERR_FILE_IO_ERROR;
	}
    }
}

int ftpGetFileDesc(int sock, char * remotename, int *filesize) {
    int dataSocket=-1;
    struct sockaddr_in dataAddress;
    int i, j;
    char * passReply;
    char * sizeReply;
    char * chptr;
    char * retrCommand;
    char sizeCommand[256];
    int rc;
    char *sptr=NULL;

    sprintf(sizeCommand,"SIZE %s\r\n", remotename);
    i=strlen(sizeCommand);
    if (write(sock, sizeCommand, i) != i) {
        return FTPERR_SERVER_IO_ERROR;
    }

    if ((rc = ftpCheckResponse(sock, &sizeReply))) {
/*	close(dataSocket); */
	    logMessage("SIZE: Server does not support 'SIZE'.");
/*	    return rc; */
	    *filesize=0;
    } else {
	    sptr=index(sizeReply,' ');
	    sptr++;
	    *filesize=atoi(sptr);
    }
    
    if (write(sock, "PASV\r\n", 6) != 6) {
        return FTPERR_SERVER_IO_ERROR;
    }
    if ((rc = ftpCheckResponse(sock, &passReply)))
	return FTPERR_PASSIVE_ERROR;

    chptr = passReply;
    while (*chptr && *chptr != '(') chptr++;
    if (*chptr != '(') return FTPERR_PASSIVE_ERROR; 
    chptr++;
    passReply = chptr;
    while (*chptr && *chptr != ')') chptr++;
    if (*chptr != ')') return FTPERR_PASSIVE_ERROR;
    *chptr-- = '\0';

    while (*chptr && *chptr != ',') chptr--;
    if (*chptr != ',') return FTPERR_PASSIVE_ERROR;
    chptr--;
    while (*chptr && *chptr != ',') chptr--;
    if (*chptr != ',') return FTPERR_PASSIVE_ERROR;
    *chptr++ = '\0';
    
    /* now passReply points to the IP portion, and chptr points to the
       port number portion */

    dataAddress.sin_family = AF_INET;
    if (sscanf(chptr, "%d,%d", &i, &j) != 2) {
	return FTPERR_PASSIVE_ERROR;
    }
    dataAddress.sin_port = htons((i << 8) + j);

    chptr = passReply;
    while (*chptr++) {
	if (*chptr == ',') *chptr = '.';
    }

    if (!inet_aton(passReply, &dataAddress.sin_addr)) 
	return FTPERR_PASSIVE_ERROR;

    dataSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (dataSocket < 0) {
        return FTPERR_FAILED_CONNECT;
    }

    
    retrCommand = alloca(strlen(remotename) + 20);
    sprintf(retrCommand, "RETR %s\r\n", remotename);
    i = strlen(retrCommand);
   
    if (write(sock, retrCommand, i) != i) {
        return FTPERR_SERVER_IO_ERROR;
    }

    if (connect(dataSocket, (struct sockaddr *) &dataAddress, 
	        sizeof(dataAddress))) {
	close(dataSocket);
        return FTPERR_FAILED_DATA_CONNECT;
    }

    if ((rc = ftpCheckResponse(sock, NULL))) {
	close(dataSocket);
	return rc;
    }

    return dataSocket;
}

int ftpGetFileDone(int sock) {
    if (ftpCheckResponse(sock, NULL)) {
	return FTPERR_BAD_SERVER_RESPONSE;
    }

    return 0;
}

void ftpOpenScaleWindow(newtComponent *getForm,const char *filename) {
	char s0[256];

	logMessage("Opening scale window...");
/*	scaletextbox=properTextbox(_("Receiving '%s'"),filename); */
	sprintf(s0,_("Receiving '%s'"),filename);
	getLabel=newtLabel(0,0,s0);
	getScale=newtScale(0,0,50,100);
	newtScaleSet(getScale,0);
	*getForm=askUser(layout,_("Receiving File"),NULL,
			  compGroup(vertical,center,
				    groupMargine(0,1,
						 compGroup(vertical,center,
							   getLabel,
							   NULL)
						 ),
				    groupMargine(0,1,
						 compGroup(vertical,center,
							   getScale,
							   NULL)
						 ),
				    NULL)
			  );
	newtDrawForm(*getForm);
	newtRefresh();
}

int ftpGetRPM(int sock, char *remotename, int dest, newtComponent *label, newtComponent *sizelabel, newtComponent *scale) {
	int dataSocket,rc;
	int filesize=0;
	char s0[256];
	char *s1=NULL;
	char s2[256];

	s1=rindex(remotename,'/');
	s1++;
	logMessage("FTP: Retrieving RPM: '%s'",s1);
	sprintf(s2,"%s/%s",installer ? RPMSDIR_FTP_INSTALLER : RPMSDIR_FTP,s1);
	logMessage("FTP: Filename is '%s'",s2);
	dataSocket=ftpGetFileDesc(sock,remotename,&filesize);
	if (dataSocket < 0) return dataSocket;
	if (autocontrol) {
#ifdef AUTOCONTROL
		sprintf(s0,"FTPSTART:%s:%d\n",s1,filesize);
		autoMessage(controlsocket,s0);
#endif
	} else {
		sprintf(s0,_("FTP: %s"),s1);
		newtLabelSetText(*label,s0);
		if (filesize == 0) {
			sprintf(s0,_("(SIZE not available)"));
		} else {
			sprintf(s0,_("%d kb"),(filesize / 1024));
		}
		newtLabelSetText(*sizelabel,s0);
		newtRefresh();
	}
	rc=ftpReadData(dataSocket,dest,filesize,1,s2,scale);
	close(dataSocket);
	if (rc != 0) return rc;
	return ftpGetFileDone(sock);
}
	
int ftpGetFile(int sock, char * remotename, int dest, int displaystatus) {
    int dataSocket, rc;
    int filesize=0;
    char s0[256];
    char *s1=NULL;
    newtComponent getForm;
    char rootText[256];

    if (displaystatus == 1) {
	    s1=rindex(remotename,'/');
	    s1++;
	    logMessage("FTP: Retrieving '%s' as '%s'",remotename,s1);
	    if (!autocontrol) {
		    ftpOpenScaleWindow(&getForm,s1);
	    }
    }
    dataSocket = ftpGetFileDesc(sock, remotename,&filesize);
    if (dataSocket < 0) {
	    if (!autocontrol) {
		    releaseForm(&getForm);
	    }
	    return dataSocket;
    }

    if (autocontrol) {
#ifdef AUTOCONTROL
	    sprintf(s0,"FTP:%s:%d:0\n",s1,filesize);
	    logMessage(s0);
	    autoMessage(controlsocket,s0);
#endif
    } else {
	    if (filesize == 0) {
		    sprintf(s0,_("(SIZE not available)"));
	    } else {
		    sprintf(s0,_("Receiving '%s' (%d kb)"),s1,(filesize / 1024));
	    }
	    newtLabelSetText(getLabel,s0);
	    newtRefresh();
    }
    rc = ftpReadData(dataSocket, dest, filesize, displaystatus, (displaystatus == 1) ? s1 : NULL,&getScale);
    close(dataSocket);
        

    if (displaystatus == 1) {
	    if (!autocontrol) {
		    releaseForm(&getForm);
		    newtCls();
		    sprintf(rootText,_("TurboPkg v%s %s - (C) 1998 Pacific HiTech"),VERSION,ARCH);
		    newtDrawRootText(0,0,rootText);
		    newtPushHelpLine(_("<TAB>/<Alt-TAB> between elements  | <SPACE> Selects | <F1> = Help"));
		    newtRefresh();
/*	    newtPopWindow(); */
	    }
    }
    if (rc) return rc;
    return ftpGetFileDone(sock);
}

int ftpGetFileByName(int sock, char *remotename, char *localname, int displaystatus) {
	int fd;
	int rc;

	fd=open(localname,O_WRONLY | O_CREAT, FTP_CREATE_MODE);
	if (fd < 0) {
		logMessage("FTP: GetFileByName (create) failed for '%s':");
		logMessage("\t(%s)",strerror(fd));
		if (displaystatus == 1) {
			errWindow(FALSE,_("Could not create destination file: %s"),strerror(fd));
		}
		return FTPERR_FILE_IO_ERROR;
	}
	rc=ftpGetFile(sock,remotename,fd,displaystatus);
	close(fd);
	return rc;
}

void ftpClose(int sock) {
    close(sock);
}

const char *ftpStrerror(int errorNumber) {
  switch (errorNumber) {
    case FTPERR_BAD_SERVER_RESPONSE:
      return ("Bad FTP server response");

    case FTPERR_SERVER_IO_ERROR:
      return("FTP I/O error");

    case FTPERR_SERVER_TIMEOUT:
      return("FTP server timeout");

    case FTPERR_BAD_HOST_ADDR:
      return("Nameserver Lookup Failure: Bad FTP server host address");

    case FTPERR_BAD_HOSTNAME:
      return("Nameserver Lookup Failure: Bad FTP server host name");

    case FTPERR_FAILED_CONNECT:
      return("Failed to connect to FTP server");

    case FTPERR_FAILED_DATA_CONNECT:
      return("Failed to establish data connection to FTP server");

    case FTPERR_FILE_IO_ERROR:
      return("I/O error to local file");

    case FTPERR_PASSIVE_ERROR:
      return("Error setting remote server to passive mode");

    case FTPERR_FILE_NOT_FOUND:
      return("File not found on server");
  case FTPERR_USER_CANCELLED:
	  return("User Cancelled");

    case FTPERR_UNKNOWN:
    default:
      return("FTP Unknown or unexpected error");
  }
}

char *proxyHelpText=__("If you are behind a firewall, you may need to use a proxy host.  Ask your system administrator for these values if you do not know them.");

int ftpOK(void);

int selectMediaWindow(void) {
	newtComponent mForm, mTextbox, mListbox, mOk, mCancel;
	selectionList list=NULL;
	int rc=0;

	mTextbox=properTextbox(_("Select Source Media Type:"));
	addToSelectionList(&list,_("Local Filesystem or CD-ROM"),(void *)1,NULL);
	addToSelectionList(&list,_("FTP Server"),(void *)2,NULL);
	addToSelectionList(&list,_("No Source Media"),(void *)3,NULL);
	mOk=newtButton(0,0,_("Ok"));
	mCancel=newtButton(0,0,_("Cancel"));

	mForm=askUser(ask_keep,_("Source Media"),NULL,
		      compGroup(vertical,center,
				groupMargine(0,1,
					     compGroup(vertical,center,
						       mTextbox,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,center,
						       selectionListbox(&mListbox,FALSE,4,0,list),
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(horizontal,bottomRight,
						       mOk,
						       mCancel,
						       NULL)
					     ),
				NULL)
		      );
	if (newtFormGetCurrent(mForm) == mCancel) {
		logMessage("MEDIA: Cancelled.");
		releaseForm(&mForm);
		return 99; /* special return code */
	}
	rc = (int)newtListboxGetCurrent(mListbox);
	logMessage("MEDIA: Selected mediatype '%d'",rc);
	releaseForm(&mForm);
	return rc;
}
	

int askFTP(ftpConfig *f, int in_installer) {
	int ichoice;
	bool choice,choice1;
	static newtComponent ftpform,ftpcancel;
	char *newProxyHost=NULL;
	char *newPort=NULL;
	int fdone=0;
	static struct newtExitStruct event;

	f->proxy=NULL;
	f->port=FTP_DEFAULT_PORT;
	f->username=strdup("anonymous");
	f->password=strdup("tl-install@");
	if (in_installer) {
		if (!ftpinstall) return 0;
		choice=TRUE;
	} else {
		if (!ftpOK()) {
			errWindow(FALSE,_("You have no usable network devices available."));
			return 0;
		}
		/*
		choice=yesnoWindow(_("Use FTP?"),FALSE,_("Use FTP"),_("Use Local Filesystem"),_("Will you be installing packages from an FTP site or will you be installing packages from a local filesystem (ie, a CD-ROM)?"));
		*/
		ichoice=selectMediaWindow();
		if (ichoice == 2) {
			choice = TRUE;
		} else {
			choice = FALSE;
		}
		if (ichoice == 99) {
			return 99; /* special return code */
		}
		if (ichoice == 3) {
			return 3;
		}
	}
	if (choice == TRUE) {
		choice1=yesnoWindow(_("Use Proxy?"),FALSE,_("Use Proxy"),_("Do Not Use Proxy"),_("Do you need to use a proxy for outgoing FTP connections?"));
		if (choice1 == TRUE) {
			ftpcancel=newtButton(0,0,_("Cancel"));
			ftpform=askUser(layout,_("FTP Proxy Setup"),NULL,
					compGroup(vertical,center,
						  groupMargine(0,1,
							       compGroup(vertical,center,
									 properTextbox(_("Enter your FTP Proxy settings below:")),
									 NULL)
							       ),
						  groupMargine(0,1,
							       compGroup(vertical,topLeft,
									 compGroup(horizontal,topLeft,
										   newtLabel(0,0,_("FTP Proxy Host:  ")),
										   newtEntry(0,0,NULL,50,&newProxyHost,0),
										   NULL),
									 compGroup(horizontal,topLeft,
										   newtLabel(0,0,_("FTP Proxy Port:  ")),
										   newtEntry(0,0,NULL,6,&newPort,0),
										   NULL),
									 NULL)
							       ),
						  groupMargine(0,1,
							       compGroup(horizontal,bottomRight,
									 newtButton(0,0,_("Ok")),
									 ftpcancel,
									 NULL)
							  ),
						  NULL)
					);
			newtFormAddHotKey(ftpform,NEWT_KEY_F1);
			while (!fdone) {
				newtFormRun(ftpform,&event);
				if (event.reason == NEWT_EXIT_HOTKEY) {
					if (event.u.key == NEWT_KEY_F1) {
						helpWindow(_("FTP Proxy"),1,&proxyHelpText);
						/*
						msgWindow(_("Help"),_("Ok"),proxyHelpText);
						*/
					}
				} else if (event.reason == NEWT_EXIT_COMPONENT) {
					if (event.u.co == ftpcancel) {
						releaseForm(&ftpform);
						return (choice == TRUE) ? 1 : 0;
					}
					fdone=1;
				}
			}
			if ((newProxyHost) && (strlen(newProxyHost))) {
				free(f->proxy);
				f->proxy=strdup(newProxyHost);
				logMessage("FTP: Proxy Host is '%s'",f->proxy);
			}
			if ((newPort) && (strlen(newPort))) {
				f->port=atoi(newPort);
				logMessage("FTP: Proxy Port is '%d'",f->port);
			}
			releaseForm(&ftpform);
		}
	}
	return (choice == TRUE) ? 1 : 0;
}

char *ftpCustomHelpText = __("You may now manually enter the FTP host and directory from which to perform the FTP installation.  The directory should ideally be a TurboLinux distribution's top-level directory (ie, it has a TurboLinux/base subdirectory), although you can also specify a directory that just contains binary RPM packages.");

int getCustomFTP(ftpConfig *f) {
	static newtComponent ftpcform,ftpccancel;
	char *newHost=NULL;
	char *newDir=NULL;
	char *newUsername=NULL;
	char *newPassword=NULL;
	static struct newtExitStruct cevent;
	int cancelled=0;
	int fdone=0;

	ftpccancel=newtButton(0,0,_("Cancel"));
	ftpcform=askUser(layout,_("FTP Directory Setup"),NULL,
			compGroup(vertical,center,
				  groupMargine(0,1,
					       compGroup(vertical,center,
							 properTextbox(_("Enter your FTP Host/Dir settings below:")),
							 NULL)
					  ),
				  groupMargine(0,1,
					       compGroup(vertical,topLeft,
							 compGroup(horizontal,topLeft,
								   newtLabel(0,0,_("FTP Host:  ")),
								   newtEntry(0,0,NULL,50,&newHost,0),
								   NULL),
							 compGroup(horizontal,topLeft,
								   newtLabel(0,0,_("FTP Dir :  ")),
								   newtEntry(0,0,NULL,50,&newDir,0),
								   NULL),
							 compGroup(horizontal,topLeft,
								   newtLabel(0,0,_("FTP User:  ")),
								   newtEntry(0,0,"anonymous",50,&newUsername,0),
								   NULL),
							 compGroup(horizontal,topLeft,
								   newtLabel(0,0,_("FTP Pass:  ")),
								   newtEntry(0,0,"tl-install@",50,&newPassword,0),
								   NULL),
							 NULL)
					  ),
				  groupMargine(0,1,
					       compGroup(horizontal,bottomRight,
							 newtButton(0,0,_("Ok")),
							 ftpccancel,
							 NULL)
					  ),
				  NULL)
		);
	newtFormAddHotKey(ftpcform,NEWT_KEY_F1);
	while (!fdone) {
		newtFormRun(ftpcform,&cevent);
		if (cevent.reason == NEWT_EXIT_HOTKEY) {
			if (cevent.u.key == NEWT_KEY_F1) {
				/*
				msgWindow(_("Help"),_("Ok"),ftpCustomHelpText);
				*/
				helpWindow(_("FTP Custom Parameters"),1,&ftpCustomHelpText);
			}
		} else if (cevent.reason == NEWT_EXIT_COMPONENT) {
			if (cevent.u.co == ftpccancel) {
				cancelled=1;
				releaseForm(&ftpcform);
				if (cancelled) return CANCELLED;
			}
			fdone=1;
		}
	}
	if ((newHost) && (strlen(newHost))) {
		free(f->ftphost);
		f->ftphost=strdup(newHost);
		logMessage("FTP: Host is '%s'",f->ftphost);
	}
	if ((newDir) && (strlen(newDir))) {
		free(f->ftpdir);
		f->ftpdir=strdup(newDir);
		logMessage("FTP: Dir is '%s'",f->ftpdir);
	}
	if ((newUsername) && (strlen(newUsername))) {
		free(f->username);
		f->username=strdup(newUsername);
		logMessage("FTP: Username is '%s'",f->username);
	}
	if ((newPassword) && (strlen(newPassword))) {
		free(f->password);
		f->password=strdup(newPassword);
		logMessage("FTP: Password exists.");
/*		logMessage("FTP: Password is '%s'",f->password); */
	}
	releaseForm(&ftpcform);
	if (cancelled) return CANCELLED;
	return SUCCESS;
}

int getFTPPaths(ftpConfig *f) {
	newtComponent fform,ftextbox,flistbox,fok,fcancel;
	selectionList list=NULL;
	ftpMirror *mptr=NULL;
	char *mastersite=NULL;
	char s0[256];
	char s1[256];
	char *s2=NULL;
	int rc=0;
	bool choice;
	int isupdate=0;
	int h=10;
	int mcount=0;

	if (!f) return FAILURE;
#ifndef INSTALLER
	choice=yesnoWindow(_("FTP Host Type"),TRUE,_("Updates"),_("Full Distribution"),
			   _("Do you want to connect to a host containing updates or a full TurboLinux distribution?  Note that if you want to connect to a custom location, it does not matter which of these options you select."));
	if (choice == TRUE) {
		isupdate=1;
		logMessage("FTP: 'Updates' chosen.");
	}
#endif
	mastersite=strdup(isupdate ? f->uftphost : f->ftphost);
	ftextbox=properTextbox(_("Select the location from which you would like to perform the FTP installation:"));
	sprintf(s0,_("Master Site (%s)"),mastersite);
/*	free(mastersite); */
	if (isupdate) {
		sprintf(s1,"%s:%s",f->uftphost,f->uftpdir);
	} else {
		sprintf(s1,"%s:%s",f->ftphost,f->ftpdir);
	}
	addToSelectionList(&list,strdup(s0),(void *)strdup(s1),NULL);
	addToSelectionList(&list,_("Custom"),"custom",NULL);
	mcount=2;
	mptr=isupdate ? f->umirrors->next : f->mirrors->next;
	while(mptr) {
		sprintf(s0,"Mirror: %s (%s)",mptr->hostname,mptr->location);
		sprintf(s1,"%s:%s",mptr->hostname,mptr->directory);
		addToSelectionList(&list,strdup(s0),(void *)strdup(s1),NULL);
		mptr=mptr->next;
		mcount++;
	}
	fok=newtButton(0,0,_("Ok"));
	fcancel=newtButton(0,0,_("Cancel"));

	if (mcount < h) {
		h=mcount;
	}
	if (h < 2) {
		h=2;
	}
	fform=askUser(ask_keep,_("Select FTP Install Location"),NULL,
		      compGroup(vertical,center,
				groupMargine(0,1,
					     compGroup(vertical,center,
						       ftextbox,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,center,
						       selectionListbox(&flistbox,FALSE,h,0,list),
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(horizontal,bottomRight,
						       fok,
						       fcancel,
						       NULL)
					     ),
				NULL)
		      );
	if (newtFormGetCurrent(fform) == fcancel) {
		rc=CANCELLED;
	} else {
		s2=(char *)newtListboxGetCurrent(flistbox);
		if (!strcmp(s2,"custom")) {
			releaseForm(&fform);
			return (getCustomFTP(f));
		} else {
			free(f->ftphost);
			free(f->ftpdir);
			f->ftphost=strdup(strtok(s2,":"));
			f->ftpdir=strdup(strtok(NULL,":"));
		}
		rc=SUCCESS;
	}
	releaseForm(&fform);
	return rc;
}
		
int getHDListByFTP(ftpConfig *f) {
	int fd;
	int ofd;
	int rc;
	char s0[256];
	char **sptr;
	int found=0;

	sprintf(s0,_("Connecting to remote FTP host (%d second timeout)"),TIMEOUT_SECS);
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatus(controlsocket,"hdlist",s0);
#endif
	} else {
		statusWindow(_("Retrieving Header List"),s0);
	}
	fd=ftpOpen(f->ftphost,f->username,f->password,f->proxy,f->port);
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatusEnd(controlsocket);
#endif
	} else {
		newtPopWindow();
	}
	unlink(installer ? HDLIST_FTP_INSTALLER : HDLIST_FTP);
	ofd=open(installer ? HDLIST_FTP_INSTALLER : HDLIST_FTP,O_WRONLY | O_CREAT, FTP_CREATE_MODE);
	if (fd < 0) {
		if (autocontrol) {
#ifdef AUTOCONTROL
			autoErrMsg(controlsocket,_("FTP Connection Failed."),0);
#endif
		} else {
			errWindow(FALSE,_("FTP Connection Failed: %s"),ftpStrerror(fd));
		}
		logMessage("FTP: Open failed: '%s'",ftpStrerror(fd));
		return FAILURE;
	} else {
		logMessage("FTP: Open to '%s' successful.",f->ftphost);
	}
	sptr=tryPaths;
	while ((*sptr) && (!found)) {
		sprintf(s0,"%s/%s/hdlist",f->ftpdir,*sptr);
		rc=ftpGetFile(fd,s0,ofd,1);
		if (rc >= 0) {
			found=1;
			logMessage("FTP: Retrieved '%s' as '%s'",s0,installer ? HDLIST_FTP_INSTALLER : HDLIST_FTP);
		} else {
			logMessage("FTP: Retrieval of '%s' returned '%s'",s0,ftpStrerror(rc));
		}
		sptr++;
	}
	close(ofd);
	ftpClose(fd);
	if (!found) {
		return FAILURE;
	}
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoMessage(controlsocket,"FTP:Complete");
#endif
	}
	return SUCCESS;
}

int getCompsByFTP(ftpConfig *f, tlInstall *i) {
	int fd;
	int ofd;
	int rc;
	char s0[256];
	char **sptr;
	int found=0;
	char *localfn=strdup(i->compsfile);

	sprintf(s0,_("Connecting to remote FTP host (%d second timeout)"),TIMEOUT_SECS);
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatus(controlsocket,"hdlist",s0);
#endif
	} else {
		statusWindow(_("Retrieving Component List"),s0);
	}
	fd=ftpOpen(f->ftphost,f->username,f->password,f->proxy,f->port);
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatusEnd(controlsocket);
#endif
	} else {
		newtPopWindow();
	}
	unlink(localfn);
	ofd=open(localfn,O_WRONLY | O_CREAT, FTP_CREATE_MODE);
	if (fd < 0) {
		if (autocontrol) {
#ifdef AUTOCONTROL
			autoErrMsg(controlsocket,_("FTP Connection Failed."),0);
#endif
		} else {
			errWindow(FALSE,_("FTP Connection Failed: %s"),ftpStrerror(fd));
		}
		logMessage("FTP: Open failed: '%s'",ftpStrerror(fd));
		return FAILURE;
	} else {
		logMessage("FTP: Open to '%s' successful.",f->ftphost);
	}
	sptr=tryPaths;
	while ((*sptr) && (!found)) {
		sprintf(s0,"%s/%s/comps",f->ftpdir,*sptr);
		if (i->language) {
			strcat(s0,".");
			strcat(s0,i->language);
		}
		rc=ftpGetFile(fd,s0,ofd,1);
		if (rc >= 0) {
			found=1;
			logMessage("FTP: Retrieved '%s' as '%s'",s0,localfn);
		} else {
			logMessage("FTP: Could not get '%s' : '%s'",s0,ftpStrerror(rc));
		}
		sptr++;
	}
	close(ofd);
	ftpClose(fd);
	if (!found) {
		return FAILURE;
	}
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoMessage(controlsocket,"FTP:Complete");
#endif
	}
	return SUCCESS;
}

int getSkelByFTP(ftpConfig *f) {
	int fd;
	int ofd;
	int rc;
	char s0[256];
	int found=0;

	sprintf(s0,_("Connecting to remote FTP host (%d second timeout)"),TIMEOUT_SECS);
	statusWindow(_("Retrieving Base System Tree"),s0);
	fd=ftpOpen(f->ftphost,f->username,f->password,f->proxy,f->port);
	newtPopWindow();
	unlink(SKELETON_LOCAL_FN);
	ofd=open(SKELETON_LOCAL_FN, O_WRONLY | O_CREAT, FTP_CREATE_MODE);
	if (fd < 0) {
		logMessage("FTP: Open failed: '%s'",ftpStrerror(fd));
		newtPopWindow();
		return FAILURE;
	} else {
		logMessage("FTP: Open to '%s' successful.",f->ftphost);
	}
	sprintf(s0,"%s/%s",f->ftpdir,SKELETON_REMOTE_FN);
	rc=ftpGetFile(fd,s0,ofd,1);
	if (rc >= 0) {
		found=1;
		logMessage("FTP: Retrieved '%s' as '%s'",s0,SKELETON_LOCAL_FN);
	} else {
		logMessage("FTP: Could not get '%s' : '%s'",s0,ftpStrerror(rc));
	}

	close(ofd);
	ftpClose(fd);
	if (!found) {
		return FAILURE;
	}
	return SUCCESS;
}

int netDeviceAvailable(char * device) {
    struct ifreq req;
    int s;
    
    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0) {
        close(s);
        errWindow(TRUE, _("Socket Error!"));
        return 1;
    }

    strcpy(req.ifr_name, device);

    if (ioctl(s, SIOCGIFFLAGS, &req)) {
        /* if we can't get the flags, the networking device isn't available */
        close(s);
        return 0;
    }
  
    return 1;
}


int ftpOK(void) {
	char **sptr=installer ? netdevices_installer : netdevices;

	while(*sptr) {
		if (netDeviceAvailable(*sptr)) {
			logMessage("NETWORK: Interface '%s' is available, FTP is ok.",*sptr);
			return 1;
		}
		sptr++;
	}
	return 0;
}

#ifdef INSTALLER
int ftpForceXServers(pkgSet *theSet) {
	struct packageInfo *pptr=NULL;
	static int zz;

	pptr=getPackageByName(theSet,"XFree86");
	if (!pptr) {
		return FAILURE;
	}
	if (!pptr->selected) {
		logMessage("FTPXFORCE: X not installed, don't need to force.");
		return SUCCESS;
	}
	for (zz=0;zz < theSet->numPackages;zz++) {
		if ((strstr(theSet->packages[zz]->name,"XFree86")) ||
		    (strstr(theSet->packages[zz]->name,"XBF_"))) {
			theSet->packages[zz]->selected=1;
			theSet->packages[zz]->priority=(DEFAULT_PRIORITY + 100);
			logMessage("FTPXFORCE: Added '%s', priority %d",theSet->packages[zz]->name,theSet->packages[zz]->priority);
		}
	}
	return SUCCESS;
}
#endif
