/*
#### Copyright (c) 1998 Pacific HiTech, Inc.
####              All Rights Reserved
####              Distributable under the terms of the GPL.
#### File:        genhdlist.c
#### Author:      Scott Stone <sstone@turbolinux.com>
#### Version:     2.0TL
#### Date:        May 1, 1998
#### Description:
# Source code for TurboLinux
#### ChangeLog:
# (none)
*/

#include <alloca.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <rpm/rpmlib.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define FILENAME_TAG 1000000

int tags[] =  { RPMTAG_NAME, RPMTAG_VERSION, RPMTAG_RELEASE, RPMTAG_GROUP,
		RPMTAG_REQUIREFLAGS, RPMTAG_REQUIRENAME, RPMTAG_REQUIREVERSION,
		RPMTAG_DESCRIPTION, RPMTAG_SUMMARY, RPMTAG_PROVIDES,
		RPMTAG_SIZE, RPMTAG_COPYRIGHT, RPMTAG_ARCH, RPMTAG_SERIAL };
int numTags = sizeof(tags) / sizeof(int);

int main(int argc, char ** argv) {
    char buf[300];
    char buf0[300];
    DIR *dir=NULL;
    int outfd=0;
    struct dirent * ent;
    int fd, rc, isSource;
    Header h;
    Header newh=NULL;
    int count, type;
    int i=0;
    void * ptr;
    int debugfoo=0;

    if (argc != 3) {
	fprintf(stderr, "usage: genhdlist <dir> <file>\n");
	exit(1);
    }

    strcpy(buf0, argv[1]);
/*     strcat(buf, "/TurboLinux/RPMS"); */
    
    dir = opendir(buf0);
    if (!dir) {
	fprintf(stderr,"error opening directory %s: %s\n", buf,
		strerror(errno));
	return 1;
    }    

    strcpy(buf, argv[2]);
/*    strcat(buf, "/TurboLinux/base/hdlist"); */
    
    outfd = open(buf, O_WRONLY | O_TRUNC | O_CREAT, 0644);
    if (outfd < 0) {
	fprintf(stderr,"error creating file %s: %s\n", buf, strerror(errno));
	return 1;
    }

    chdir(buf0);
    errno = 0;
    ent = readdir(dir);
    if (errno) {
	perror("readdir");
	return 1;
    }

    while (ent) {
	if ((!(ent->d_name[0] == '.' && (ent->d_name[1] == '\0' || 
	    ((ent->d_name[1] == '.') && (ent->d_name[2] == '\0'))))) &&
	    (strstr(ent->d_name,".rpm"))) {
		if (strstr(ent->d_name,"glibc")) {
			debugfoo=1;
		}
	    fd = open(ent->d_name, O_RDONLY);

	    fprintf(stderr,"Reading rpm package: %s\n",ent->d_name);
	    if (fd < 0) {
		perror("open");
		exit(1);
	    }

	    rc = rpmReadPackageHeader(fd, &h, &isSource, NULL, NULL);

	    close(fd);
	    if (rc) {
		fprintf(stderr, "failed to rpmReadPackageHeader %s\n", ent->d_name);
		exit(1);
	    }

	    /* Regular bit */
	    newh = headerNew();
	    for (i = 0; i < numTags; i++) {
		if (!headerGetEntry(h, tags[i], &type, &ptr, &count)) {
		    switch (tags[i]) {
		    case RPMTAG_REQUIREFLAGS:
		    case RPMTAG_REQUIRENAME:
		    case RPMTAG_REQUIREVERSION:
		    case RPMTAG_SUMMARY:
		    case RPMTAG_PROVIDES:
		    case RPMTAG_COPYRIGHT:
		    case RPMTAG_SERIAL:
			break;
		      default:
			fprintf(stderr, "tag %d not found in file %s\n", 
				tags[i], ent->d_name);
			exit(1);
		    }
		} else {
			if ((type == 6) && (count != 1)) {
				printf("oops? count was '%d'\n",count);
				count=1;
			}
			headerAddEntry(newh, tags[i], type, ptr, count);
		}		       
	    }
	    headerAddEntry(newh, FILENAME_TAG, RPM_STRING_TYPE, ent->d_name, 1);
	    headerWrite(outfd, newh, HEADER_MAGIC_YES);
	    headerFree(newh);

	    headerFree(h);
	}

	errno = 0;
	ent = readdir(dir);
	if (errno) {
	    perror("readdir");
	    return 1;
	}
    } 

    closedir(dir);
    close(outfd);

    return 0;
}
