#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/time.h>
#include "defines.h"
#include "log.h"
#include "prototypes.h"
#include "autowin.h"

extern int debug;
extern int doSuggest;
extern int ignoreConflicts;
extern int autoDeps;
extern int autoDepsModified;
extern int groupSelectType;
extern int returnReason;
extern int ignoreLocal;
extern int installer;
extern int dodowngrades;
extern int dynamicsize;
extern int progmountedfloppy;
extern int autocontrol;
extern int enforcereqs;
extern int controlsocket;
extern fd_set in_set;
extern fd_set out_set;
extern DESCRIPTOR_DATA *mydesc;

extern int nohdlist;
extern int norpmdir;
extern int nocompsfile;

extern int completeinstallindex;
extern int custominstallindex;
extern int filelistinstallindex;

#define NEWT_KEY_CONTROL_S NEWT_KEY_F4

char **removemarked=NULL;

extern helptextTable *itypesTable;
char *customHelpText=NULL;
char *customSymbolKey1=NULL;
char *customSymbolKey2=NULL;
char *customHotKeys=NULL;
pkgSet *savedSelections=NULL;


void defineHelpText(void) {
	customHelpText = strdup(
		_("The TurboLinux packages are categorized into several package groups, listed on the package selection screen.  Underneath these category names are the names of the individual packages that belong to the category.  Selecting an entire category will select all of the packages inside it.  However, you may also select individual packages without selecting every other package in that group.  By default, all the groups and packages are marked as selected, so you may simply DE-select the ones that you do NOT want, and then press the OK button to continue with the installation."));	
	customSymbolKey1 = strdup(
		_("Key to Symbols:\n"
		  "        Group Checkboxes:\n"
		  "                '*' = Selected\n"
		  "                '-' = Partially Selected\n"
		  "                ' ' = Not Selected\n"
		  "        Package Checkboxes:\n"
		  "                '*' = Selected for Install/Upgrade\n"
		  "                'R' = Selected for Removal\n"
		  "                ' ' = Not Selected\n"
		  "                'X' = Package Missing\n"
			)
		);
	customSymbolKey2 = strdup(
		_("Key to Symbols (continued):\n"
		  "        Right-hand Column: (not for initial install)\n"
		  "                '+' = Package on media is newer than current.\n"
		  "                '-' = Package on media is older than current.\n"
		  "                '=' = Package on media is the same as current.\n"
		  "                ' ' = Package is not currently installed.\n"
		  "                'R' = Package is scheduled for removal.\n"
			)
		);
	customHotKeys = strdup(
		_("Hot Keys Available:\n"
		  "        'A' = Select ALL Packages\n"
		  "        'C' = Clear Selection\n"
		  "        'N' = Select All Upgrades and New Packages\n"
		  "        'R' = Mark Package for Removal\n"
		  "        'U' = Select All Upgrades (no new packages)\n"
		  )
		);
}


char **resolveConflicts(char **slist, char **clist) {
	char **tmp=(char **)malloc(MAX_PKGS * sizeof(char *));
	char **ptr=NULL;
	int tmpindex=0;
	static int zz;

	for (zz=0;zz < MAX_PACKAGES;zz++) {
		tmp[zz]=NULL;
	}
	ptr=slist;
	while (*ptr) {
		if (!memberofStrArray(clist,*ptr)) {
			tmp[tmpindex]=strdup(*ptr);
			tmpindex++;
		}
		ptr++;
	}	
	return tmp;
}

int getCurrentLine(newtComponent theListbox) {
	int rc=-1;
	int xx=0;
	char *theData=NULL;
	char *stmp=NULL;
	char *stmp2=NULL;
	void *vtmp=NULL;

	theData=(char *)newtListboxGetCurrent(theListbox);
	while(rc == -1) {
		newtListboxGetEntry(theListbox,xx,&stmp,&vtmp);
		stmp2=(char *)vtmp;
		if (!strcmp(theData,stmp2)) {
			rc=xx;
		}
		xx++;
	}
	return rc;
}

char **getSelectedPackageNames(installType *itype, tlComponentGroup *masterlist) {
	char **sptr=(char **)malloc(MAX_PKGS * sizeof(char *));
	char **sptr2=NULL;
	char **sptr3=NULL;
	char **clist=(char **)malloc(MAX_PKGS * sizeof(char *));
	char **rlist=NULL;
	tlComponentGroup *gptr=NULL;
	tlComponentGroup *gptr2=NULL;
	int spindex=0;
	int clindex=0;
	static int zz;
	static int yy;

	if (!itype || !masterlist) return NULL;
	for (zz=0;zz < MAX_PKGS;zz++) {
		sptr[zz]=NULL;
		clist[zz]=NULL;
	}
	sptr2=itype->compList;
	while (*sptr2) {
		gptr=getGroupByName(masterlist,*sptr2);
		logMessage("Processing Group: '%s'",*sptr2);

		for (yy=0;yy < gptr->packageSet.numPackages;yy++) {
			if (!memberofStrArray(sptr,gptr->packageSet.packages[yy]->name)) {
				sptr[spindex]=strdup(gptr->packageSet.packages[yy]->name);
				spindex++;
			}
		}
		sptr3=gptr->conflicts;
		while (*sptr3) {
			gptr2=getGroupByName(masterlist,*sptr3);
			for (yy=0;yy < gptr2->packageSet.numPackages;yy++) {
				if (!memberofStrArray(clist,gptr2->packageSet.packages[yy]->name)) {
					clist[clindex]=strdup(gptr2->packageSet.packages[yy]->name);
					clindex++;
				}
			}
			sptr3++;
		}
		sptr3=gptr->requires;
		while (*sptr3) {
			gptr2=getGroupByName(masterlist,*sptr3);
			for (yy=0;yy < gptr2->packageSet.numPackages;yy++) {
				if (!memberofStrArray(sptr,gptr2->packageSet.packages[yy]->name)) {
					sptr[spindex]=strdup(gptr2->packageSet.packages[yy]->name);
					spindex++;
				}
			}
			sptr3++;
		}
		if (doSuggest) {
			sptr3=gptr->suggests;
			while (*sptr3) {
				gptr2=getGroupByName(masterlist,*sptr3);
				for (yy=0;yy < gptr2->packageSet.numPackages;yy++) {
					if (!memberofStrArray(sptr,gptr2->packageSet.packages[yy]->name)) {
						sptr[spindex]=strdup(gptr2->packageSet.packages[yy]->name);
						spindex++;
					}
				}
				sptr3++;
			}
		}
		sptr2++;
	}
	if (!ignoreConflicts) {
		rlist=resolveConflicts(sptr,clist);
		for (zz=0;zz < MAX_PACKAGES;zz++) {
			free(sptr[zz]);
			free(clist[zz]);
		}
		free(sptr);
		free(clist);
		return rlist;
	}
	return sptr;
}

struct packageInfo *getPackageByName(pkgSet *pkgList, char *pkgname) {
	int pindex=0;
	struct packageInfo *pptr=pkgList->packages[0];
	
	while ((pptr) && (pindex < pkgList->numPackages)) {
		if (!pptr->name) continue;
		if (!strcmp(pptr->name,pkgname)) return pptr;
		pindex++;
		pptr=pkgList->packages[pindex];
	}
	return NULL;
}

int comparePackageInfo(void *xp1, void *xp2) {
	/*
	struct packageInfo *p1=(struct packageInfo *)xp1;
	struct packageInfo *p2=(struct packageInfo *)xp2;
	*/
	struct packageInfo **p1=xp1;
	struct packageInfo **p2=xp2;

	if ((*p1)->priority < (*p2)->priority) return -1;
	if ((*p1)->priority > (*p2)->priority) return 1;
	if ((*p1)->priority == (*p2)->priority) {
		return (strcmp((*p1)->name,(*p2)->name));
	}
	return 0;
}

void logPackageList(pkgSet *theList) {
	static int zz;
	int selcount=0;
	
	logMessage("Package List size is %d packages total:",theList->numPackages);
	for (zz=0;zz < theList->numPackages;zz++) {
		if (theList->packages[zz]->selected) {
			logMessage("\t%s (%d)",theList->packages[zz]->name,theList->packages[zz]->priority);
			selcount++;
		}
	}
	logMessage("%d selected packages.",selcount);
}

pkgSet *prepareNormal(installType *itype, tlComponentGroup *masterlist, pkgSet *selectedlist, pkgSet *mpkglist, char **forcelist, Header *installedpackages, char **ipkglist) {
	char **plist=NULL;
	pkgSet *tmp=(pkgSet *)malloc(sizeof(pkgSet));
	pkgSet *tmp2=NULL;
	static int zz;
	int rc=0;
	bool choice;
	int xx=0;
	int yy=0;
	int xa=0;
	char s0[256];
	char s1[256];

	statusWindow(_("Preparing"),_("Preparing Installation..."));
	logMessage("Preparing Installation (normal): '%s'",itype->name);
	if (!masterlist) return NULL;
	plist=getSelectedPackageNames(itype,masterlist);
	if (!plist) return NULL;
	tmp->numPackages=0;
	tmp->packages=(struct packageInfo **)malloc(MAX_PACKAGES * sizeof(struct packageInfo *));
	for (zz=0;zz < mpkglist->numPackages;zz++) {
		tmp->packages[zz]=(struct packageInfo *)malloc(sizeof(struct packageInfo));
		tmp->packages[zz]->name=NULL;
		tmp->packages[zz]->name=strdup(mpkglist->packages[zz]->name);
		tmp->packages[zz]->group=NULL;
		tmp->packages[zz]->priority=mpkglist->packages[zz]->priority;
		tmp->packages[zz]->size=mpkglist->packages[zz]->size;
		tmp->packages[zz]->selected=0;
		tmp->packages[zz]->data=NULL;
		if (memberofStrArray(plist,tmp->packages[zz]->name)) {
			if (tmp->packages[zz]->priority > 0) {
				tmp->packages[zz]->selected=1;
			} else {
				logMessage("PRI0: '%s'",tmp->packages[zz]->name);
			}
		}
		if (memberofStrArray(forcelist,tmp->packages[zz]->name)) {
			tmp->packages[zz]->selected=1;
			if (tmp->packages[zz]->priority == 0) {
				tmp->packages[zz]->priority = DEFAULT_PRIORITY;
			}
		}
		tmp->numPackages++;
	}
	if (debug) {
		logMessage("PRESORT:");
		logPackageList(tmp);
	}
	
	rc=getPackageHeaders2(NULL,mpkglist,tmp);
	if (rc == FAILURE) return NULL;
	if (debug) {
		logMessage("POSTSORT:");
		logPackageList(tmp);
	}
	newtPopWindow();
	rc=verifyDependencies(tmp,installedpackages,removemarked);	
	if (rc != SUCCESS) {
		if (installer) {
			returnReason=1;
		} else {
			returnReason=2;
			savedSelections=tmp;
		}
		return NULL;
	}
	qsort(tmp->packages,tmp->numPackages,sizeof(*tmp->packages),(void *)comparePackageInfo);
	choice=FALSE;
	while (choice == FALSE) {
		statusWindow(_("Analyzing"),_("Analyzing Package Sizes and Disk Space..."));
		xx=getPkgSetSize2(tmp,installedpackages);
		yy=checkFreeSpace((xx * 1024),&xa);
		sprintf(s0,_("megabytes (WARNING - this installation requires %d MB, but you only have %d MB available!"),xx,(xa / 1024));
		sprintf(s1,_("MB (%d MB free)"),(xa/1024));
		newtPopWindow();	
		choice=yesnoWindow(_("Confirm Selection"),TRUE,_("Proceed"),_("Customize"),_("The total size of the selected install (after dependencies have been verified) is %d %s.  Would you like to proceed or customize this installation?"),xx,(yy == 1) ? s1 : s0);
		if (choice == FALSE) {
			tmp2=prepareCustom(itype,masterlist,tmp,mpkglist,forcelist,installedpackages,ipkglist);
			if (!tmp2) return NULL;
			/* free(tmp); */
			/*
			logMessage("MALLOC: Freeing old package set.");
			freePkgSet(&tmp);
			*/
			tmp=tmp2;
			tmp2=NULL;
		}
	}
	return tmp;
}

int getPkgSetSize(pkgSet *theSet) {
	int tmp=0;
	static int zz;

	logMessage("PKGSETSIZE: Running getPkgSetSize (basic)");
	for (zz=0;zz < theSet->numPackages;zz++) {
		if (theSet->packages[zz]->selected) {
			if (removemarked) {
				if (!memberofStrArray(removemarked,theSet->packages[zz]->name)) {
					tmp += theSet->packages[zz]->size;
				}
			} else {
				tmp += theSet->packages[zz]->size;
			}
		}
	}
	tmp=(tmp / (1024*1024));
	return tmp;
}

int getPkgSetSizeInBytes(pkgSet *theSet) {
	int tmp=0;
	static int zz;

	logMessage("PKGSETSIZE: Running getPkgSetSizeInBytes (bytes)");
	for (zz=0;zz < theSet->numPackages;zz++) {
		if (theSet->packages[zz]->selected) {
			if (removemarked) {
				if (!memberofStrArray(removemarked,theSet->packages[zz]->name)) {
					tmp += theSet->packages[zz]->size;
				}
			} else {
				tmp += theSet->packages[zz]->size;
			}
		}
	}
	return tmp;
}

/* returns size entry if present, or 0 if not */
unsigned int memberofSizeCache(headerSizeCacheEntry *cache, char *pkgname) {
	static int zz;

	if (!cache || !pkgname) return 0;
	for (zz=0;zz < MAX_PACKAGES;zz++) {
		if (!strcmp(cache[zz].name,pkgname)) {
			return cache[zz].size;
		}
	}
	return 0;
}

int getPkgSetSize2(pkgSet *theSet, Header *ihlist) {
	int tmp=0;
	static int zz;
	char xc;
	int xtype,xC;
	int hindex=0;
	Header htmp;
	char *tmpname=NULL;
	int x=0;
	int y=0;
	unsigned int *sizeptr;
	int newsize=0;
	int upgsize=0;
	headerSizeCacheEntry *sizecache=NULL;
	int cacheindex=0;


	logMessage("PKGSETSIZE: Running getPkgSetSize2 (advanced)");
	if ((installer) || (!ihlist)) {
		return getPkgSetSize(theSet);
	}
	if (!sizecache) {
		sizecache=(headerSizeCacheEntry *)
			  malloc(MAX_PACKAGES * sizeof(headerSizeCacheEntry));
		for (zz=0;zz < MAX_PACKAGES;zz++) {
			sizecache[zz].name=NULL;
			sizecache[zz].size=0;
		}
		logMessage("PKGSETSIZE: Building header size cache...");
		htmp=ihlist[hindex];
		while(htmp) {
			headerGetEntry(htmp,RPMTAG_NAME,&xtype,(void **)&tmpname,&xC);
			headerGetEntry(htmp,RPMTAG_SIZE,&xtype,(void **)&sizeptr,&xC);
			sizecache[cacheindex].name=strdup(tmpname);
			sizecache[cacheindex].size=*sizeptr;
			cacheindex++;
			hindex++;
			htmp=ihlist[hindex];
			/*
			free(tmpname);
			free(sizeptr);
			*/
		}
		logMessage("PKGSETSIZE: Added %d entries to size cache.",cacheindex);
	}
	for (zz=0;zz < theSet->numPackages;zz++) {
		if (theSet->packages[zz]->selected) {
			xc=isPackageInstalled(ihlist,NULL,theSet,theSet->packages[zz],removemarked ? removemarked : NULL);
			switch(xc) {
			case ' ':
				tmp += theSet->packages[zz]->size;
				newsize += theSet->packages[zz]->size;
				break;
			case '=':
			case 'R':
				break;
			case '-':
			case '+':
				if ((xc == '-') && (!dodowngrades)) break;
				logMessage("PKGSETSIZE-UPG: Upgrade: '%s'",theSet->packages[zz]->name);
				*sizeptr=memberofSizeCache(sizecache,theSet->packages[zz]->name);
				x=theSet->packages[zz]->size;
				y=(x - *sizeptr);
				tmp += y;
				logMessage("PKGSETSIZE-UPG: mod. size by '%d' (%d, %d)",y,x,*sizeptr);
				upgsize += y;
				break;
			default:
				break;
			}
		}
	}
	logMessage("MALLOC: Freeing sizecache...");
	for (zz=0;zz < MAX_PACKAGES;zz++) {
		free(sizecache[zz].name);
	}
	free(sizecache);
	logMessage("MALLOC: Free complete.");
	tmp=(tmp / (1024*1024));
	newsize=(newsize / (1024*1024));
	upgsize=(upgsize / (1024*1024));
	logMessage("PKGSETSIZE: Total required: %d MB - New: %d MB, Upg: %d MB.",tmp,newsize,upgsize);
	return tmp;
}
				
				
	
char togglePackage(pkgSet *theSet, char *pkgName) {
	static int zz;
	int pkgselected=0;

	if (norpmdir) {
		errWindow(FALSE,_("You did not specify a path to any RPM packages - you cannot add or upgrade any packages."));
		return ' ';
	}
	for (zz=0;zz < theSet->numPackages;zz++) {
		if (!strcmp(theSet->packages[zz]->name,pkgName)) {
			if (theSet->packages[zz]->size == 0) {
				return 'X';
			}
			theSet->packages[zz]->selected=(abs(1-(theSet->packages[zz]->selected)));
			pkgselected=theSet->packages[zz]->selected;
		}
	}
	if (removemarked) {
		removemarked=removeFromStrArray(removemarked,pkgName,MAX_PACKAGES);
	}
	return (pkgselected ? '*' : ' ');
}

char toggleRemovePackage(pkgSet *theSet, char *pkgName, Header *installedpackages) {
	int rflag=0;
	struct packageInfo *pptr=NULL;
	int pkgselected;
	char ctest;
	char s0[256];

	pptr=getPackageByName(theSet,pkgName);
	if (pptr->size == 0) return 'X';
	ctest=isPackageInstalled(installedpackages,NULL,theSet,pptr,NULL);
	pkgselected=pptr->selected;
	if (ctest == ' ') {
		if (ignoreLocal) {
			sprintf(s0,_("Cannot perform remove functions - no local package info was loaded."));
		} else {
			sprintf(s0,_("Cannot remove '%s' - it is not installed."),pptr->name);
		}
		errWindow(FALSE,s0);
		return (pkgselected ? '*' : ' ');
	}
	if (memberofStrArray(removemarked,pkgName)) {
		removemarked=removeFromStrArray(removemarked,pkgName,MAX_PACKAGES);
		rflag=0;
	} else {
		addToStrArray(removemarked,pkgName);
		rflag=1;
	}
	if (rflag == 1) {
		return 'R';
	} else {
		return (pkgselected ? '*' : ' ');
	}
	return ' ';
}

int getNextGroupLine(int *glines, int groupnum) {
	return glines[(groupnum + 1)];
}

int getPrevGroupLine(int *glines, int linenumber) {
	int zz=0;

	while (1) {
		if ((zz > 0) && (!glines[zz])) {
			return glines[zz-1];
		} else {
			if (glines[zz] > linenumber) {
				return glines[zz-1];
			}
			zz++;
		}
	}
	return 0;
}

int listboxIndexOfGroup(newtComponent listbox,char *groupname) {
	int zz=0;
	int found=0;
	char *s1=NULL;
	void *v1=NULL;

	while (!found) {
		newtListboxGetEntry(listbox,zz,&s1,&v1);
		if (strstr(s1,groupname)) return zz;
		/*
		free(s1);
		free(v1);
		*/
		zz++;
	}
	return 0;
}
		

char toggleGroup(char isselected, pkgSet *theSet, tlComponentGroup *masterlist, char *grpName) {
	tlComponentGroup *gptr=NULL;
	struct packageInfo *pptr=NULL;
	static int zz;
	int sv;

	if (norpmdir) {
		errWindow(FALSE,_("You did not specify a path to any RPM packages - you cannot add or upgrade any packages."));
		return ' ';
	}
	switch(isselected) {
	case ' ':
		sv=0;
		break;
	default:
		sv=1;
		break;
	}

	sv=abs(1-sv);
	gptr=getGroupByName(masterlist,grpName);
	for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
		pptr=getPackageByName(theSet,gptr->packageSet.packages[zz]->name);
		if (pptr) {
			pptr->selected=sv;
		}
	}
	return (sv ? '*' : ' ');
}

char getGroupStatus(tlComponentGroup *masterlist, char *grpName, pkgSet *p) {
	tlComponentGroup *gptr=NULL;
	struct packageInfo *pptr=NULL;
	char rc='*';
	static int zz;
	int numSelected=0;

	if (!masterlist || !grpName || !p) return ' ';
	gptr=getGroupByName(masterlist,grpName);
	for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
		pptr=getPackageByName(p,gptr->packageSet.packages[zz]->name);
		if (pptr) {
			if ((!pptr->selected) || ((removemarked) && memberofStrArray(removemarked,pptr->name))) {
				rc='-';
			} else {
				numSelected++;
			}
		} else {
			rc='-';
		}
	}
	if (numSelected == 0) {
		rc=' ';
	}
	if (debug) {
		logMessage("GROUPSTATUS: '%s' has %d selected, code '%c'",grpName,numSelected,rc);
	}
	return rc;
}


char isPackageSelected(pkgSet *theSet, char *pkgname) {
	static int zz;

	for (zz=0;zz < theSet->numPackages;zz++) {
		if (!strcmp(theSet->packages[zz]->name,pkgname)) {
			if (debug) {
				logMessage("SELECTED: '%s' == '%d'",pkgname,theSet->packages[zz]->selected);
			}
			if (theSet->packages[zz]->size == 0) {
				return 'X';
			}
			return (theSet->packages[zz]->selected) ? '*' : ' ';
		}
	}
	return ' ';
}

pkgSet *prepareCustomByRPMGroup(installType *itype, tlComponentGroup **masterlist, pkgSet *selectedlist, pkgSet *mpkglist, char **forcelist, Header *installedpackages, char **ipkglist) {
/*	freeComponentGroup(masterlist); */
	*masterlist=getRPMGroups(mpkglist);
	
	return mpkglist;
}



pkgSet *prepareCustom(installType *itype, tlComponentGroup *masterlist, pkgSet *selectedlist, pkgSet *mpkglist, char **forcelist, Header *installedpackages, char **ipkglist) {
	pkgSet *tmp=(pkgSet *)malloc(sizeof(pkgSet));
	static int zz;
	int rc=0;
	newtComponent cform,clistbox,ctextbox,csizelabel,cokbutton,coptionsbutton,ccancelbutton;
	int myrows, mycols;
	int extrarows=0;
	int compindex=0;
	int gindex=0;
	int lineindex=0;
	/*
	int numgroups=0;
	char **sptr=NULL;
	*/
	char groupValues[MAX_GROUPS];
	tlComponentGroup *gptr=NULL;
	char s0[256];
	char s1[256];
	char s1a[256];
	char *s2=NULL;
	char *s3=NULL;
	char *s4=NULL;
	char *s5=NULL;
	char *s6=NULL;
	char *s7=NULL;
	char *s8=NULL;
	void *v1=NULL;
	void *v2=NULL;
	selectionList list=NULL;
	struct newtExitStruct cevent;
	int cformdone=0;
	int xx;
	int yy;
	int aa;
	char ctemp;
	int *groupstartlines=(int *)malloc(MAX_GROUPS * sizeof(int));
	int gsi=0;
	float pkgsize;
	int thisgroupon=0;
	int displaypackages=1;
	int oldlineindex=-1;
	char *language=NULL;
	int jflag=0;
	struct packageInfo *pptr=NULL;
	bool choice;
	int redrawsize=0;
	char **customHelpList=NULL;
	char *curdata=NULL;
	char *curgroup=NULL;
	char *prevSearch=NULL;
	int searchFound=0;
	int previousLine;
	int tmpMode=-1;
#ifdef AUTOCONTROL
	char acstr[4096];
#endif

	ctextbox=NULL;
	csizelabel=NULL;
	if (autoDeps) {
		logMessage("CUSTOM: Turning off autoDeps.");		
		autoDeps=0;
		autoDepsModified=1;
	}
	defineHelpText();
	removemarked=(char **)malloc(MAX_PACKAGES * sizeof(char *));
	if (norpmdir) {
		dynamicsize=0;
	}
	for (zz=0;zz < MAX_PACKAGES;zz++) {
		removemarked[zz]=NULL;
	}
	language=getenv("LANG");
	if ((language) && (strstr(language,"ja_JP"))) {
		jflag=1;
	}
	if (groupSelectType == 1) {
		if (!nocompsfile) { /* if it's equal to 1, this is already done. */
			if ((prepareCustomByRPMGroup(itype,&masterlist,selectedlist,mpkglist,forcelist,installedpackages,ipkglist)) == NULL) return NULL;
		}
	}
	logMessage("Preparing Installation (custom)");
	if (savedSelections) {
		if (selectedlist) {
			choice=yesnoWindow(_("Previous Selections"),
					   TRUE,
					   _("Previous Selections"),
					   _("Defaults"),
					   _("You have chosen to customize a pre-existing install type.  However, you also have saved selections from a previous customization attempt.  Which do you wish to use?"));
			if (choice == TRUE) {
				logMessage("Using Saved Selections...");
				tmp=savedSelections;
			}
		} else {
			logMessage("Using Saved Selections...");
			tmp=savedSelections;
		}
	} else {
		if (!selectedlist) {
			tmp->numPackages=0;
			tmp->packages=(struct packageInfo **)malloc(mpkglist->numPackages * sizeof(struct packageInfo *));
			for (zz=0;zz<mpkglist->numPackages;zz++) {
				tmp->packages[zz]=(struct packageInfo *)malloc(sizeof(struct packageInfo));
				tmp->packages[zz]->name=NULL;
				tmp->packages[zz]->name=strdup(mpkglist->packages[zz]->name);
				tmp->packages[zz]->priority=mpkglist->packages[zz]->priority;
				tmp->packages[zz]->size=mpkglist->packages[zz]->size;
/* 			tmp->packages[zz]->selected=norpmdir ? 0 : 1; */
				tmp->packages[zz]->selected=installer ? 1 : 0;
				if (tmp->packages[zz]->priority == 0) {
					tmp->packages[zz]->selected=0;
				}
				tmp->numPackages++;
			}
		} else {
			tmp=selectedlist;
			for (zz=0;zz < tmp->numPackages;zz++) {
				if (memberofStrArray(forcelist,tmp->packages[zz]->name)) {
					tmp->packages[zz]->selected=1;
					if (tmp->packages[zz]->priority == 0) {
						tmp->packages[zz]->priority=DEFAULT_PRIORITY;
					}
				}
			}
		}
	}
	rc=getPackageHeaders2(NULL,mpkglist,tmp);
	if (rc == FAILURE) return NULL;
	/*
	sptr=itype->compList;
	if (sptr) {
		while (*sptr) {
			numgroups++;
			sptr++;
		}
	}
	*/
	compindex=0;
	if (!autocontrol) {
		newtGetScreenSize(&mycols,&myrows);
		extrarows=(myrows - 25);
	}
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatus(controlsocket,"analyzing",_("Analyzing Package Information..."));
#endif
	} else {
		if (!installer && !ignoreLocal) {
			statusWindow(_("Analyzing System"),_("Checking Installed Versions..."));
		} else {
			statusWindow(_("Analyzing System"),_("Generating Package List..."));
		}
	}
	if (autocontrol) {
#ifdef AUTOCONTROL
		sprintf(s0,"TOTALSIZE:%d:MB",getPkgSetSize2(tmp,installedpackages));
		autoMessage(controlsocket,s0);
#endif
	}
	sprintf(s0,_("Installation Size: %d Pkgs (%d MB)"),
		getNumPackages(tmp,removemarked),
		getPkgSetSize2(tmp,installedpackages));
	if (!autocontrol) {
		csizelabel=newtLabel(1,1,s0);
		ctextbox=properTextbox(_("Select the component groups/individual packages which you would like to install.  ENTER toggles a package or group's status.  F1 for help.  F2 to expand/collapse the tree.  F3 for information on the selected package.  Press F4 to search."));
	}
	gptr=masterlist->next;
	while(gptr) {
		/*
		if (memberofStrArray(itype->compList,gptr->name)) {
			groupValues[compindex]='*';
			thisgroupon=1;
		} else {
			groupValues[compindex]=' ';
			thisgroupon=0;
		}
		*/
#ifdef AUTOCONTROL
		if (autocontrol) {
			if ( FD_ISSET( mydesc->descriptor, &in_set ) )
			{
				if ( !read_from_descriptor( mydesc ) )
				{
					FD_CLR( mydesc->descriptor, &out_set );
					mydesc->outtop	= 0;
					close_socket( mydesc );
					continue;
				}
			}
/* input stuff is here */
			read_from_buffer( mydesc );
			if ( mydesc->incomm[0] != '\0' )
			{
				if (strstr(mydesc->incomm,"@CANCEL")) {
					autoMessage(mydesc->descriptor,"CANCELLED:");
					mydesc->incomm[0]='\0';
					return NULL;
				}
				mydesc->incomm[0]='\0';
			}
		}
#endif
		groupValues[compindex]=getGroupStatus(masterlist,gptr->name,tmp);
		if (groupValues[compindex] == '*') {
			thisgroupon=1;
		} else {
			thisgroupon=0;
		}
		sprintf(s0,"G:%d:%d:%s",lineindex,compindex,gptr->name);
		groupstartlines[gsi]=lineindex;
		gsi++;
		sprintf(s1,"[%c] %s",groupValues[compindex],gptr->name);
		compindex++;
		if (autocontrol) {
#ifdef AUTOCONTROL
			sprintf(acstr,"GROUP:%s%s",s1,s0);
			autoMessage(controlsocket,acstr);
#endif
		}
		addToSelectionList(&list,strdup(s1),strdup(s0),NULL);
		lineindex++;
		for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
			pkgsize=(float)((gptr->packageSet.packages[zz]->size * 1.0) / (1024));
			sprintf(s1a,"(%.2f kB)",pkgsize);
			sprintf(s1,"        [%c] %-30s %-15s %c",isPackageSelected(tmp,gptr->packageSet.packages[zz]->name),gptr->packageSet.packages[zz]->name,s1a,isPackageInstalled(installedpackages,ipkglist,mpkglist,gptr->packageSet.packages[zz],removemarked));
			sprintf(s0,"P:%d:%.2f:%s",lineindex,pkgsize,gptr->packageSet.packages[zz]->name);
			if (autocontrol) {
#ifdef AUTOCONTROL
				sprintf(acstr,"PACKAGE:%s%s",s1,s0);
				autoMessage(controlsocket,acstr);
#endif
			}
			addToSelectionList(&list,strdup(s1),strdup(s0),NULL);
			lineindex++;
		}
		gptr=gptr->next;
	}
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatusEnd(controlsocket);
#endif
	} else {
		if (!installer && !ignoreLocal) {
			newtPopWindow();
		} else {
			newtPopWindow();
		}
	}
#ifdef AUTOCONTROL
	if (autocontrol) {
		autoMessage(controlsocket,"ENDINITIALDATA:");
		return tmp;
	}
#endif
	groupstartlines[gsi]=lineindex;
	cokbutton=newtButton(0,0,_("Ok"));
	coptionsbutton=newtButton(0,0,_("Options..."));
	ccancelbutton=newtButton(0,0,_("Cancel"));

	cform=askUser(layout,_("Custom Package Selection"),NULL,
		      compGroup(vertical,center,
				groupMargine(0,1,
					     compGroup(vertical,topLeft,
						       csizelabel,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,center,
						       ctextbox,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,center,
						       selectionListbox(&clistbox,FALSE,6+extrarows,0,list),
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(horizontal,bottomRight,
						       cokbutton,
						       coptionsbutton,
						       ccancelbutton,
						       NULL)
					     ),
				NULL)
		      );
	newtFormAddHotKey(cform,NEWT_KEY_F1);
	newtFormAddHotKey(cform,NEWT_KEY_F2);
	newtFormAddHotKey(cform,NEWT_KEY_F3);
	newtFormAddHotKey(cform,NEWT_KEY_F4);
	newtFormAddHotKey(cform,NEWT_KEY_CONTROL_S);
	newtFormAddHotKey(cform,' ');
	if (!installer) {
		newtFormAddHotKey(cform,'r');
		newtFormAddHotKey(cform,'R');
		newtFormAddHotKey(cform,'u');
		newtFormAddHotKey(cform,'U');
		newtFormAddHotKey(cform,'n');
		newtFormAddHotKey(cform,'N');
	}
	newtFormAddHotKey(cform,'a');
	newtFormAddHotKey(cform,'A');
	newtFormAddHotKey(cform,'c');
	newtFormAddHotKey(cform,'C');
	
	while (!cformdone) {
		redrawsize=0;
		newtFormRun(cform,&cevent);
		if (cevent.reason == NEWT_EXIT_HOTKEY) {
			if (cevent.u.key == ' ') {
				cevent.u.co = clistbox;
				goto listboxselected;
			}
			if (cevent.u.key == NEWT_KEY_F1) {
				customHelpList=(char **)malloc(4 * sizeof(char *));
				customHelpList[0]=strdup(customHelpText);
				customHelpList[1]=strdup(customSymbolKey1);
				customHelpList[2]=strdup(customSymbolKey2);
				customHelpList[3]=strdup(customHotKeys);
				helpWindow(_("Custom Install"),4,customHelpList);
				free(customHelpList[0]);
				free(customHelpList[1]);
				free(customHelpList[2]);
				free(customHelpList[3]);
				free(customHelpList);
				customHelpList=NULL;
			} else if (cevent.u.key == NEWT_KEY_F3) {
				s3=newtListboxGetCurrent(clistbox);
				if (s3[0] == 'G') {
					msgWindow(_("Package Information"),_("Ok"),_("Please select an individual RPM package before pressing F3.  %s"),displaypackages ? "" : _("You need to press F2 to display the individual packages."));
				} else {
					displayRPMInfo(getRPMInfo(mpkglist,s3),mpkglist,s3);
				}
			} else if ((cevent.u.key == NEWT_KEY_F4) ||
				   (cevent.u.key == NEWT_KEY_CONTROL_S)) {
				s3=entryWindow(_("Search String"),40,prevSearch ? prevSearch : "",_("Search"),_("Cancel"),_("Enter Search String (case insensitive):"));
				if (s3) {
					logMessage("SEARCH: key is '%s'",s3);
					
					v1=newtListboxGetCurrent(clistbox);
					yy=0;
					for (zz=0;zz < lineindex;zz++) {
						newtListboxGetEntry(clistbox,zz,&s4,&v2);
						if (!strcmp((char *)v1,(char *)v2)) {
							yy=zz;
							logMessage("SEARCH: current line is '%d'",yy);
						}
					}
					searchFound=0;
					if ((prevSearch) && (!strcmp(s3,prevSearch))) {
						yy++;
					}
					free(prevSearch);
					prevSearch=strdup(s3);
topofsearch:
					xx=0;
					for (zz=yy;zz < lineindex;zz++) {
						newtListboxGetEntry(clistbox,zz,&s4,&v2);
						if (debug) {
							logMessage("SEARCH-DEBUG: Comparing '%s' to '%s'",s3,s4);
						}
						if (strcasestr(s4,s3)) {
							xx=zz;
							logMessage("SEARCH: Match found at line '%d'",xx);
							searchFound=1;
						}
						if (xx > 0) break;
					}
					if (!searchFound) {
						if (yesnoWindow(_("Not Found"),TRUE,_("Ok"),_("Cancel"),_("Search string not found - search from top?"))) {
							yy=0;
							goto topofsearch;
						}
					} else {
						newtListboxSetCurrent(clistbox,xx);
					}
				} else {
					logMessage("SEARCH: cancelled by user.");
				}
			} else if (((cevent.u.key == 'r') || (cevent.u.key == 'R')) && (!installer)) {
				if (dynamicsize) {
					redrawsize=1;
				}
				s2=strdup((char *)newtListboxGetCurrent(clistbox));
				if (s2[0] == 'P') {
					strtok(s2,":");
					xx=atoi(strtok(NULL,":"));
					pkgsize=atof(strtok(NULL,":"));
					s3=strtok(NULL,":");
					ctemp=toggleRemovePackage(tmp,s3,installedpackages);
					sprintf(s1a,"(%.2f kB)",pkgsize);
					sprintf(s1,"        [%c] %-30s %-15s %c",ctemp,s3,s1a,isPackageInstalled(installedpackages,ipkglist,mpkglist,getPackageByName(mpkglist,s3),removemarked));
					newtListboxSetText(clistbox,xx,s1);
					yy=getPrevGroupLine(groupstartlines,xx);
					newtListboxGetEntry(clistbox,yy,&s3,(void **)&s4);
					s5=rindex(s4,':');
					s5++;
					s6=strdup(s3);
					s6[1]=getGroupStatus(masterlist,s5,tmp);
					newtListboxSetText(clistbox,yy,s6);
					s7=strdup(s4);
					strtok(s7,":");
					strtok(NULL,":");
					s8=strtok(NULL,":");
					yy=atoi(s8);
					groupValues[yy]=s6[1];
				}
			} else if (((cevent.u.key == 'c') || (cevent.u.key == 'C'))) {
				if (dynamicsize) {
					redrawsize=1;
				}
				if (!installer && !ignoreLocal) {
					statusWindow(_("Analyzing System"),_("Clearing Selections..."));
				}
				for (zz=0;zz < tmp->numPackages;zz++) {
					tmp->packages[zz]->selected=0;
					if (memberofStrArray(removemarked,tmp->packages[zz]->name)) {
						removemarked=removeFromStrArray(removemarked,tmp->packages[zz]->name,MAX_PACKAGES);
					}
				}
				previousLine=getCurrentLine(clistbox);
				newtListboxClear(clistbox);
				if (displaypackages) {
					oldlineindex=lineindex;
					lineindex=0;
					compindex=0;
					gsi=0;
					gptr=masterlist->next;
					while(gptr) {
						groupValues[compindex]=getGroupStatus(masterlist,gptr->name,tmp);
						if (groupValues[compindex] == '*') {
							thisgroupon=1;
						} else {
							thisgroupon=0;
						}
						sprintf(s0,"G:%d:%d:%s",lineindex,compindex,gptr->name);
						groupstartlines[gsi]=lineindex;
						gsi++;
						sprintf(s1,"[%c] %s",groupValues[compindex],gptr->name);
						compindex++;
						newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
						lineindex++;
						for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
							pkgsize=(float)((gptr->packageSet.packages[zz]->size * 1.0) / (1024));
							sprintf(s1a,"(%.2f kB)",pkgsize);
							sprintf(s1,"        [%c] %-30s %-15s %c",isPackageSelected(tmp,gptr->packageSet.packages[zz]->name),gptr->packageSet.packages[zz]->name,s1a,isPackageInstalled(installedpackages,ipkglist,mpkglist,gptr->packageSet.packages[zz],removemarked));
							sprintf(s0,"P:%d:%.2f:%s",lineindex,pkgsize,gptr->packageSet.packages[zz]->name);
							newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
							lineindex++;
						}
						gptr=gptr->next;
					}
					if (!installer && !ignoreLocal) {
						newtPopWindow();
					}
					groupstartlines[gsi]=lineindex;
				} else {
					oldlineindex=lineindex;
					lineindex=0;
					compindex=0;
					gsi=0;
					gptr=masterlist->next;
					while(gptr) {
						groupValues[compindex]=getGroupStatus(masterlist,gptr->name,tmp);
						if (groupValues[compindex] == '*') {
							thisgroupon=1;
						} else {
							thisgroupon=0;
						}
						sprintf(s0,"G:%d:%d:%s",lineindex,compindex,gptr->name);
						groupstartlines[gsi]=lineindex;
						gsi++;
						sprintf(s1,"[%c] %s",groupValues[compindex],gptr->name);
						compindex++;
						newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
						lineindex++;
						gptr=gptr->next;
					}
					if (!installer && !ignoreLocal) {
						newtPopWindow();
					}
					groupstartlines[gsi]=lineindex;
				}
				newtListboxSetCurrent(clistbox,previousLine);
				newtRefresh();
			} else if (((cevent.u.key == 'u') ||
				    (cevent.u.key == 'U') ||
				    (cevent.u.key == 'n') ||
				    (cevent.u.key == 'N') ||
				    (cevent.u.key == 'a') ||
				    (cevent.u.key == 'A'))) {
				if (toupper(cevent.u.key) == 'U') {
					tmpMode=0;
				}
				else if (toupper(cevent.u.key) == 'N') {
					tmpMode=1;
				}
				else if (toupper(cevent.u.key) == 'A') {
					tmpMode=2;
				}
				if (dynamicsize) {
					redrawsize=1;
				}
				if (!installer && !ignoreLocal) {
					statusWindow(_("Analyzing System"),_("Selecting All Updates..."));
				}
				for (zz=0;zz < tmp->numPackages;zz++) {
					if (tmpMode == 2) {
						tmp->packages[zz]->selected=1;
					} else {
						if (isPackageInstalled(installedpackages,ipkglist,mpkglist,tmp->packages[zz],removemarked) == '+') {
							tmp->packages[zz]->selected=1;
						}
						if (tmpMode == 1) {
							if (isPackageInstalled(installedpackages,ipkglist,mpkglist,tmp->packages[zz],removemarked) == ' ') {
								tmp->packages[zz]->selected=1;
							}
						}
						if ((isPackageInstalled(installedpackages,ipkglist,mpkglist,tmp->packages[zz],removemarked) == '-') && (dodowngrades == 1)) {
							tmp->packages[zz]->selected=1;
						}
					}
					if (memberofStrArray(removemarked,tmp->packages[zz]->name)) {
						removemarked=removeFromStrArray(removemarked,tmp->packages[zz]->name,MAX_PACKAGES);
					}
				}
				previousLine=getCurrentLine(clistbox);
				newtListboxClear(clistbox);
				if (displaypackages) {
					oldlineindex=lineindex;
					lineindex=0;
					compindex=0;
					gsi=0;
					gptr=masterlist->next;
					while(gptr) {
						groupValues[compindex]=getGroupStatus(masterlist,gptr->name,tmp);
						if (groupValues[compindex] == '*') {
							thisgroupon=1;
						} else {
							thisgroupon=0;
						}
						sprintf(s0,"G:%d:%d:%s",lineindex,compindex,gptr->name);
						groupstartlines[gsi]=lineindex;
						gsi++;
						sprintf(s1,"[%c] %s",groupValues[compindex],gptr->name);
						compindex++;
						newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
						lineindex++;
						for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
							pkgsize=(float)((gptr->packageSet.packages[zz]->size * 1.0) / (1024));
							sprintf(s1a,"(%.2f kB)",pkgsize);
							sprintf(s1,"        [%c] %-30s %-15s %c",isPackageSelected(tmp,gptr->packageSet.packages[zz]->name),gptr->packageSet.packages[zz]->name,s1a,isPackageInstalled(installedpackages,ipkglist,mpkglist,gptr->packageSet.packages[zz],removemarked));
							sprintf(s0,"P:%d:%.2f:%s",lineindex,pkgsize,gptr->packageSet.packages[zz]->name);
							newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
							lineindex++;
						}
						gptr=gptr->next;
					}
					if (!installer && !ignoreLocal) {
						newtPopWindow();
					}
					groupstartlines[gsi]=lineindex;
				} else {
					oldlineindex=lineindex;
					lineindex=0;
					compindex=0;
					gsi=0;
					gptr=masterlist->next;
					while(gptr) {
						groupValues[compindex]=getGroupStatus(masterlist,gptr->name,tmp);
						if (groupValues[compindex] == '*') {
							thisgroupon=1;
						} else {
							thisgroupon=0;
						}
						sprintf(s0,"G:%d:%d:%s",lineindex,compindex,gptr->name);
						groupstartlines[gsi]=lineindex;
						gsi++;
						sprintf(s1,"[%c] %s",groupValues[compindex],gptr->name);
						compindex++;
						newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
						lineindex++;
						gptr=gptr->next;
					}
					if (!installer && !ignoreLocal) {
						newtPopWindow();
					}
					groupstartlines[gsi]=lineindex;
				}
				newtListboxSetCurrent(clistbox,previousLine);
				newtRefresh();
			} else if (cevent.u.key == NEWT_KEY_F2) {
				displaypackages=abs(1-displaypackages);
				free(curgroup);
				curdata=strdup((char *)newtListboxGetCurrent(clistbox));
				if (curdata[0] == 'G') {
					s3=rindex(curdata,':');
					s3++;
					curgroup=strdup(s3);
				} else if (curdata[0] == 'P') {
/*					curgroup=NULL; */
					strtok(curdata,":");
					xx=atoi(strtok(NULL,":"));
					yy=getPrevGroupLine(groupstartlines,xx);
					newtListboxGetEntry(clistbox,yy,&s3,(void **)&s4);
					s5=rindex(s4,':');
					s5++;
					curgroup=strdup(s5);
				}
					
				newtListboxClear(clistbox);
				if (displaypackages) {
					if (!installer && !ignoreLocal) {
						statusWindow(_("Analyzing System"),_("Checking Installed Versions..."));
					}
					oldlineindex=lineindex;
					lineindex=0;
					compindex=0;
					gsi=0;
					gptr=masterlist->next;
					while(gptr) {
						/*
						if (memberofStrArray(itype->compList,gptr->name)) {
							groupValues[compindex]='*';
							thisgroupon=1;
						} else {
							groupValues[compindex]=' ';
							thisgroupon=0;
						}
						*/
						groupValues[compindex]=getGroupStatus(masterlist,gptr->name,tmp);
						if (groupValues[compindex] == '*') {
							thisgroupon=1;
						} else {
							thisgroupon=0;
						}
						sprintf(s0,"G:%d:%d:%s",lineindex,compindex,gptr->name);
						groupstartlines[gsi]=lineindex;
						gsi++;
						sprintf(s1,"[%c] %s",groupValues[compindex],gptr->name);
						compindex++;
						newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
						lineindex++;
						for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
							pkgsize=(float)((gptr->packageSet.packages[zz]->size * 1.0) / (1024));
							sprintf(s1a,"(%.2f kB)",pkgsize);
							sprintf(s1,"        [%c] %-30s %-15s %c",isPackageSelected(tmp,gptr->packageSet.packages[zz]->name),gptr->packageSet.packages[zz]->name,s1a,isPackageInstalled(installedpackages,ipkglist,mpkglist,gptr->packageSet.packages[zz],removemarked));
							sprintf(s0,"P:%d:%.2f:%s",lineindex,pkgsize,gptr->packageSet.packages[zz]->name);
							newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
							lineindex++;
						}
						gptr=gptr->next;
					}
					if (!installer && !ignoreLocal) {
						newtPopWindow();
					}
					groupstartlines[gsi]=lineindex;
					if (curgroup) {
						newtListboxSetCurrent(clistbox,listboxIndexOfGroup(clistbox,curgroup));
					}
				} else {
					oldlineindex=lineindex;
					lineindex=0;
					compindex=0;
					gsi=0;
					gptr=masterlist->next;
					while(gptr) {
						/*
						if (memberofStrArray(itype->compList,gptr->name)) {
							groupValues[compindex]='*';
							thisgroupon=1;
						} else {
							groupValues[compindex]=' ';
							thisgroupon=0;
						}
						*/
						groupValues[compindex]=getGroupStatus(masterlist,gptr->name,tmp);
						if (groupValues[compindex] == '*') {
							thisgroupon=1;
						} else {
							thisgroupon=0;
						}
						sprintf(s0,"G:%d:%d:%s",lineindex,compindex,gptr->name);
						groupstartlines[gsi]=lineindex;
						gsi++;
						sprintf(s1,"[%c] %s",groupValues[compindex],gptr->name);
						compindex++;
						newtListboxAddEntry(clistbox,strdup(s1),(void *)strdup(s0));
						lineindex++;
						gptr=gptr->next;
					}
					groupstartlines[gsi]=lineindex;
					if (curgroup) {
						newtListboxSetCurrent(clistbox,listboxIndexOfGroup(clistbox,curgroup));
					}
				}
				/*
				for (zz=lineindex;zz < oldlineindex;zz++) {
					newtListboxGetEntry(clistbox,zz,&s4,&v1);
					if (v1) {
						newtListboxDeleteEntry(clistbox,v1);
					}
				}
				*/
			}
			if (redrawsize) {
				statusWindow(_("Calculating"),_("Recalculating installation size..."));
				sprintf(s0,_("Installation Size: %d Pkgs (%d MB)"),
					getNumPackages(tmp,removemarked),
					getPkgSetSize2(tmp,installedpackages));
				newtLabelSetText(csizelabel,s0);
				newtPopWindow();
			}
		} else if (cevent.reason == NEWT_EXIT_COMPONENT) {
 listboxselected:
			if (dynamicsize) {
				redrawsize=1;
			}
			if (cevent.u.co == ccancelbutton) {
				/*
				for (zz=0;zz < tmp->numPackages;zz++) {
					free(tmp->packages[zz]);
				}
				free(tmp);
				*/
				savedSelections=tmp;
				tmp=NULL;
				returnReason=1;
				cformdone=1;
			} else if (cevent.u.co == coptionsbutton) {
				doOptionsMenu();
			} else if (cevent.u.co == clistbox) {
				s2=strdup((char *)newtListboxGetCurrent(clistbox));
				if (s2[0] == 'G') {
					strtok(s2,":");
					xx=atoi(strtok(NULL,":"));
					yy=atoi(strtok(NULL,":"));
					s3=strtok(NULL,":");
					ctemp=toggleGroup(groupValues[yy],tmp,masterlist,s3);
					groupValues[yy]=ctemp;
					sprintf(s1,"[%c] %s",ctemp,s3);
					newtListboxSetText(clistbox,xx,s1);
					if (displaypackages) {
						aa=getNextGroupLine(groupstartlines,yy);
						for (zz=(xx+1);zz < aa;zz++) {
							newtListboxSetCurrent(clistbox,zz);
							s3=strdup((char *)newtListboxGetCurrent(clistbox));
							strtok(s3,":");
							yy=atoi(strtok(NULL,":"));
							pkgsize=atof(strtok(NULL,":"));
							s4=strtok(NULL,":");
							sprintf(s1a,"(%.2f kB)",pkgsize);
							sprintf(s1,"        [%c] %-30s %-15s %c",ctemp,s4,s1a,isPackageInstalled(installedpackages,ipkglist,mpkglist,getPackageByName(mpkglist,s4),removemarked));
							newtListboxSetText(clistbox,yy,strdup(s1));
/*							free(s3); */
						}
						newtListboxSetCurrent(clistbox,xx);
					}
				} else if (s2[0] == 'P') {
					strtok(s2,":");
					xx=atoi(strtok(NULL,":"));
					pkgsize=atof(strtok(NULL,":"));
					s3=strtok(NULL,":");
					ctemp=togglePackage(tmp,s3);
					sprintf(s1a,"(%.2f kB)",pkgsize);
					sprintf(s1,"        [%c] %-30s %-15s %c",ctemp,s3,s1a,isPackageInstalled(installedpackages,ipkglist,mpkglist,getPackageByName(mpkglist,s3),removemarked));
					newtListboxSetText(clistbox,xx,s1);
					yy=getPrevGroupLine(groupstartlines,xx);
					newtListboxGetEntry(clistbox,yy,&s3,(void **)&s4);
					s5=rindex(s4,':');
					s5++;
					s6=strdup(s3);
					s6[1]=getGroupStatus(masterlist,s5,tmp);
					newtListboxSetText(clistbox,yy,s6);
					s7=strdup(s4);
					strtok(s7,":");
					strtok(NULL,":");
					s8=strtok(NULL,":");
					yy=atoi(s8);
					groupValues[yy]=s6[1];
				}
				if (redrawsize) {
					statusWindow(_("Calculating"),_("Recalculating installation size...")); 
					sprintf(s0,_("Installation Size: %d Pkgs (%d MB)"),
						getNumPackages(tmp,removemarked),
						getPkgSetSize2(tmp,installedpackages));
					newtLabelSetText(csizelabel,s0);
					newtPopWindow();
				}
			} else {
				logMessage("Finished w/ custom package selection.");
				cformdone=1;
			}
		}
	}
	releaseForm(&cform);
	if (returnReason == 1) return NULL;
	statusWindow(_("Preparing"),_("Preparing Installation..."));
	gptr=masterlist->next;
	gindex=0;
	while(installer && gptr) {
		if ((groupValues[gindex] != '*') && (gptr->flags & CGROUP_FLAG_REQUIRED)) {
			logMessage("REQ: Selecting group '%s'",gptr->name);
			msgWindow(_("Failed Requirement"),_("Ok"),_("The component group '%s' must be installed in its entirety in order for your system to function.  It has now been auto-selected."),gptr->name);
			for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
				pptr=getPackageByName(tmp,gptr->packageSet.packages[zz]->name);
				if (pptr) {
					if (pptr->selected == 0) {
						logMessage("REQ: Selecting package '%s'",pptr->name);
					}
					pptr->selected=1;
					if (removemarked) {
						removemarked=removeFromStrArray(removemarked,pptr->name,MAX_PACKAGES);
					}
				} else {
					sprintf(s0,_("Error - no information found for '%s' - cannot install.  %s"),gptr->packageSet.packages[zz]->name,enforcereqs ? _("Package is required, cannot continue.") : _("WARNING - Package is flagged as 'required'.  Installing anyway."));
					errWindow(FALSE,s0);
					return NULL;
				}
			}
		} else if ((groupValues[gindex] != '*') && (gptr->flags & CGROUP_FLAG_WARN) && (itype->index == custominstallindex)) {
			choice=yesnoWindow(_("Warning"),TRUE,_("Yes"),_("No"),_("The component group '%s' has not been completely selected for this installation.  While it is not absolutely required, it is highly recommended, and should be installed.  Do you want to install it?"),gptr->name);
			if (choice == TRUE) {
				logMessage("REQ-WARN: Selecting group '%s'",gptr->name);
				for (zz=0;zz < gptr->packageSet.numPackages;zz++) {
					pptr=getPackageByName(tmp,gptr->packageSet.packages[zz]->name);
					if (pptr) {
						if (pptr->selected == 0) {
							logMessage("REQ: Selecting package '%s'",pptr->name);
						}
						pptr->selected=1;
						if (removemarked) {
							removemarked=removeFromStrArray(removemarked,pptr->name,MAX_PACKAGES);
						}
					} else {
						sprintf(s0,_("Warning - no information found for '%s' - cannot install."),gptr->packageSet.packages[zz]->name);
						msgWindow(_("Warning"),_("Ok"),s0);
					}
				}
			} else {
				logMessage("REQ-WARN: Decided AGAINST '%s' !",gptr->name);
			}
		}
		gptr=gptr->next;
		gindex++;
	}

	newtPopWindow();
	rc=verifyDependencies(tmp,installedpackages,removemarked);
	if (rc != SUCCESS) {
		returnReason=1;
		return NULL;
	}
		
	xx=getPkgSetSize2(tmp,installedpackages);
	if (!checkFreeSpace((xx * 1024),&yy)) {
		sprintf(s0,_("Not enough disk space is available to install everything.  Please try again. (you have %d megabytes available, and you need %d)."),(yy / 1024),xx);
		errWindow(FALSE,s0);
		if (!testing) {
			if (!installer) {
				returnReason=2;
				savedSelections=tmp;
				tmp=NULL;
			} else {
				returnReason=1;
			}			
			return NULL;
		}
	}
	qsort(tmp->packages,tmp->numPackages,sizeof(*tmp->packages),(void *)comparePackageInfo);
	rc=getPackageHeaders2(NULL,mpkglist,tmp);
	if (rc == FAILURE) return NULL;
	logPackageList(tmp);
	return tmp;
}

/* only seen in initial install */
pkgSet *prepareFromList(installType *itype, tlComponentGroup *masterlist, pkgSet *selectedlist, pkgSet *mpkglist, char **forcelist, Header *installedpackages, char **ipkglist) {
	pkgSet *tmp=(pkgSet *)malloc(sizeof(pkgSet));
	pkgSet *tmp2=NULL;
	static int zz,zy;
	int xx,yy,xa;
	bool choice;
	int rc;
	DIR *mydirptr;
	struct dirent *mydir=NULL;
	char *listpath=NULL;
	char s0[256];
	customList *mylists=NULL;
	int listindex=0;
	int pindex=0;
	FILE *fp;
	char s1[256];
	char *s2=NULL;
	struct stat mystat;
	char **plist=NULL;
	char **sptr=NULL;
	struct packageInfo *pptr=NULL;
	char sxs[256];
	
	logMessage("Preparing Installation (from list): '%s'",itype->name);
	rc=loadLocationMenu();
	switch(rc) {
	case 0:
		returnReason=1;
		return NULL;
		break;
	case 1:
		choice=yesnoWindow(_("Insert Floppy"),TRUE,_("Ok"),_("Cancel"),_("Insert the floppy disk containing the file list(s) into your first floppy drive (/dev/fd0, DOS A:), now."));
		if (choice == FALSE) {
			returnReason=1;
			return NULL;
		}
		rc=mountFloppy();
		if (rc == FAILURE) return NULL;
		listpath=strdup(installer ? CUSTOMLISTDIR_INSTALLER_FLOPPY : CUSTOMLISTDIR_FLOPPY);
		break;
	case 2:
		listpath=strdup(CUSTOMLISTDIR_INSTALLER);
		break;
	case 99:
		listpath=strdup(CUSTOMLISTDIR_TESTING);
		break;
	default:
		break;
	}
	mylists=(customList *)malloc(MAX_LISTS * sizeof(customList));
	for (zz=0;zz < MAX_LISTS;zz++) {
		mylists[zz].name=NULL;
		mylists[zz].pathname=NULL;
		mylists[zz].packages=NULL;
	}
	mydirptr=opendir(listpath);
	if (!mydirptr) {
		errWindow(FALSE,_("Could not get directory listing for '%s': %s"),listpath,strerror(errno));
		returnReason=1;
		if (progmountedfloppy) {
			unmountFloppy();
		}
		return NULL;
	}
	while ((mydir=readdir(mydirptr))) {
		if ((mydir->d_name) && (strcmp(mydir->d_name,".")) && (strcmp(mydir->d_name,".."))) {
			sprintf(s0,"%s/%s",listpath,mydir->d_name);
			stat(s0,&mystat);
			if (mystat.st_mode & S_IFREG) {
				fp=fopen(s0,"r");
				while(fgets(s1,256,fp)) {
					strchop(s1);
					if (strlen(s1)) {
						if (s1[0] == '!') {
							switch (s1[1]) {
							case 'N':
								if (mylists[listindex].packages) {
									listindex++;
									pindex=0;
								}
								strtok(s1,":");
								mylists[listindex].name=strdup(strtok(NULL,":"));
								mylists[listindex].pathname=strdup(s0);
								mylists[listindex].packages=(char **)malloc(MAX_PACKAGES * sizeof(char *));
								for (zz=0;zz < MAX_PACKAGES;zz++) {
									mylists[listindex].packages[zz]=NULL;
								}
								break;
							}
						} else {
							s2=rindex(s1,'-');
							s2[0]='\0';
							s2=rindex(s1,'-');
							s2[0]='\0';
							addToStrArray(mylists[listindex].packages,s1);
						}
					}
				}
				fclose(fp);
			}
		}
	}
	closedir(mydirptr);
	if (progmountedfloppy) {
		rc=unmountFloppy();
		if (rc == SUCCESS) {
			msgWindow(_("Floppy"),_("Ok"),_("You may now eject the floppy disk from your drive."));
		}
	}
	logMessage("MALLOC: Freeing listpath string.");
	free(listpath);
	logMessage("CUSTOMLIST: Read %d lists.",mylists[0].packages ? (listindex+1) : 0);
	plist=chooseCustomList(mylists);
	if (!plist) {
		returnReason=1;
		return NULL;
	}
	tmp->numPackages=0;
	tmp->packages=(struct packageInfo **)malloc(mpkglist->numPackages * sizeof(struct packageInfo *));
	for (zz=0;zz<mpkglist->numPackages;zz++) {
		tmp->packages[zz]=(struct packageInfo *)malloc(sizeof(struct packageInfo));
		tmp->packages[zz]->name=NULL;
		tmp->packages[zz]->name=strdup(mpkglist->packages[zz]->name);
		tmp->packages[zz]->priority=mpkglist->packages[zz]->priority;
		tmp->packages[zz]->size=mpkglist->packages[zz]->size;
		tmp->packages[zz]->selected=0;
		tmp->numPackages++;
	}
	
	rc=getPackageHeaders2(NULL,mpkglist,tmp);
	if (rc == FAILURE) return NULL;
	sptr=plist;
	while(*sptr) {
		pptr=getPackageByName(tmp,*sptr);
		if (pptr) {
			pptr->selected=1;
		} else {
			errWindow(FALSE,_("Cannot select package '%s': It does not exist on the distribution media."),*sptr);
		}
		sptr++;
	}
	logMessage("MALLOC: Freeing mylists - not needed anymore.");
	for (zz=0;zz < MAX_LISTS;zz++) {
		free(mylists[zz].name);
		free(mylists[zz].pathname);
		if (mylists[zz].packages) {
			for (zy=0;zy < MAX_PACKAGES;zy++) {
				free(mylists[zz].packages[zy]);
			}
			free(mylists[zz].packages);
		}
	}
	free(mylists);
	logMessage("MALLOC: Free complete!");
	rc=verifyDependencies(tmp,installedpackages,removemarked);
	if (rc != SUCCESS) {
		returnReason=1;
		return NULL;
	}
	qsort(tmp->packages,tmp->numPackages,sizeof(*tmp->packages),(void *)comparePackageInfo);
	xx=getPkgSetSize2(tmp,installedpackages);
	yy=checkFreeSpace((xx * 1024),&xa);
	sprintf(s0,_("megabytes (WARNING - this installation requires %d MB, but you only have %d MB available!"),xx,(xa / 1024));
	sprintf(sxs,_("MB (%d MB free)"),(xa/1024));
	choice=yesnoWindow(_("Confirm Selection"),TRUE,_("Proceed"),_("Customize"),_("The total size of the selected install (after dependencies have been verified) is %d %s.  Would you like to proceed or customize this installation?"),xx,(yy == 1) ? sxs : s0);
	if (choice == FALSE) {
		tmp2=prepareCustom(itype,masterlist,tmp,mpkglist,forcelist,installedpackages,ipkglist);
		if (!tmp2) return NULL;
		return tmp2;
	}
	return tmp;
}

int initInstallTypes(tlInstall *theInstall) {
	helptextTable *tmp=itypesTable;
	installType *ptr=NULL;
	static int zz;
	int iindex=0;

	theInstall->numInstallTypes=0;
	while(tmp->key) {
		for (ptr=theInstall->installTypes;ptr->next;ptr=ptr->next);
		ptr->next=(installType *)malloc(sizeof(installType));
		if (!ptr->next) return FAILURE;
		ptr->next->next=NULL;
		ptr->next->name=strdup(tmp->key);
		ptr->next->index=iindex;
		ptr->next->inmenu=1;
		ptr->next->flags=tmp->flags;
		ptr->next->helptext=strdup(tmp->helptext);
		ptr->next->compList=(char **)malloc(MAX_PPG * sizeof(char *));
		for (zz=0;zz < MAX_PPG;zz++) {
			ptr->next->compList[zz]=NULL;
		}
		ptr->next->size=0;
/*		if (!strcmp(tmp->key,__("Complete Install"))) { */
		if (iindex == completeinstallindex) {
			ptr->next->flags |= ITYPE_FLAG_IGNORECONFLICTS;
		}
/*		if (!strcmp(tmp->key,__("Custom Install"))) { */
		if (iindex == custominstallindex) {
			ptr->next->prepareFunction=prepareCustom;
		}
/*		else if (!strcmp(tmp->key,__("Use Package List"))) { */
		else if (iindex == filelistinstallindex) {
			ptr->next->prepareFunction=prepareFromList;
		} else {
			ptr->next->prepareFunction=prepareNormal;
		}
		logMessage("Added Install Type '%s'",ptr->next->name);
		theInstall->numInstallTypes++;
		if (debug) {
			logMessage("\tpreparefunction = 0x%X",ptr->next->prepareFunction);
		}
		iindex++;
		tmp++;
	}

	return SUCCESS;
}

int buildITypeCompList(tlInstall *theInstall) {
	tlComponentGroup *m=NULL;
	tlComponentGroup *mptr=NULL;
	installType *iptr=NULL;
	int iindex=1;
	int clindex=0;

	if (!theInstall) return FAILURE;
	m=theInstall->masterComponentList;
	iptr=theInstall->installTypes->next;
	if (!m || !iptr) return FAILURE;
	while(iptr) {
		clindex=0;
		mptr=m->next;
		if (!mptr) return FAILURE;
		while(mptr) {
			if (mptr->installtypes[iindex - 1] == '1') {
				iptr->compList[clindex]=strdup(mptr->name);
				clindex++;
			}
			mptr=mptr->next;
		}
		iptr=iptr->next;
		iindex++;
	}
	return SUCCESS;
}

int getInstallSize(installType *i, tlComponentGroup *m, int addSuggestions, int ignoreConflicts, char **forcelist, pkgSet *masterpackageset) {
	tlComponentGroup *gPtr=NULL;
	tlComponentGroup *gPtr2=NULL;
	char **sptr=NULL;
	char **gsptr=NULL;
	char **selected=NULL;
	char **conflicting=NULL;
	int selindex=0;
	int confindex=0;
	long long mysize=0;
	static int zz;
	char **sptr2=NULL;
	struct packageInfo *pptr=NULL;

	if (!installer && (!(testing && ignoreLocal))) return 0;
	if (!i || !m) return 0;
	selected=(char **)malloc(MAX_PPG * sizeof(char *));
	conflicting=(char **)malloc(MAX_PPG * sizeof(char *));
	for (zz=0;zz < MAX_PPG;zz++) {
		selected[zz]=NULL;
		conflicting[zz]=NULL;
	}
	logMessage("Calculating size of install '%s'...",i->name);
	sptr=i->compList;
	while (*sptr) {
		if (!memberofStrArray(selected,*sptr)) {
			gPtr2=getGroupByName(m,*sptr);
			if (debug) {
				logMessage("\tAdding group: '%s' (%d)",*sptr,gPtr2->size);
			}
			mysize += gPtr2->size;
			selected[selindex]=strdup(*sptr);
			selindex++;
		}
		gPtr=getGroupByName(m,*sptr);
		gsptr=gPtr->conflicts;
		while (*gsptr) {
			if (!memberofStrArray(conflicting,*gsptr)) {
				if (debug) {
					logMessage("\tMarking conflicting group '%s'",*gsptr);
				}
				conflicting[confindex]=strdup(*gsptr);
				confindex++;
			}
			gsptr++;
		}
		gsptr=gPtr->requires;
		while (*gsptr) {
			if (!memberofStrArray(selected,*gsptr)) {				
				gPtr2=getGroupByName(m,*gsptr);
				if (debug) {
					logMessage("\tAdding required group: '%s' (%d)",*gsptr,gPtr2->size);
				}
				mysize += gPtr2->size;
				selected[selindex]=strdup(*gsptr);
				selindex++;
			}
			gsptr++;
		}
		if (addSuggestions) {
			gsptr=gPtr->suggests;
			while (*gsptr) {
				if (!memberofStrArray(selected,*gsptr)) {
					gPtr2=getGroupByName(m,*gsptr);
					if (debug) {
						logMessage("\tAdding suggested group: '%s' (%d)",*gsptr,gPtr2->size);
					}
					mysize += gPtr2->size;
					selected[selindex]=strdup(*gsptr);
					selindex++;
				}
				gsptr++;
			}
		}
		sptr++;
	}
	if ((ignoreConflicts) || (i->flags & ITYPE_FLAG_IGNORECONFLICTS)) {
		logMessage("Ignoring Conflicts....(%d)(%d)",ignoreConflicts,i->flags);
	} else {
		gsptr=conflicting;
		while (*gsptr) {
			gPtr2=getGroupByName(m,*gsptr);
			mysize -= gPtr2->size;
			gsptr++;
		}
	}
	if (forcelist) {
		sptr2=forcelist;
		while (*sptr2) {
			pptr=getPackageByName(masterpackageset,*sptr2);
			mysize += pptr->size;
			logMessage("Adding 'force install' package: '%s'",*sptr2);
			sptr2++;
		}
	}
	logMessage("\tSize: %ld",mysize);
	return mysize;
}

int freeComponentGroup(tlComponentGroup *g) {
	int rc;
	
	if (!g) return SUCCESS;
	while (g->next) {
		rc=freeComponentGroup(g->next);
		if (rc == FAILURE) return FAILURE;
	}
	free(g->name);
	free(g->installtypes);
	free(g->helptext);
	free(g->requires);
	free(g->suggests);
	free(g->conflicts);
	free(g);

	return SUCCESS;
}

int getNumPackages(pkgSet *theSet, char **rlist) {
	static int zz;
	int nCount=0;
	for (zz=0;zz < theSet->numPackages;zz++) {
		if ((theSet->packages[zz]->selected == 1) &&
		    (!memberofStrArray(removemarked,
				       theSet->packages[zz]->name))) {
			nCount++;
		}
	}
	return nCount;
}

int freePackage(struct packageInfo **p) {
	struct packageInfo *pptr=NULL;
	
	if (!p) return FAILURE;
	if (!(*p)) return FAILURE;
	pptr=*p;
	headerFree(pptr->h);
	free(pptr->name);
	free(pptr->group);
	free(pptr->data);
	free(*p);
	return SUCCESS;
}

int freePkgSet(pkgSet **ps) {
	int zz;
	pkgSet *psptr=NULL;
	int rc=0;

	if (!ps) return FAILURE;
	if (!(*ps)) return FAILURE;
	psptr=*ps;
	for (zz=0;zz < psptr->numPackages;zz++) {
		rc=freePackage(&psptr->packages[zz]);
		if (rc == FAILURE) {
			logMessage("FREEPKGS: Failure to free package index '%d'",zz);
			return FAILURE;
		}
	}
	free(psptr->packages);
	free(*ps);
	return SUCCESS;
}
	
