/*
#### Copyright (c) 1998 Pacific HiTech, Inc.
####              All Rights Reserved
####              Distributable under the terms of the GPL.
#### File:        log.c
#### Author:      Scott Stone <sstone@turbolinux.com>
#### Version:     1.0J
#### Date:        January 28, 1998
#### Description:
# Source code for TurboLinux
#### ChangeLog:
# (none)
*/

#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

#include "log.h"

extern int noLog;
extern int installer;

FILE * logfile = NULL;
FILE * logfile2 = NULL;
int logfd;

void doLogMessage(char * s, ...) {
    va_list args;

    if (noLog) return;
    if (!logfile) return;

    va_start(args, s);
    fprintf(logfile, "* ");
    vfprintf(logfile, s, args);
    fprintf(logfile, "\n");
    if (installer) {
	    fprintf(logfile2, "* ");
	    vfprintf(logfile2, s, args);
	    fprintf(logfile2, "\n");
    }
    va_end(args);

    fflush(logfile);
    if (installer) {
	    fflush(logfile2);
    }
}

void openLog(void) {
	char cmd[256];
	struct stat logstat;
	
	if (noLog) return;
	if (installer) {
		logfile = fopen("/dev/tty3", "w");
		logfile2 = fopen(LOGFN_INSTALLER, "w");
		if (logfile)
			logfd = open("/dev/tty3", O_WRONLY);
		else {
			logfile = fopen("/tmp/install.log", "a");
			logfd = open("/tmp/install.log", O_WRONLY| O_APPEND);
		}
	} else {
		if (!access(LOGFN,F_OK)) {
			stat(LOGFN,&logstat);
			if (!(S_ISREG(logstat.st_mode))) {
				printf("*** SECURITY WARNING ***\n");
				printf("%s is a symbolic link to another file.\n",LOGFN);
				exit(-1);
			}
			if (logstat.st_nlink > 1) {
				printf("*** SECURITY WARNING ***\n");
				printf("%s has more than one hard link entry - possible security violation.\n",
				       LOGFN);
				exit(-1);
			}
		}
		if (!access(LOGFN,F_OK)) {
			sprintf(cmd,"rm -f %s",LOGFN);
			system(cmd);
		}
		logfile = fopen(LOGFN, "w");
		logfd = open(LOGFN, O_WRONLY);
	}
}

void closeLog(void) {
    if (logfile) {
	fclose(logfile);
	close(logfd);
    }
}
