#include "defines.h"
#include "autowin.h"
#include "log.h"
#include "prototypes.h"

extern int debug;
extern int installer;
extern int testing;
extern int newtRunning;
extern int autoDeps;
extern int ignoreDependencies;
extern int doSuggest;
extern int ignoreConflicts;
extern int groupSelectType;
extern int returnReason;
extern int dodowngrades;
extern int dynamicsize;
extern int unattended;
extern int enforcereqs;

extern int autocontrol;
extern int controlsocket;

extern ftpConfig ftpconf;

char *unresolvHelp=__("This is a list of unresolvable dependencies.  If you are using one of the standard install types (an install type other than Custom or 'From File List'), please report these dependency problems to support@turbolinux.com");

void doWelcomeScreen(void) {
	msgWindow(_("Welcome"),_("Ok"),_("Welcome to TurboPkg.  This program will allow you to view, add, modify, and remove the RPM packages on your system.  Please report all bugs/suggestions to support@turbolinux.com"));      	
}

void doInstallerWelcomeScreen(void) {
	msgWindow(_("Welcome"),_("Ok"),_("Install Packages - This step will allow you to choose the components and individual packages that will be installed on your TurboLinux system.  This package selection process can be accessed after the installation by running the 'turbopkg' program as the super-user (usually root)."));
}

char *getITypeByNumber(tlInstall *theInstall, int itnumber) {
	static int zz;
	installType *itptr=theInstall->installTypes->next;
	for (zz=1;zz < itnumber;zz++) {
		itptr=itptr->next;
	}
	return itptr->name;
}
	
	
int doMainMenu(tlInstall *theInstall) {
	newtComponent form,textbox,listbox,installbutton,optionsbutton,cancelbutton;
	struct newtExitStruct event;
	selectionList list=NULL;
	installType *i=NULL;
	installType *iptr=NULL;
	int itindex=1;
	int iSize=0;
	char s0[256];
	int lbHeight=0;
	int menuDone=0;
	int xx=0;
	static int zz;
	int rc=0;
	bool choice;

	if (!theInstall) return FAILURE;
	if ((!installer) && (!testing)) {
		theInstall->itypeSelected=CUSTOM_INDEX;
		rc=doInstall(theInstall,&ftpconf);
		return rc;
	}
	list=NULL;
	menuDone=0;
	i=theInstall->installTypes;
	if (!i) return FAILURE;
	textbox=properTextbox(_("Choose an Installation Type from the list below.  Pressing F1 will display extended information about whichever installation type is currently highlighted.  The sizes listed are approximations - they may change after dependencies have been satisfied."));
	i=i->next;
	while(i) {
		iSize=0;
		iSize=getInstallSize(i,theInstall->masterComponentList,1,0,theInstall->forcepackages,&theInstall->masterPackageSet);
		iSize = (iSize / (1024 * 1024));
		if (i->flags & ITYPE_FLAG_HIDESIZE) {
			sprintf(s0,"%s",i->name);
		} else {
			sprintf(s0,"%s (%d MB)",i->name,iSize);
		}
		addToSelectionList(&list,strdup(s0),(void *)itindex,NULL);
		itindex++;
		i=i->next;
	}
	lbHeight=(theInstall->numInstallTypes < 10) ? theInstall->numInstallTypes : 10;
	installbutton=newtButton(0,0,_("Install"));
	optionsbutton=newtButton(0,0,_("Options..."));
	cancelbutton=newtButton(0,0,_("Cancel"));
	form=askUser(layout,_("Installation Type"),NULL,
		     compGroup(vertical,center,
			       groupMargine(0,1,
					    compGroup(vertical,center,
						      textbox,
						      NULL)
					    ),
			       groupMargine(0,1,
					    compGroup(vertical,center,
						      selectionListbox(&listbox,FALSE,lbHeight,0,list),
						      NULL)
					    ),
			       groupMargine(0,1,
					    compGroup(horizontal,bottomRight,
						      installbutton,
						      optionsbutton,
						      cancelbutton,
						      NULL)
					    ),
			       NULL)
		     );
	newtFormAddHotKey(form,NEWT_KEY_F1);
	while (!menuDone) {
		newtFormRun(form,&event);
		if (event.reason == NEWT_EXIT_HOTKEY) {
			if (event.u.key == NEWT_KEY_F1) {
				xx=(int)newtListboxGetCurrent(listbox);
				logMessage("Selected help for itype %d.",xx);
				xx--;
				iptr=theInstall->installTypes->next;
				for (zz=0; zz < xx;zz++) {
					iptr=iptr->next;
				}
				/*
				msgWindow(_("Help"),_("Ok"),"'%s':\n\n%s",iptr->name,iptr->helptext);
				*/
				helpWindow(iptr->name,1,&iptr->helptext);
			}
		} else if (event.reason == NEWT_EXIT_COMPONENT) {
			if (event.u.co == cancelbutton) {
				rc=FORCECANCEL;
				menuDone=1;
			} else if (event.u.co == optionsbutton) {
				rc=doOptionsMenu();
				newtRefresh();
			} else {
/*				choice=yesnoWindow(_("Confirm"),TRUE,_("Proceed"),_("Cancel"),_("Are you sure that you want to proceed with the '%s' installation?"),getITypeByNumber(theInstall,(int)newtListboxGetCurrent(listbox))); */
				choice = TRUE;
				if (choice == TRUE) {
					menuDone=1;
					theInstall->itypeSelected=(int)newtListboxGetCurrent(listbox);
					rc=SUCCESS;
				}
			}
		}
	}
	releaseForm(&form);
	if (rc != SUCCESS) return rc;
	rc=doInstall(theInstall,&ftpconf);
	return rc;
}

int doOptionsMenu(void) {
	newtComponent oform, otextbox, oignoreDeps,oautoDeps,odoSuggest,ododowngrades,oignoreConflicts,ogroupSelectType,odynamicsize,ounattended;
#ifdef INSTALLER
	newtComponent oenforcereqs;
#endif
	newtComponent okbutton,cancelbutton;
	struct newtExitStruct oevent;
	int omenudone=0;
	char myignoreDeps,myautoDeps,mydoSuggest,mydodowngrades,myignoreConflicts,mygroupSelectType,mydynamicsize,myunattended;
#ifdef INSTALLER
	char myenforcereqs;
#endif
	int rc=-1;
	char **helptexts=NULL;
	
	otextbox=properTextbox(_("Please select which options you would like to use (F1 for help):"));

	oignoreDeps=newtCheckbox(0,0,_("Ignore Dependencies (not recommended)"),ignoreDependencies ? '*' : ' ',NULL,&myignoreDeps);
	oautoDeps=newtCheckbox(0,0,_("Automatically Resolve Dependencies"),autoDeps ? '*' : ' ',NULL,&myautoDeps);
	odoSuggest=newtCheckbox(0,0,_("Automatically Include 'Suggested' Groups"),doSuggest ? '*' : ' ',NULL,&mydoSuggest);
	ododowngrades=newtCheckbox(0,0,_("Downgrade Packages when Applicable"),dodowngrades ? '*' : ' ',NULL,&mydodowngrades);
	oignoreConflicts=newtCheckbox(0,0,_("Ignore Group Conflicts"),ignoreConflicts ? '*' : ' ',NULL,&myignoreConflicts);
	ogroupSelectType=newtCheckbox(0,0,_("Use RPM Groups for Custom Package Selection"),groupSelectType ? '*' : ' ',NULL,&mygroupSelectType);
	odynamicsize=newtCheckbox(0,0,_("Show Real-Time Size Info in Custom Install"),dynamicsize ? '*' : ' ',NULL,&mydynamicsize);
	ounattended=newtCheckbox(0,0,_("Unattended Mode - Continue if/when errors occur."),unattended ? '*' : ' ',NULL,&myunattended);
#ifdef INSTALLER
	oenforcereqs=newtCheckbox(0,0,_("Enforce Package/Group Requirements"),enforcereqs ? '*' : ' ',NULL,&myenforcereqs);
#endif
	okbutton=newtButton(0,0,_("Ok"));
	cancelbutton=newtButton(0,0,_("Cancel"));
	oform=askUser(layout,_("Configure Options"),NULL,
		      compGroup(vertical,center,
				groupMargine(0,1,
					     compGroup(vertical,center,
						       otextbox,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,topLeft,
						       oignoreDeps,
						       oautoDeps,
#ifdef INSTALLER
						       oenforcereqs,
#endif
						       odoSuggest,
						       ododowngrades,
						       oignoreConflicts,
						       ogroupSelectType,
						       odynamicsize,
						       ounattended,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(horizontal,bottomRight,
						       okbutton,
						       cancelbutton,
						       NULL)
					     ),
				NULL)
		      );
	newtFormAddHotKey(oform,NEWT_KEY_F1);
	while (!omenudone) {
		newtFormRun(oform,&oevent);
		if (oevent.reason == NEWT_EXIT_HOTKEY) {
			if (oevent.u.key == NEWT_KEY_F1) {
				helptexts=(char **)malloc(3 * sizeof(char *));
#ifdef INSTALLER
				helptexts[0]=strdup(_("'Automatically Resolve Dependencies':\n"
						      "Selecting this option will automatically resolve any RPM\n"
						      "package dependency problems.\n\n"
						      "'Enforce Requirements':\n"
						      "Selecting this option will enforce package/group "
						      "requirements.  Experts may turn it off for more "
						      "control over package installation.\n\n"
						      "'Automatically Include Suggested Groups':\n"
						      "Selecting this option will automatically include\n"
						      "'Suggested' as well as 'Required' additional groups.\n\n"
						      "'Do Downgrades':\n"
						      "If this is selected, packages will be downgraded if the package on the distribution media is older than the currently installed version.\n\n"));
#else
				helptexts[0]=strdup(_("'Automatically Resolve Dependencies':\n"
						      "Selecting this option will automatically resolve any RPM\n"
						      "package dependency problems.\n\n"
						      "'Automatically Include Suggested Groups':\n"
						      "Selecting this option will automatically include\n"
						      "'Suggested' as well as 'Required' additional groups.\n\n"
						      "'Do Downgrades':\n"
						      "If this is selected, packages will be downgraded if the package on the distribution media is older than the currently installed version.\n\n"));
#endif
				helptexts[1]=strdup(_("'Ignore Group Conflicts':\n"
						      "This option will cause TurboPkg to ignore\n"
						      "group conflicts.  Group conflicts are\n"
						      "ignored by default in the Complete install.\n\n"
						      "'Use RPM Groups for Custom...':\n"
						      "By default, this program groups packages\n"
						      "using simplified categories.  If you would prefer\n"
						      "to see them grouped by RPM group, select this option.\n"
						      "(affects Custom Install only).\n\n"));
				helptexts[2]=strdup(_("'Show Dynamic Size':\n"
						      "If this option is checked, the size of the\n"
						      "custom/postinstall package selection will be\n"
						      "shown in real-time in the package selection window.\n"
						      "Turn it off to increase speed.\n\n"
						      "'Unattended Mode':\n"
						      "If this is checked and a package fails to install,\n"
						      "instead of waiting for the user to hit OK,\n"
						      "the program will instead wait five seconds\n"
						      "and then continue."));
				helpWindow(_("Options"),3,helptexts);
				free(helptexts[0]);
				free(helptexts[1]);
				free(helptexts[2]);
				free(helptexts);
				helptexts=NULL;
			}
		} else if (oevent.reason == NEWT_EXIT_COMPONENT) {
			if (oevent.u.co == cancelbutton) {
				rc=CANCELLED;
				omenudone=1;
			} else {
				omenudone=1;
				ignoreDependencies=(myignoreDeps == '*') ? 1 : 0;
				autoDeps=(myautoDeps == '*') ? 1 : 0;
				doSuggest=(mydoSuggest == '*') ? 1 : 0;
				dodowngrades=(mydodowngrades == '*') ? 1 : 0;
				ignoreConflicts=(myignoreConflicts == '*') ? 1 : 0;
				groupSelectType=(mygroupSelectType == '*') ? 1 : 0;
				dynamicsize=(mydynamicsize == '*') ? 1 : 0;
				unattended=(myunattended == '*') ? 1 : 0;
				rc=SUCCESS;
			}
		}
	}
	releaseForm(&oform);
	newtRefresh();
	return rc;
}

void showDependencies(char **requires, char **provides) {
	newtComponent dpform,dptextbox,reqlistbox,provlistbox,dpokbutton;
	selectionList reqlist=NULL;
	selectionList provlist=NULL;
	char **sptr=NULL;
	int nr=0;
	int np=0;

	sptr=requires;
	while (*sptr) {
		addToSelectionList(&reqlist,strdup(*sptr),(void *)1,NULL);
		sptr++;
		nr++;
	}
	sptr=provides;
	while (*sptr) {
		addToSelectionList(&provlist,strdup(*sptr),(void *)1,NULL);
		sptr++;
		np++;
	}
	if (nr == 0) {
		addToSelectionList(&reqlist,"(none)",1,NULL);
	}
	if (np == 0) {
		addToSelectionList(&provlist,"(none)",1,NULL);
	}
	dptextbox=properTextbox(_("The list on the left is a list of items required by this package.  The list on the right is a list of items provided by this package."));
	dpokbutton=newtButton(0,0,_("Ok"));

	dpform=askUser(ask_keep,_("Package Dependencies"),NULL,
		       compGroup(vertical,center,
				 groupMargine(0,1,
					      compGroup(vertical,center,
							dptextbox,
							NULL)
					 ),
				 groupMargine(0,1,
					      compGroup(horizontal, topLeft,
							selectionListbox(&reqlistbox,FALSE,5,0,reqlist),
							properTextbox("       "),
							selectionListbox(&provlistbox,FALSE,5,0,provlist),
							NULL)
					 ),
				 groupMargine(0,1,
					      compGroup(vertical,center,
							dpokbutton,
							NULL)
					 ),
				 NULL)
		       );
	releaseForm(&dpform);
}
	

void displayRPMInfo(char **info, pkgSet *mpkgset, char *pkgdatastr) {
	newtComponent dform,dtextbox1,dtextbox2,okbutton,depbutton;
	int okdisp=1;
	char **requires=NULL;
	char **provides=NULL;

	if (!info) okdisp=0;
	if (okdisp) {
		if ((!info[0]) || (!info[1])) okdisp=0;
	}
	if (!okdisp) {
		msgWindow(_("No Information"),_("Ok"),_("No information is available about the specified package."));
		return;
	}
	dtextbox1=properTextbox(info[0]);
	dtextbox2=properTextbox(info[1]);
	okbutton=newtButton(0,0,_("Ok"));
	depbutton=newtButton(0,0,_("Dependencies"));

	dform=askUser(ask_keep,_("Package Information"),NULL,
		      compGroup(vertical,center,
				groupMargine(0,1,
					     compGroup(vertical,topLeft,
						       dtextbox1,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,topLeft,
						       dtextbox2,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(horizontal,bottomRight,
						       okbutton,
						       depbutton,
						       NULL)
					     ),
				NULL)
		      );
	if (newtFormGetCurrent(dform) == depbutton) {
		requires=uniquifyStrArray(getRequires(mpkgset,pkgdatastr),1);
		provides=uniquifyStrArray(getProvides(mpkgset,pkgdatastr),1);
		showDependencies(requires,provides);
	}
	releaseForm(&dform);
}

int showUnresolvableDependencies(char **udlist) {
	newtComponent uform,utextbox,ulistbox,uoverride,ucancel;
	selectionList list=NULL;
	char **sptr=NULL;
	int rc=SUCCESS;
	char *s1=NULL;
	char *s2=NULL;
	char s3[256];
	char *stmp=NULL;
	struct newtExitStruct event;
	int fdone=0;
#ifdef AUTOCONTROL
	char sa[4096];
#endif

	utextbox=NULL;
	if (!autocontrol) {
		utextbox=properTextbox(_("The following packages are required by the current installation, but are not available on the distribution media: (F1 for help)"));
	}
	sptr=udlist;
	while(*sptr) {
		stmp=strdup(*sptr);
		s1=strtok(stmp,"\001");
		s2=strtok(NULL,"\001");
		sprintf(s3,_("%s (required by %s)"),s1,s2);
		if (autocontrol) {
#ifdef AUTOCONTROL
			sprintf(sa,"UDEP:%s:%s::0\n",s2,s1);
			autoMessage(controlsocket,sa);
#endif
		} else {
			addToSelectionList(&list,strdup(s3),(void *)1,NULL);
		}
		sptr++;
	}
	if (autocontrol) {
		return SUCCESS;
	}
	uoverride=newtButton(0,0,_("Install Anyway"));
	ucancel=newtButton(0,0,_("Cancel"));
	uform=askUser(layout,_("Unresolvable Dependencies"),NULL,
		      compGroup(vertical,center,
				groupMargine(0,1,
					     compGroup(vertical,center,
						       utextbox,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,center,
						       selectionListbox(&ulistbox,FALSE,10,0,list),
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(horizontal,bottomRight,
						       uoverride,
						       ucancel,
						       NULL)
					     ),
				NULL)
		      );
	newtFormAddHotKey(uform,NEWT_KEY_F1);
	while (!fdone) {
		newtFormRun(uform,&event);
		if (event.reason == NEWT_EXIT_HOTKEY) {
			if (event.u.key == NEWT_KEY_F1) {
				helpWindow(_("Unresolvable Dependencies"),1,&unresolvHelp);
				/*
				msgWindow(_("Help"),_("Ok"),_("This is a list of unresolvable dependencies.  If you are using one of the standard install types (an install type other than Custom or 'From File List'), please report these dependency problems to support@turbolinux.com"));
				*/
			}
		} else if (event.reason == NEWT_EXIT_COMPONENT) {
			if (event.u.co == ucancel) {
				rc=CANCELLED;
			}
			fdone=1;
		}
	}
	releaseForm(&uform);
	return rc;
}

void helpWindow(char *title, int numPages, char **helpPages) {
	newtComponent helpForm,helpLabel,helpTextbox,helpBackPage,helpForwardPage,helpDoneButton;
	int pagenumber=1;
	int pageindex=0;
	int helpdone=0;
	char s0[256];

	while (!helpdone) {
		sprintf(s0,_("'%s' Help : Page %d of %d"),title,pagenumber,numPages);
		helpLabel=newtLabel(0,0,s0);
		helpTextbox=properTextbox(helpPages[pageindex]);
		helpBackPage=newtButton(0,0,_("<< Page"));
		helpForwardPage=newtButton(0,0,_(">> Page"));
		helpDoneButton=newtButton(0,0,_("Done"));
		helpForm=askUser(layout,_("Help"),helpDoneButton,
				 compGroup(vertical,center,
					   groupMargine(0,1,
							compGroup(vertical,center,
								  helpLabel,
								  NULL)
						   ),
					   groupMargine(0,1,
							compGroup(vertical,center,
								  helpTextbox,
								  NULL)
						   ),
					   groupMargine(0,1,
							compGroup(horizontal,bottomRight,
								  helpBackPage,
								  helpDoneButton,
								  helpForwardPage,
								  NULL)
						   ),
					   NULL)
			);
		newtFormSetCurrent(helpForm,helpDoneButton);
		newtRunForm(helpForm);
		if (newtFormGetCurrent(helpForm) == helpDoneButton) {
			helpdone=1;
			releaseForm(&helpForm);
		}
		else if (newtFormGetCurrent(helpForm) == helpBackPage) {
			if (pageindex > 0) {
				pagenumber--;
				pageindex--;				
			}
		}
		else if (newtFormGetCurrent(helpForm) == helpForwardPage) {
			if (pagenumber < numPages) {
				pagenumber++;
				pageindex++;				
			}
		}
		releaseForm(&helpForm);
	}
	newtRefresh();
}
				   
int loadLocationMenu(void) {
	newtComponent lForm,lTextbox,lListbox,lOk,lCancel;
	selectionList list=NULL;
	int rc=0;
	
	lTextbox=properTextbox(_("Where would you like to load the custom list from?"));
	lOk=newtButton(0,0,_("Ok"));
	lCancel=newtButton(0,0,_("Cancel"));

	addToSelectionList(&list,_("Load From Floppy (/dev/fd0)"),(void *)1,NULL);
	addToSelectionList(&list,_("Load From CD-ROM"),(void *)2,NULL);
	if (testing) {
		addToSelectionList(&list,_("TESTING"),(void *)99,NULL);
	}

	lForm=askUser(ask_keep,_("List Location"),NULL,
		      compGroup(vertical,center,
				groupMargine(0,1,
					     compGroup(vertical,center,
						       lTextbox,
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(vertical,center,
						       selectionListbox(&lListbox,FALSE,5,0,list),
						       NULL)
					     ),
				groupMargine(0,1,
					     compGroup(horizontal,bottomRight,
						       lOk,
						       lCancel,
						       NULL)
					     ),
				NULL)
		      );
	if (newtFormGetCurrent(lForm) == lCancel) {
		rc=0;
	} else {
		rc=(int)newtListboxGetCurrent(lListbox);
	}
	releaseForm(&lForm);
	return rc;
}

char *cclHelp = __("Each of the lists below was generated with the 'gencustomlist' program (which is provided inside the turbopkg RPM package and also in the bin/ directory on the TurboLinux CD-ROM #1).  The lists are snapshots of the packages installed on a running TurboLinux system.  The installer will attempt to match your new installation to the snapshot that you select.  The lists provided in the customlists/ directory on the TurboLinux CD-ROM have been certified by Pacific HiTech - use others at your own risk!");

char **chooseCustomList(customList *lists) {
	newtComponent clForm,clTextbox,clListbox,clOk,clCancel;
	selectionList list=NULL;
	int lindex=0;
	int formdone=0;
	struct newtExitStruct event;
	int isCancelled=0;
	int clindex=0;

	if (!lists[0].packages) {
		errWindow(FALSE,_("No lists available!"));
		returnReason=1;
		return NULL;
	}
	clTextbox=properTextbox(_("Which list do you want to use? (F1 for help)"));
	clOk=newtButton(0,0,_("Ok"));
	clCancel=newtButton(0,0,_("Cancel"));

	while (lists[lindex].packages) {
		addToSelectionList(&list,lists[lindex].name,(void *)(lindex + 1),NULL);
		lindex++;
	}

	clForm=askUser(layout,_("Choose List"),NULL,
		       compGroup(vertical,center,
				 groupMargine(0,1,
					      compGroup(vertical,center,
							clTextbox,
							NULL)
					      ),
				 groupMargine(0,1,
					      compGroup(vertical,center,
							selectionListbox(&clListbox,FALSE,(lindex > 10) ? 10 : lindex,0,list),
							NULL)
					      ),
				 groupMargine(0,1,
					      compGroup(horizontal,bottomRight,
							clOk,
							clCancel,
							NULL)
					      ),
				 NULL)
		       );
	newtFormAddHotKey(clForm,NEWT_KEY_F1);
	while (!formdone) {
		newtFormRun(clForm,&event);
		if (event.reason == NEWT_EXIT_HOTKEY) {
			if (event.u.key == NEWT_KEY_F1) {
				helpWindow(_("Custom Package Lists"),1,&cclHelp);
			}
		}
		else if (event.reason == NEWT_EXIT_COMPONENT) {
			formdone=1;
			if (event.u.co == clCancel) {
				isCancelled=1;
			} else {
				clindex=(int)newtListboxGetCurrent(clListbox);
				clindex--;
			}
		}
	}
	releaseForm(&clForm);
	if (isCancelled) {
		returnReason=1;
		return NULL;
	}
	return lists[clindex].packages;
}
		
