#include "defines.h"
#include "log.h"
#include "prototypes.h"
#include "autowin.h"
#include <rpm/rpmlib.h>
#include <unistd.h>
#include <fcntl.h>

extern int debug;
extern int newtRunning;

int addPackage(tlComponentGroup *mygroup, char *pkgname, int priority, int flags, int selected) {
	int psIndex=0;
	struct packageInfo *pptr=NULL;

	if (flags == 0) {
		while (mygroup->packageSet.packages[psIndex]) {
			psIndex++;
		}
		if (!memberofPkgSet(&mygroup->packageSet,pkgname)) {
			mygroup->packageSet.packages[psIndex]=(struct packageInfo *)malloc(sizeof(struct packageInfo));
			mygroup->packageSet.packages[psIndex]->group=NULL;
			mygroup->packageSet.packages[psIndex]->name=strdup(pkgname);
			mygroup->packageSet.packages[psIndex]->priority=priority;
			mygroup->packageSet.packages[psIndex]->selected=selected ? '*' : ' ';
			logMessage("PACKAGE:  Added '%s' (priority %d) to '%s'",mygroup->packageSet.packages[psIndex]->name,mygroup->packageSet.packages[psIndex]->priority,mygroup->name);
			mygroup->packageSet.numPackages++;
			psIndex++;
		} else {
			logMessage("PACKAGE:  '%s' already in package list for '%s'",pkgname,mygroup->name);
			pptr=getPackageByName(&mygroup->packageSet,pkgname);
			if (pptr) {
				if ((pptr->priority != priority) && (priority != DEFAULT_PRIORITY)) {
					logMessage("PACKAGE: New priority is not '%d'",DEFAULT_PRIORITY);
					logMessage("PACKAGE: Updating priority to %d.",priority);
					pptr->priority=priority;
				}
			}
		}
			      
	}
	return SUCCESS;
}

int addPackageToSet(pkgSet *theSet, char *pkgname, int priority, int flags, int selected) {
	int psIndex=0;
	struct packageInfo *pptr=NULL;

	if ((!theSet) || (!pkgname)) {
		logMessage("OOPS: theSet or pkgname was NULL in addPackageToSet!");
		return FAILURE;
	}
	if (flags == 0) {
		while (theSet->packages[psIndex]) {
			psIndex++;
		}
		if (!memberofPkgSet(theSet,pkgname)) {
			if (debug) {
				logMessage("DEBUG: Adding '%s' (%d) to unknown set.",pkgname,priority);
			}
			theSet->packages[psIndex]=(struct packageInfo *)malloc(sizeof(struct packageInfo));
			theSet->packages[psIndex]->group=NULL;
			theSet->packages[psIndex]->name=strdup(pkgname);
			theSet->packages[psIndex]->priority=priority;
			theSet->packages[psIndex]->selected=selected;
			theSet->numPackages++;
			psIndex++;
			if (debug) {
				logMessage("\t(done)");
			}
		} else {
			pptr=getPackageByName(theSet,pkgname);
			if (pptr) {
				if ((pptr->priority != priority) && (priority != DEFAULT_PRIORITY)) {
					logMessage("PACKAGE: Updating priority to %d.",priority);
					pptr->priority=priority;
				}
			}
		}
			      
	}
	return SUCCESS;
}

int addPackageToSetByHeader(pkgSet *theSet, Header h, int priority, int flags, int selected) {
	int psIndex=0;
	char *pkgname=NULL;
	unsigned int *theSize;
	int xtype,xc;

	if ((!theSet) || (!h)) {
		logMessage("OOPS: theSet or header was NULL in addPackageToSetByHeader!");
		return FAILURE;
	}
	headerGetEntry(h,RPMTAG_NAME,&xtype,(void **)&pkgname,&xc);
	headerGetEntry(h,RPMTAG_SIZE,&xtype,(void **)&theSize,&xc);

	if (!pkgname) return FAILURE;
	if (flags == 0) {
		while (theSet->packages[psIndex]) {
			psIndex++;
		}
		if (!memberofPkgSet(theSet,pkgname)) {
			if (debug) {
				logMessage("DEBUG: Adding '%s' (%d) by header to unknown set.",pkgname,priority);
			}
			theSet->packages[psIndex]=(struct packageInfo *)malloc(sizeof(struct packageInfo));
			theSet->packages[psIndex]->h=h;
			headerGetEntry(h,RPMTAG_GROUP,&xtype,(void **)&theSet->packages[psIndex]->group,&xc);
			theSet->packages[psIndex]->name=strdup(pkgname);
			theSet->packages[psIndex]->size=*theSize;
			theSet->packages[psIndex]->priority=priority;
			theSet->packages[psIndex]->selected=selected;
			theSet->numPackages++;
			psIndex++;
			if (debug) {
				logMessage("\t(done)");
			}
		} else {
		}
			      
	}
	return SUCCESS;
}

int addGroupDep(char **grouplist, const char *groupname) {
	int gindex=0;
	while(grouplist[gindex]) {
		gindex++;
	}
	grouplist[gindex]=strdup(groupname);
	return SUCCESS;
}

int parseCompsFile(tlInstall *theInstall) {
	tlComponentGroup *m=theInstall->masterComponentList;
	tlComponentGroup *ptr=NULL;
	tlComponentGroup *tmp=NULL;
	FILE *fp;
	int lindex=0;
	char s0[256];
	char *s1;
	char *s2;
	int incomp=0;
	static int zz;
	int xx;
	int fpindex=0;

	theInstall->forcepackages=(char **)malloc(MAX_PPG * sizeof(char *));
	for (zz=0;zz < MAX_PPG;zz++) {
		theInstall->forcepackages[zz]=NULL;
	}
	fp=fopen(theInstall->compsfile,"r");
	if (!fp) return FAILURE;
	while(fgets(s0,256,fp)) {
		s0[(strlen(s0) - 1)]='\0';
		if (!strlen(s0)) continue; /* skip blank lines */
		if (s0[0] == '#') continue; /* skip # - comment lines */
		if (lindex == 0) {
			if (strcmp(s0,COMPSFILE_REQUIRED_VERSION)) {
				logMessage("ERROR: Components file is not version '%s'",COMPSFILE_REQUIRED_VERSION);
				logMessage("\t(value was '%s')",s0);
				fclose(fp);
				if (newtRunning) {
					errWindow(FALSE,_("Incorrect component file version - cannot use."));
				}
				return FAILURE;
			}
			lindex++;
		} else {
			if (!incomp) {
				if ((s0[0] == '0') || (s0[0] == '1')) {
					incomp=1;
					tmp=(tlComponentGroup *)malloc(sizeof(tlComponentGroup));
					tmp->next=NULL;
					tmp->priority=30000;
					s1=strtok(s0,":");
					if (!s1) {
						logMessage("PARSE ERROR: Line %d in comps file.",lindex);
						return FAILURE;
					}
					tmp->installtypes=strdup(s1);
					s1=strtok(NULL,":");
					if (!s1) {
						logMessage("PARSE ERROR: Line %d in comps file.",lindex);
						return FAILURE;
					}
					tmp->flags=atoi(s1);
					s1=strtok(NULL,":");
					if (!s1) {
						logMessage("PARSE ERROR: Line %d in comps file.",lindex);
						return FAILURE;
					}
					if (atoi(s1) != 0) {
						tmp->priority=atoi(s1);
					}
					s1=strtok(NULL,":");
					if (!s1) {
						logMessage("PARSE ERROR: Line %d in comps file.",lindex);
						return FAILURE;
					}
					tmp->name=strdup(s1);
					tmp->packageSet.packages=(struct packageInfo **)malloc(MAX_PPG * sizeof(struct packageInfo*));
					tmp->packageSet.numPackages=0;
					for (zz=0;zz < MAX_PPG;zz++) {
						tmp->packageSet.packages[zz]=NULL;
						tmp->requires[zz]=NULL;
						tmp->suggests[zz]=NULL;
						tmp->conflicts[zz]=NULL;
					}
					logMessage("GROUP: '%s' [%s], priority %d",tmp->name,tmp->installtypes,tmp->priority);
				}
			} else {
				if (!strcmp(s0,"end")) {
					incomp=0;
					for (ptr=m;ptr->next;ptr=ptr->next);
					ptr->next=tmp;
					tmp=NULL;
				} else {
					if (s0[0] != '!') {
						addPackage(tmp,s0,tmp->priority,0,0);
						addPackageToSet(&theInstall->masterPackageSet,s0,tmp->priority,0,0);
					} else {
						switch (s0[1]) {
						case 'O':
							strtok(s0,":");
							xx=atoi(strtok(NULL,":"));
							s2=strtok(NULL,":");
							addPackage(tmp,s2,xx,0,0);
							addPackageToSet(&theInstall->masterPackageSet,s2,xx,0,0);
							break;
						case 'R':
							strtok(s0,":");
							s2=strtok(NULL,":");
							addGroupDep(tmp->requires,s2);
							break;
						case 'S':
							strtok(s0,":");
							s2=strtok(NULL,":");
							addGroupDep(tmp->suggests,s2);
							break;
						case 'G':
							strtok(s0,":");
							s2=strtok(NULL,":");
							addGroupDep(tmp->conflicts,s2);
							break;
						case 'X':
							strtok(s0,":");
							s2=strtok(NULL,":");
							theInstall->forcepackages[fpindex]=strdup(s2);
							fpindex++;
							addPackage(tmp,s2,tmp->priority,0,0);
							addPackageToSet(&theInstall->masterPackageSet,s2,tmp->priority,0,0);
							break;
						default:
							break;
						}
					}
				}
			}
			lindex++;
		}
	}
	fclose(fp);
	logMessage("COMPSFILE: Successfully parsed.  Contained %d lines.",lindex);
	return SUCCESS;
}
		
