#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>

#include "defines.h"
#include "autowin.h"
#include "prototypes.h"
#include "log.h"

extern int debug;
extern int installer;
extern int testing;
extern int newtRunning;
extern int returnReason;
extern int autoDeps;
extern int autoDepsModified;
extern int ignoreConflicts;
extern int doSuggest;
extern int ignoreLocal;
extern int ftpinstall;
extern int dodowngrades;
extern int unattended;

extern int autocontrol;
extern int controlsocket;

extern char **removemarked;
extern int dosSource;
extern int filenames_scanned;
extern fnBinding *myfnBindings;

struct statusWindowInfo {
    newtComponent form, packageLabel, sizeLabel, summaryText;
    newtComponent pkgScale, globalScale;
    newtComponent pkgDoneLabel, pkgRemainsLabel;
    newtComponent sizeDoneLabel, sizeRemainsLabel;
    newtComponent timeDoneLabel, timeRemainsLabel, timeTotalLabel;
    int numPackages, packagesDone;
    unsigned int sizePackages, sizeDone;
    int thisPackageSize;
    time_t timeStarted;
} si;

static void swOpen(int,int,char *);
static void swPackage(Header);
static void swPackageComplete(void);
static void swCallback(const unsigned long amount, const unsigned long total);
static void swClose(void);

char *rpmDirs_installer[] = {
	"TurboLinux/RPMS/",
	"turbolin/rpms/",
	"TurboUpgrade/RPMS/",
	"RPMS/",
	ARCH_PATH,
	"/",
	NULL,
};

char *rpmDirs[] = {
	"/",
	"TurboLinux/RPMS/",
	"turbolin/rpms/",
	"TurboUpgrade/RPMS/",
	"RPMS/",
	ARCH_PATH,	
	NULL,
};

/* only for the installer */
char *noscriptlist[] = {
	"diald-config",
	"gpm",
	"gtk+-devel",
	"kdelibs",
	"mysql",
	"libg++-devel",
	"netatalk",
	"glibc-devel",
	NULL,
};

static void formatTime(char * buf, time_t t) {
    int hours, minutes, secs;

    hours = t / 60 / 60;
    t %= (60 * 60);

    minutes = t / 60;
    t %= 60;

    secs = t;

    sprintf(buf, "%01d:%02d.%02d", hours, minutes, secs);
}

int removeSelectedPackages(char **rlist) {
	bool choice;
	int rcount=0;
	char **sptr;
	rpmdb db=NULL;
	dbiIndexSet myindexset;
	unsigned int myoffset;

	if (installer) return SUCCESS;
	if (!rlist) return SUCCESS;
	sptr=rlist;
	while(*sptr) {
		rcount++;
		sptr++;
	}
	if (rcount == 0) return SUCCESS;
	sptr=rlist;
	if (!autocontrol) {
		choice=yesnoWindow(_("Remove Packages"),
				   TRUE,
				   _("Remove"),
				   _("Cancel"),
				   _("You have marked %d packages for removal - are you sure you want to remove them?"),
				   rcount);
		if (choice == FALSE) {
			logMessage("REMOVEPACKAGES: Cancelled.");
			return CANCELLED;
		}
	}
	rpmReadConfigFiles(NULL, NULL, NULL, 0);
	if (rpmdbOpen("/", &db, O_RDWR | O_CREAT, 0644)) {
		logMessage("REMOVEPACKAGES: Fatal error opening RPM db.\n");
		if (!autocontrol) {
			errWindow(FALSE,_("Error - could not open RPM database!  Cannot uninstall packages."));
		}
		return FAILURE;
	} else {
		logMessage("REMOVEPACKAGES: RPM db open");
	}
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatus(controlsocket,"removing",_("Removing Packages..."));
#endif
	} else {
		statusWindow(_("Removing"),_("Removing Packages..."));
	}
	sptr=rlist;
	while(*sptr) {
		rpmdbFindPackage(db,*sptr,&myindexset);
		myoffset=myindexset.recs[0].recOffset;
		if (!testing) {
			rpmRemovePackage("/",db,myoffset,0);
		}
		logMessage("REMOVEPACKAGES: Removed '%s' %s",*sptr,testing ? "testing" : "");
		sptr++;
	}
	rpmdbClose(db);
	logMessage("REMOVEPACKAGES: RPM db closed");
	if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatusEnd(controlsocket);
#endif
	} else {
		newtPopWindow();
	}
	return SUCCESS;
}

char *getPackageFilename(tlInstall *i, ftpConfig *f, int ftpsocket, struct packageInfo *p) {
	int rc;
	char **sptr=NULL;
	char *xName=NULL;
	char *xVer=NULL;
	char *xRel=NULL;
	int xtype,xc;
	char rpmfn[256];
	char altrpmfn[256];
	char remotefn[256];
	int retrieved=0;
	char *localfn=NULL;
	int outfd;

	if (!i || !p) return NULL;
	localfn=(char *)malloc(256);
	headerGetEntry(p->h,RPMTAG_NAME,&xtype,(void **)&xName,&xc);
	headerGetEntry(p->h,RPMTAG_VERSION,&xtype,(void **)&xVer,&xc);
	headerGetEntry(p->h,RPMTAG_RELEASE,&xtype,(void **)&xRel,&xc);
	if (!dosSource) {
		sprintf(rpmfn,"%s-%s-%s.%s.rpm",xName,xVer,xRel,ARCH);
		sprintf(altrpmfn,"%s-%s-%s.noarch.rpm",xName,xVer,xRel);
		logMessage("DOINSTALL: Need filename '%s' or '%s'",rpmfn,altrpmfn);
	} else {
		localfn=lookupFilename(myfnBindings,xName);
		if (localfn) {
			strcpy(rpmfn,localfn);
		} else {
			return NULL;
		}
		strcpy(altrpmfn,rpmfn);
		logMessage("DOINSTALL: Filename looked up as '%s'",rpmfn);
	}
	if (ftpinstall) {
		sprintf(localfn,"%s/%s",installer ? "/mnt/tmp" : "/tmp",rpmfn);
		unlink(localfn);
		outfd=creat(localfn,FTP_CREATE_MODE);
		if (!outfd) {
			if (!autocontrol) {
				errWindow(FALSE,_("Cannot create temporary file '%s' : %s"),strerror(errno));
			}
			logMessage("Cannot create temporary file '%s' : %s",strerror(errno));
			return NULL;
		}
		sptr=installer ? rpmDirs_installer : rpmDirs;
		while(*sptr) {
			sprintf(remotefn,"%s/%s%s",f->ftpdir,*sptr,rpmfn);
			if (!autocontrol) {				
				newtTextboxSetText(si.summaryText,_("Receiving file by FTP."));
				newtDrawForm(si.form);
				newtRefresh();
			}
			
			rc=ftpGetRPM(ftpsocket,remotefn,outfd,&si.packageLabel,&si.sizeLabel,&si.pkgScale);
			if (rc == 0) {
				retrieved=1;
				close(outfd);
				return localfn;
			} else {
				logMessage("FTP: Retrieve of '%s' returned '%s'",remotefn,ftpStrerror(rc));
			}
			sptr++;
		}
		close(outfd);
		if (!retrieved) {
			unlink(localfn);
			logMessage("FTP: Trying alternate filename '%s'",altrpmfn);
			sprintf(localfn,"%s/%s",installer ? "/mnt/tmp" : "/tmp",altrpmfn);
			unlink(localfn);
			outfd=creat(localfn,FTP_CREATE_MODE);
			if (!outfd) {
				errWindow(FALSE,_("Cannot create temporary file '%s' : %s"),strerror(errno));
				logMessage("Cannot create temporary file '%s' : %s",strerror(errno));
				return NULL;
			}
			sptr=installer ? rpmDirs_installer : rpmDirs;
			while(*sptr) {
				if (!autocontrol) {
					newtTextboxSetText(si.summaryText,_("Receiving file by FTP."));
					newtDrawForm(si.form);
					newtRefresh();
				}
				sprintf(remotefn,"%s/%s%s",f->ftpdir,*sptr,altrpmfn);
				rc=ftpGetRPM(ftpsocket,remotefn,outfd,&si.packageLabel,&si.sizeLabel,&si.pkgScale);
				if (rc == 0) {
					retrieved=1;
					close(outfd);
					return localfn;
				} else {
					logMessage("FTP: Retrieve of '%s' returned '%s'",remotefn,ftpStrerror(rc));
				}
				sptr++;
			}
		}
		if (!retrieved) {
			unlink(localfn);
			logMessage("FTP: Cannot find the requested file: '%s' or '%s'",rpmfn,altrpmfn);
			if (!autocontrol) {
				errWindow(FALSE,_("Error: Cannot find the '%s' package on the remote server."),xName);
			}
			return NULL;
		}
	} else {
		if (dosSource) {
			return localfn;
		}
		sptr=installer ? rpmDirs_installer : rpmDirs;
		while(*sptr) {
			sprintf(localfn,"%s/%s/%s",i->rpmdir,*sptr,rpmfn);
			if (!access(localfn,R_OK)) {
				retrieved=1;
				return localfn;
			}
			sptr++;
		}
		if (!retrieved) {
			logMessage("DOINSTALL: Trying alternate filename '%s'",altrpmfn);
			sptr=installer ? rpmDirs_installer : rpmDirs;
			while(*sptr) {
				sprintf(localfn,"%s/%s/%s",i->rpmdir,*sptr,altrpmfn);
				if (!access(localfn,R_OK)) {
					retrieved=1;
					return localfn;
				}
				sptr++;
			}
		}
		if (!retrieved) {
			if (!filenames_scanned) {
				logMessage("DOINSTALL: Need to scan filenames...");
				scanFilenames(i->rpmdir,myfnBindings);
			}
			localfn=lookupFilename(myfnBindings,xName);
			if (localfn) {
				retrieved=1;
				return localfn;
			}
		}
		if (!retrieved) {
			logMessage("DOINSTALL: Cannot find file '%s' or '%s'.",rpmfn,altrpmfn);
			errWindow(FALSE,_("Error: Cannot find the package '%s' on the distribution media!"),xName);
			return NULL;
		}
	}
	return NULL;
}

int shouldIgnoreScripts(char *pkgname) {
	char **sptr;

	if (!pkgname) return 0;
	if (!installer) return 0;
	sptr=noscriptlist;
	while(*sptr) {
		if (!strcmp(pkgname,*sptr)) return 1;
		sptr++;
	}
	return 0;
}

int doInstall(tlInstall *theInstall, ftpConfig *f) {
	installType *itptr=NULL;
	static int zz;
	rpmdb db=NULL;
	int fd=0;
	int rc;
	int rpmflags=0;
	int ftpsocket=0;
	char *thisPkg=NULL;
	char xc;
	int okinst=0;
	int tmpcount=0;
	FILE *testfd=NULL;
#ifdef AUTOCONTROL
	char sa[4096];
#endif

	logMessage("AUTODEPS: %s",autoDeps ? "enabled" : "disabled");
	if (testing) {
		testfd=fopen("/tmp/turbopkg.order","w");
	}
	logMessage("DOINSTALL: Performing preparation/installation.");
	if (installer || (testing && ignoreLocal)) {
		if (!theInstall->installTypes) return FAILURE;
		itptr=theInstall->installTypes->next;
		for (zz=1;zz < theInstall->itypeSelected;zz++) {
			itptr=itptr->next;
		}
	} else {
		itptr=(installType *)malloc(sizeof(installType));
		itptr->flags=0;
		itptr->prepareFunction=prepareCustom;
	}

	if (!autocontrol) {
 tryPrepareAgain:
		theInstall->selectedPackageSet=itptr->prepareFunction(itptr,theInstall->masterComponentList,NULL,&theInstall->masterPackageSet,theInstall->forcepackages,theInstall->installedPackages,theInstall->ipkglist);
	}
	if (autoDepsModified) {
		autoDepsModified=0;
		autoDeps=1;
	}
	if (!theInstall->selectedPackageSet) {
		if (returnReason == 1) {
			returnReason = 0;
			if ((installer) || (testing)) {
				return CANCELLED;
			} else {
				return FORCECANCEL;
			}
		} else if (returnReason == 2) {
			if (!autocontrol) {
				goto tryPrepareAgain;
			}
		}
		if (newtRunning) {
			errWindow(FALSE,_("DOINSTALL: ERROR Creating selected package set!  Aborting program."));
		} else {
			logMessage("DOINSTALL: ERROR Creating selected package set! (fatal)");
		}
		return FAILURE;
	}
	logMessage("DOINSTALL: Preparation Complete.");
#ifdef INSTALLER
	if (installer) {
		logMessage("DOINSTALL: Initial Install - installing skeleton...");
		rc = installSkeleton(f);
		if ((rc != SUCCESS) && (!testing)) return rc;
	}
#endif
	if (!autocontrol) {
		logMessage("DOINSTALL: Removing packages marked for removal...");
		removeSelectedPackages(removemarked);
#ifdef INSTALLER
		if (installer && ftpinstall) {
			ftpForceXServers(theInstall->selectedPackageSet);
		}
#endif
	}
	logMessage("DOINSTALL: Installing packages marked for install/upgrade...");
	rpmReadConfigFiles(NULL, NULL, NULL, 0);
	if (rpmdbOpen(installer ? "/mnt" : "/", &db, O_RDWR | O_CREAT, 0644)) {
		logMessage("DOINSTALL: Fatal error opening RPM db: %s\n",rpmErrorString());
		return FAILURE;
	} else {
		logMessage("DOINSTALL: RPM db open");
	}
	logMessage("DOINSTALL: Installing %d packages.",theInstall->selectedPackageSet->numPackages);

	if (!installer) {
		if (autocontrol) {
#ifdef AUTOCONTROL
			autoStatus(controlsocket,"analyzingActions",_("Analyzing Needed Actions..."));
#endif
		} else {
			statusWindow(_("Analyzing"),_("Analyzing Needed Actions..."));
		}
		tmpcount=0;
		for (zz=0;zz < theInstall->selectedPackageSet->numPackages;zz++) {
			xc=isPackageInstalled(theInstall->installedPackages,NULL,&theInstall->masterPackageSet,theInstall->selectedPackageSet->packages[zz],removemarked);
			okinst=0;
			switch(xc) {
			case ' ':
			case '+':
				okinst=1;
				break;
			case '-':
				if (dodowngrades) {
					okinst=1;
				}
				break;
			default:
				break;
			}
			if (!okinst) {
				logMessage("DOINSTALL: DE-selecting '%s'",theInstall->selectedPackageSet->packages[zz]->name);
				theInstall->selectedPackageSet->packages[zz]->selected=0;
			}
			
		}
		if (autocontrol) {
#ifdef AUTOCONTROL
			autoStatusEnd(controlsocket);
#endif
		} else {
			newtPopWindow();
		}
	}
	if (ftpinstall) {
		logMessage("FTP: Connect: %s (%s : %d), as %s (%s)",f->ftphost,f->proxy ? f->proxy : "No Proxy", f->port,f->username,"password not shown");
		if (!autocontrol) {
			statusWindow(_("FTP"),_("Connecting to FTP Server: '%s' ..."),f->ftphost);
		} else {
#ifdef AUTOCONTROL
			autoStatus(controlsocket,"ftpconn",_("Connecting to FTP Server"));
#endif
		}
		ftpsocket=ftpOpen(f->ftphost,f->username,f->password,f->proxy,f->port);
		if (!autocontrol) {
			newtPopWindow();
		} else {
#ifdef AUTOCONTROL
			autoStatusEnd(controlsocket);
#endif
		}
		if (ftpsocket < 0) {
			logMessage("FTP: Cannot connect to host '%s': %s",f->ftphost,ftpStrerror(ftpsocket));
			if (autocontrol) {
#ifdef AUTOCONTROL
				sprintf(sa,_("FTP Error - Cannot connect to host '%s' : %s"),
					f->ftphost,ftpStrerror(ftpsocket));
				autoErrMsg(controlsocket,sa,1);
#endif
			} else {
				errWindow(FALSE,
					  _("FTP Error - Cannot connect to host '%s' : %s"),
					  f->ftphost,ftpStrerror(ftpsocket));
			}
			return FAILURE;
		}
	}
	for (zz=0;zz < theInstall->selectedPackageSet->numPackages;zz++) {
		if (theInstall->selectedPackageSet->packages[zz]->selected == 0) {
			tmpcount++;
		}
	}
	logMessage("DOINSTALL: Only %d packages to install.",(theInstall->selectedPackageSet->numPackages - tmpcount));
	if (autocontrol) {
#ifdef AUTOCONTROL
		sprintf(sa,"INSTINFO:%d:%d:%s\n",(theInstall->selectedPackageSet->numPackages - tmpcount),getPkgSetSizeInBytes(theInstall->selectedPackageSet), theInstall->language);
		autoMessage(controlsocket,sa);
#endif
	} else {
		swOpen((theInstall->selectedPackageSet->numPackages - tmpcount),getPkgSetSizeInBytes(theInstall->selectedPackageSet), theInstall->language);
	}
	for (zz=0;zz < theInstall->selectedPackageSet->numPackages;zz++) {
		if (theInstall->selectedPackageSet->packages[zz]->selected == 0) continue;
		rpmflags=DEFAULT_RPM_FLAGS;
		if (dodowngrades) {
			rpmflags |= RPMINSTALL_UPGRADETOOLD;
		}
		if (!strcmp(theInstall->selectedPackageSet->packages[zz]->name,"kernel-modules")) {
			logMessage("DOINSTALL: Updating flags for kernel-modules");
			rpmflags=RPM_FLAGS_KERNELMODULES;
		}
		if (shouldIgnoreScripts(theInstall->selectedPackageSet->packages[zz]->name)) {
			rpmflags |= RPMINSTALL_NOSCRIPTS;
		}
		thisPkg=getPackageFilename(theInstall,f,ftpsocket,theInstall->selectedPackageSet->packages[zz]);
		if (thisPkg) {
			swPackage(theInstall->selectedPackageSet->packages[zz]->h);
			logMessage("DOINSTALL: Installing from filename '%s'",thisPkg);
			fd=open(thisPkg,O_RDONLY);
			if (!fd) {
				if (autocontrol) {
#ifdef AUTOCONTROL
					sprintf(sa,"NOINSTALL:%s:%s\n",thisPkg,strerror(errno));
					autoMessage(controlsocket,sa);
#endif
				} else {
					errWindow(FALSE,_("Cannot open RPM package '%s' : %s"),thisPkg,strerror(errno));
				}
				if (ftpinstall) {
					logMessage("CLEANUP: Removing '%s'",thisPkg);
					unlink(thisPkg);
				}
				swPackageComplete();
			}
		} else {
			logMessage("DOINSTALL: Package not found, skipping it.");
		}
		if ((!thisPkg) || (!fd)) {
			logMessage("DOINSTALL: Continuing Loop...");
			continue;
		}
		
		if (!testing) {
#ifdef RPM25
			rc=rpmInstallPackage(
				installer ? "/mnt" : "/",
				db,
				fd,
				NULL,
				rpmflags,
				swCallback,
				NULL);
#else
			rc = rpmInstallPackage(installer ? "/mnt" : "/",db,fd,NULL,rpmflags,swCallback,NULL,NULL);
#endif
			if (rc) {
				logMessage("DOINSTALL: RPM Install error for '%s' : %s",theInstall->selectedPackageSet->packages[zz]->name,rpmErrorString());
				if (autocontrol) {
#ifdef AUTOCONTROL
					sprintf(sa,"NOINSTALL:%s:%s\n",theInstall->selectedPackageSet->packages[zz]->name,rpmErrorString());
					autoMessage(controlsocket,sa);
#endif
				} else {
					if (unattended) {
						statusWindow(_("Error - Please Wait"),
							     _("Error installing '%s' : %s - waiting 5 sec to continue."),theInstall->selectedPackageSet->packages[zz]->name,rpmErrorString());
						sleep(5);
					} else {
						errWindow(FALSE,_("Error installing '%s' : %s"),theInstall->selectedPackageSet->packages[zz]->name,rpmErrorString());
					}
				}
			}
		} else {
			fprintf(testfd,"%d : %s (%d)\n",theInstall->selectedPackageSet->packages[zz]->priority,theInstall->selectedPackageSet->packages[zz]->name,theInstall->selectedPackageSet->packages[zz]->selected);
		}
		close(fd);
		if (ftpinstall) {
			logMessage("CLEANUP: Removing '%s'",thisPkg);
			unlink(thisPkg);
		}
		swPackageComplete();
	}
	swClose();
	rpmdbClose(db);
	if (testing) {
		fclose(testfd);
	}
	if (ftpinstall) {
		logMessage("FTP: Closing connection.");
		ftpClose(ftpsocket);
	}
	logMessage("DOINSTALL: RPM Database closed.");
	return SUCCESS;
}


static void swOpen(int numPackages, int sizePackages, char *language) {
    char buf[50];

    openCenteredWindow(60, 15, testing ? _("Installation Status (TESTING)") : _("Installation Status"));

    si.form = newtForm(NULL, NULL, 0);
    newtFormAddComponent(si.form, newtLabel(1, 1, _("Package:")));
    newtFormAddComponent(si.form, newtLabel(1, 2, _("Size   :")));
    newtFormAddComponent(si.form, newtLabel(1, 3, _("Summary:")));

    si.packageLabel = newtLabel(13, 1, "");
    si.sizeLabel    = newtLabel(13, 2, "");
    si.summaryText  = newtTextbox(13, 3, 45, 2, NEWT_TEXTBOX_WRAP);

    si.pkgScale = newtScale(3, 6, 54, 100);

    newtFormAddComponent(si.form, 
        newtLabel(1,8,_("             Packages       Bytes           Time")));
        /*               12345678901234567890123456789012345678901234567
                                  1         2         3         4 */
    newtFormAddComponent(si.form, newtLabel(1, 9,  _("Total    :")));
    newtFormAddComponent(si.form, newtLabel(1, 10, _("Completed:")));
    newtFormAddComponent(si.form, newtLabel(1, 11, _("Remaining:")));

    si.numPackages = numPackages;
    si.sizePackages = sizePackages;
    si.packagesDone = 0;
    si.sizeDone = 0;
    si.timeStarted = time(NULL);

    sprintf(buf, "%8d", numPackages);
    newtFormAddComponent(si.form, newtLabel(14, 9, buf));
    si.pkgDoneLabel = newtLabel(14, 10, "");
    si.pkgRemainsLabel = newtLabel(14, 11, "");

    sprintf(buf, "%4uM", sizePackages / (1024 * 1024));
    newtFormAddComponent(si.form, newtLabel(29, 9, buf));
    si.sizeDoneLabel = newtLabel(29, 10, "");
    si.sizeRemainsLabel = newtLabel(29, 11, "");

    si.timeTotalLabel = newtLabel(42, 9, "");
    si.timeDoneLabel = newtLabel(42, 10, "");
    si.timeRemainsLabel = newtLabel(42, 11, "");

    if ((language) && (strstr(language,"ja_JP"))) {
	    si.globalScale = newtScale(3, 13, 54, sizePackages);
    } else {
	    si.globalScale = newtScale(1, 13, 58, sizePackages);
    }

    newtFormAddComponents(si.form, si.packageLabel, si.sizeLabel, 
                          si.summaryText, si.pkgScale, si.globalScale, 
                          si.pkgDoneLabel, si.pkgRemainsLabel, 
                          si.sizeDoneLabel, si.sizeRemainsLabel, 
                          si.timeDoneLabel, si.timeRemainsLabel,
                          si.timeTotalLabel, NULL);
}

static void swPackage(Header h) {
    char * name, * version, * release, * summary;
    char buf[50];
#ifdef AUTOCONTROL
    char sa[4096];
#endif
    uint_32 * size;

    headerGetEntry(h, RPMTAG_NAME, NULL, (void *) &name, NULL);
    headerGetEntry(h, RPMTAG_VERSION, NULL, (void *) &version, NULL);
    headerGetEntry(h, RPMTAG_RELEASE, NULL, (void *) &release, NULL);
    headerGetEntry(h, RPMTAG_SIZE, NULL, (void *) &size, NULL);

    if (!headerGetEntry(h, RPMTAG_SUMMARY, NULL, (void *) &summary, NULL))
            summary = _("(no summary)");

    if (autocontrol) {
#ifdef AUTOCONTROL
	    sprintf(sa,
		    "PKGSTART:%s-%s-%s:%d:%s\n",
		    name,
		    version,
		    release,
		    *size,
		    summary);
	    autoMessage(controlsocket,sa);
#endif
    } else {		    
	    sprintf(buf, "%s-%s-%s", name, version, release);
	    newtLabelSetText(si.packageLabel, buf);
	    
	    sprintf(buf, "%dk", (*size) / 1024);
	    newtLabelSetText(si.sizeLabel, buf);
	    
	    newtTextboxSetText(si.summaryText, summary);
	    
	    si.thisPackageSize = *size;
	    
	    newtScaleSet(si.pkgScale, 0);
	    
	    newtDrawForm(si.form);
	    newtRefresh();
    }
}

static void swPackageComplete(void) {
    char buf[50];
    time_t now, finishTime, elapsedTime, remainingTime;

    if (autocontrol) {
#ifdef AUTOCONTROL
	    autoMessage(controlsocket,"PKGCOMPLETE:\n");
	    return;
#endif
    }
    si.packagesDone++;
    si.sizeDone += si.thisPackageSize;

    sprintf(buf, "%8d", si.packagesDone);
    newtLabelSetText(si.pkgDoneLabel, buf);

    sprintf(buf, "%8d", si.numPackages - si.packagesDone);
    newtLabelSetText(si.pkgRemainsLabel, buf);

    sprintf(buf, "%4dM", si.sizeDone / (1024 * 1024));
    newtLabelSetText(si.sizeDoneLabel, buf);

    sprintf(buf, "%4dM", (si.sizePackages - si.sizeDone) / (1024 * 1024));
    newtLabelSetText(si.sizeRemainsLabel, buf);

    now = time(NULL);
    elapsedTime = now - si.timeStarted;
    formatTime(buf, elapsedTime);
    newtLabelSetText(si.timeDoneLabel, buf);

    finishTime = (((float) si.sizePackages) / si.sizeDone) * elapsedTime;
    formatTime(buf, finishTime);
    newtLabelSetText(si.timeTotalLabel, buf);
    
    remainingTime = finishTime - elapsedTime;
    formatTime(buf, remainingTime);
    newtLabelSetText(si.timeRemainsLabel, buf);

    newtScaleSet(si.globalScale, si.sizeDone);

    newtRefresh();
}

static void swCallback(const unsigned long amount, const unsigned long total) {
#ifdef AUTOCONTROL
	char sa[4096];
	int v=0;
#endif
	if (autocontrol) {
#ifdef AUTOCONTROL
		if (total == 0) {
			v=100;
		} else {
			v=((amount * 100) / total);
		}
		sprintf(sa,"PKG:%d\n",v);
		autoMessage(controlsocket,sa);
#endif
	} else {
		if (total == 0)
			newtScaleSet(si.pkgScale, 100);
		else
			newtScaleSet(si.pkgScale, (amount * 100) / total);
		
		newtRefresh();
	}
}

static void swClose(void) {
	if (!autocontrol) {
		newtPopWindow();
	}
}

int installSkeleton(ftpConfig *f) {
	int rc;
	char cmd[256];
	char skelpath[256];
	
	if (ftpinstall) {
		logMessage("SKELETON: Calling getSkelByFTP()");
		rc=getSkelByFTP(f);
	} else {
		if (dosSource == 1) {
			strcpy(skelpath,SKELETON_CDROM_FN2);
		} else {
			strcpy(skelpath,SKELETON_CDROM_FN);
		}
		logMessage("SKELETON: Using '%s'",skelpath);
		if (access(skelpath,R_OK)) {
			rc=FAILURE;
		} else {
			if (symlink(skelpath,SKELETON_LOCAL_FN)) {
				rc = FAILURE;
			} else {
				rc = SUCCESS;
			}
		}
	}
	if (rc == FAILURE) {
		errWindow(FALSE,_("ERROR: Cannot access base system tree!"));
		logMessage("SKELETON: Cannot access skeleton.cgz!");
		return FAILURE;
	}
	statusWindow(_("Installing"),_("Installing Base System Tree..."));
	chdir("/mnt");
	sprintf(cmd, "gunzip < %s | cpio -iumd --quiet",SKELETON_LOCAL_FN);
	logMessage("SKELETON: '%s'",cmd);
	rc=system(cmd);
	if (rc) {
		newtPopWindow();
		logMessage("SKELETON: Unpacking archive failed: %s",strerror(errno));
		errWindow(FALSE,_("ERROR: Cannot unpack base system archive!"));
		return FAILURE;
	}
	unlink(SKELETON_LOCAL_FN);
	newtPopWindow();
	return SUCCESS;
}

int scanFilenames(char *dirname, fnBinding *fblist) {
	DIR *dir;
	struct dirent *ent;
	int fd,rc,isSource,type,count;
	Header h;
	fnBinding *newfb=NULL;
	fnBinding *fbptr=NULL;
	char *name;
#ifdef INSTALLER
	char *group;
#endif
	char *filename;
#ifdef INSTALLER
	FILE *fp;
#endif

	logMessage("scanning %s for package filenames...", dirname);
	dir = opendir(dirname);
	if (!dir) {
		errWindow(TRUE, _("Cannot open directory '%s'!"),dirname);
		return FAILURE;
	}

	errno = 0;
	ent = readdir(dir);
	if (errno) {
		errWindow(TRUE, _("Error reading from directory '%s'!"),dirname);
		closedir(dir);
		return FAILURE;
	}
	filename = alloca(strlen(dirname) + 500);
	statusWindow(_("Scanning"), _("Scanning Available Packages..."));

#ifdef INSTALLER
	fp=fopen(INSTALLER_XPKGS_LIST,"w");
#endif
	while (ent) {
		if (!(ent->d_name[0] == '.' &&
		      (ent->d_name[1] == '\0' || 
		       ((ent->d_name[1] == '.') && (ent->d_name[2] == '\0'))))) {
			sprintf(filename, "%s/%s", dirname, ent->d_name);
			fd = open(filename, O_RDONLY);
			
			if (fd < 0) {
				logMessage("failed to open %s: %s", filename,
					   strerror(errno));
			} else {
				rc = rpmReadPackageHeader(fd, &h, &isSource, NULL, NULL);
				close(fd);
				if (rc) {
					logMessage("failed to rpmReadPackageHeader %s:", 
						   ent->d_name);
				} else {
					headerGetEntry(h, RPMTAG_NAME, &type, (void **) &name, 
						       &count);
#ifdef INSTALLER
					headerGetEntry(h, RPMTAG_GROUP, &type, (void **) &group,
						       &count);
#endif
					newfb=(fnBinding *)malloc(sizeof(fnBinding));
					newfb->next=NULL;
					newfb->packagename=strdup(name);
					newfb->filename=strdup(filename);
					for(fbptr=fblist;fbptr->next;fbptr=fbptr->next);
					fbptr->next=newfb;
#ifdef INSTALLER					
					if ((group) && (strstr(group,"X11")) &&
					    (strstr(group,"Servers"))) {
						fprintf(fp,"%s:%s\n",name,filename);
					}
#endif
				}
			}
			
		}
		errno = 0;
		ent = readdir(dir);
		if (errno) {
			newtPopWindow();
			errWindow(TRUE, _("Error reading from directory!"));
			closedir(dir);
#ifdef INSTALLER
			fclose(fp);
#endif
			return FAILURE;
		}
	}	
	closedir(dir);
#ifdef INSTALLER
	fclose(fp);
#endif
	newtPopWindow();
	filenames_scanned=1;
	return SUCCESS;
}

char *lookupFilename(fnBinding *fblist, char *packagename) {
	fnBinding *fbptr=NULL;

	if ((!fblist) || (!packagename)) return NULL;
	fbptr=fblist->next;
	while(fbptr) {
		if (!strcmp(fbptr->packagename,packagename)) {
			return fbptr->filename;
		}
		fbptr=fbptr->next;
	}
	return NULL;
}
       
