#include "defines.h"
#include "log.h"
#include "autowin.h"
#include "prototypes.h"
#include <rpm/rpmlib.h>
#include <newt.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>

extern int debug;
extern int newtRunning;
extern int ignoreLocal;
extern int autoDeps;
extern int ignoreDependencies;
extern int ftpinstall;
extern int autocontrol;
extern int controlsocket;
extern int filenames_scanned;
extern fnBinding *myfnBindings;

int pkgCompare(void * first, void * second) {
    struct packageInfo ** a = first;
    struct packageInfo ** b = second;

    /* put packages w/o names at the end */
    if (!(*a)->name) return 1;
    if (!(*b)->name) return -1;

    return strcmp((*a)->name, (*b)->name);
};

int skipPackage(char *pkgname) {
	return 0;
}

int getPackageHeaders(char *dirname, pkgSet *psp, int usingNewt) {
    DIR * dir;
    struct dirent * ent;
    int fd, rc, isSource;
    Header h;
    int packagesAlloced;
    pkgSet ps;
    int count, type;
    unsigned int * sizeptr;
    char * name, * group;
    char * filename;
    fnBinding *newfb=NULL;
    fnBinding *fbptr=NULL;

    ps.numPackages = 0;
    packagesAlloced = 5;
    ps.packages = malloc(sizeof(*ps.packages) * packagesAlloced);

    logMessage("scanning %s for packages", dirname);
    dir = opendir(dirname);
    if (!dir) {
	    if (usingNewt) {
		    errWindow(TRUE, _("Cannot open directory!"));
	    }
	    return FAILURE;
    }

    errno = 0;
    ent = readdir(dir);
    if (errno) {
        free(ps.packages);
	if (usingNewt) {
		errWindow(TRUE, _("Error reading from directory!"));
	}
        closedir(dir);
        return FAILURE;
    }

    filename = alloca(strlen(dirname) + 500);

    if (usingNewt) {
	    statusWindow(_("Running"), _("Scanning Available Packages..."));
    }

    while (ent) {
        if (!(ent->d_name[0] == '.' && (ent->d_name[1] == '\0' || 
            ((ent->d_name[1] == '.') && (ent->d_name[2] == '\0'))))) {
            sprintf(filename, "%s/%s", dirname, ent->d_name);
            fd = open(filename, O_RDONLY);

	    if ((!usingNewt) && (debug)) {
		    fprintf(stderr,"%s\n",filename);
	    }
            if (fd < 0) {
                logMessage("failed to open %s: %s", filename,
                           strerror(errno));
            } else {
                rc = rpmReadPackageHeader(fd, &h, &isSource, NULL, NULL);

                close(fd);
                if (rc) {
                    logMessage("failed to rpmReadPackageHeader %s:", 
                                ent->d_name);
                } else {
                    if (ps.numPackages == packagesAlloced) {
                        packagesAlloced += 5;
                        ps.packages = realloc(ps.packages,
                                sizeof(*ps.packages) * packagesAlloced);
                    }

                    ps.packages[ps.numPackages] = 
                                malloc(sizeof(struct packageInfo));
                    ps.packages[ps.numPackages]->h = h;
                    ps.packages[ps.numPackages]->selected = 0;
                    ps.packages[ps.numPackages]->data = strdup(ent->d_name);

                    headerGetEntry(h, RPMTAG_NAME, &type, (void **) &name, 
                                        &count);
		    newfb=(fnBinding *)malloc(sizeof(fnBinding));
		    newfb->next=NULL;
		    newfb->packagename=strdup(name);
		    newfb->filename=strdup(filename);
		    for (fbptr=myfnBindings;fbptr->next;fbptr=fbptr->next);
		    fbptr->next=newfb;
		    
                    if (headerGetEntry(h, RPMTAG_SIZE, &type, (void **) 
                                  &sizeptr, &count))
                        ps.packages[ps.numPackages]->size = *sizeptr;
                    else
                        ps.packages[ps.numPackages]->size = 0;

                    if (skipPackage(name)) 
                        ps.packages[ps.numPackages]->inmenu = 0;
                    else
                        ps.packages[ps.numPackages]->inmenu = 1;

                    if (!headerGetEntry(h, RPMTAG_GROUP, &type, 
                                        (void **) &group, &count)) {
                        group = _("(unknown group)");
                    } else if (!strlen(group)) {
                        group = _("(unknown group)");
                    }

                    ps.packages[ps.numPackages]->name = name;
                    ps.packages[ps.numPackages]->group = group;
        
                    ps.numPackages++;
                }
            }
        }

        errno = 0;
        ent = readdir(dir);
        if (errno) {
		if (usingNewt) {
			newtPopWindow();
			errWindow(TRUE, _("Error reading from directory!"));
		}
		logMessage("Error reading from directory!");
		free(ps.packages);
		closedir(dir);
		return FAILURE;
        }
    } 

    *psp = ps;

    qsort(ps.packages, ps.numPackages, sizeof(*ps.packages), 
          (void *) pkgCompare);
    
    closedir(dir);

    if (usingNewt) {
	    newtPopWindow();
    }

    filenames_scanned=1;
    return SUCCESS;
}

int getPackageHeadersFromHDListDesc(int fd, pkgSet * psp, int noSeek) {
    pkgSet ps;
    int end = 0;
    int type, count;
    unsigned int * sizeptr;
    Header h;
    int packagesAlloced;
    static char *name;
    static char *group;
    int done = 0;

    ps.numPackages = 0;
    packagesAlloced = 5;
    ps.packages = malloc(sizeof(*ps.packages) * packagesAlloced);

    if (!noSeek) {
        count = lseek(fd, 0, SEEK_CUR);
        end = lseek(fd, 0, SEEK_END); 
        lseek(fd, count, SEEK_SET); 
    }

    while (!done) {
        h = headerRead(fd, HEADER_MAGIC_YES);
        if (!h && noSeek) {
            done = 1;
        } else if (!h) {
		if (newtRunning) {
			errWindow(FALSE, _("Error reading header at: %d"),
				  (int) lseek(fd, 0, SEEK_CUR));
		}
		logMessage("Error reading header at: %d",(int)lseek(fd,0,SEEK_CUR));
		free(ps.packages);
		return FAILURE;
        } else {
            headerGetEntry(h, RPMTAG_NAME, &type, (void **) &name, &count);

	    if (debug) {
		    logMessage("READHEADER: Got header for '%s'",name);
	    }
            if (!headerGetEntry(h, RPMTAG_GROUP, &type, (void **) &group, 
                                &count)) {
                group = _("(unknown group)");
            } else if (!strlen(group))
                group = _("(unknown group)");

            if (ps.numPackages == packagesAlloced) {
                packagesAlloced += 5;
                ps.packages = realloc(ps.packages,
                        sizeof(*ps.packages) * packagesAlloced);
            }

            ps.packages[ps.numPackages] = 
                        malloc(sizeof(struct packageInfo));
            ps.packages[ps.numPackages]->h = h;
            ps.packages[ps.numPackages]->selected = 0;

            if (headerGetEntry(h, RPMTAG_SIZE, &type, (void **) &sizeptr, 
                                &count))
                ps.packages[ps.numPackages]->size = *sizeptr;
            else
                ps.packages[ps.numPackages]->size = 0;

            if (skipPackage(name)) 
                ps.packages[ps.numPackages]->inmenu = 0;
            else
                ps.packages[ps.numPackages]->inmenu = 1;

            ps.packages[ps.numPackages]->name = name;
            ps.packages[ps.numPackages]->group = group;
            headerGetEntry(h, FILENAME_TAG, &type,
                     (void **) &ps.packages[ps.numPackages]->data, 
                     &count);

            ps.numPackages++;
        }

        if (!noSeek) {
            if (end <= lseek(fd, 0, SEEK_CUR)) 
                done = 1;
        }
    }

    logMessage("psFromHeaderListDesc read %d headers", ps.numPackages);

    *psp = ps;

    qsort(ps.packages, ps.numPackages, sizeof(*ps.packages), 
          (void *) pkgCompare);
    
    return SUCCESS;
}

int getPackageHeadersFromHDList(char * file, pkgSet * psp) {
    int fd, rc;

    fd = open(file, O_RDONLY, 0644);
    if (fd < 0) {
        errWindow(TRUE, _("Error reading header at:"));
        return FAILURE;
    }

    rc = getPackageHeadersFromHDListDesc(fd,psp,0);
    close(fd);

    return rc;
}


Header getHeaderPointer(pkgSet *m, char *pkgname) {
	static int zz;

	for (zz=0;zz < m->numPackages;zz++) {
		if (!strcmp(m->packages[zz]->name,pkgname)) {
			return m->packages[zz]->h;
		}
	}
	return NULL;
}

int getPackageSize(pkgSet *m, char *pkgname) {
	static int zz;

	for (zz=0;zz < m->numPackages;zz++) {
		if (!strcmp(m->packages[zz]->name,pkgname)) {
			return m->packages[zz]->size;
		}
	}
	return 0;
}

	
int getPackageHeaders2(char *dirname, pkgSet *m, pkgSet *psp) {
	static int zz;

	for (zz=0;zz < psp->numPackages;zz++) {
		if (memberofPkgSet(m,psp->packages[zz]->name)) {
			psp->packages[zz]->h=getHeaderPointer(m,psp->packages[zz]->name);
			psp->packages[zz]->size=getPackageSize(m,psp->packages[zz]->name);
		} else {
			psp->packages[zz]->inmenu=0;
			psp->packages[zz]->size=0;
		}
	}
	return SUCCESS;
}

tlComponentGroup *getGroupByName(tlComponentGroup *grouplist,char *gname) {
	tlComponentGroup *ptr;

	ptr=grouplist->next;
	while(ptr) {
		if (!ptr->name) continue;
		if (!strcmp(ptr->name,gname)) return ptr;
		ptr=ptr->next;
	}
	return NULL;
}

int getGroupSize(tlComponentGroup *grouplist,tlComponentGroup *g) {
	int mysize=0;
/*	char **sptr; */
	static int zz;
/*	tlComponentGroup *gptr=NULL; */

	for (zz=0;zz < g->packageSet.numPackages;zz++) {
		mysize += g->packageSet.packages[zz]->size;
	}
	/*
	sptr=g->requires;
	while (*sptr) {
		gptr=getGroupByName(grouplist,*sptr);
		if (gptr) {
			mysize += getGroupSize(grouplist,gptr);
		}
		sptr++;
	}
	*/
	g->size=mysize;
	return SUCCESS;
}

char **getRPMInfo(pkgSet *pkgset,const char *pkgtext) {
	char **tmp=NULL;
	char *s1=NULL;
	char *s2=NULL;
	char *s3=NULL;
	char *s4=NULL;
	char *s5=NULL;
	char *sx=NULL;
	struct packageInfo *pptr=NULL;
	int rc;
	int xType;
	int xC;
	int sLength;
	int padsize=512;
	static int zz;
	rpmdb db=NULL;
	int myoffset=-1;
	dbiIndexSet myindex;
	Header hTemp;
	char *iVersion=NULL;
	char *iRelease=NULL;

	
	tmp=(char **)malloc(2 * sizeof(char *));
	tmp[0]=NULL;
	tmp[1]=NULL;
	if (autocontrol) {
#ifdef AUTOCONTROL
		s1=strdup(pkgtext);
#endif
	} else {
		sx=strdup(pkgtext);
		strtok(sx,":");
		strtok(NULL,":");
		strtok(NULL,":");
		s1=strtok(NULL,":");
	}
	pptr=getPackageByName(pkgset,s1);
	if (!pptr) return NULL;
	if (!installer && !ignoreLocal) {
		rpmReadConfigFiles(NULL, NULL, NULL, 0);
		if (rpmdbOpen("/", &db, O_RDWR | O_CREAT, 0644))
			db = NULL;
		if (db) {
			if (!rpmdbFindPackage(db,s1,&myindex)) {
				myoffset=myindex.recs[0].recOffset;
				hTemp=rpmdbGetRecord(db,myoffset);
				rc=headerGetEntry(hTemp,RPMTAG_VERSION,&xType,(void **)&iVersion,&xC);
				rc=headerGetEntry(hTemp,RPMTAG_RELEASE,&xType,(void **)&iRelease,&xC);
			}
			rpmdbClose(db);
		}
	}
	rc=headerGetEntry(pptr->h,RPMTAG_NAME,&xType,(void **)&s1,&xC);
	rc=headerGetEntry(pptr->h,RPMTAG_DESCRIPTION,&xType,(void **)&s2,&xC);
	rc=headerGetEntry(pptr->h,RPMTAG_VERSION,&xType,(void **)&s3,&xC);
	rc=headerGetEntry(pptr->h,RPMTAG_RELEASE,&xType,(void **)&s4,&xC);
	rc=headerGetEntry(pptr->h,RPMTAG_COPYRIGHT,&xType,(void **)&s5,&xC);

	if (!s1) {
		s1=strdup(_("N/A"));
	}
	if (!s2) {
		s2=strdup(_("N/A"));
	}
	if (!s3) {
		s3=strdup(_("N/A"));
	}
	if (!s4) {
		s4=strdup(_("N/A"));
	}
	if (!s5) {
		s5=strdup(_("N/A"));
	}
	for (zz=0;zz < strlen(s2);zz++) {
		if (s2[zz] == '\n') {
			s2[zz] = ' ';
		}
	}
	sLength=(strlen(s1)) + (strlen(s3)) + (strlen(s4)) + (strlen(s5));
	if (iVersion && iRelease) {
		sLength += (strlen(iVersion) + strlen(iRelease));
	}
	tmp[0]=(char *)malloc(sLength + padsize);
	if (iVersion && iRelease) {
		sprintf(tmp[0],"%s (%s: %s-%s %s: %s-%s)\n%s : '%s'",s1,_("New"),s3,s4,_("Installed"),iVersion,iRelease,_("Copyright"),s5);
	} else {
		sprintf(tmp[0],"%s (%s-%s)\n%s : '%s'",s1,s3,s4,_("Copyright"),s5);
	}
	tmp[1]=strdup(s2);
	/*
	free(s1);
	free(s2);
	free(s3);
	free(s4);
	free(s5);
	*/
	return tmp;
}

char **getRequires(pkgSet *pkgset,const char *pkgtext) {
	char *s1=NULL;
	char *sx=NULL;
	struct packageInfo *pptr=NULL;
	int xType=0;
	int xC=0;
	int xC2=0;
	int rc;
	void *vptr_names=NULL;
	void *vptr_versions=NULL;
	void *vptr_flags=NULL;
	char **reqlist_names=NULL;
	char **reqlist_versions=NULL;
	int *reqlist_flags=NULL;
	static int zz;
	char **reqlist=NULL;
	char *tmpname=NULL;
	char opstr[10];

	if (autocontrol) {
#ifdef AUTOCONTROL
		s1=strdup(pkgtext);
#endif
	} else {
		sx=strdup(pkgtext);
		strtok(sx,":");
		strtok(NULL,":");
		strtok(NULL,":");
		s1=strtok(NULL,":");
	}
	logMessage("REQUIRES: Finding requirements for '%s'",s1);
	pptr=getPackageByName(pkgset,s1);
	if (!pptr) return NULL;
	rc=headerGetEntry(pptr->h,RPMTAG_NAME,&xType,(void **)&tmpname,&xC);
	rc=headerGetEntry(pptr->h,RPMTAG_REQUIRENAME,&xType,(void **)&vptr_names,&xC);
	reqlist_names=(char **)vptr_names;
	rc=headerGetEntry(pptr->h,RPMTAG_REQUIREVERSION,&xType,(void **)&vptr_versions,&xC2);
	reqlist_versions=(char **)vptr_versions;
	rc=headerGetEntry(pptr->h,RPMTAG_REQUIREFLAGS,&xType,(void **)&vptr_flags,&xC2);
	reqlist_flags=(int *)vptr_flags;
	reqlist=(char **)malloc((xC+1) * sizeof(char *));
	reqlist[xC]=NULL;
	if (!reqlist_names) {
		reqlist[0]=NULL;
		return reqlist;
	}
	for (zz=0;zz < xC;zz++) {
		reqlist[zz]=(char *)malloc(100);
		strcpy(reqlist[zz],reqlist_names[zz]);
		if ((reqlist_versions[zz]) && (strlen(reqlist_versions[zz]))) {
			strcat(reqlist[zz]," ");
			strcpy(opstr,"");
			if (reqlist_flags[zz] & RPMSENSE_LESS) {
				strcat(opstr,"<");
			}
			if (reqlist_flags[zz] & RPMSENSE_GREATER) {
				strcat(opstr,">");
			}
			if (reqlist_flags[zz] & RPMSENSE_EQUAL) {
				strcat(opstr,"=");
			}
			if (reqlist_flags[zz]) {
				strcat(reqlist[zz],opstr);
				strcat(reqlist[zz]," ");
			}
			strcat(reqlist[zz],reqlist_versions[zz]);
		}
		if (debug) {
			logMessage("REQUIRES: %s requires %s",s1,reqlist[zz]);
		}
	}
	return reqlist;
}

char **getProvides(pkgSet *pkgset,const char *pkgtext) {
	char *s1=NULL;
	char *sx=NULL;
	struct packageInfo *pptr=NULL;
	int xType=0;
	int xC=0;
	int rc;
	void *vptr=NULL;
	char **provlist=NULL;
	char **provlist2=NULL;
	static int zz;

	if (autocontrol) {
#ifdef AUTOCONTROL
		s1=strdup(pkgtext);
#endif
	} else {
		sx=strdup(pkgtext);
		strtok(sx,":");
		strtok(NULL,":");
		strtok(NULL,":");
		s1=strtok(NULL,":");
	}
	logMessage("PROVIDES: Finding provides for '%s'",s1);
	pptr=getPackageByName(pkgset,s1);
	if (!pptr) return NULL;
	rc=headerGetEntry(pptr->h,RPMTAG_PROVIDES,&xType,(void **)&vptr,&xC);
	if (debug) {
		logMessage("PROVIDES: %s provides %d things.",s1,xC);
	}
	provlist=(char **)vptr;
/*	provlist=realloc(provlist,(xC + 1)*sizeof(char *)); */
	provlist2=(char **)malloc((xC + 1)*sizeof(char *));
	for (zz=0;zz < xC;zz++) {
		provlist2[zz]=strdup(provlist[zz]);
	}
	provlist2[xC]=NULL;
	for (zz=0;zz < xC;zz++) {
		if (debug) {
			logMessage("PROVIDES: %s provides %s",s1,provlist2[zz]);
		}
	}
	return provlist2;
}

#ifndef INSTALLER
int getCurrentlyInstalled(tlInstall *theInstall) {
	rpmdb db = NULL;
	unsigned int dbindex=0;
	int pindex=0;
	static int zz;
	int *rtype=NULL;
	int *c=NULL;
	char *s0=NULL;	

	if (installer) return SUCCESS;
	if (newtRunning) {
		statusWindow(_("Analyzing"),_("Analyzing System...")); 
	}
	if (!access("/var/lib/rpm/packages.rpm", R_OK)) {
		rpmReadConfigFiles(NULL, NULL, NULL, 0);
		if (rpmdbOpen("/", &db, O_RDWR | O_CREAT, 0644))
			db = NULL;
	}
	if (!db) {
		if (newtRunning) {
			newtPopWindow();
		}
		return FAILURE;
	}

	theInstall->installedPackages=(Header *)malloc(MAX_PACKAGES * sizeof(Header));
	for (zz=0;zz < MAX_PACKAGES;zz++) {
		theInstall->installedPackages[zz]=NULL;
	}
	dbindex=rpmdbFirstRecNum(db);
	while (dbindex > 0) {
		theInstall->installedPackages[pindex]=rpmdbGetRecord(db,dbindex);
		pindex++;
		dbindex=rpmdbNextRecNum(db,dbindex);
	}
	logMessage("Installed Packages: %d of them.",pindex);
	if (debug) {
		for (zz=0;zz < pindex;zz++) {
			headerGetEntry(theInstall->installedPackages[zz],RPMTAG_NAME,rtype,(void **)&s0,c);
			logMessage("\t%s",s0);
		}
	}
	rpmdbClose(db);
	if (newtRunning) {
		newtPopWindow(); 
	}
	return SUCCESS;
}
#endif

int getPackageSetFromSystem(tlInstall *i) {
	pkgSet *tmp=NULL;
	rpmdb db=NULL;
	static int zz;
	unsigned int dbindex=0;
	int pindex=0;
	int xtype=0;
	int xc=0;
	unsigned int *sizeptr;

	if (!i) return FAILURE;
	if (newtRunning) {
		statusWindow(_("Analyzing"),_("Analyzing installed packages..."));
	}
	if (!access("/var/lib/rpm/packages.rpm", R_OK)) {
		rpmReadConfigFiles(NULL, NULL, NULL, 0);
		if (rpmdbOpen("/", &db, O_RDWR | O_CREAT, 0644))
			db = NULL;
	}
	if (!db) return FAILURE;
	/*
	tmp=(pkgSet *)malloc(sizeof(pkgSet));
	*/
	tmp=&i->masterPackageSet;
	if (tmp->numPackages > 0) {
		logMessage("MALLOC: there's already packages in the set I'm about to overwrite.");
	}
	if (!tmp) return FAILURE;
	i->installedPackages=(Header *)malloc(MAX_PACKAGES * sizeof(Header));
	tmp->numPackages=0;
	tmp->packages=(struct packageInfo **)malloc(MAX_PACKAGES * sizeof(struct packageInfo *));
	if (!tmp->packages) return FAILURE;
	for (zz=0;zz < MAX_PACKAGES;zz++) {
		tmp->packages[zz]=(struct packageInfo *)malloc(sizeof(struct packageInfo));
		if (!tmp->packages[zz]) return FAILURE;
	}
	dbindex=rpmdbFirstRecNum(db);
	while (dbindex > 0) {
		tmp->packages[pindex]->h=rpmdbGetRecord(db,dbindex);
		i->installedPackages[pindex]=tmp->packages[pindex]->h;
		tmp->packages[pindex]->name=NULL;
		tmp->packages[pindex]->group=NULL;
		tmp->packages[pindex]->inmenu=1;
		tmp->packages[pindex]->selected=0;
		tmp->packages[pindex]->priority=DEFAULT_PRIORITY;
		headerGetEntry(tmp->packages[pindex]->h,RPMTAG_NAME,&xtype,(void **)&tmp->packages[pindex]->name,&xc);
		headerGetEntry(tmp->packages[pindex]->h,RPMTAG_GROUP,&xtype,(void **)&tmp->packages[pindex]->group,&xc);
		headerGetEntry(tmp->packages[pindex]->h,RPMTAG_SIZE,&xtype,(void **)&sizeptr,&xc);
		tmp->packages[pindex]->size=*sizeptr;
		tmp->numPackages++;
		pindex++;
		dbindex=rpmdbNextRecNum(db,dbindex);
	}
	rpmdbClose(db);
	if (newtRunning) {
		newtPopWindow();
	}
	return SUCCESS;
}

Header getHeaderByName(Header *hlist, char *pkgname) {
	Header tmp;
	int hindex=0;
	char *tmpname=NULL;
	int xtemp,xc;

	tmp=hlist[hindex];
	while(tmp) {
		headerGetEntry(tmp,RPMTAG_NAME,&xtemp,(void **)&tmpname,&xc);
		if ((tmpname) && (!strcmp(tmpname,pkgname))) {
			return tmp;
		}
		hindex++;
		tmp=hlist[hindex];
	}
	return NULL;
}

int verifyDependencies(pkgSet *psp, Header *ihlist, char **rlist) {
    rpmdb db = NULL;
    rpmDependencies rpmdeps;
    int i;
    struct rpmDependencyConflict * conflicts;
    struct packageInfo * package=NULL;
    int numConflicts;
    newtComponent okay, listbox, cancel, form;
    char buf[80];
    char selectPackages;
    int anyrealconflicts=0;
    int fixup=autoDeps;
    selectionList list=NULL;
    char **udeps=(char **)malloc(MAX_PACKAGES * sizeof(char *));
    int udindex=0;
    static int zz;
    int rc;
    char *s0;
    char s1[256];
    dbiIndexSet myindex;
    int myoffset=0;
#ifdef AUTOCONTROL
    char sa[4096];
#endif

    if (ignoreDependencies) return SUCCESS;
    if (autocontrol) {
#ifdef AUTOCONTROL
	    autoStatus(controlsocket,"verifydeps",_("Verifying Dependencies..."));
#endif
    } else {
	    statusWindow(_("Dependencies"),_("Verifying Dependencies..."));
    }
    logMessage("DEPENDENCIES: Verifying Dependencies...");
    for (zz=0;zz < MAX_PACKAGES;zz++) {
	    udeps[zz]=NULL;
    }
    if ((!installer) && (!ignoreLocal)) {
	    if (!access("/var/lib/rpm/packages.rpm", R_OK)) {
		    rpmReadConfigFiles(NULL, NULL, NULL, 0);
		    if (rpmdbOpen("/", &db, O_RDWR | O_CREAT, 0644)) {
			    db = NULL;
		    }
	    }
    }

    rpmdeps = rpmdepDependencies(db);

    for (i = 0; i < psp->numPackages; i++) {
	    if (psp->packages[i]->selected) {
		    if (memberofStrArray(rlist,psp->packages[i]->name)) {
			    if (!rpmdbFindPackage(db,psp->packages[i]->name,&myindex)) {
				    myoffset=myindex.recs[0].recOffset;
				    rpmdepRemovePackage(rpmdeps,myoffset);
				    psp->packages[i]->selected=0;
				    logMessage("DEPS: Removed '%s' from rpmdeplist.",psp->packages[i]->name);
			    }
		    } else {
			    rpmdepAddPackage(rpmdeps, psp->packages[i]->h, psp->packages[i]);
		    }
	    } else {
		    if (memberofStrArray(rlist,psp->packages[i]->name)) {
			    if (!rpmdbFindPackage(db,psp->packages[i]->name,&myindex)) {
				    myoffset=myindex.recs[0].recOffset;
				    rpmdepRemovePackage(rpmdeps,myoffset);
				    logMessage("DEPS: Removed '%s' from rpmdeplist.",psp->packages[i]->name);
			    }
		    } else {
			    rpmdepAvailablePackage(rpmdeps, psp->packages[i]->h, 
						   psp->packages[i]);
		    }
	    }
    }

    rpmdepCheck(rpmdeps, &conflicts, &numConflicts);

    rpmdepDone(rpmdeps);
    if (db) rpmdbClose(db);

    if (!numConflicts) {
        return SUCCESS;
    }

    if (fixup) {
        for (i = 0; i < numConflicts; i++) {
		if (conflicts[i].suggestedPackage) {
			package = conflicts[i].suggestedPackage;
			package->selected = 1;
			if (package->priority == 0) {
				package->priority = DEFAULT_PRIORITY;
			}
			logMessage("AUTODEPS: '%s' requires '%s', adding.",conflicts[i].byName,package->name);
		} else {
			if (conflicts[i].needsName[0] != '/') {
				s0=(char *)malloc(strlen(conflicts[i].byName) + strlen(conflicts[i].needsName) + 10);
				logMessage("AUTODEPS: unresolvable: '%s' requires '%s'.",conflicts[i].byName,conflicts[i].needsName);
				sprintf(s1,"%s",conflicts[i].needsName);
				if ((conflicts[i].needsVersion) && (strlen(conflicts[i].needsVersion))) {
					sprintf(s1,"%s-%s",s1,conflicts[i].needsVersion);
				}
				sprintf(s0,"%s\001%s",s1,conflicts[i].byName);
				if (!memberofStrArray(udeps,s0)) {
					udeps[udindex]=strdup(s0);
					udindex++;
				}
				free(s0);
			}
		}
        }
	if (!udindex) {
		rpmdepFreeConflicts(conflicts, numConflicts);
		if (autocontrol) {
#ifdef AUTOCONTROL
			autoStatusEnd(controlsocket);
#endif
		} else {
			newtPopWindow();
		}
		return SUCCESS;
	} else {
		logMessage("DEPS: Unresolvable dependencies exist!");
		if (autocontrol) {
#ifdef AUTOCONTROL
			autoStatusEnd(controlsocket);
#endif
		} else {
			newtPopWindow();
		}
		rc=showUnresolvableDependencies(udeps);
		return rc;
	}
    }
    if (autocontrol) {
#ifdef AUTOCONTROL
	    autoStatusEnd(controlsocket);
#endif
    } else {
	    newtPopWindow();
    }
    for (i = 0; i < numConflicts; i++) {
        package = conflicts[i].suggestedPackage;
        if (conflicts[i].needsName[0] != '/') {                 
                if (package) {
                        sprintf(buf, "%-20s %-20s %-20s",
                            conflicts[i].byName,conflicts[i].needsName,
                            conflicts[i].needsVersion);
                        anyrealconflicts=1;
                } else {
                        sprintf(buf, "%-20s %-20s %-20s",
                            conflicts[i].byName,conflicts[i].needsName,
                            conflicts[i].needsVersion);
                        anyrealconflicts=1;
                }
		if (autocontrol) {
#ifdef AUTOCONTROL
			sprintf(sa,"UDEP:%s:%s:%s:%s\n",conflicts[i].byName,
				conflicts[i].needsName,
				conflicts[i].needsVersion,
				package ? package->name : "0");
			autoMessage(controlsocket,sa);
#endif
		} else {
			addToSelectionList(&list,strdup(buf),buf,NULL);
		}
        }
    }
    if (!anyrealconflicts) {
            return SUCCESS;
    }

    if (!autocontrol) {
	    form = askUser(ask_keep, _("Unresolved Dependencies"), okay,
			   compGroup(vertical, center,
				     groupMargine(0, 1, compGroup(horizontal, topLeft,
								  properTextbox(_(
									  "Some of the packages you have selected to install "
									  "require packages you have not selected. If you just "
									  "select Ok all of those required packages will be "
									  "installed.")),
								  NULL)),
				     properTextbox(
					     "%-20s %-20s %-20s", 
					     _("Package"), _("Requirement"), _("Required Ver")),
				     selectionListbox(&listbox,FALSE,5,0,list),
				     newtCheckbox(0, 0, 
						  _("Install Packages to Satisfy Dependencies"), '*',
						  NULL, &selectPackages),
				     groupMargine(0, 1, compGroup(horizontal, bottomRight,
								  okay   = newtButton(0, 0, _("Ok")),
								  cancel = newtButton(0, 0, _("Cancel")),
								  NULL)),
				     NULL));
	    
	    if (newtFormGetCurrent(form) == cancel) {
		    free(conflicts);
		    releaseForm(&form);
		    return CANCELLED;
	    }
    } else {
	    selectPackages = ' ';
    }

    if (selectPackages != ' ') {
        for (i = 0; i < numConflicts; i++) {
            package = conflicts[i].suggestedPackage;
            if (package) package->selected = 1;
        }
    }

    free(conflicts);
    if (!autocontrol) {
	    releaseForm(&form);
    }
    return SUCCESS;
}


struct headerTagTableEntry headerChecks[3] = {
	{ "RPMTAG_NAME",1 },
	{ "RPMTAG_VERSION",1 },
	{ "RPMTAG_RELEASE",1 }
};

char isPackageInstalled(Header *hlist, char **mycachelist, pkgSet *mpkgset, struct packageInfo *p, char **rlist) {
	Header hptr;
	int hindex=0;
	int rc=0;
	char *name0=NULL;
	char *name1=NULL;
	char *ver0=NULL;
	char *ver1=NULL;
	char *irel0=NULL;
	char *irel1=NULL;
	char *rel0=NULL;
	char *rel1=NULL;
#ifndef INSTALLER
	int *ser0;
	int *ser1;
	int src0,src1;
#endif
	int xtype;
	int xc;
	char crc=' ';
	int clindex=0;
	char *stmp=NULL;
	char *stmp2=NULL;
	int one_has_tl=0;

	if (installer || ignoreLocal) return ' ';

	if (!hlist) return ' ';
	if (rlist) {
		if (memberofStrArray(rlist,p->name)) {
			return 'R';
		}
	}
	hptr=hlist[0];
	if (debug) {
		logMessage("INST: Checking installation of '%s'...",p->name);
	}
	if (mycachelist) {
		stmp=memberofStrArray2(mycachelist,p->name);
		if (stmp) {
			if (debug) {
				logMessage("INST-CACHE: Cache hit for package '%s'",p->name);
			}
			return stmp[0];
		}
	}
	
	if (!memberofPkgSet(mpkgset,p->name)) {
		return ' ';
	}
	if (!p->h) {
		return ' ';
	}
	if (mycachelist) {
		while (mycachelist[clindex]) {
			clindex++;
		}
	}
	while (hptr) {
		rc=headerGetEntry(hptr,RPMTAG_NAME,&xtype,(void **)&name0,&xc);
		rc=headerGetEntry(hptr,RPMTAG_VERSION,&xtype,(void **)&ver0,&xc);
		rc=headerGetEntry(hptr,RPMTAG_RELEASE,&xtype,(void **)&rel0,&xc);
#ifndef INSTALLER
		src0=headerGetEntry(hptr,RPMTAG_SERIAL,&xtype,(void **)&ser0,&xc);
#endif
		rc=headerGetEntry(p->h,RPMTAG_NAME,&xtype,(void **)&name1,&xc);
		rc=headerGetEntry(p->h,RPMTAG_VERSION,&xtype,(void **)&ver1,&xc);
		rc=headerGetEntry(p->h,RPMTAG_RELEASE,&xtype,(void **)&rel1,&xc);
#ifndef INSTALLER		
		src1=headerGetEntry(p->h,RPMTAG_SERIAL,&xtype,(void **)&ser1,&xc);
#endif
		if ((!name0) || (!name1)) continue;
		if (!strcmp(name0,name1)) {
#ifndef INSTALLER
			if ((src1) && (!src0)) {
				crc='+';
			}
			else if ((src1) && (src0)) {
				if (*ser1 > *ser0) crc='+';
				else if (*ser1 < *ser0) crc='-';
				else crc='=';
				logMessage("SERIAL: Setting crc to '%c' for pkg '%s'",crc,name0);
			}
#else
			if (0) {
			}
#endif
			else {
				if (pkgVersionCompare(ver1,'>',ver0)) crc='+';
				else if (pkgVersionCompare(ver1,'<',ver0)) crc='-';
				else if (pkgVersionCompare(ver1,'=',ver0)) {
					irel0=strdup(rel0);
					irel1=strdup(rel1);
					if ((strstr(rel0,"TL")) || (strstr(rel1,"TL"))) {
						if (!((strstr(rel0,"TL")) &&
						      (strstr(rel1,"TL")))) {
							one_has_tl=1;
						}
						stmp2=index(irel0,'T');
						if (stmp2) {
							stmp2[0]='\0';
						}
						stmp2=index(irel1,'T');
						if (stmp2) {
							stmp2[0]='\0';
						}
					}
					if (pkgVersionCompare(irel1,'>',irel0)) crc='+';
					else if (pkgVersionCompare(irel1,'<',irel0)) crc='-';
					else crc='=';
					if ((crc == '=') && (one_has_tl == 1)) {
						if (strstr(rel1,"TL")) {
							crc='+';
						}
						else if (strstr(rel0,"TL")) {
							crc='-';
						}
					}
					free(irel0);
					free(irel1);
				}
			}
		}
		hindex++;
		hptr=hlist[hindex];
		/*
#ifndef INSTALLER
		free(ser0);
		free(ser1);
		ser0=NULL;
		ser1=NULL;
#endif
		free(name0);
		free(name1);
		free(ver0);
		free(ver1);
		free(rel0);
		free(rel1);
		name0=NULL;
		name1=NULL;
		ver0=NULL;
		ver1=NULL;
		rel0=NULL;
		rel1=NULL;
		*/
	}
	if (mycachelist) {
		mycachelist[clindex]=(char *)malloc(strlen(name1) + 2);
		sprintf(mycachelist[clindex],"%c%s",crc,name1);
		if (debug) {
			logMessage("INST-CACHE: Adding entry '%s' at index '%d'",name1,clindex);
		}
	}
	return crc;
}

int pkgVersionCompare(const char *iv1, const char operator, const char *iv2) {
    char *s1=strdup(iv1);
    char *s2=strdup(iv2);
    int checktype=0;  /* 1 = return on first fail, 2 = return on first success */
    int isAlpha=0;
    char *stmp=NULL;
    char **v1=(char **)malloc(MAX_VERSION_FIELDS * sizeof(char *));
    char **v2=(char **)malloc(MAX_VERSION_FIELDS * sizeof(char *));
    static int zz;
    int vindex=0;
    int a1,a2;
    int xx,yy;

    for (zz=0;zz < MAX_VERSION_FIELDS;zz++) {
	    v1[zz]=NULL;
	    v2[zz]=NULL;
    }
    v1[0]=strdup(strtok(s1,"."));
    vindex++;
    while ((stmp=strtok(NULL,"."))) {
	    v1[vindex]=strdup(stmp);
	    vindex++;
    }
    a1=vindex;
    vindex=0;
    stmp=NULL;
    v2[0]=strdup(strtok(s2,"."));
    vindex++;
    while ((stmp=strtok(NULL,"."))) {
	    v2[vindex]=strdup(stmp);
	    vindex++;
    }
    a2=vindex;
    if (a2 > a1) {
	    for(zz=a1;zz < a2;zz++) {
		    v1[zz]=strdup("0");
	    }
	    a1=a2;
    }
    
    for (zz=0;zz < a1;zz++) {
	    if (!v2[zz]) {
		    v2[zz]=strdup("0");
	    }
	    if ((containsNonDigits(v1[zz])) || (containsNonDigits(v2[zz]))) {
		    isAlpha=1;
		    if (operator == '>') {
			    checktype=2;
		    }
		    if (operator == '<') {
			    checktype=2;
		    }
		    if (operator == '=') {
			    checktype=1;
		    }
		    if (operator == '!') {
			    checktype=2;
		    }
	    } else {
		    isAlpha=0;
		    if (operator == '>') {
			    checktype=2;
		    }
		    if (operator == '<') {
			    checktype=2;
		    }
		    if (operator == '=') {
			    checktype=1;
		    }
		    if (operator == '!') {
			    checktype=2;
		    }
	    }
	    if (isAlpha == 1) {
		    if (!v2[zz]) {
			    xx=1;
		    } else {
			    xx=strcmp(v1[zz],v2[zz]);
			    if (xx > 0) xx=1;
			    if (xx < 0) xx=-1;
		    }
	    } else {
		    if (!v2[zz]) {
			    xx=1;
		    } else {
			    if (atoi(v1[zz]) > atoi(v2[zz])) xx=1;
			    else if (atoi(v2[zz]) > atoi(v1[zz])) xx=-1;
			    else xx=0;
		    }
	    }
	    switch(operator) {
	    case '>':
		    if (xx == 1) {
			    yy=1;
		    } else if (xx == -1) {
			    yy=-1;
		    } else {
			    yy=0;
		    }
		    break;
	    case '<':
		    if (xx == -1) {
			    yy=1;
		    } else if (xx == 1) {
			    yy=-1;
		    } else {
			    yy=0;
		    }
		    break;
	    case '=':
		    if (xx == 0) {
			    yy=1;
		    } else {
			    yy=0;
		    }
		    break;
	    default:
		    yy=0;
		    break;
	    }
		    
	    if (checktype == 1) {
		    if (yy == 0) {
			    return 0;
		    }
		    if (zz == (a1 - 1)) {
			    return 1;
		    }
	    }
	    if (checktype == 2) {
		    if (yy == 1) {
			    return 1;
		    }
		    if (yy == -1) {
			    return 0;
		    }
		    if (zz == (a1 - 1)) {
			    return 0;
		    }
	    }
    }
    for (zz=0;zz < a1;zz++) {
	    free(v1[zz]);
	    free(v2[zz]);
    }
    free(v1);
    free(v2);
    free(s1);
    free(s2);
    return 0;	
}

int addToRPMGroupList(tlComponentGroup *list, char *groupname, Header h) {
	tlComponentGroup *gptr=NULL;
	static int zz;
	int rc;
	
	if (!memberofComponentGroup(list,groupname)) {
		for (gptr=list;gptr->next;gptr=gptr->next);
		gptr->next=(tlComponentGroup *)malloc(sizeof(tlComponentGroup));
		gptr->next->next=NULL;
		gptr->next->name=strdup(groupname);
		gptr->next->installtypes=strdup("00000001");
		gptr->next->flags=CGROUP_FLAG_INMENU;
		gptr->next->priority=DEFAULT_PRIORITY;
		gptr->next->helptext=NULL;
		gptr->next->packageSet.numPackages=0;
		gptr->next->packageSet.packages=(struct packageInfo **)malloc(MAX_PPG * sizeof(struct packageInfo));
		for (zz=0;zz < MAX_PPG;zz++) {
			gptr->next->packageSet.packages[zz]=NULL;
			gptr->next->requires[zz]=NULL;
			gptr->next->suggests[zz]=NULL;
			gptr->next->conflicts[zz]=NULL;
		}
	}
	for (gptr=list->next;strcmp(gptr->name,groupname);gptr=gptr->next);
	rc=addPackageToSetByHeader(&gptr->packageSet,h,DEFAULT_PRIORITY,0,1);
	if (rc == FAILURE) return FAILURE;
		
	return SUCCESS;
}

int compareGptrs(const void *v1, const void *v2) {
	tlComponentGroup **pg1=(tlComponentGroup **)v1;
	tlComponentGroup **pg2=(tlComponentGroup **)v2;
	tlComponentGroup *g1=*pg1;
	tlComponentGroup *g2=*pg2;
	int x;

	if (!v1) return -1;
	if (!v2) return 1;
	if (debug) {
		logMessage("DEBUG: GPTRS: comparing '%s' to '%s'",g1->name,g2->name);
	}
	x=strcmp(g1->name,g2->name);
	return x;
}
	

tlComponentGroup *getRPMGroups(pkgSet *i) {
	tlComponentGroup *newlist=NULL;
	tlComponentGroup **gArray=NULL;
	tlComponentGroup *gptr=NULL;
	int gindex=0;
	int numgroups=0;
	static int zz;
	int xtype,xc;
	char *tmpgroup=NULL;
	int rc;

	newlist=(tlComponentGroup *)malloc(sizeof(tlComponentGroup));
	newlist->next=NULL;
	if (!newlist || !i) return NULL;
	for (zz=0;zz < i->numPackages;zz++) {
		headerGetEntry(i->packages[zz]->h,RPMTAG_GROUP,&xtype,(void **)&tmpgroup,&xc);
		logMessage("RPMGROUP: '%s'",tmpgroup);
		rc=addToRPMGroupList(newlist,tmpgroup,i->packages[zz]->h);
		if (rc == FAILURE) return NULL;
	}
	logMessage("Finished with RPM groups - sorting...");
	gArray=(tlComponentGroup **)malloc(MAX_GROUPS * sizeof(tlComponentGroup *));
	for (zz=0;zz < MAX_GROUPS;zz++) {
		gArray[zz]=NULL;
	}
	gptr=newlist->next;
	while(gptr) {
		gArray[gindex]=gptr;
		gindex++;
		gptr=gptr->next;
	}
	numgroups=gindex;
	qsort(gArray,numgroups,sizeof(tlComponentGroup *),compareGptrs);
	newlist->next=NULL;
	for (zz=0;zz < numgroups;zz++) {
		for (gptr=newlist;gptr->next;gptr=gptr->next);
		gptr->next=gArray[zz];
		gptr->next->next=NULL;
	}
	return newlist;
}

int extractPictureFromPackage(tlInstall *i, char *pkgname, char *outbasefn) {
	rpmdb db=NULL;
	dbiIndexSet myindex;
	int myoffset=-1;
	Header hTemp;
	int ofd;
	int rfd;
	void *gifdata=NULL;
	void *xpmdata=NULL;
	char filename[128];
	int xType;
	int xC;
	int rcGif;
	int rcXpm;
	int rc;
	int dbFound=0;
	int isSource;
	char *rpmfn=NULL;

	rpmReadConfigFiles(NULL, NULL, NULL, 0);
	if (rpmdbOpen("/", &db, O_RDWR | O_CREAT, 0644))
		db = NULL;
	if (db) {
		if (!rpmdbFindPackage(db,pkgname,&myindex)) {
			myoffset=myindex.recs[0].recOffset;
			hTemp=rpmdbGetRecord(db,myoffset);
			rcGif=headerGetEntry(hTemp,RPMTAG_GIF,&xType,(void **)&gifdata,&xC);
			rcXpm=headerGetEntry(hTemp,RPMTAG_XPM,&xType,(void **)&xpmdata,&xC);
			dbFound=1;
		}
		rpmdbClose(db);
	}
	if (!dbFound) {
		if (ftpinstall) return 0;
		rpmfn=getFilenameByPkgName(i->rpmdir,pkgname);
		if (!rpmfn) return 0;
		rfd=open(rpmfn,O_RDONLY);
		if (rfd < 0) return 0;
		rc = rpmReadPackageHeader(rfd, &hTemp, &isSource, NULL, NULL);
		close(rfd);
		if (rc) return 0;
		rcGif=headerGetEntry(hTemp,RPMTAG_GIF,&xType,(void **)&gifdata,&xC);
		rcXpm=headerGetEntry(hTemp,RPMTAG_XPM,&xType,(void **)&xpmdata,&xC);
	}

	strcpy(filename,outbasefn);
	if (gifdata) {
		strcat(filename,".gif");
	} else if (xpmdata) {
		strcat(filename,".xpm");
	} else {
		return 0;
	}
	ofd=open(filename, O_CREAT | O_WRONLY, 0644);
	if (!ofd) return 0;
	if (gifdata) {
		write(ofd,gifdata,xC);
	} else if (xpmdata) {
		write(ofd,xpmdata,xC);
	}
	close(ofd);
	if (gifdata) {
		return 1;
	} else if (xpmdata) {
		return 2;
	}
	return 0;
}
	
