#include "defines.h"
#include "config.h"
#include "autowin.h"
#include "log.h"
#include "prototypes.h"
#include <errno.h>
#include <fcntl.h>
#include <rpm/rpmlib.h>
#include <sys/stat.h>                /* for mkdir(2) ?!? */
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#ifndef GLIBC
#include <linux/fs.h>
#endif
#include <linux/cdrom.h>
#include <sys/mount.h>


extern int debug;
extern int doSuggest;
extern int ignoreConflicts;
extern int autoDeps;
extern int groupSelectType;
extern int returnReason;
extern int ignoreLocal;
extern int installer;
extern int ftpinstall;
extern int newtRunning;
extern int progmountedcdrom;
extern int progmountedfloppy;
extern int nocompsfile;
extern int norpmdir;
extern int nohdlist;
extern int ejectok;
extern int autocontrol;
extern int controlsocket;
extern int basepath_selected;

char *tryFilePaths[] = {
	"/TurboLinux/base/",
	"/TurboUpgrade/base/",
	"/base/",
	"/",
	NULL,
};

char *tryDirs[] = {
	"/TurboLinux/RPMS/",
	"/TurboUpgrade/RPMS/",
	"/RPMS/",
	ARCH_PATH,
	"/",
	NULL,
};

int askLocations(tlInstall *i, int comps, int rpms, int hdlist) {
	char *newComps;
	char *newRPMs;
	char *newHDList;

	newComps=strdup(COMPSFILE);
	newRPMs=strdup(RPMSDIR);
	newHDList=strdup(HDLIST);
	
	if (!comps && !rpms && !hdlist) {
		logMessage("We don't need to ask for locations.");
		return SUCCESS;
	}
	if (comps) {
		newComps=pathExpandHomeDir(entryWindow(_("Components File"),60,newComps,_("Ok"),_("Cancel"),_("Enter the location of the components file.  If you do not have one, simply hit ENTER and then select 'Ok'.")));
		if ((!newComps) || (access(newComps,R_OK))) {
			free(i->compsfile);
			i->compsfile=NULL;
			nocompsfile=1;
		} else {
			free(i->compsfile);
			i->compsfile=strdup(newComps);
		}
		if (!newComps) {
			returnReason=1;
			return CANCELLED;
		}
	}
	if (rpms) {
		newRPMs=pathExpandHomeDir(entryWindow(_("RPMs Directory"),60,newRPMs,_("Ok"),_("Cancel"),_("Enter the location of the RPMs directory.  If you do not have one, simply hit ENTER and then select 'Ok'.")));
		if ((!newRPMs) || (access(newRPMs,F_OK))) {
			norpmdir=1;
			free(i->rpmdir);
			i->rpmdir=NULL;
		} else {
			free(i->rpmdir);
			i->rpmdir=strdup(newRPMs);
		}
		if (!newRPMs) {
			returnReason=1;
			return CANCELLED;
		}
	}
	if (hdlist) {
		newHDList=pathExpandHomeDir(entryWindow(_("HDList File"),60,newHDList,_("Ok"),_("Cancel"),_("Enter the location of the RPM Header List file.  If you do not have one, simply hit ENTER and then select 'Ok'.")));
		if ((!newHDList) || (access(newHDList,R_OK))) {
			free(i->hdlist);
			i->hdlist=NULL;
			nohdlist=1;
		} else {
			free(i->hdlist);
			i->hdlist=strdup(newHDList);
		}
		if (!newHDList) {
			returnReason=1;
			return CANCELLED;
		}
	}
	return SUCCESS;
}

int getLocationsByBasePath(tlInstall *i, char *path) {
	char **sptr=NULL;
	char s0[256];
	int comps;
	int rpms;
	int hdlist;
	int rc;
	bool choice;
	int canAskLocations=1;

	comps=1;
	rpms=1;
	hdlist=1;
	sptr=tryFilePaths;
	while (*sptr) {
		sprintf(s0,"%s/%s/hdlist",path,*sptr);
		if ((!access(s0,R_OK)) && (hdlist)) {
			free(i->hdlist);
			i->hdlist=strdup(s0);
			hdlist=0;
			logMessage("BASEPATH: hdlist found as '%s'",i->hdlist);
		}
		sprintf(s0,"%s/%s/comps",path,*sptr);
		if ((!access(s0,R_OK)) && (comps)) {
			free(i->compsfile);
			i->compsfile=strdup(s0);
			comps=0;
			logMessage("BASEPATH: compsfile found as '%s'",i->compsfile);
		}
		sptr++;
	}	
	sptr=tryDirs;
	while (*sptr) {
		sprintf(s0,"%s/%s",path,*sptr);
		logMessage("BASEPATH: Trying RPM dir '%s'",s0);
		if ((!access(s0,F_OK)) && (rpms)) {
			free(i->rpmdir);
			i->rpmdir=strdup(s0);
			rpms=0;
			logMessage("BASEPATH: RPM dir found as '%s'",i->rpmdir);
		}
		sptr++;
	}
	if (!autocontrol) {
		if (comps || rpms || hdlist) {
			if (!basepath_selected) {
				msgWindow(_("File(s) Not Found"),_("Ok"),_("One or more support files were not found under the tree starting at the specified base path.  You will have to either manually enter the path(s) to these file(s), or, if the files do not exist and you wish for them to be ignored, simply select 'OK' at the following dialogs."));
			} else {
				choice=yesnoWindow(_("Confirm Paths"),
						   TRUE,
						   _("Ok"),
						   _("Incorrect"),
						   _("Please confirm the correctness of these selections:\n\n"
						     "Components File: %s\n"
						     "Header List    : %s\n"
						     "RPM Directory  : %s\n"),
						   comps ? _("Ignored") : i->compsfile,
						   hdlist ? _("Ignored") : i->hdlist,
						   rpms ? _("Ignored") : i->rpmdir);
				if (choice == TRUE) {
					if (comps) nocompsfile=1;
					if (hdlist) nohdlist=1;
					if (rpms) norpmdir=1;
					canAskLocations=0;
				}
				if (choice == FALSE) {
					comps=1;
					rpms=1;
					hdlist=1;
				}
			}
		}
		if (canAskLocations) {
			rc=askLocations(i,comps,rpms,hdlist);
		} else {
			rc=SUCCESS;
		}
		return rc;
	} else {
		/*
		if (comps || rpms || hdlist) {
			return FAILURE;
		}
		*/
		if (comps) {
			free(i->compsfile);
			i->compsfile=strdup(_("Ignored"));
		}
		if (hdlist) {
			free(i->hdlist);
			i->hdlist=strdup(_("Ignored"));
		}
		if (rpms) {
			free(i->rpmdir);
			i->rpmdir=strdup(_("Ignored"));
		}
	}
	return SUCCESS;
}

int cdromMounted(void) {
	FILE *fp;
	char s0[256];

	if (testing) return 1;
	fp=fopen("/etc/mtab","r");
	while(fgets(s0,256,fp)) {
		if (strstr(s0,"/mnt/cdrom")) return 1;
	}
	fclose(fp);
	return 0;
}

int askCDROM(void) {
	FILE *fp;
	char s0[256];
	char s1[40];
	char s2[40];
	char s3[40];
#ifdef AUTOCONTROL
	char stmp[256];
#endif
	char *cddev=NULL;
	char *cdmountpoint=NULL;
	bool choice;
	int rc;
	int fd;
	char upc[8];
	unsigned long rwflag = MS_MGC_VAL | MS_RDONLY;

	fp=fopen("/etc/fstab","r");
	while(fgets(s0,256,fp)) {
		if (s0[0] == '/') {
			sscanf(s0,"%s %s %s",s1,s2,s3);
			if (!strcmp(s3,"iso9660")) {
				cddev=strdup(s1);
				cdmountpoint=strdup(s2);
			}
		}
	}
	fclose(fp);
	if (!cddev) return FAILURE;
	if (strcmp(cdmountpoint,"/mnt/cdrom")) return SUCCESS;
	if (autocontrol) {
		choice=TRUE;
	} else {
		choice=yesnoWindow(_("Mount CD-ROM?"),TRUE,_("Yes"),_("No"),_("Do you want to mount a CD-ROM at this time?  If so, put the CD into the drive and select 'Yes'."));
	}
	if (choice == TRUE) {
		if (autocontrol) {
#ifdef AUTOCONTROL
			autoStatus(controlsocket,"mounting",_("Mounting CD-ROM..."));
#endif
		} else {
			statusWindow(_("Mounting"),_("Mounting CD-ROM..."));
		}
		if (access("/mnt/cdrom",F_OK)) {
			rc=mkdir("/mnt/cdrom",0755);
			if (rc < 0) {
				if (autocontrol) {
#ifdef AUTOCONTROL
					autoStatusEnd(controlsocket);
					sprintf(stmp,_("Could not create mountpoint /mnt/cdrom: %s"),strerror(errno));
					autoErrMsg(controlsocket,stmp,0);
#endif
				} else {
					newtPopWindow();
					errWindow(FALSE,_("Could not create mountpoint /mnt/cdrom: %s"),strerror(errno));
				}
				logMessage("Could not create mountpoint /mnt/cdrom: %s",strerror(errno));
				return FAILURE;
			}
		}
		rc=mount(cddev,cdmountpoint,"iso9660",rwflag,NULL);
		if (rc < 0) {
			fd=open(cddev,O_RDONLY);
			if (ioctl(fd, CDROM_GET_UPC, &upc)) {
				newtPopWindow();
				errWindow(FALSE,_("Could not mount CD-ROM: no CD present!"));
				returnReason=1;
				close(fd);
				return FAILURE;
			} else {
				newtPopWindow();
				close(fd);
				errWindow(FALSE,_("Could not mount CD-ROM on /mnt/cdrom : %s"),strerror(errno));
			}
			logMessage("Could not mount CD-ROM on /mnt/cdrom : %s",strerror(errno));
			return FAILURE;
		} else {
			progmountedcdrom=1;
			logMessage("CDROM: Successfully mounted CD-ROM '%s'",cddev);
		}
		newtPopWindow();
	}
	if (choice == FALSE) return 99;  /* special return code */
	return SUCCESS;
}

void unmountCDROM(void) {
	int rc;
	int fd;
	
	if (!progmountedcdrom) return;
	progmountedcdrom=0;
	if (newtRunning) {
		statusWindow(_("Unmounting"),_("Unmounting CD-ROM..."));
	}
	logMessage("CDROM: Unmounting CD-ROM.");
	rc=umount("/mnt/cdrom");
	if (rc < 0) {
		errWindow(FALSE,_("Error unmounting CD: %s"),strerror(errno));
	}
#ifdef EJECT_CDROM
	if (ejectok) {
		fd=open("/dev/cdrom",O_RDONLY);
		if (ioctl(fd,CDROMEJECT,NULL)) {
			logMessage("Could not eject CD-ROM (too bad).");
		}
		close(fd);
	}
#endif
	if (newtRunning) {
		newtPopWindow();
	}
}

char *floppyFS[] = {
	"ext2",
	"msdos",
	NULL,
};
	

int mountFloppy(void) {
	char *mountpoint=NULL;
	int rc;
	char **sptr=NULL;
	int mounted=0;
	unsigned long flag=MS_MGC_VAL | MS_RDONLY;
	void *data=NULL;

	mountpoint=strdup(installer ? FLOPPY_MNTPOINT_INSTALLER : FLOPPY_MNTPOINT);
	logMessage("FLOPPY: Will try to mount on '%s'",mountpoint);
	if (access(FLOPPY_MNTPOINT_INSTALLER,F_OK)) {
		rc=mkdir(FLOPPY_MNTPOINT_INSTALLER,0755);
		if (rc < 0) {
			errWindow(FALSE,_("Cannot create mount point: %s"),strerror(errno));
			return FAILURE;
		}
	}
	sptr=floppyFS;
	while((*sptr) && (!mounted)) {
		rc=mount("/dev/fd0",mountpoint,*sptr,flag,data);
		if (rc < 0) {
			logMessage("\tFLOPPY: Mounting as '%s' returned '%s'.",*sptr,strerror(errno));
		} else {
			logMessage("\tFLOPPY: Mount successful, fstype '%s'",*sptr);
			mounted=1;
		}
		sptr++;
	}
	if (!mounted) {
		errWindow(FALSE,_("Unable to mount floppy disk."));
		return FAILURE;
	} else {
		progmountedfloppy=1;
	}
	return SUCCESS;
}

int unmountFloppy(void) {
	char *mountpoint=NULL;
	int rc;
	
	mountpoint=strdup(installer ? FLOPPY_MNTPOINT_INSTALLER : FLOPPY_MNTPOINT);
	logMessage("FLOPPY: Will try to unmount '%s'",mountpoint);
	rc=umount(mountpoint);
	if (rc < 0) {
		logMessage("\tFLOPPY: Error unmounting '%s' : %s",mountpoint,strerror(errno));
		errWindow(FALSE,_("Error unmounting '%s' : %s"),mountpoint,strerror(errno));
		return FAILURE;
	} else {
		logMessage("\tFLOPPY: Unmount successful.");
	}
	return SUCCESS;
}
