/* tlcomps.h - header file for new component list format */

#ifndef TLCOMPS_H
#define TLCOMPS_H

#define CGROUP_FLAG_INMENU 1
#define CGROUP_FLAG_REQUIRED 2
#define CGROUP_FLAG_WARN 4
#define CGROUP_FLAG_OPTIONAL 8

#define DEFAULT_PRIORITY 30000
#define MAX_GROUPS 512
#define MAX_PPG 256
#define MAX_PACKAGES 1000
#define MAX_PKGS MAX_PACKAGES

struct packageInfo {
	Header h;
	char *name;
	char *group;
	char selected;
	char inmenu;
	int size;
	int priority;
	void *data;	
};

struct _pkgSet {
	struct packageInfo ** packages;
	int numPackages;
};

typedef struct _fnBinding {
	struct _fnBinding *next;
	char *packagename;
	char *filename;
} fnBinding;

typedef struct _pkgSet pkgSet;

typedef struct _tlComponentGroup {
	struct _tlComponentGroup *next;
	char *name;
	char *installtypes;
	char flags;
	int priority;
	char *helptext;
	pkgSet packageSet;
	int size;
	int selected;
	char *requires[MAX_PPG];
	char *suggests[MAX_PPG];
	char *conflicts[MAX_PPG];
} tlComponentGroup;

#define ITYPE_FLAG_DOREQUIRED 1
#define ITYPE_FLAG_DOSUGGESTED 2
#define ITYPE_FLAG_SATISFYDEPS 4
#define ITYPE_FLAG_IGNORECONFLICTS 8
#define ITYPE_FLAG_HIDESIZE 16

typedef struct _installType installType;
typedef pkgSet *(*itypeFunction)(installType *, tlComponentGroup *, pkgSet *, pkgSet *, char **, Header *, char **);

struct _installType {       
	struct _installType *next;
	char *name;
	int index;
	int inmenu;
	char **compList;
	int size;
	char flags;
	itypeFunction prepareFunction;
	char *helptext;
};

struct _helptextTable {
	char *key;
	char *helptext;
	char flags;
};

typedef struct _helptextTable helptextTable;



#endif
