/* Main turbopkg.c program file - this program contains
 * the start of execution (main func).
 *
 * by Scott Stone <sstone@turbolinux.com>
 * (C) 1998 Pacific HiTech
 * Distributable under the terms of the GNU GPL v2.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <newt.h>
#include <fcntl.h>
#include <rpm/rpmlib.h>

#define MAINCODE
#include "tlcomps.h"
#include "defines.h"
#include "prototypes.h"
#include "log.h"
#include "autowin.h"
#include "mycolors.h"

int testing=0;
int debug=0;
int installer=0;
int dumpinfo=0;
char *dumpfn=NULL;
int newtRunning=0;
int doSuggest=0;
int ignoreConflicts=0;
int autoDeps=1;
int autoDepsModified=0;
int groupSelectType=0;
int returnReason=0;
int noLog=0;
int ignoreLocal=0;
int ftpinstall=0;
int progmountedcdrom=0;
int progmountedfloppy=0;
int dynamicsize=1;
int dodowngrades=0;
ftpConfig ftpconf;
int nocompsfile=0;
int norpmdir=0;
int nohdlist=0;
int ejectok=1;
int autocontrol=0;
int controlsocket;
int basepath_selected=0;
int ignoreDependencies=0;
int unattended=0;
int enforcereqs=1;
/* int ftpPassiveOK=1; */

int numinstalltypes=0;
int completeinstallindex=0;
int custominstallindex=0;
int filelistinstallindex=0;

int needCompsLocation=0;
int needRPMLocation=0;
int needHDListLocation=0;
int dosSource=0;
int filenames_scanned=0;
fnBinding *myfnBindings=NULL;

int mystderr;

helptextTable *itypesTable=NULL;

typedef struct _pTemp {
	char *name;
	int priority;
} pTemp;

typedef struct _dlTrans {
	char *realextension;
	char *dosextension;
} dlTrans;

dlTrans dlTransTable[] = {
	{ "ja_JP.ujis","_ja.uji" },
	{ NULL, NULL}
};

tlInstall *theInstall;
pTemp ptemplist[MAX_PACKAGES];

int trappedSignals[] = { 1,3,4,6,7,8,10,11,12,14,15,16,19,23,24,-1 };

void handle_signals(int i) {
#ifdef AUTOCONTROL
	char s0[256];
#endif
	if (newtRunning) {
		newtFinished();
		newtRunning=0;
	}
	logMessage("FATAL ERROR: Program terminated by signal %d: '%s'",
		   i,strsignal(i));
	closeLog();
	unmountCDROM();
	printf("Program terminated by signal %d: '%s'\n",i,strsignal(i));
	if (autocontrol) {
#ifdef AUTOCONTROL
		sprintf(s0,"ServerTermBySignal/%d/%s",i,strsignal(i));
		autoErrMsg(controlsocket,s0,1);
#endif
	}
	exit(-1);
}

void showUsage(void) {
	printf("\nTurboPkg v%s %s by Scott Stone <sstone@turbolinux.com>\n",VERSION,ARCH);
	printf("Most of these options can be overridden in the Options menu.\n\n");
	printf("--debug     = Turn on debugging\n");
	printf("--test      = Run in TEST mode (does not actually install packages)\n");
/*	printf("--installer = Tells turbopkg that it is being run by the TL installer.\n"); */
/*	printf("--dumpinfo (file) = Dump package/component info to (file).\n"); */
/*	printf("--dosuggest = Used w/ dumpinfo.  if ON, suggested groups will be included.\n"); */
/*	printf("--ignoreconflicts = Used w/ dumpinfo.  Ignore group conflicts.\n"); */
	printf("--ignorelocal = Ignore installed packages.\n");
	printf("--ignoredeps  = Ignore all dependencies (not recommended).\n");
	printf("--dodowngrades = Downgrade packages if applicable.\n");
	printf("--nodynamicsize = Don't dynamically show install size in custom inst.\n");
	printf("--unattended = Don't prompt for 'ok' on failed pkg installs.\n");
/*	printf("--ftpinstall  = Perform FTP install (for installer only)\n"); */
/*	printf("--ftpnopassive = Do not use passive mode w/ FTP transfers.\n"); */
	printf("--nolog     = Do not log output to /tmp/turbopkg.log\n");
	printf("--noeject   = Do not eject CD when unmounting\n");
	printf("--version   = Show version info.\n");
	printf("--help      = Show this usage info.\n");
	printf("\n");
}

int getPriorityByName(int l,char *pkgname) {
	static int zz;

	for (zz=0;zz < l;zz++) {
		if (!ptemplist[zz].name) continue;
		if (!strcmp(ptemplist[zz].name,pkgname)) return ptemplist[zz].priority;
	}
	return 0;
}			

char *getDosName(char *realname) {
	dlTrans *dptr=NULL;

	dptr=&dlTransTable[0];
	while (dptr->realextension) {
		if (!strcmp(dptr->realextension,realname)) {
			return (dptr->dosextension);
		}
		dptr++;
	}
	return NULL;
}

tlInstall *initInstall(char *acfn) {
	char *s0;
	char *s1=NULL;
	tlInstall *tmp;
	int rc;
	static int zz;
	tlComponentGroup *cPtr=NULL;
	int ptindex=0;
	ftpMirror *mptr=NULL;
	int nosrc=0;
	bool choice;
	char *basepath=NULL;
	int cdismounted=0;

	tmp=(tlInstall *)malloc(sizeof(tlInstall));
	tmp->numInstallTypes=0;
	tmp->language=NULL;
	tmp->installTypes=(installType *)malloc(sizeof(installType));
	tmp->installTypes->next=NULL;
	tmp->masterComponentList=(tlComponentGroup *)malloc(sizeof(tlComponentGroup));
	tmp->masterComponentList->next=NULL;
	tmp->installedPackages=NULL;
	tmp->ipkglist=NULL;
	tmp->ipkglist=(char **)malloc(MAX_PACKAGES * sizeof(char *));
	for (zz=0;zz < MAX_PACKAGES;zz++) {
		tmp->ipkglist[zz]=NULL;
	}
	tmp->masterPackageSet.numPackages=0;
	tmp->masterPackageSet.packages=(struct packageInfo **)malloc(MAX_PACKAGES * sizeof(struct packageInfo *));
	for (zz=0;zz < MAX_PACKAGES;zz++) {
		tmp->masterPackageSet.packages[zz]=NULL;
	}
	tmp->compsfile=NULL;
	tmp->rpmdir=NULL;
	tmp->forcepackages=NULL;
	s0=getenv("LANG");
	if (s0 && (strcmp(s0,"C"))) {
		tmp->language=strdup(s0);
		logMessage("LANGUAGE: '%s'",tmp->language);
	} else {
		logMessage("LANGUAGE: (unset)");
	}
	ftpconf.mirrors=NULL;
	ftpconf.umirrors=NULL;
	rc=parsePkgRCFile(&ftpconf,tmp->language,0);
	if (rc == FAILURE) return NULL;
	if (autocontrol) {
#ifdef AUTOCONTROL
		rc=autoLoadConfig(acfn,&tmp->compsfile,&tmp->rpmdir,&tmp->hdlist,&ftpconf,&ftpinstall,&nocompsfile,&norpmdir,&nohdlist);
		if (rc != SUCCESS) {
			autoError(controlsocket,AUTO_ERR_SUPPORTFILES,AUTO_ERRFLAG_FATAL);
			return NULL;
		}
		goto autocontrol_end;
#endif
	} else {
		if (installer) {
#ifdef INSTALLER
			s1=strdup(COMPSFILE_INSTALLER);
			if (access(s1,R_OK)) {
				free(s1);
				s1=strdup(COMPSFILE_INSTALLER2);
				if (!ftpinstall) {
					dosSource=1;
				}
			}
#endif
		} else if (testing) {
			s1=strdup(COMPSFILE_TESTING);
		} else {
			s1=strdup(COMPSFILE);
		}
	}
	if (tmp->language) {
		s0=(char *)malloc(256);
		if (dosSource) {
			sprintf(s0,"%s%s",s1,getDosName(tmp->language));
		} else {
			sprintf(s0,"%s.%s",s1,tmp->language);
		}
		tmp->compsfile=strdup(s0);
		free(s0);
	} else {
		tmp->compsfile=strdup(s1);
		free(s1);
	}
	logMessage("COMPSFILE: '%s'",tmp->compsfile);
	if (access(tmp->compsfile,R_OK)) {
		logMessage("Cannot access '%s' !",tmp->compsfile);
		needCompsLocation=1;
	}
	if (installer) {
#ifdef INSTALLER
		tmp->rpmdir=strdup(RPMSDIR_INSTALLER);
		logMessage("RPMSDIR: Trying '%s'",RPMSDIR_INSTALLER);
		if ((access(tmp->rpmdir,F_OK)) && (!ftpinstall)) {
			free(tmp->rpmdir);
			tmp->rpmdir=strdup(RPMSDIR_INSTALLER2);
			logMessage("RPMSDIR: Trying '%s'",RPMSDIR_INSTALLER2);
			if (access(tmp->rpmdir,F_OK)) {
				free(tmp->rpmdir);
				logMessage("RPMSDIR: Trying '%s'",RPMSDIR_INSTALLER3);
				tmp->rpmdir=strdup(RPMSDIR_INSTALLER3);
				if (!ftpinstall) {
					dosSource=2;
				}
			} else {
				if (!ftpinstall) {
					dosSource=1;
				}
			}
			if (access(tmp->rpmdir,F_OK)) {
				logMessage("RPMSDIR: Cannot find RPMSDIR! Aborting.");
				returnReason=1;
				return NULL;
			}
			logMessage("RPMSDIR: Directory mounted on %s partition.",
				   (dosSource == 2) ? "Win95" : "DOS");
			logMessage("RPMSDIR: Scanning directory '%s' for package names.",
				   tmp->rpmdir);
			scanFilenames(tmp->rpmdir,myfnBindings);
		}
		logMessage("RPMSDIR: Using '%s'",tmp->rpmdir);
#endif
	} else if (testing) {
		s0=RPMSDIR_TESTING;
		if (s0) {
			tmp->rpmdir=strdup(RPMSDIR_TESTING);
		} else {
			logMessage("ERROR: TPI_RPMS has no value.");
/*			printf("ERROR: Environment variable 'TPI_RPMS' has no value!\n"); */
/*			return NULL; */
			needRPMLocation=1;
		}
	} else {
		tmp->rpmdir=strdup(RPMSDIR);
	}
	logMessage("RPMSDIR: '%s'",tmp->rpmdir);
	if (access(tmp->rpmdir,R_OK)) {
		logMessage("Cannot find '%s' - Trying '%s'...",tmp->rpmdir,RPMSDIR_ALT);
		if (!access(RPMSDIR_ALT,R_OK)) {
			free(tmp->rpmdir);
			tmp->rpmdir=strdup(RPMSDIR_ALT);
			logMessage("RPMDIR: Using '%s'",RPMSDIR_ALT);
		} else {
			logMessage("Cannot find '%s'.. need to ask for it...",RPMSDIR_ALT);
/*			printf("Cannot find RPM Directory - fatal error!\n"); */
			needRPMLocation=1;
		}
	}
	if (installer) {
#ifdef INSTALLER
		tmp->hdlist=strdup(HDLIST_INSTALLER);
		if (access(tmp->hdlist,R_OK)) {
			free(tmp->hdlist);
			tmp->hdlist=strdup(HDLIST_INSTALLER2);
			if (!ftpinstall) {
				dosSource=1;
			}
		}
#endif
	} else if (testing) {
		s0=HDLIST_TESTING;
		if (s0) {
			tmp->hdlist=strdup(HDLIST_TESTING);
		} else {
			logMessage("ERROR: TPI_HDLIST has no value.");
/*			printf("ERROR: Environment variable 'TPI_HDLIST' has no value!\n"); */
			needHDListLocation=1;
		}
	} else {
		tmp->hdlist=strdup(HDLIST);
	}
	logMessage("HDLIST: '%s'",tmp->hdlist);
	if (access(tmp->hdlist,R_OK)) {
		logMessage("Cannot find '%s' - Trying '%s'...",tmp->hdlist,HDLIST_ALT);
		if (!access(HDLIST_ALT,R_OK)) {
			free(tmp->hdlist);
			tmp->hdlist=strdup(HDLIST_ALT);
			logMessage("HDLIST: Using '%s'",HDLIST_ALT);
		} else {
			logMessage("Cannot find '%s'...",HDLIST_ALT);
			needHDListLocation=1;
		}
	}

	if (newtRunning) {
/*		newtPopWindow(); */
	}
 top_ask:
	rc=askFTP(&ftpconf,installer);
	if (rc == 99) {
		returnReason=1;
		return NULL;
	} else if (rc == 3) {  /* no source media */
		needCompsLocation=0;
		needRPMLocation=0;
		needHDListLocation=0;
		nocompsfile=1;
		norpmdir=1;
		nohdlist=1;
		nosrc=1;
	} else {
		ftpinstall=rc;
	}

	logMessage("FILES: Checking for FTP installation...");
	if (!ftpinstall) {
		if (installer) {
#ifdef INSTALLER
			if ((needCompsLocation) || (needRPMLocation) || (needHDListLocation)) {
				logMessage("FILES: Determining from base path '/tmp/tlimage'");
				rc=getLocationsByBasePath(tmp,"/tmp/tlimage");
			} else {
				logMessage("FILES: All necessary files found.");
				rc = SUCCESS;
			}
#endif
		} else {
			if ((!nosrc) && !cdromMounted()) {
				returnReason=0;
				cdismounted=askCDROM();
				if (cdismounted == SUCCESS) {
					rc=getLocationsByBasePath(tmp,DEFAULT_CD_MOUNTPOINT);
				} 
				if (returnReason) {
					returnReason=0;
					if (progmountedcdrom) {
						unmountCDROM();
						progmountedcdrom=0;
					}
					goto top_ask;
				}
			}
			if ((cdismounted != SUCCESS) && (!nosrc)) {
				choice=yesnoWindow(_("Find Files by Base Path"),TRUE,_("Use Base Path"),_("Select Individually"),_("Do you want to have TurboPkg find its files based on a single top-level path, or do you want to individually specify the locations of the support files?"));
				if (choice == TRUE) {
					basepath=pathExpandHomeDir(entryWindow(_("Enter Base Path"),50,DEFAULT_BASEPATH,_("Ok"),_("Cancel"),_("Enter the Base Path:")));
				}
				if ((choice == TRUE) && (basepath)) {
					basepath_selected=1;
					rc=getLocationsByBasePath(tmp,basepath);
				} else {
					rc=askLocations(tmp,needCompsLocation,needRPMLocation,needHDListLocation);
				}
				if (returnReason) {
					returnReason=0;
					goto top_ask;
				}
			}
		}
		if (rc == FAILURE) {
			if (newtRunning) {
				errWindow(FALSE,_("ERROR: Invalid path(s)!"));
			}
			logMessage("ERROR locating required files!");
			return NULL;
		}		
	} else {
		logMessage("Setting directory constants for FTP inst.\n");
/*		groupSelectType=1; */
		free(tmp->compsfile);
		free(tmp->rpmdir);
		free(tmp->hdlist);
		tmp->compsfile=strdup(installer ? COMPSFILE_FTP_INSTALLER : COMPSFILE_FTP);
		if (tmp->language && strcmp(tmp->language,"C")) {
			s0=(char *)malloc(256);		
			sprintf(s0,"%s.%s",tmp->compsfile,tmp->language);
			free(tmp->compsfile);
			tmp->compsfile=strdup(s0);
			free(s0);
			logMessage("COMPSFILE: Updated to '%s'",tmp->compsfile);
		}
		tmp->rpmdir=strdup(installer ? RPMSDIR_FTP_INSTALLER : RPMSDIR_FTP);
		tmp->hdlist=strdup(installer ? HDLIST_FTP_INSTALLER : HDLIST_FTP);
	}	
	if (ftpinstall) {		
		if (rc == FAILURE) {
			errWindow(FALSE,_("Error parsing turbopkgrc file."));
			logMessage("Error parsing rc file.");
			goto top_ask;
		}
		rc=getFTPPaths(&ftpconf);
		if (rc == FAILURE) {
			errWindow(FALSE,_("Error getting FTP paths."));
			logMessage("Error getting FTP paths.");
			goto top_ask;
		} else if (rc == CANCELLED) {
			logMessage("Cancelled FTP install.");
			goto top_ask;
		}
		logMessage("FTP: Host is '%s'",ftpconf.ftphost);
		logMessage("FTP: Directory is '%s'",ftpconf.ftpdir);
		if (ftpconf.language) {
			logMessage("FTP: Using language '%s'",ftpconf.language);
		}
		mptr=ftpconf.mirrors->next;
		while(mptr) {
			logMessage("FTP: Mirror: '%s:%s', location '%s'",mptr->hostname,mptr->directory,mptr->location);
			mptr=mptr->next;
		}
	}
#ifdef AUTOCONTROL
autocontrol_end:
#endif
	rc=initInstallTypes(tmp);
	if (rc == FAILURE) {
		logMessage("ERROR: initInstallTypes returned failure!");
		return NULL;
	}
	tmp->itypeSelected=0;
	logMessage("NumInstallTypes: %d",tmp->numInstallTypes);

	if (ftpinstall) {
		rc=getHDListByFTP(&ftpconf);
		if (rc != SUCCESS) {
			if (newtRunning) {
				errWindow(FALSE,_("Could not get RPM header list!"));
			} else if (autocontrol) {
#ifdef AUTOCONTROL
				autoErrorMsg(controlsocket,_("Could not get RPM header list!"),AUTO_ERRFLAG_FATAL);
#endif
			}
			logMessage("Failed to gethdlistbyftp!");
			return NULL;
		}
		rc=getCompsByFTP(&ftpconf,tmp);
		if (rc != SUCCESS) {
			if (installer) {
#ifdef INSTALLER
				if (newtRunning) {
					errWindow(FALSE,_("Could not get Components list!"));
				}
				logMessage("Failed to getcompsbyftp!");
				return NULL;
#endif
			} else {
				logMessage("FTP: Warning, no comps file - that's okay.");
				nocompsfile=1;
			}
		}
	}

	if (newtRunning) {
		statusWindow(_("Loading"),_("Loading Data, Please Wait..."));
	} else if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatus(controlsocket,"load1",_("Loading Data, Please Wait..."));
#endif
	}
	if (!nocompsfile) {
		rc=parseCompsFile(tmp);
		if (rc != SUCCESS) {
			nocompsfile=1;
		}
	} else {
		groupSelectType=1;
	}
	for (zz=0; zz < MAX_PACKAGES;zz++) {
		ptemplist[zz].name=NULL;
	}
	for (zz=0;zz < tmp->masterPackageSet.numPackages;zz++) {
		ptemplist[ptindex].name=strdup(tmp->masterPackageSet.packages[zz]->name);
		ptemplist[ptindex].priority=tmp->masterPackageSet.packages[zz]->priority;
		ptindex++;
	}
	logMessage("Getting package headers for master package list...");
	if ((nohdlist) || (access(tmp->hdlist,R_OK))) {
		if (!ftpinstall) {
			if (norpmdir) {
				logMessage("Grabbing from installed system...");
				rc=getPackageSetFromSystem(tmp);
				if (rc == FAILURE) {
					if (newtRunning) {
						newtPopWindow();
						errWindow(FALSE,_("Cannot get package information from the installed system.  Cannot continue."));
					} else if (autocontrol) {
#ifdef AUTOCONTROL
						autoStatusEnd(controlsocket);
						autoErrMsg(controlsocket,_("Cannot get package information from the installed system.  Cannot continue."),AUTO_ERRFLAG_FATAL);
#endif
					}
					return NULL;
				}
			} else {					
				logMessage("Grabbing from individual RPM packages...");
				rc=getPackageHeaders(tmp->rpmdir,&tmp->masterPackageSet,0);
				logMessage("FILENAMES_SCANNED = %d",filenames_scanned);
			}
		} else {
			logMessage("No header list file, aborting.");
			if (newtRunning) {
				newtPopWindow();
				errWindow(FALSE,_("There is no header list file on the specified remote FTP site, cannot continue."));
			} else {
#ifdef AUTOCONTROL
				autoStatusEnd(controlsocket);
				autoErrMsg(controlsocket,_("There is no header list file on the specified remote FTP site, cannot continue."),AUTO_ERRFLAG_FATAL);
#endif
			}
			return NULL;
		}
	} else {
		logMessage("Using header list file...");
		rc=getPackageHeadersFromHDList(tmp->hdlist,&tmp->masterPackageSet);
	}
	if (rc == FAILURE) {
		logMessage("\tFAILED!");
		if (newtRunning) {
			newtPopWindow();
		}
		return NULL;
	} else {
		logMessage("\tOK!");
	}
	if (!norpmdir) {
		for (zz=0;zz < tmp->masterPackageSet.numPackages;zz++) {
			tmp->masterPackageSet.packages[zz]->priority=getPriorityByName(tmp->masterPackageSet.numPackages,tmp->masterPackageSet.packages[zz]->name);
		}
		for (zz=0;zz < MAX_PACKAGES;zz++) {
			if (debug) {
				logMessage("MALLOC: Freeing ptemplist[%d]",zz);
			}
			free(ptemplist[zz].name);
		}
	} else {
		for (zz=0;zz < MAX_PACKAGES;zz++) {
			if (debug) {
				logMessage("MALLOC: Freeing ptemplist[%d]",zz);
			}
			free(ptemplist[zz].name);
		}
	}
	for (zz=0;zz < tmp->masterPackageSet.numPackages;zz++) {
		tmp->masterPackageSet.packages[zz]->group=NULL;
	}
	if (!nocompsfile) {
		cPtr=tmp->masterComponentList->next;
		while(cPtr) {
			logMessage("Getting package headers for group '%s'...",cPtr->name);
			rc=getPackageHeaders2(tmp->rpmdir,&tmp->masterPackageSet,&cPtr->packageSet);
			cPtr->size=0;
			rc=getGroupSize(tmp->masterComponentList,cPtr);
			logMessage("Group size is [%d]",cPtr->size);
			cPtr=cPtr->next;
		}
		logMessage("Creating component lists for install types...");
		rc=buildITypeCompList(tmp);
		if (rc == FAILURE) {
			logMessage("\tFAILED!");
			return NULL;
		}
	} else {
		logMessage("Getting master component list from RPM groups...");
		tmp->masterComponentList=getRPMGroups(&tmp->masterPackageSet);
	}
	if (newtRunning) {
		newtPopWindow();
	} else if (autocontrol) {
#ifdef AUTOCONTROL
		autoStatusEnd(controlsocket);
#endif
	}
	return tmp;
}

void suspend_callback(void) {
	char xch;
	newtSuspend();
	printf("TurboLinux Package Installer, Version %s %s\n",VERSION,ARCH);
	printf("This version released on %s\n",VER_DATE);
	printf("Built on: %s (%s)\n",__DATE__,__TIME__);
	printf("\nPress a key to continue: ");
	xch=getchar();
	newtResume();
}

void drawBackground(char *msg) {
	char s0[256];
	sprintf(s0,_("TurboPkg v%s %s - (C) 1998 Pacific HiTech"),VERSION,ARCH);
	newtDrawRootText(0,0,s0);
	newtDrawRootText((79-strlen(msg)),1,msg);
	newtRefresh();
}

int check_environment(void) {
	if (getuid() != 0) {
		if (newtRunning) {
			errWindow(FALSE,_("This program must be run as root."));
		}
		logMessage("FATAL: Not run as root.  Aborting.");
		return FAILURE;
	}
	return SUCCESS;
}
	

int main(int argc, char **argv) {
	static int zz;
	char **argptr=(argv+1);
	int rc;
	char rootText[80];

	for (zz=0;trappedSignals[zz] != -1;zz++) {
		signal(trappedSignals[zz],handle_signals);
	}
	while (*argptr) {
		if (!strcmp(*argptr,"--debug")) {
			debug=1;
		}
#ifdef AUTOCONTROL
		else if (!strcmp(*argptr,"--autocontrol")) {
			autocontrol=1;
		}
#endif
		else if (!strcmp(*argptr,"--test")) {
			testing=1;
			doSuggest=0;
		}
		else if (!strcmp(*argptr,"--testing")) {
			testing=1;
			doSuggest=1;
		}
		else if (!strcmp(*argptr,"--installer")) {
			installer=1;
			doSuggest=1;
		}
		else if (!strcmp(*argptr,"--dosuggest")) {
			doSuggest=1;
		}
		else if (!strcmp(*argptr,"--ignoreconflicts")) {
			ignoreConflicts=1;
		}
		else if (!strcmp(*argptr,"--unattended")) {
			unattended=1;
		}
		else if (!strcmp(*argptr,"--ignorelocal")) {
			ignoreLocal=1;
		}
		else if (!strcmp(*argptr,"--ignoredeps")) {
			ignoreDependencies=1;
		}
		else if (!strcmp(*argptr,"--dodowngrades")) {
			dodowngrades=1;
		}
		else if (!strcmp(*argptr,"--nodynamicsize")) {
			dynamicsize=0;
		}
		else if (!strcmp(*argptr,"--ftpinstall")) {
			ftpinstall=1;
		}
		/*
		else if (!strcmp(*argptr,"--ftpnopassive")) {
			ftpPassiveOK=0;
		}
		*/
		else if (!strcmp(*argptr,"--nolog")) {
			noLog=1;
		}
		else if (!strcmp(*argptr,"--noeject")) {
			ejectok=0;
		}
		/*
		else if (!strcmp(*argptr,"--dumpinfo")) {
			dumpinfo=1;
			if (*(argptr+1)) {
				dumpfn=strdup(*(argptr+1));
				argptr++;
			} else {
				printf("--dumpinfo requires an argument.\n");
				exit(-1);
			}
		}
		*/
		else if (!strcmp(*argptr,"--version")) {
			printf("\nTurboPkg v%s %s by Scott Stone <sstone@turbolinux.com>\n",VERSION,ARCH);
			printf("(C) 1998 Pacific HiTech\n\n");
			printf("Built on: %s (%s)\n\n",__DATE__,__TIME__);
			return 0;
		}
		else if ((!strcmp(*argptr,"--help")) || (!strcmp(*argptr,"-h"))) {
			showUsage();
			return 0;
		}
		else {
			printf("Unknown option '%s'\n",*argptr);
			return -1;
		}
		argptr++;
	}	
	if ((getenv("DUMMYINSTALL")) && (!strcmp(getenv("DUMMYINSTALL"),"yes"))) {
		testing=1;
	}
	openLog();
	logMessage("TurboPkg v%s %s - Starting...",VERSION,ARCH);
	if (debug) {
		logMessage("Running in DEBUG mode.");
	}
#ifdef AUTOCONTROL
	if (autocontrol) {		
		logMessage("Running in AUTOCONTROL mode (xturbopkg).");
	}
#endif
	if (installer) {
		logMessage("Running from the TurboLinux Installer.");
	}
	if (testing) {
		logMessage("Running in TESTING mode - no packages will install.");
	}
	if (dumpinfo) {
		logMessage("Running in DUMPINFO mode - dumping to file '%s'",dumpfn);
	}
	if (check_environment() != SUCCESS) {
		printf("You must run %s as root!\n",argv[0]);
		return -1;
	}
	setlocale(LC_ALL,"");
	bindtextdomain(PACKAGE, installer ? LOCALEDIR_INSTALLER : LOCALEDIR);
	textdomain(PACKAGE);
	logMessage("REDIRECT: Redirecting STDERR to /dev/null");
	mystderr=open("/dev/null",O_RDWR);
	close(2);
	dup2(mystderr,2);
	close(mystderr);
	myfnBindings=(fnBinding *)malloc(sizeof(fnBinding));
	myfnBindings->next=NULL;
#ifdef AUTOCONTROL
	if (autocontrol) {
		/*
		if (!access(AUTOCONTROL_UNIXSOCK,F_OK)) {
			logMessage("STARTUP: Unlinking '%s'",AUTOCONTROL_UNIXSOCK);
			unlink(AUTOCONTROL_UNIXSOCK);
		}
		*/
		autoMainHandler(argc,argv);
		/*
		unlink(AUTOCONTROL_UNIXSOCK);
		*/
		return 0;
	}
#endif
	if (!dumpinfo) {
		newtRunning=1;
		newtInit();
		newtSetColors(myColors);
		newtCls();
		
		newtSetSuspendCallback(suspend_callback);
		
		sprintf(rootText,_("TurboPkg v%s %s - (C) 1998 Pacific HiTech"),VERSION,ARCH);
		newtDrawRootText(0,0,rootText);
		newtPushHelpLine(_("<TAB>/<Alt-TAB> between elements  | <SPACE> Selects | <F1> = Help"));
/*		statusWindow(_("Analyzing"),_("Analyzing package info, please wait...")); */
	}
install_init:
	theInstall=initInstall(NULL);
	if (!theInstall) {
		if (!returnReason) {
			/*printf("Fatal error initializing package information!\n"); */
			free(ftpconf.mirrors);
			ftpconf.mirrors=NULL;
			ftpconf.umirrors=NULL;
			goto install_init;
		}
		if (newtRunning) {
			newtFinished();
			newtRunning=0;
		}
		
		unmountCDROM();
		return -1;
	}
	if ((!installer) && (!ignoreLocal) && (!norpmdir)) {
#ifndef INSTALLER
		getCurrentlyInstalled(theInstall);
#endif
	}

	if (dumpinfo) {
		dumpInfo(theInstall);
		logMessage("TurboPkg v%s %s - Finished.",VERSION,ARCH);
		closeLog();
		return 0;
	}

/*    newtPopWindow(); */
	if (!installer) {
		doWelcomeScreen();
	} else {
		doInstallerWelcomeScreen();
	}
	do {
		rc=doMainMenu(theInstall);
	} while (rc == CANCELLED);
	if ((rc == FORCECANCEL) && (!installer)) {
		msgWindow(_("Cancelled"),_("Ok"),_("You have chosen to cancel turbopkg - you may re-run it at any time from your installed system."));
	} else {
		if (!installer) {
			newtCls();
			newtRefresh();
			sprintf(rootText,_("TurboPkg v%s %s - (C) 1998 Pacific HiTech"),VERSION,ARCH);
			newtDrawRootText(0,0,rootText);
			newtPushHelpLine(_("<TAB>/<Alt-TAB> between elements  | <SPACE> Selects | <F1> = Help"));
			msgWindow(_("Finished"),_("Ok"),_("TurboPkg has finished installing/removing the requested packages.  A log of operations has been kept as %s."),LOGFN);
		}
	}
	newtFinished();
	newtRunning=0;
	logMessage("TurboPkg v%s %s - Finished (%d).",VERSION,ARCH,rc);
	closeLog();
	if (!installer) {
		printf("turbopkg: finished.\n");
	}
	unmountCDROM();
	if (rc != SUCCESS) {
		return 1;
	}
	return 0;
}

