#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include "defines.h"
#include "autowin.h"
#include "log.h"
#include "prototypes.h"

extern int debug;

extern int numinstalltypes;
extern int completeinstallindex;
extern int custominstallindex;
extern int filelistinstallindex;

extern helptextTable *itypesTable;

void strchop(char *s) {
	int l;
	if (!s) return;
	l=strlen(s);
	s[(l-1)]='\0';
}

int memberofStrArray(char **array, const char *element) {
	char **sptr=array;

	if (!sptr || !element) return 0;
	while (*sptr) {
		if (!strcmp(element,*sptr)) return 1;
		sptr++;
	}
	return 0;
}

char *memberofStrArray2(char **array, const char *element) {
	char **sptr=array;

	if (!sptr || !element) return 0;
	while (*sptr) {
		if (!strcmp(element,(*sptr + 1))) return *sptr;
		sptr++;
	}
	return NULL;
}

int memberofStrArrayBySubstring(char **array, const char *element) {
	char **sptr=array;

	if (!sptr || !element) return 0;
	while (*sptr) {
		if (strstr(*sptr,element)) return 1;
		sptr++;
	}
	return 0;
}

int basicStrCompare(const void * a, const void * b) {
    const char * const * one = a;
    const char * const * two = b;

    return strcmp(*one, *two);
}

char **uniquifyStrArray(char **array, int sort) {
	char **tmp=(char **)malloc(sizeof(char *));
	char **sptr=NULL;
	int eindex=0;

	tmp[0]=NULL;
	sptr=array;
	while (*sptr) {		
		if (!memberofStrArray(tmp,*sptr)) {
			tmp=realloc(tmp,((eindex + 2) * sizeof(char *)));
			tmp[eindex]=strdup(*sptr);
			tmp[eindex + 1]=NULL;
			eindex++;
		}
		sptr++;
	}
	if (sort) {
		qsort(tmp,eindex,sizeof(char *),basicStrCompare);
	}
	return tmp;
}

int memberofPkgSet(pkgSet *theSet, char *pkgname) {
	static int zz;

	if (!theSet || !pkgname) return 0;
	for (zz=0;zz < theSet->numPackages;zz++) {
		if (!strcmp(theSet->packages[zz]->name,pkgname)) return 1;
	}
	return 0;
}

int memberofPkgSetBySubstring(pkgSet *theSet, char *pkgname) {
	static int zz;

	if (!theSet || !pkgname) return 0;
	for (zz=0;zz < theSet->numPackages;zz++) {
		if (strstr(theSet->packages[zz]->name,pkgname)) return 1;
	}
	return 0;
}

int memberofComponentGroup(tlComponentGroup *grouplist, char *groupname) {
	tlComponentGroup *gptr=NULL;
	
	if (!grouplist || !groupname) return 0;
	gptr=grouplist->next;
	while (gptr) {
		if (!strcmp(gptr->name,groupname)) return 1;
		gptr=gptr->next;
	}
	return 0;
}

char *strcasestr(const char *is1, const char *is2) {
	char *s1=strdup(is1);
	char *s2=strdup(is2);
	static int zz;

	for (zz=0;zz < strlen(s1);zz++) {
		if (isalpha(s1[zz])) {
			if (s1[zz] >= 97) {
				s1[zz] -= 32;
			}
		}
	}
	for (zz=0;zz < strlen(s2);zz++) {
		if (isalpha(s2[zz])) {
			if (s2[zz] >= 97) {
				s2[zz] -= 32;
			}
		}
	}
	
	return strstr(s1,s2);
}

int containsNonDigits(char *s) {
	int l=strlen(s);
	static int zz;

	for (zz=0;zz < l;zz++) {
		if (!isdigit(s[zz])) return 1;
	}
	return 0;
}

int addMirror(ftpConfig *f, char *loc, char *res, int isupdate) {
	ftpMirror *mptr=NULL;

	if (!f->mirrors) return FAILURE;
	if (!loc || !res) return FAILURE;
	if (!isupdate) {
		for (mptr=f->mirrors;mptr->next;mptr=mptr->next);
	} else {
		for (mptr=f->umirrors;mptr->next;mptr=mptr->next);
	}
	mptr->next=(ftpMirror *)malloc(sizeof(ftpMirror));
	mptr->next->next=NULL;
	mptr->next->location=strdup(loc);
	mptr->next->hostname=strdup(strtok(res,":"));
	mptr->next->directory=strdup(strtok(NULL,":"));
	return SUCCESS;
}

int parsePkgRCFile(ftpConfig *f, char *language,int forcedefs) {
	FILE *fp;
	char s0[256];
	char *s1=NULL;
	char *s2=NULL;
	char s3[256];
	char s4[256];
	char *s5=NULL;
	char *s6=NULL;
	char s7[256];
	char s8[256];
	char s9[256];
	char s10[256];
	char *defHost=NULL;
	char *defDir=NULL;
	char *defUHost=NULL;
	char *defUDir=NULL;
	int rc=0;
	int defNum=0;
	int defComplete=0;
	int defCustom=0;
	int defList=0;
	int in_itype=0;
	char itypekey[256];
	int itypeindex=0;

	completeinstallindex=-1;
	custominstallindex=-1;
	filelistinstallindex=-1;
	logMessage("TURBOPKGRC: Parsing '%s'",CONFIGFILE);
	if (language) {
		if (strcmp(language,"C")) {
			logMessage("TURBOPKGRC: Language is '%s'",language);
			f->language=strdup(language);
		} else {
			f->language=NULL;
		}
	} else {
		f->language=NULL;
	}
	sprintf(itypekey,"ITYPE");
	if ((f->language) && (!forcedefs)) {
		sprintf(itypekey,"ITYPE.%s",f->language);
	}
	logMessage("TURBOPKGRC: Install-type Key will be '%s'",itypekey);
	f->ftphost=NULL;
	f->ftpdir=NULL;
	f->uftphost=NULL;
	f->uftpdir=NULL;
	f->mirrors=(ftpMirror *)malloc(sizeof(ftpMirror));
	f->mirrors->next=NULL;
	f->umirrors=(ftpMirror *)malloc(sizeof(ftpMirror));
	f->umirrors->next=NULL;
	sprintf(s3,"FTPHOST.%s",f->language ? f->language : "null");
	sprintf(s4,"FTPDIR.%s",f->language ? f->language : "null");
	sprintf(s7,"MIRROR.%s",f->language ? f->language : "null");
	sprintf(s8,"UFTPHOST.%s",f->language ? f->language : "null");
	sprintf(s9,"UFTPDIR.%s",f->language ? f->language : "null");
	sprintf(s10,"UMIRROR.%s",f->language ? f->language : "null");
	fp=fopen(CONFIGFILE,"r");
	while (fgets(s0,256,fp)) {
		strchop(s0);
		if ((strlen(s0)) && (s0[0] != '#')) {
			if (!in_itype) {
				s1=strtok(s0,"");
				s2=strtok(NULL,"");
				if (!strcmp(s1,"FTPHOST")) {
					if (s2) {
						defHost=strdup(s2);
					}
				}
				else if (!strcmp(s1,"FTPDIR")) {
					if (s2) {
						defDir=strdup(s2);
					}
				}
				if (!strcmp(s1,"UFTPHOST")) {
					if (s2) {
						defUHost=strdup(s2);
					}
				}
				else if (!strcmp(s1,"UFTPDIR")) {
					if (s2) {
						defUDir=strdup(s2);
					}
				}
				else if (!strcmp(s1,s3)) {
					if (s2) {
						f->ftphost=strdup(s2);
					}
				}
				else if (!strcmp(s1,s4)) {
					if (s2) {
						f->ftpdir=strdup(s2);
					}
				}
				else if (!strcmp(s1,s8)) {
					if (s2) {
						f->uftphost=strdup(s2);
					}
				}
				else if (!strcmp(s1,s9)) {
					if (s2) {
						f->uftpdir=strdup(s2);
					}
				}
				else if ((!strcmp(s1,"MIRROR")) && (!f->language)) {
					if (s2) {
						s5=strdup(s2);
						s6=strtok(NULL,"");
						rc=addMirror(f,s5,s6,0);
						if (rc != SUCCESS) {
							logMessage("Could not add mirror: '%s' '%s'",s5,s6);
							return FAILURE;
						}
					}
				}
				else if ((!strcmp(s1,s7)) && (f->language)) {
					if (s2) {
						s5=strdup(s2);
						s6=strtok(NULL,"");
						rc=addMirror(f,s5,s6,0);
						if (rc != SUCCESS) {
							logMessage("Could not add mirror: '%s' '%s'",s5,s6);
							return FAILURE;
						}
					}
				}
				else if ((!strcmp(s1,"UMIRROR")) && (!f->language)) {
					if (s2) {
						s5=strdup(s2);
						s6=strtok(NULL,"");
						rc=addMirror(f,s5,s6,1);
						if (rc != SUCCESS) {
							logMessage("Could not add umirror: '%s' '%s'",s5,s6);
							return FAILURE;
						}
					}
				}
				else if ((!strcmp(s1,s10)) && (f->language)) {
					if (s2) {
						s5=strdup(s2);
						s6=strtok(NULL,"");
						rc=addMirror(f,s5,s6,1);
						if (rc != SUCCESS) {
							logMessage("Could not add umirror: '%s' '%s'",s5,s6);
							return FAILURE;
						}
					}
				}
				else if (strstr(s1,"NUMINSTALLTYPES") == s1) {
					logMessage("TURBOPKGRC: numinstalltypes read.");
					s5=(s1 + strlen("NUMINSTALLTYPES"));
					if (s5[0] > 0) {
						s5++;
						if ((f->language) && (!strcmp(s5,f->language))) {
							if (s2) {
								numinstalltypes=atoi(s2);
							}
						}
					} else {
						if (s2) {
							defNum=atoi(s2);
						}
					}
				}
				else if (strstr(s1,"COMPLETEINSTALL") == s1) {
					logMessage("TURBOPKGRC: completeinstallindex read.");
					s5=(s1 + strlen("COMPLETEINSTALL"));
					if (s5[0] > 0) {
						s5++;
						if ((f->language) && (!strcmp(s5,f->language))) {
							if (s2) {
								completeinstallindex=atoi(s2);
							}
						}
					} else {
						if (s2) {
							defComplete=atoi(s2);
						}
					}
				}
				else if (strstr(s1,"CUSTOMINSTALL") == s1) {
					logMessage("TURBOPKGRC: custominstallindex read.");
					s5=(s1 + strlen("CUSTOMINSTALL"));
					if (s5[0] > 0) {
						s5++;
						if ((f->language) && (!strcmp(s5,f->language))) {
							if (s2) {
								custominstallindex=atoi(s2);
							}
						}
					} else {
						if (s2) {
							defCustom=atoi(s2);
						}
					}
				}
				else if (strstr(s1,"FILELISTINSTALL") == s1) {
					logMessage("TURBOPKGRC: filelistinstallindex read.");
					s5=(s1 + strlen("FILELISTINSTALL"));
					if (s5[0] > 0) {
						s5++;
						if ((f->language) && (!strcmp(s5,f->language))) {
							if (s2) {
								filelistinstallindex=atoi(s2);
							}
						}
					} else {
						if (s2) {
							defList=atoi(s2);
						}
					}
				}
				else if (!strcmp(s1,itypekey)) {
					if (!itypesTable) {
						if (!numinstalltypes) numinstalltypes=defNum;
						if (!numinstalltypes) {
							logMessage("TURBOPKGRC: Parse error: itypes defined before numinstalltypes!");
							errWindow(FALSE,_("Error parsing turbopkgrc - cannot continue!"));
							return FAILURE;
						}
						itypesTable=(helptextTable *)malloc((numinstalltypes + 1) * sizeof(helptextTable));
						if (!itypesTable) {
							logMessage("TURBOPKGRC: Cannot allocate table!");
							errWindow(FALSE,_("Memory allocation problem, cannot continue!"));
							return FAILURE;
						}
					}
					itypesTable[itypeindex].key=strdup(s2);
					itypesTable[itypeindex].flags=atoi(strtok(NULL,""));
					itypesTable[itypeindex].helptext=(char *)malloc(1024);
					itypesTable[itypeindex].helptext[0]='\0';
					in_itype=1;
				}
			} else {
				if (strstr(s0,"ENDITYPE") == s0) {
					logMessage("TURBOPKGRC: Finished itype '%s'",itypesTable[itypeindex].key);
					in_itype=0;
					itypeindex++;
				} else {
/*					itypesTable[itypeindex].helptext=realloc(itypesTable[itypeindex].helptext,(strlen(s0) + 5));*/
					if ((strlen(itypesTable[itypeindex].helptext) + strlen(s0) + 1) <= 1024) {
						strcat(itypesTable[itypeindex].helptext,s0);
						strcat(itypesTable[itypeindex].helptext," ");
						if (debug) {
							logMessage("TURBOPKGRC-DEBUG: Added line to '%s':\n(%s)",itypesTable[itypeindex].key,s0);
						}
					}
				}
			}
		}
				
	}

	if (!itypeindex) {
		errWindow(FALSE,_("No install types available for the language '%s'!"),f->language ? f->language : _("(default)"));
		logMessage("TURBOPKGRC: Error: No install types available for '%s' !",f->language ? f->language : _("(default)"));
		if (!forcedefs) {
			return parsePkgRCFile(f,language,1);
		} else {
			return FAILURE;
		}
	}
	
	itypesTable[itypeindex].key=NULL;
	itypesTable[itypeindex].helptext=NULL;
	itypesTable[itypeindex].flags=0;
	
	if (!numinstalltypes) numinstalltypes=defNum;
	if (!f->language) {
		if (completeinstallindex == -1) completeinstallindex=defComplete;
		if (custominstallindex == -1) custominstallindex=defCustom;
		if (filelistinstallindex == -1) filelistinstallindex=defList;
	}
	logMessage("TURBOPKGRC: numinstalltypes = %d",numinstalltypes);
	if (itypeindex != numinstalltypes) {
		logMessage("TURBOPKGRC: WARNING - numinstalltypes != number of types defined!");
	}
	logMessage("TURBOPKGRC: completeinstallindex = %d",completeinstallindex);
	logMessage("TURBOPKGRC: custominstallindex = %d",custominstallindex);
	logMessage("TURBOPKGRC: filelistinstallindex = %d",filelistinstallindex);
	fclose(fp);
	if (!f->ftphost) {
		if (defHost) {
			f->ftphost=strdup(defHost);
		} else {
			return FAILURE;
		}
	}
	if (!f->ftpdir) {
		if (defDir) {
			f->ftpdir=strdup(defDir);
		} else {
			return FAILURE;
		}
	}
	if (!f->uftphost) {
		if (defUHost) {
			f->uftphost=strdup(defUHost);
		} else {
			return FAILURE;
		}
	}
	if (!f->uftpdir) {
		if (defUDir) {
			f->uftpdir=strdup(defUDir);
		} else {
			return FAILURE;
		}
	}
	free(defHost);
	free(defDir);
	return SUCCESS;
}

char *binLocs[] = {
	"/bin",
	"/sbin",
	"/usr/bin",
	"/usr/sbin",
	"/usr/jp/bin",
	"/usr/jp/sbin",
	"/usr/local/bin",
	"/usr/local/sbin",
	NULL,
};

int checkFreeSpace(int requiredkb, int *freeptr) {
	char s0[256];
	char cmd[256];
	char **sptr=NULL;
	char *dflocation=NULL;
	FILE *pp=NULL;
	char s1[20];
	char s2[20];
	char s3[20];
	char s4[20];
	char s5[20];
	char s6[20];
	int freeSpace;
	
	if (installer) {
		sprintf(s0,"%s/df",INSTALLER_BINARY_DIR);
		if (access(s0,X_OK)) {
			sprintf(s0,"%s/df",INSTALLER_BINARY_DIR2);
			if (access(s0,X_OK)) {
				logMessage("WARNING: Cannot find 'df' - assuming free space is okay.");
				return 1;
			} else {
				dflocation=strdup(s0);
			}
		} else {
			dflocation=strdup(s0);
		}
	} else {
		sptr=binLocs;
		while((*sptr) && (!dflocation)) {
			sprintf(s0,"%s/df",*sptr);
			if (!access(s0,X_OK)) {
				dflocation=strdup(s0);
			}
			sptr++;
		}
	}
	if (installer) {
		if (!access("/mnt/usr",F_OK)) {
			sprintf(cmd,"%s %s",dflocation,"/mnt/usr");
		} else {
			sprintf(cmd,"%s %s",dflocation,"/mnt");
		}
	} else {
		sprintf(cmd,"%s %s",dflocation,"/usr");
	}
	pp=popen(cmd,"r");
	while(fgets(s0,256,pp)) {
		if (s0[0] == '/') {
			sscanf(s0,"%s %s %s %s %s %s",s1,s2,s3,s4,s5,s6);
		}
	}
	pclose(pp);
	freeSpace=atoi(s4);
	*freeptr=freeSpace;
	if (freeSpace >= requiredkb) return 1;
	return 0;
}

char **removeFromStrArray(char **array, char *element, int size) {
	char **sptr=NULL;
	char **newarray=NULL;
	int aindex=0;
	static int zz;

	newarray=(char **)malloc(size * sizeof(char *));
	if (!newarray) return NULL;
	for (zz=0;zz < size;zz++) {
		newarray[zz]=NULL;
	}
	sptr=array;
	while(*sptr) {
		if (strcmp(*sptr,element)) {
			newarray[aindex]=strdup(*sptr);
			aindex++;
		}
		sptr++;
	}
	for (zz=0;zz < size;zz++) {
		free(array[zz]);
	}
	free(array);
	return newarray;
}

int addToStrArray(char **array,char *element) {
	char **sptr;
	int aindex=0;

	sptr=array;
	while(*sptr) {
		aindex++;
		sptr++;
	}
	array[aindex]=strdup(element);
	return SUCCESS;
}

char *pathExpandHomeDir(const char *thePath) {
	char *s1=NULL;
	char *s2=NULL;
	char *sx=NULL;
	FILE *fp;
	char s0[256];
	char *homedir=NULL;
	char *tmp=NULL;

	if (!thePath) return NULL;
	if (thePath[0] != '~') return (char *)thePath;
	sx=strdup(thePath);
	s1=sx;
	s1++;
	s2=index(s1,'/');
	if (s2) {
		s2[0]='\0';  
	}
	tmp=(char *)malloc(256);
	strcpy(tmp,s1);
	strcat(tmp,":");
	s1=strdup(tmp);
	free(tmp);
	fp=fopen(PASSWDFILE,"r");
	while(fgets(s0,256,fp)) {
		if (strlen(s0)) {
			if (strstr(s0,s1) == s0) {
				strtok(s0,":");
				strtok(NULL,":");
				strtok(NULL,":");
				strtok(NULL,":");
				strtok(NULL,":");
				homedir=strdup(strtok(NULL,":"));
			}
		}
	}
	if (!homedir) return NULL;
	fclose(fp);
	s1=index(thePath,'/');
	if (s1) {
		s1++;
	}
	tmp=(char *)malloc(256);
	strcpy(tmp,homedir);
	strcat(tmp,"/");
	if (s1) {
		strcat(tmp,s1);
	}
	if (debug) {
		logMessage("PATHEXP: Expanded path is '%s'",tmp); 
	}
	return tmp;
}

char *checkVersions(void) {
	rpmdb db=NULL;
	int myoffset=-1;
	dbiIndexSet myindex;
	Header hTemp;
	char *iVersion=NULL;
	char *distName=NULL;
	FILE *pp;
	char *str_rc=NULL;
	char *sysinfo=NULL;
	char s0[256];
	int xType,xC;
	int rc;
	
	rpmReadConfigFiles(NULL, NULL, NULL, 0);
	if (rpmdbOpen("/", &db, O_RDWR | O_CREAT, 0644))
		db = NULL;
	if (db) {
		if (!rpmdbFindPackage(db,"turbolinux-release",&myindex)) {
			myoffset=myindex.recs[0].recOffset;
			hTemp=rpmdbGetRecord(db,myoffset);
			rc=headerGetEntry(hTemp,RPMTAG_VERSION,&xType,(void **)&iVersion,&xC);
			distName=strdup("TurboLinux");
		}
		else if (!rpmdbFindPackage(db,"redhat-release",&myindex)) {
			myoffset=myindex.recs[0].recOffset;
			hTemp=rpmdbGetRecord(db,myoffset);
			rc=headerGetEntry(hTemp,RPMTAG_VERSION,&xType,(void **)&iVersion,&xC);
			distName=strdup("RedHat");
		}
		else {
			distName=strdup("Slackware/Debian/Other");
			iVersion=strdup("(unknown version)");
		}
	} else {
		distName=strdup("Slackware/Debian/Other");
		iVersion=strdup("(unknown version)");
	}
	rpmdbClose(db);
	pp=popen("uname -mrs","r");
	while(fgets(s0,256,pp)) {
		s0[strlen(s0)-1]='\0';
		free(sysinfo);
		sysinfo=strdup(s0);
	}
	pclose(pp);
	if (!sysinfo) {
		sysinfo=strdup("Unknown System");
	}
	str_rc=(char *)malloc((strlen(distName) + strlen(iVersion) + strlen(sysinfo) + 20));
	sprintf(str_rc,"%s %s - %s",distName,iVersion,sysinfo);
	return str_rc;
}
	
char *getFilenameByPkgName(char *dirname, char *pkgname) {
	DIR *dirp;
	struct dirent *ent;
	int fd;
	int rc;
	int isSource;
	Header h;
	int xType,xC;
	char *xpkgname=NULL;
	char filename[128];
	char *outfn=NULL;

	dirp=opendir(dirname);
	if (!dirp) return NULL;
	while ((ent=readdir(dirp))) {
		if (!(ent->d_name[0] == '.' &&
		      (ent->d_name[1] == '\0' || 
		       ((ent->d_name[1] == '.') &&
			(ent->d_name[2] == '\0'))))) {
			sprintf(filename, "%s/%s", dirname, ent->d_name);
			fd = open(filename, O_RDONLY);
			if (fd < 0) {
				return NULL;
			}
			rc = rpmReadPackageHeader(fd, &h, &isSource, NULL, NULL);
			close(fd);
			if (rc) {
				return NULL;
			}
			rc=headerGetEntry(h,RPMTAG_NAME,&xType,(void **)&xpkgname,&xC);
			if ((xpkgname) && (!strcmp(xpkgname,pkgname))) {
				outfn=strdup(filename);
				closedir(dirp);
				return outfn;
			}
		}
	}
	closedir(dirp);
	return NULL;
}
