#!/usr/bin/perl

#### Copyright (c) 1997, 1998 Pacific HiTech, Inc. 
####              All Rights Reserved.
####		  Distributable under the terms of the GPL.
#### File:        xturbopkg.pl (formerly turbopkgmgr.pl, xpkgupdate.pl)
#### Author:      Scott Stone <sstone@turbolinux.com>
#### Version:     5.1
#### Date:        Wed Jul 29, 1998
#### Description:
# This is the X-based version of the TurboLinux
# package management system.
####
#### ChangeLog:
#
# ChangeLog (begun 8/8/97):
#
# (see file ChangeLog for more recent entries)
#

use English;
use Tk;
use Tk::Dialog;
use Socket;

# initialization for internationalization.
use Locale::GetText;
import Locale::GetText I_;
Locale::GetText::textdomain("turbodesk", "/usr/lib/turbodesk/text");
Locale::GetText::textdomain("xturbopkg", "/usr/lib/turbolinux/turbopkg/text");

# Global Variables

$VERSION="5.3.2";
$VER_DATE="Oct 20, 1998";
$CONFIGFN="xturbopkg.conf";
$ARCH="unknown";
$OSTYPE="unknown";

$PREFSFILE="/etc/sysconfig/xturbopkg.prefs";
$ERRCOLOR="red";

$DEBUG=0;
$PROGRAM_STATE=0;
$OKSTART=1;
$MAIL_LOG=1;
$SOURCEMEDIA=0;
$GLIBC=0;
$AUTOPKG=0;
$HAVEHDLIST=0;
$HAVELOCALDEPS=0;
$FASTLOAD=0;
$SERVER_TIMEOUT=2;
$DISP_LINES=20;
$AUTODEPS=1;
$DODOWNGRADES=0;
$DYNRESIZE=1;
$UNATTENDED=0;
$SETSIZE=0;
$SETCOUNT=0;
$IFTYPE=1;
$SORTGROUPS=1;
$SHOWPACKAGES=1;
$CASCADE=1;
$CASCADE_DISTANCE=50;
$SHOWGRAPHICS=1;
$IGNOREDEPS=0;
$FIRSTENTRY="";

$PROGMOUNTEDCDROM=0;
$NEWS_SHOWN=0;
$CANCELLED=0;

$BINOVERRIDE="";

require '/usr/lib/turbodesk/xturbopkg_dir/defines.ph';

$COMPSFILEPATH=&COMPSFILE;
$RPMSPATH=&RPMSDIR;
$HDLISTPATH=&HDLIST;
$FTPHOST="";
$FTPDIR="";
$FTPUSERNAME="anonymous";
$FTPPASSWORD="tlinstall\@";
$USE_PROXY=0;
$FTPPROXY="";
$FTPPORT=21;
$ISFTP=0;
$NOCOMPSFILE=0;
$NORPMDIR=0;
$NOHDLIST=0;
$MWGEOMETRY="";

$searchkey="";

undef %RADIOBUTTONS;
undef %CHECKBOXES;
undef %VALUES;
undef %WIDGETS;



$DEBUG=0;
$TEST=0;
for ($argc=0;$argc <= $#ARGV;$argc++) {
    if ($ARGV[$argc] eq "-debug") {
	$DEBUG++;
    }
    if ($ARGV[$argc] eq "-nostart") {
	$OKSTART=0;
    }
    if ($ARGV[$argc] eq "-test") {
	$TEST=1;
    }
    if ($ARGV[$argc] eq "-autopkg") {
	$AUTOPKG=1;
    }
    if ($ARGV[$argc] eq "-geometry") {
	$MWGEOMETRY=$ARGV[($argc + 1)];
    }
    if ($ARGV[$argc] eq "-serverbin") {	
	$BINOVERRIDE=$ARGV[($argc + 1)];
	if ($DEBUG) {
	    print "Setting binoverride '$BINOVERRIDE'\n";
	}
    }
}

if ($TEST) {
    require '/usr/lib/turbodesk/turbodesklib.pl';
} else {
    require '/usr/lib/turbodesk/turbodesklib.pl';
}

# Color defs/etc are now external for 2.3

$TMPDIR=$ENV{TMPDIR} || "/tmp";
if ($DEBUG) {
    print "TEMP: $TMPDIR\n";
}

# Processing starts here

# Redirect STDERR,STDOUT 

if (!($DEBUG)) {
    close STDERR;
    open(STDERR,">/dev/null");
    close STDOUT;
    open(STDOUT,">/dev/null");
}

# Find out OSType and Architecture

open(osfp,"uname -s -m |");
while(<osfp>) {
    chop;
    ($OSTYPE,$ARCH)=split(/\s+/);
}
close osfp;
if ($DEBUG) {
    print "OSType: [$OSTYPE]\nArchitecture: [$ARCH]\n";
    if ($ARCH =~ m/i[0-9]86/) {
	print "\t(Intel Platform)\n";
    }
}
if ($ARCH =~ m/i[0-9]86/) {
    $ARCH="i386";
}

if ($TEST) {
    $CONFIGPATH="./";
} else {
    $CONFIGPATH="/usr/lib/turbodesk/";
}

if (!(-f "$CONFIGPATH/$CONFIGFN")) {
    print I_("* ERROR loading config file.\n");
    die;
}

if (length($BINOVERRIDE)) {
    if ($DEBUG) {
	print "BINOVERRIDE: $BINOVERRIDE\n";
    }
    $SERVERBIN=$BINOVERRIDE;
}

&load_configuration;
if (length($BINOVERRIDE)) {
    if ($DEBUG) {
	print "BINOVERRIDE: $BINOVERRIDE\n";
    }
    $SERVERBIN=$BINOVERRIDE;
}
&connectToServer || die "Cannot connect to server!\n";
&config_backcompat;
&load_preferences;
    

    
$MW = MainWindow->new;

$TitleText=sprintf(I_("XTurboPkg v%s"), $VERSION);
$MW->title($TitleText);
$MW->configure(-background => $FRAMEBGCOLOR);
if (length($MWGEOMETRY)) {
    $MW->geometry($MWGEOMETRY);
}

$MW->iconbitmap("@/usr/lib/turbodesk/bitmaps/xturbopkg.xbm");
$MW->iconname("Xturbopkg");

$FONT = '-*-helvetica-bold-r-*-*-14-*-*-*-*-*-*-*';

$MW->bind('<Alt-KeyPress-q>' => \&exit_program);
$MW->bind('<KeyPress-F1>' => sub {
    &mainhelp_window;
} );
$MW->bind('<KeyPress-F2>' => sub {
    &treeExpand;
} );
$MW->bind('<KeyPress-F3>' => sub {
    if ($PROGRAM_STATE == 1) {
	&doQuery;
    }
} );
$MW->bind('<KeyPress-F4>' => sub {
    &doSearch;
} );
$MW->bind('<Control-KeyPress-s>' => sub {
    &doSearch;
} );
$MW->bind('<Control-KeyPress-i>' => sub {
    &doQuery;
} );
$MW->bind('<KeyPress-r>' => sub {
    if ($PROGRAM_STATE == 1) {
	&removePackage;
    }
} );
$MW->bind('<KeyPress-c>' => sub {
    if ($PROGRAM_STATE == 1) {
	&clearEntries;
    }
} );
$MW->bind('<KeyPress-u>' => sub {
    if ($PROGRAM_STATE == 1) {
	&selectUpgrades;
    }
} );


$mb=$MW->Frame(-relief => 'raised', -borderwidth => '1', -bg => $SECBGCOLOR);
$mb->pack(-side => 'top', -fill => 'x', -anchor => 'center');
$mb_file=$mb->Menubutton( -tearoff => false, -text => I_('File'), -underline => -1, -bg => $SECBGCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR, -fg => $TEXTCOLOR);
$mb_file->command(-label => I_('Quit (Alt-Q)'),-underline => -1, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR, -command => \&exit_program);
$mb_file->pack(-side => 'left');

$mb_opts=$mb->Menubutton( -tearoff => false, -text => I_('Options'), -underline => -1, -bg => $SECBGCOLOR, -fg => $TEXTCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR);
$mb_opts->command(-label => I_('Preferences...'), -underline => -1, -command => \&preferences_window, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_opts->pack(-side => 'left');

$mb_actions=$mb->Menubutton( -tearoff => false, -text => I_('Actions'), -underline => -1, -bg => $SECBGCOLOR, -fg => $TEXTCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR);
$mb_actions->command(-label => I_('Expand/Collapse Tree'), -underline => -1, -command => \&treeExpand, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_actions->command(-label => I_('Search'), -underline => -1, -command => \&doSearch, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_actions->command(-label => I_('Reset Display'), -underline => -1, -command => \&resetScreen, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_actions->command(-label => I_('Restart Program'),-underline => -1, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR, -command => \&restart_program);

$mb_shortcuts=$mb->Menubutton( -tearoff => false, -text => I_('Shortcuts'), -underline => -1, -bg => $SECBGCOLOR, -fg => $TEXTCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR);
$mb_shortcuts->command(-label => I_('Clear Package/Group Selections'), -underline => -1, -command => \&clearEntries, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_shortcuts->command(-label => I_('Select all Upgrades/New Packages'), -underline => -1, -command => \&selectUpgrades, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);

$mb_help=$mb->Menubutton( -tearoff => false, -text => I_('Help'), -underline => -1, -bg => $SECBGCOLOR, -fg => $TEXTCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR);
$mb_help->command(-label => I_('About'), -underline => -1, -command => \&about_window, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_help->command(-label => I_('Instructions'), -underline => -1, -command => \&mainhelp_window, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_help->command(-label => I_('Server Information'), -underline => -1, -command => \&servinfo_window, -background => $SECBGCOLOR, -foreground => $TEXTCOLOR);
$mb_help->pack(-side => 'right');

$topmsgFrame=$MW->Frame(-bg => $FRAMEBGCOLOR);
$topmsgFrame->pack(-side => 'top', -fill => 'x', -pady => '6');
$topmsgPic=$topmsgFrame->Label(-image => $topmsgFrame->Pixmap(-file => '/usr/lib/turbodesk/pixmaps/turbodesk.xpm'), -borderwidth => 0, -bg => $FRAMEBGCOLOR);
$topmsgPic->pack(-side => 'left', -padx => '4');
$topmsg_text=I_("XTurboPkg - Interactive Package Management");
$topmsg=$topmsgFrame->Message(-fg => $TITLECOLOR, -background => $FRAMEBGCOLOR, -width => '600', -font => $FONT, -text => $topmsg_text);
$topmsg->pack(-side => 'left', -padx => '4');

$MainFrame=$MW->Frame(-bg => $FRAMEBGCOLOR);
$MainFrame->pack(-side => 'top', -fill => 'both', -pady => '6');
$BottomFrame=$MW->Frame(-bg => $FRAMEBGCOLOR);
$BottomFrame->pack(-side => 'bottom', -pady => '6');



&parseTurboPkgRC;
looptop:
$ISFTP=0;
$BottomFrame->configure(-height => 1);
$MW->update;
&getSourceMedia;
&getMediaOptions;
&writeINIFile;
if (!(&initializeData)) {
    &choiceWin(I_("Error"),I_("Ok"),0,0,0,I_("Unable to Initialize Data."));
    &exit_program(1);
}
&getInitialPackageData;
&drawPackageList;
&remapMainButtons;

$PROGRAM_STATE=1;
$mb_actions->pack(-side => 'left');
$mb_shortcuts->pack(-side => 'left');
MainLoop;

&exit_program(1);

## Subroutines

sub remapMainFrame {
    $MainFrame->DESTROY if Exists($MainFrame);
    $MainFrame=$MW->Frame(-bg => $FRAMEBGCOLOR);
    $MainFrame->pack(-side => 'top', -fill => 'both');
}

sub remapMainButtons {
    $ButtonFrame=$MainFrame->Frame(-bg => $FRAMEBGCOLOR);
    $ButtonFrame->pack(-side => 'right', -padx => '4');
    $DoneButton=$ButtonFrame->Button(-width => 5, -bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Proceed'), -command => sub {
	&doInstall;
    } );
    $QuitButton=$ButtonFrame->Button(-width => 5, -bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Quit'), -command => sub {    
	&exit_program;
    } );
    $DoneButton->pack(-side => 'top', -pady => '4');
    $QuitButton->pack(-side => 'top', -pady => '4');
}
    

sub config_backcompat {
    if (!(length($FRAMEBGCOLOR))) {
	if ($DEBUG) {
	    print "** Assuming FRAMEBGCOLOR = MAINBGCOLOR\n";
	}
	$FRAMEBGCOLOR=$MAINBGCOLOR;
    }
    if (!(length($BUTTONTEXTCOLOR))) {
	if ($DEBUG) {
	    print "** Assuming BUTTONTEXTCOLOR = TEXTCOLOR\n";
	}
	$BUTTONTEXTCOLOR=$TEXTCOLOR;
    }
}

sub load_preferences {
    open(ifp,"$PREFSFILE");
    while(<ifp>) {
	chop;
	my $line=$_;
	if ((length($line)) && (substr($line,0,1) ne "#")) {
	    my ($key,$value)=split(//,$line);
	    $$key=$value;
	    if ($DEBUG) {
		print "PREFS: $key = $value\n";
	    }
	}
    }
    close ifp;
}
		
sub preferences_window {
    if ($DEBUG) {
	print "PREFS: Doing prefs window...\n";
    }
    my @PREF_ENTRIES=("DISP_LINES","CASCADE_DISTANCE");
    my @PREF_CHECKBOXES=("IFTYPE","AUTODEPS","DODOWNGRADES","DYNRESIZE",
			 "SORTGROUPS","CASCADE","SHOWGRAPHICS",
			 "IGNOREDEPS","UNATTENDED");
    $PREFLABELS{"IGNOREDEPS"}=I_("Ignore Dependencies");
    $PREFLABELS{"SHOWGRAPHICS"}=I_("Show Package Graphics");
    $PREFLABELS{"CASCADE"}=I_("Cascade Windows");
    $PREFLABELS{"CASCADE_DISTANCE"}=I_("Cascade Distance (pixels)");
    $PREFLABELS{"SORTGROUPS"}=I_("Sort Listing by Group");
    $PREFLABELS{"IFTYPE"}=I_("Double-Click Interface");
    $PREFLABELS{"DISP_LINES"}=I_("Display Height (lines)");
    $PREFLABELS{"AUTODEPS"}=I_("Automatic Dependency Resolution");
    $PREFLABELS{"DODOWNGRADES"}=I_("Do Package Downgrades");
    $PREFLABELS{"DYNRESIZE"}=I_("Display Size Info in Real-Time");
    $PREFLABELS{"UNATTENDED"}=I_("Unattended Mode - autocontinue on error");
    $PREFWIDTHS{"DISP_LINES"}=10;
    $PREFWIDTHS{"CASCADE_DISTANCE"}=10;

    foreach $prf (@PREF_ENTRIES) {
	$BACKUPVALUES{$prf}=$$prf;
    }
    foreach $prf (@PREF_CHECKBOXES) {
	$BACKUPVALUES{$prf}=$$prf;
    }
    $prefw->DESTROY if Exists($prefw);
    $prefw=$MW->Toplevel;
    $prefw->configure(-bg => $FRAMEBGCOLOR);
    $prefw->title(I_("Preferences"));
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$prefw->geometry($newGeometry);
    }

    $prefwTopText=$prefw->Label(-fg => $HEADINGCOLOR, -bg => $FRAMEBGCOLOR, -text => I_("Edit Preferences"));
    $prefwTopText->pack(-side => 'top', -pady => '6');
    
    foreach $eprf (sort(@PREF_ENTRIES)) {
	my $framename="$eprf" . "_frame";
	my $labelname="$eprf" . "_label";
	my $entryname="$eprf" . "_entry";
	$WIDGETS{$framename}=$prefw->Frame(-bg => $MAINBGCOLOR);
	$WIDGETS{$framename}->pack(-side => 'top', -pady => '2', -fill => 'x');
	$WIDGETS{$labelname}=$WIDGETS{$framename}->Label(-font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => $PREFLABELS{$eprf});
	$WIDGETS{$entryname}=$WIDGETS{$framename}->Entry(-font => 'fixed', -width => ($PREFWIDTHS{$eprf} || 40), -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $SECBGCOLOR, -highlightbackground => $SECBGCOLOR, -textvariable => \$$eprf);
	$WIDGETS{$labelname}->pack(-side => 'left');
	$WIDGETS{$entryname}->pack(-side => 'right');
    }
    
    foreach $cprf (sort(@PREF_CHECKBOXES)) {
	my $framename="$cprf" . "_frame";
	my $cboxname="$cprf" . "_cbox";

	$WIDGETS{$framename}=$prefw->Frame(-bg => $MAINBGCOLOR);
	$WIDGETS{$framename}->pack(-side => 'top', -pady => '2', -fill => 'x');
	$WIDGETS{$cboxname}=$WIDGETS{$framename}->Checkbutton(-font => 'fixed', -selectcolor => $SPECIALTEXTCOLOR, -text => $PREFLABELS{$cprf}, -variable => \$$cprf, -underline => -1, -background => $SECBGCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR, -foreground => $TEXTCOLOR, -highlightbackground => $MAINBGCOLOR);
	$WIDGETS{$cboxname}->pack(-side => 'left');
    }
    
    $prefwButtons=$prefw->Frame(-bg => $FRAMEBGCOLOR);
    $prefwButtons->pack(-side => 'bottom', -pady => '6');
    $prefwOkButton=$prefwButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('OK (Save)'), -command => sub {
	open(ofp,">$PREFSFILE");
	print ofp "# XTurboPkg Preferences File - do not edit.\n";
	print ofp "# XTurboPkg is (C) 1998 Pacific HiTech, Inc.\n";
	print ofp "\n";
	foreach $prf (@PREF_ENTRIES) {
	    print ofp "$prf$$prf\n";
	}
	foreach $prf (@PREF_CHECKBOXES) {
	    print ofp "$prf$$prf\n";
	}
	close ofp;
	&msgWin(I_("Warning"),I_("Ok"),I_("You must restart XTurboPkg for your changes to take effect."));
	$prefw->DESTROY;
    } );
    $prefwCancelButton=$prefwButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Cancel'), -command => sub {
	foreach $prf (@PREF_ENTRIES) {
	    $$prf=$BACKUPVALUES{$prf};
	}
	foreach $prf (@PREF_CHECKBOXES) {
	    $$prf=$BACKUPVALUES{$prf};
	}
	$prefw->DESTROY;
    } );
    $prefwOkButton->pack(-side => 'left', -padx => '4');
    $prefwCancelButton->pack(-side => 'left', -padx => '4');    
}

sub unmountCDROM {
    if (!($PROGMOUNTEDCDROM)) {
	return;
    }
    my $mcmd="umount /mnt/cdrom";
    system($mcmd);
}

sub exit_program {
    my $force=shift;
    my $confirm=2;
    if ($force == 1) {
	$confirm=1;
    } else {
	$confirm=&choiceWin(I_("Confirm"),I_("Yes"),I_("No"),0,0,I_("Are you sure you want to quit?"));
    }
    if ($DEBUG) {
	print "EXIT: Confirm is '$confirm'\n";
    }
    if ($confirm == 2) {
	return;
    }
    if ($DEBUG) {
	print "EXIT: Unmounting CD (if applicable) and exiting...\n";
    }
    &unmountCDROM;
    exit(0);
}

sub restart_program {
    &remapMainFrame;
    &statusWin("none",I_("Restarting XTurboPkg..."));
    $MW->update;
    close SOCK;
    &connectToServer || die "Cannot connect to server!\n";
    &load_preferences;    
    $swMsg->DESTROY if Exists($swMsg);       
    goto looptop;
}

sub parseTurboPkgRC {
    if (!(-e &CONFIGFILE)) {
	die "Turbopkgrc does not exist!\n";
    }
    open(ifp,&CONFIGFILE);
    while(<ifp>) {
	chop;
	my $line=$_;
	if ((length($line)) && (substr($line,0,1) ne "#")) {
	    if ($line =~ m/^MIRROR/) {
		my @f=split(//,$line);
		$MIRRORS{$f[0]}.="$f[1]$f[2]";
	    } elsif ($line =~ m/^UMIRROR/) {
		my @f=split(//,$line);
		$UMIRRORS{$f[0]}.="$f[1]$f[2]";
	    } else {
		my @f=split(//,$line);
		my $varname=$f[0];
		$$varname=$f[1];
	    }
	}
    }
    foreach $mkey (keys(%MIRRORS)) {
	if (substr($MIRRORS{$mkey},-1,1) eq "") {
	    chop $MIRRORS{$mkey};
	}
    }
    foreach $mkey (keys(%UMIRRORS)) {
	if (substr($UMIRRORS{$mkey},-1,1) eq "") {
	    chop $UMIRRORS{$mkey};
	}
    }
    close ifp;
    if ($DEBUG) {
	print "TURBOPKGRC: Main FTP Host/Dir: [$FTPHOST, $FTPDIR]\n";
	print "TURBOPKGRC: Updates FTP Host/Dir: [$UFTPHOST, $UFTPDIR]\n";
    }
}

sub getSourceMedia {
    my $gsmDone=0;
    $gsmTitle=$MainFrame->Message(-width => 300, -fg => $HEADINGCOLOR, -bg => $FRAMEBGCOLOR, -text => I_("Select Source Media Type:"));
    $gsmTitle->pack(-side => 'top', -pady => '10');
    $gsmCDROMFrame=$MainFrame->Frame(-bg => $MAINBGCOLOR);
    $gsmCDROMFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $gsmCDROM=$gsmCDROMFrame->Radiobutton(-selectcolor => $SPECIALTEXTCOLOR,-highlightthickness => '0', -background => $MAINBGCOLOR, -foreground => $TEXTCOLOR, -activeforeground => $TEXTCOLOR, -activebackground => $SECBGCOLOR, -text => I_('Use CD-ROM'), -value => 1, -variable => \$SOURCEMEDIA);
    $gsmLocalFrame=$MainFrame->Frame(-bg => $MAINBGCOLOR);
    $gsmLocalFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $gsmLocal=$gsmLocalFrame->Radiobutton(-selectcolor => $SPECIALTEXTCOLOR,-highlightthickness => '0', -background => $MAINBGCOLOR, -foreground => $TEXTCOLOR, -activeforeground => $TEXTCOLOR, -activebackground => $SECBGCOLOR, -text => I_('Use Local Filesystem'), -value => 2, -variable => \$SOURCEMEDIA);
    $gsmFTPFrame=$MainFrame->Frame(-bg => $MAINBGCOLOR);
    $gsmFTPFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $gsmFTP=$gsmFTPFrame->Radiobutton(-selectcolor => $SPECIALTEXTCOLOR,-highlightthickness => '0', -background => $MAINBGCOLOR, -foreground => $TEXTCOLOR, -activeforeground => $TEXTCOLOR, -activebackground => $SECBGCOLOR, -text => I_('Use FTP'), -value => 3, -variable => \$SOURCEMEDIA);
    $gsmNoneFrame=$MainFrame->Frame(-bg => $MAINBGCOLOR);
    $gsmNoneFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $gsmNone=$gsmNoneFrame->Radiobutton(-selectcolor => $SPECIALTEXTCOLOR,-highlightthickness => '0', -background => $MAINBGCOLOR, -foreground => $TEXTCOLOR, -activeforeground => $TEXTCOLOR, -activebackground => $SECBGCOLOR, -text => I_('No Source Media'), -value => 0, -variable => \$SOURCEMEDIA);
    $gsmCDROM->pack(-side => 'left');
    $gsmLocal->pack(-side => 'left');
    $gsmFTP->pack(-side => 'left');
    $gsmNone->pack(-side => 'left');
    $gsmButtons=$MainFrame->Frame(-bg => $FRAMEBGCOLOR);
    $gsmButtons->pack(-side => 'bottom', -pady => '6');
    $gsmContinueButton=$gsmButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Continue'), -command => sub {
	$gsmDone=1;
    } );
    $gsmContinueButton->pack(-side => 'left', -padx => '4');
    $gsmCDROM->invoke;
    while (!($gsmDone)) {
	$MW->update;
    }
}

sub connectToServer {
    $sockport="/usr/lib/turbolinux/turbopkgsock";
    my $waiting=0;
    my $rcmd="rm -f $sockport";
    if ($OKSTART) {
	system($rcmd);
    }
    if ($DEBUG) {
	print "CONNECT: Starting server...\n";
    }
    my $cmd="$SERVERBIN --autocontrol &";
    if ($DEBUG) {
	print "CONNECT: '$cmd'\n";
	$cmd="$SERVERBIN --debug --autocontrol &";
    }
    if ($OKSTART) {
	my $rc=system($cmd);
	if ($rc) {
	    return 0;
	}
    }
    if ($DEBUG) {
	print "CONNECT: Waiting for server...\n";
    }
    while (!(-e $sockport)) {}
    if ($DEBUG) {
	print "CONNECT: Creating socket...\n";
    }    
    socket(SOCK,AF_UNIX,SOCK_STREAM,0) || return 0;
    if ($DEBUG) {
	print "CONNECT: Connecting to socket...\n";
    }
    connect(SOCK,sockaddr_un($sockport)) || return 0;
    if ($DEBUG) {
	print "CONNECT: Waiting for Response...\n";
    }
    $waiting=1;
    while ($waiting) {
	if ($line = <SOCK>) {
	    $waiting=0;
	}
    }
    if (!($line =~ m/HELLO\:TurboPkg/)) {
	die "This is not a TurboPkg server!\n";
    }
    my $line2=substr($line,6);
    $SERVERINFO=$line2;
    if ($DEBUG) {	
	print "CONNECT: Connected to '$line2'.\n";
    }
    return 1;
}

sub getMediaOptions {
    my $rc=-1;
    if ($SOURCEMEDIA == 1) {
	$rc=&getCDROMOptions;
    }
    if ($SOURCEMEDIA == 2) {
	$rc=&getLocalOptions;
    }
    if ($SOURCEMEDIA == 3) {
	$rc=&getFTPOptions;
    }
    if ($SOURCEMEDIA == 0) {
	$rc=&getNoneOptions;
    }
    &remapMainFrame;
    if ($rc == 2) {
	goto looptop;
    }
}

sub getNoneOptions {
    $NOCOMPSFILE=1;
    $NORPMDIR=1;
    $NOHDLIST=1;
}

sub getCDROMOptions {
    open(ifp,"/etc/mtab");
    while(<ifp>) {
	chop;
	my $line=$_;
	if ($line =~ m/\/mnt\/cdrom/) {
	    if ($DEBUG) {
		print "CDROM: Already mounted.\n";
	    }
	    close ifp;
	    goto setcdpath;
	}
    }
    close ifp;
    my $rc=&choiceWin(I_("Mount CD-ROM"),I_("Ok"),I_("Cancel"),0,0,I_("Place the CD to be mounted into your CD drive and click 'OK'."));
    if ($rc == 1) {
	&statusWin(I_("Mounting"),I_("Mounting CD-ROM..."));
	$MW->update;
	my $mcmd="mount /mnt/cdrom 2>/dev/null";
	my $rc2=system($mcmd);
	$swMsg->DESTROY;
	if ($rc2) {
	    &choiceWin(I_("Error"),I_("Ok"),0,0,0,I_("Unable to mount CD-ROM."));
	    $rc=2;
	} else {
	    $PROGMOUNTEDCDROM=1;
	}
    }
  setcdpath:
    &getLocalOptions("/mnt/cdrom");
    return $rc;
}

sub getLocalOptions {
    my ($inpath) = @_;
    my $basepath="";
    if (length($inpath)) {
	$basepath=$inpath;
    } else {
	$basepath=&entryWin(I_("Base Path"),&DEFAULT_BASEPATH,I_("Enter the base path in which XTurboPkg should search for its data files:"));
    }
    if (!(length($basepath))) {
	&remapMainFrame;
	goto looptop;
    }
    my $scmd="\@FINDFILES:$basepath\n";
    if ($DEBUG) {
	print "LOCALOPTS: Sending file resolution request...\n";
    }
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    if (($line =~ m/PATHS\:/) || ($line =~ m/ERROR\:/)) {
		$waiting=0;
		if ($line =~ m/ERROR/) {
		    $err=1;
		    if ($DEBUG) {
			chop $line;
			print "LOCALOPTS: Received ERROR ($line)\n";
		    }
		    my @ee=split(/:/,$line);
		    if ($ee[1] =~ m/ServerTermBySignal/) {
			my @ee2=split(/\//,$ee[1]);
			$errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
			&msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
			exit(-1);
		    }
		} else {
		    if ($DEBUG) {
			chomp $line;
			print "LOCALOPTS: Received path info: '$line'\n";
		    }
		}
	    }
	}
    }
    my $rc2=0;
    chomp $line;
    my @f=split(/:/,$line);
    $COMPSFILEPATH=$f[1];
    $RPMSPATH=$f[2];
    $HDLISTPATH=$f[3];
    $rc2=&getManualPaths($err);
    if ($rc2 != 1) {
	return 2;
    }
    if (!($err)) {
	
	if ($DEBUG) {
	    print "LOCALOPTS: Paths parsed OK, returning 1\n";
	}
    }
    return 1;
}

sub mapFTPCustomOptions {
    my $parent=shift;
    if ($ftpCustomMapped == 1) {
	return;
    } else {
	$ftpCustomMapped=1;
    }
    $ftpcHostFrame=$parent->Frame(-bg => $MAINBGCOLOR);
    $ftpcHostFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $ftpcHostLabel=$ftpcHostFrame->Label(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Hostname:"));
    $ftpcHostEntry=$ftpcHostFrame->Entry(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -width => '30', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$FTPHOST);
    $ftpcHostEntry->pack(-side => 'right', -padx => '4');
    $ftpcHostLabel->pack(-side => 'right', -padx => '4');

    $ftpcDirFrame=$parent->Frame(-bg => $MAINBGCOLOR);
    $ftpcDirFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $ftpcDirLabel=$ftpcDirFrame->Label(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Remote Dir:"));
    $ftpcDirEntry=$ftpcDirFrame->Entry(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -width => '30', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$FTPDIR);
    $ftpcDirEntry->pack(-side => 'right', -padx => '4');
    $ftpcDirLabel->pack(-side => 'right', -padx => '4');

    $ftpcUsernameFrame=$parent->Frame(-bg => $MAINBGCOLOR);
    $ftpcUsernameFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $ftpcUsernameLabel=$ftpcUsernameFrame->Label(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Username:"));
    $ftpcUsernameEntry=$ftpcUsernameFrame->Entry(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -width => '30', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$FTPUSERNAME);
    $ftpcUsernameEntry->pack(-side => 'right', -padx => '4');
    $ftpcUsernameLabel->pack(-side => 'right', -padx => '4');

    $ftpcPasswordFrame=$parent->Frame(-bg => $MAINBGCOLOR);
    $ftpcPasswordFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $ftpcPasswordLabel=$ftpcPasswordFrame->Label(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Password:"));
    $ftpcPasswordEntry=$ftpcPasswordFrame->Entry(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -width => '30', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$FTPPASSWORD);
    $ftpcPasswordEntry->pack(-side => 'right', -padx => '4');
    $ftpcPasswordLabel->pack(-side => 'right', -padx => '4');
}

sub unmapFTPCustomOptions {
    my $parent=shift;
    if ($ftpCustomMapped == 0) {
	return;
    } else {
	$ftpCustomMapped=0;
    }
    foreach $ftpc ("Host","Dir","Username","Password") {
	my $framename="ftpc" . $ftpc . "Frame";
	$$framename->DESTROY;
    }
    $parent->configure(-height => 1);
    $parent->update;
}

sub mapFTPProxyOptions {
    my $parent=shift;
    if ($ftpProxyMapped == 1) {
	return;
    } else {
	$ftpProxyMapped=1;
    }
    $ftppHostFrame=$parent->Frame(-bg => $MAINBGCOLOR);
    $ftppHostFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $ftppHostLabel=$ftppHostFrame->Label(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Proxy Host:"));
    $ftppHostEntry=$ftppHostFrame->Entry(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -width => '30', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$FTPPROXY);
    $ftppHostEntry->pack(-side => 'right', -padx => '4');
    $ftppHostLabel->pack(-side => 'right', -padx => '4');

    $ftppPortFrame=$parent->Frame(-bg => $MAINBGCOLOR);
    $ftppPortFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $ftppPortLabel=$ftppPortFrame->Label(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Proxy Port:"));
    $ftppPortEntry=$ftppPortFrame->Entry(-highlightbackground => $MAINBGCOLOR, -font => 'fixed', -width => '30', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$FTPPORT);
    $ftppPortEntry->pack(-side => 'right', -padx => '4');
    $ftppPortLabel->pack(-side => 'right', -padx => '4');
}

sub unmapFTPProxyOptions {
    my $parent=shift;
    if ($ftpProxyMapped == 0) {
	return;
    } else {
	$ftpProxyMapped=0;
    }
    foreach $ftpp ("Host","Port") {
	my $framename="ftpp" . $ftpp . "Frame";
	$$framename->DESTROY;
    }
    $parent->configure(-height => 1);
    $parent->update;
}

sub getFTPOptions {
    my $optDone=0;
    my $customSelected=0;
    $ftpCustomMapped=0;
    $ftpProxyMapped=0;
    my $language=$ENV{LANG};
    my $key="MIRROR";
    my $hostkey="FTPHOST";
    my $dirkey="FTPDIR";
    my $ukey="UMIRROR";
    my $uhostkey="UFTPHOST";
    my $udirkey="UFTPDIR";
    if ((length($language)) && ($language ne "C")) {
	$key .= ".";
	$key .= $language;
	$hostkey .= ".";
	$hostkey .= $language;
	$dirkey .= ".";
	$dirkey .= $language;
    }
    my $isUpdate=&yesnobox(I_("FTP Host Type"),
			   I_("Do you want to connect to a host containing only updates instead of a full TurboLinux distribution?  Note that if you want to connect to a custom location, it does not matter which of these options you select."),
			   350);
    if ($DEBUG) {
	print "ISUPDATE: $isUpdate\n";
    }
    my $msitedata="";
    if ($isUpdate) {
	$msitedata=$UMIRRORS{$ukey};
	$hostkey=$uhostkey;
	$key=$ukey;
	$dirkey=$udirkey;
	$FTPHOST=$UFTPHOST;
	$FTPDIR=$UFTPDIR;
    } else {
	$msitedata=$MIRRORS{$key};
    }
    if ($DEBUG) {
	print "FTPINFO: $hostkey, $dirkey, $key\n";
    }
    my @msites=split(//,$msitedata);
    undef @msitenames;
    undef %FTPBINDINGS;
    foreach $msite (sort(@msites)) {
	my @mm=split(//,$msite);
	my ($mhost,$mdir)=split(/:/,$mm[1]);
	my $ftpName="$mm[0] ($mhost)";
	$FTPBINDINGS{$ftpName}=$mm[1];
    }
    my $defKey=sprintf(I_("Master Site (%s)"),$$hostkey);
    my $defVal=sprintf("%s:%s",$$hostkey,$$dirkey);
    my $customKey=I_("Custom");
    my $customVal=$customKey;
    if ($DEBUG) {
	print "FTP: Master site is '$defKey' <> '$defVal'\n";
    }
    $ftpWin->DESTROY if Exists($ftpWin);
    $ftpWin=$MW->Toplevel;
    $ftpWin->configure(-bg => $FRAMEBGCOLOR);
    $ftpWin->title(I_("FTP Options"));
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$ftpWin->geometry($newGeometry);
    }
    $ftpTopMsg=$ftpWin->Message(-width => 300, -fg => $HEADINGCOLOR, -bg => $FRAMEBGCOLOR, -text => I_("FTP Options"));
    $ftpTopMsg->pack(-side => 'top', -pady => '10');
    my @hostlist0=sort(keys(%FTPBINDINGS));
    my @hostlist=($defKey,$customKey,@hostlist0);
    $FTPBINDINGS{$defKey}=$defVal;
    $FTPBINDINGS{$customKey}=$customVal;
    $ftpHostFrame=$ftpWin->Frame(-bg => $MAINBGCOLOR);
    $ftpHostFrame->pack(-side => 'top', -pady => '4');
    $ftpHostLabel=$ftpHostFrame->Label(-font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("FTP Sites:"));
    $ftpHostLabel->pack(-side => 'left', -padx => '4');
    &createScrolledListbox($ftpHostFrame,"ftpsites",0,5,'fixed','single',@hostlist);
    $WIDGETS{ftpsites}->pack(-side => 'right', -padx => '4');
    if ($IFTYPE == 1) {
	$WIDGETS{ftpsites_listbox}->bind("<Double-Button-1>" => sub {	
	    my $selIndex=$WIDGETS{ftpsites_listbox}->curselection;
	    if ($selIndex != 1) {
		my $selection=$WIDGETS{ftpsites_listbox}->get($selIndex);
		my $selData=$FTPBINDINGS{$selection};
		($FTPHOST,$FTPDIR)=split(/:/,$selData);
		if ($DEBUG) {
		    print "FTP: Setting FTPHOST = $FTPHOST, FTPDIR = $FTPDIR\n";
		}
	    }
	} );
    } else {
	$WIDGETS{ftpsites_listbox}->bind("<Button-1>" => sub {	
	    my $selIndex=$WIDGETS{ftpsites_listbox}->curselection;
	    if ($selIndex != 1) {
		my $selection=$WIDGETS{ftpsites_listbox}->get($selIndex);
		my $selData=$FTPBINDINGS{$selection};
		($FTPHOST,$FTPDIR)=split(/:/,$selData);
		if ($DEBUG) {
		    print "FTP: Setting FTPHOST = $FTPHOST, FTPDIR = $FTPDIR\n";
		}
	    }
	} );
    }
    $WIDGETS{ftpsites_listbox}->selection('set',0);

    $ftpCustomFrame=$ftpWin->Frame(-bg => $MAINBGCOLOR);
    $ftpCustomFrame->pack(-side => 'top', -pady => '4');

    $ftpProxy1Frame=$ftpWin->Frame(-bg => $MAINBGCOLOR);
    $ftpProxy1Frame->pack(-side => 'top', -pady => '4', -fill => 'x');
    $ftpProxyCheck=$ftpProxy1Frame->Checkbutton(-font => 'fixed', -selectcolor => $SPECIALTEXTCOLOR, -text => I_("Use FTP Proxy"), -variable => \$USE_PROXY, -underline => -1, -background => $SECBGCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR, -foreground => $TEXTCOLOR, -highlightbackground => $MAINBGCOLOR);
    $ftpProxyCheck->pack(-side => 'left');

    $ftpProxy2Frame=$ftpWin->Frame(-bg => $MAINBGCOLOR);
    $ftpProxy2Frame->pack(-side => 'top', -pady => '4');

    $ftpButtons=$ftpWin->Frame(-bg => $FRAMEBGCOLOR);
    $ftpButtons->pack(-side => 'bottom', -pady => '6');
    $ftpOkButton=$ftpButtons->Button(-width => 5, -bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Ok'), -command => sub {
	$optDone=1;
	$ISFTP=1;
	$COMPSFILEPATH=&COMPSFILE_FTP;
	$RPMSPATH=&RPMSDIR_FTP;
	$HDLISTPATH=&HDLIST_FTP;
	my $cleancmd="rm -f $COMPSFILEPATH $HDLISTPATH";
	system($cleancmd);
    } );
    $ftpCancelButton=$ftpButtons->Button(-width => 5, -bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Cancel'), -command => sub {
	$optDone=2;
    } );
    $ftpOkButton->pack(-side => 'left', -padx => '4');
    $ftpCancelButton->pack(-side => 'left', -padx => '4');
	
    
    while (!($optDone)) {
	my $listSelection=$WIDGETS{ftpsites_listbox}->curselection;
	if ($listSelection == 1) {
	    &mapFTPCustomOptions($ftpCustomFrame);
	} else {
	    if (length($listSelection)) {
		&unmapFTPCustomOptions($ftpCustomFrame);
	    }
	}
	if ($USE_PROXY == 1) {
	    &mapFTPProxyOptions($ftpProxy2Frame);
	} else {
	    &unmapFTPProxyOptions($ftpProxy2Frame);
	}
	$MW->update;
    }
    $ftpWin->DESTROY;
    return $optDone;
}

sub writeINIFile {
    my $ofn=&AUTOCONTROL_INI;
    if ($DEBUG) {
	print "INI: Writing INI File: $ofn\n";
    }
    open(ofp,">$ofn");
    foreach $outvar ("COMPSFILEPATH","RPMSPATH","HDLISTPATH","FTPHOST","FTPDIR","FTPUSERNAME",
		     "FTPPASSWORD","FTPPROXY","FTPPORT","ISFTP","NOCOMPSFILE","NORPMDIR",
		     "NOHDLIST") {
	print ofp "$$outvar\n"
	}
    close ofp;
}

sub initializeData {
    my $sini=&AUTOCONTROL_INI;
    my $scmd="\@INITIALIZE:$sini\n";
    &statusWin(I_("Initializing"),I_("Initializing Data..."));
    $MW->configure(-cursor => 'watch');
    $MW->update;
    if ($DEBUG) {
	print "INI: Sending INITIALIZE command...\n";
    }
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    my $errstr="";
    while ($waiting) {
	if ($line = <SOCK>) {
	    if (($line =~ m/SUCCESS\:/) || ($line =~ m/ERROR\:/)) {
		$waiting=0;
		if ($line =~ m/ERROR/) {
		    $err=1;
		    chop $line;
		    my @ee=split(/:/,$line);
		    if ($ee[1] =~ m/ServerTermBySignal/) {
			my @ee2=split(/\//,$ee[1]);
			$errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
			&msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
			exit(-1);
		    }
		}
	    } elsif ($line =~ m/^FTP\:/) {
		chop $line;
#		if ($DEBUG) {
#		    print "FTP_RESP: '$line'\n";
#		}
		my ($foo,$filename,$filesize,$progress)=split(/:/,$line);
		if ($filename eq "Complete") {
		    $pwMsg->DESTROY;
		} else {
		    if (!(Exists($pwMsg))) {
			$FTPPROGRESS=0;
			my $prgText="";
			if ($filesize == 0) {
			    $prgText=sprintf(I_("Retrieving '%s' (SIZE not available)"),$filename);
			    $FTPSERVNOSIZE=1;
			} else {
			    $FTPSERVNOSIZE=0;
			    $prgText=sprintf(I_("Retrieving '%s' (%d kb)"),$filename,($filesize / 1024));
			}
			&progressWin("none",$prgText,$MainFrame,"FTPPROGRESS",I_("Cancel"));
		    }
		    $FTPPROGRESS=$progress;
		    $MW->update;
		}
	    }
	}
	if ($CANCELLED == 1) {
	    my $ccmd="\@CANCEL\n";
	    send(SOCK,$ccmd,0);
	    $CANCELLED=2;
	}
	if ($CANCELLED == 2) {
	    if ($line =~ m/CANCELLED\:/) {
		if ($DEBUG) {
		    print "CANCEL: Server sent cancel response.\n";
		}
		$waiting=0;
	    }
	}
	$MW->update;
    }
    if ($CANCELLED > 0) {
	if ($DEBUG) {
	    print "CANCELLED: Operation Cancelled.\n";
	}
	$CANCELLED=0;
	$swMsg->DESTROY if Exists($swMsg);
	$pwMsg->DESTROY if Exists($pwMsg);
	$MW->configure(-cursor => 'left_ptr');
	$MW->update;
	if ($DEBUG) {
	    print "RESTART: Closing SOCK and reconnecting...\n";
	}
	&statusWin("none",I_("Restarting TurboPkg Server..."));
	$MW->update;
	close SOCK;
	&connectToServer || die "Cannot connect to server!\n";
	$swMsg->DESTROY if Exists($swMsg);       
	goto looptop;
    }
    $swMsg->DESTROY;
    $MW->configure(-cursor => 'left_ptr');
    $MW->update;
    if ($err) {
	if (length($errstr)) {
	    &msgWin(I_("Error"),I_("Ok"),$errstr);
	}
	return 0;
    } else {
	if ($DEBUG) {
	    print "INI: Success.\n";
	    $scmd="\@SUPPORTFILESTATUS\n";
	    send(SOCK,$scmd,0);
	    my $waiting=1;
	    while ($waiting) {
		if ($line = <SOCK>) {
		    $waiting=0;
		    print $line;
		}
	    }
	}
	return 1;
    }
}

sub getManualPaths {
    my ($mode) = @_;
    my $title="";
    my $topmsg="";
    if ($mode == 1) {
	($title,$topmsg)=(I_("Path Entry"),I_("Please enter the paths where XTurboPkg can find its data files.  If one or more of the files/paths does not exist, simply ignore it and click OK anyway."));
    } else {
	($title,$topmsg)=(I_("Confirm Paths"),I_("Please confirm that the following path information is correct."));
    }
    my $mpewDone=0;
    my $rvalstr=$defaultentrytext;

    $mpew->DESTROY if Exists($mpew);
    $mpew=$MW->Toplevel;
    $mpew->configure(-bg => $FRAMEBGCOLOR);
    $mpew->title($title);
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$mpew->geometry($newGeometry);
    } else {
	my $newX=$mpew->pointerx;
	my $newY=$mpew->pointery;
	$newX -= 200;
	$newY -= 100;
	my $newGeometry="=" . "+" . "$newX" . "+" . "$newY";
	if ($DEBUG) {
	    print "[$newX,$newY] [$newGeometry]\n";
	}
	$mpew->geometry($newGeometry);
    }
    $mpewMsg=$mpew->Message(-fg => $HEADINGCOLOR, -bg => $MAINBGCOLOR, -width => 400, -text => $topmsg);
    $mpewMsg->pack(-side => 'top', -pady => '10');
    $mpewFrame1=$mpew->Frame(-bg => $MAINBGCOLOR);
    $mpewFrame1->pack(-side => 'top', -pady => '2', -fill => 'x');
    $mpewLabel1=$mpewFrame1->Label(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Components File:"));
    $mpewLabel1->pack(-side => 'left');
    $mpewEntry1=$mpewFrame1->Entry(-width => '40', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$COMPSFILEPATH);
    $mpewEntry1->pack(-side => 'right');
    $mpewFrame2=$mpew->Frame(-bg => $MAINBGCOLOR);
    $mpewFrame2->pack(-side => 'top', -pady => '2', -fill => 'x');
    $mpewLabel2=$mpewFrame2->Label(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Path to RPMs:"));
    $mpewLabel2->pack(-side => 'left');
    $mpewEntry2=$mpewFrame2->Entry(-width => '40', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$RPMSPATH);
    $mpewEntry2->pack(-side => 'right');
    $mpewFrame3=$mpew->Frame(-bg => $MAINBGCOLOR);
    $mpewFrame3->pack(-side => 'top', -pady => '2', -fill => 'x');
    $mpewLabel3=$mpewFrame3->Label(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Header List (hdlist):"));
    $mpewLabel3->pack(-side => 'left');
    $mpewEntry3=$mpewFrame3->Entry(-width => '40', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -textvariable => \$HDLISTPATH);
    $mpewEntry3->pack(-side => 'right');

    $mpewButtons=$mpew->Frame(-bg => $FRAMEBGCOLOR);
    $mpewButtons->pack(-side => 'bottom', -pady => '6');
    $mpewOkButton=$mpewButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_("Ok"), -command => sub {
		$mpewDone=1;
	    } );
    $mpewCancelButton=$mpewButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_("Cancel"), -command => sub {
		$mpewDone=2;
	    } );
    $mpewOkButton->pack(-side => 'left', -padx => '4');
    $mpewCancelButton->pack(-side => 'left', -padx => '4');
    while ($mpewDone == 0) {
	$MW->update;
    }
    $mpew->DESTROY;
    if (($COMPSFILEPATH eq I_("Ignored")) || (!(-e $COMPSFILEPATH))) {
	$NOCOMPSFILE=1;
    }
    if (($RPMSPATH eq I_("Ignored")) || (!(-e $RPMSPATH))) {
	$NORPMDIR=1;
    }
    if (($HDLISTPATH eq I_("Ignored")) || (!(-e $HDLISTPATH))) {
	$NOHDLIST=1;
    }
    if ($DEBUG) {
	print "NOFLAGS: [$NOCOMPSFILE $NORPMDIR $NOHDLIST]\n";
    }	
    return $mpewDone;
}

sub getInitialPackageData {
    undef %PKGLINES;
    undef %GROUPLINES;
    undef %GPKGS;
    my $scmd="\@BEGINPACKAGEMANAGEMENT\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $prevgrp="";
    $NUMGROUPS=0;
    $NUMPKGS=0;
    print "Getting package data...\n";
    &statusWin(I_("Analyzing"),I_("Analyzing Data..."),$MainFrame,I_("Cancel"));
    $MW->configure(-cursor => 'watch');
    $MW->update;
    while ($waiting) {
	if ($line = <SOCK>) {
	    if ($line =~ m/^GROUP\:/) {
		chop $line;
		my $s1=substr($line,6);
		my @ss=split(//,$s1);
		my @ss2=split(/:/,$ss[1]);
		$GROUPLINES{$ss2[$#ss2]}=$ss[0];
		$prevgrp=$ss2[$#ss2];
		$NUMGROUPS++;
		my $newStatus=sprintf(I_("Analyzing Data (%d groups, %d packages)"),$NUMGROUPS,$NUMPKGS);
		$swMsgMsg->configure(-text => $newStatus);
		$MW->update;
	    }
	    elsif ($line =~ m/^PACKAGE\:/) {
		chop $line;
		my $s1=substr($line,(index($line,"[")));
		my @ss=split(//,$s1);
		my $ss2=substr($ss[0],4);
		my @ss3=split(/\s+/,$ss2);
		my $namegrp=$ss3[0] . "|" . $prevgrp;
		$ALLPKGS{$ss3[0]}=$prevgrp;
		$PKGLINES{$namegrp}=$ss[0];
		$GPKGS{$prevgrp} .= $ss3[0];
		$GPKGS{$prevgrp} .= ":";
		$NUMPKGS++;
		my $newStatus=sprintf(I_("Analyzing Data (%d groups, %d packages)"),$NUMGROUPS,$NUMPKGS);
		$swMsgMsg->configure(-text => $newStatus);
		$MW->update;
	    }
	    if ($line =~ m/ENDINITIALDATA\:/) {
		$waiting=0;
	    }
	    if ($line =~ m/ERROR\:/) {
		$waiting=0;
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    }
	}
	if ($CANCELLED == 1) {
	    my $ccmd="\@CANCEL\n";
	    send(SOCK,$ccmd,0);
	    $CANCELLED=2;
	}
	if ($CANCELLED == 2) {
	    if ($line =~ m/CANCELLED\:/) {
		if ($DEBUG) {
		    print "CANCEL: Server sent cancel response.\n";
		}
		$waiting=0;
	    }
	}
    }
    if ($CANCELLED > 0) {
	if ($DEBUG) {
	    print "CANCELLED: Operation Cancelled.\n";
	}
	$CANCELLED=0;
	$swMsg->DESTROY;
	$MW->configure(-cursor => 'left_ptr');
	$MW->update;
	goto looptop;
    }
    if ($DEBUG) {
	foreach $gname (sort(keys(%GROUPLINES))) {
	    print "GROUP: $gname ($GROUPLINES{$gname})\n";
	}
	foreach $pname (sort(keys(%PKGLINES))) {
	    print "PKG: $pname\n";
	}
    }
    foreach $gpkg (keys(%GPKGS)) {
	chop $GPKGS{$gpkg};
    }
    $swMsg->DESTROY;
    $MW->configure(-cursor => 'left_ptr');
    $MW->update;
    return 1;
}

sub resetScreen {
    &remapMainFrame;
    $pSizeLabel->DESTROY if Exists($pSizeLabel);
    $pHlist->DESTROY if Exists($pHlist);
    &drawPackageList;
    &remapMainButtons;
}

sub reSyncPackages {
    &getInitialPackageData;
    &resetScreen;
}

sub drawPackageList {
    my $force=shift;
    my $redrawpackages=1;
    undef %PKGPARENTS;
    &statusWin("none",I_("Creating List..."),$BottomFrame);
    $MW->configure(-cursor => 'watch');
    $MW->update;
    if ($DYNRESIZE) {
	if ($DEBUG) {
	    print "REDRAW: Getting size of selected package set.\n";
	}
	my $scmd="\@GETSETSIZE\n";
	send(SOCK,$scmd,0);
	my $waiting=1;
	while($waiting) {
	    if ($line = <SOCK>) {
		chop $line;
		if ($line =~ m/^SETSIZE\:/) {
		    my @ss=split(/:/,$line);
		    $SETSIZE=$ss[1];
		    $SETCOUNT=$ss[2];
		    
		    $waiting=0;
		}
		if ($line =~ m/^ERROR\:/) {
		    my @ee=split(/:/,$line);
		    if ($ee[1] =~ m/ServerTermBySignal/) {
			my @ee2=split(/\//,$ee[1]);
			$errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
			&msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
			exit(-1);
		    }
		}
	    }
	}
	if ($DEBUG) {
	    print "REDRAW: New size is $SETSIZE ($SETCOUNT)\n";
	}
    }
    my $sizeText=sprintf(I_("Total Size of Selection: %d Pkgs (%d MB)"),$SETCOUNT,$SETSIZE);
    if ($DYNRESIZE == 0) {
	$sizeText=I_("Selection Size Not Calculated");
    }
    if (Exists($pSizeLabel)) {
	$pSizeLabel->configure(-text => $sizeText);
    } else {
	$pSizeLabel=$MainFrame->Label(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => $sizeText);
	$pSizeLabel->pack(-side => 'top', -pady => '6');
    }
    if (Exists($pHlist)) {
	if ($force == 1) {
	    $pHlist->delete(all);
	    $redrawpackages=1;
	} else {
	    $redrawpackages=0;
	}
    } else {
	$pScrollbar=$MainFrame->Scrollbar(-bg => $SECBGCOLOR, -highlightbackground => $MAINBGCOLOR, -troughcolor => $TROUGHCOLOR);
	$pHlist=$MainFrame->HList(-indent => 25, -font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -drawbranch => 0, -height => $DISP_LINES, -itemtype => 'text', -width => 0, -highlightbackground => $MAINBGCOLOR, -yscrollcommand => ['set', $pScrollbar]);
	$pHlist->pack(-side => 'left', -fill => 'both', -pady => '6');
	$pScrollbar->pack(-side=> 'left', -fill => 'y', -pady => '6');
	$pScrollbar->configure(-command => ['yview', $pHlist]);
	if ($IFTYPE == 1) {
	    $pHlist->bind("<Double-Button-1>" => sub {
		&toggleItem;
		&drawPackageList;
	    } );
	} else {
	    $pHlist->bind("<Button-1>" => sub {
		&toggleItem;
		&drawPackageList;
	    } );
	}
	$pHlist->bind("<Button-3>" => sub {
	    &rightClickMenu;
	} );
    }
    if ($redrawpackages) {
	undef @GKEYS;
	if ($SORTGROUPS == 1) {
	    @GKEYS=sort(keys(%GROUPLINES));
	} else {
	    @GKEYS=keys(%GROUPLINES);
	}
	foreach $gname (@GKEYS) {
	    my $goutput=$GROUPLINES{$gname};
	    my $gdata="G:$gname";
	    $FIRSTENTRY=$gname unless (length($FIRSTENTRY));
	    $pHlist->add($gname, -data => $gdata, -text => $goutput);
	    if ($SHOWPACKAGES == 1) {
		foreach $pname (sort(keys(%PKGLINES))) {
		    my @p=split(/\|/,$pname);
		    if ($p[1] eq $gname) {
			my $poutput=$PKGLINES{$pname};
			my $pdata="P:$p[0]";
			$pHlist->addchild($gname, -data => $pdata, -text => $poutput);
		    }
		}
	    }
	}
    }
    $swMsg->DESTROY;
    $MW->configure(-cursor => 'left_ptr');
    $MW->update;
}

sub toggleItem {
    if ($SOURCEMEDIA == 0) {
	&msgWin(I_("Error"),I_("Ok"),I_("You have not selected any source media, therefore you cannot add or update packages."));
	return;
    }
    my $pkgselection=$pHlist->info(selection);
    my $pkgdata=$pHlist->entrycget($pkgselection,-data);
    if (substr($pkgdata,0,1) eq "P") {
	&togglePackage;
    }
    elsif (substr($pkgdata,0,1) eq "G") {
	&toggleGroup;
    }
    else {
	(&msgWin(I_("Error"),I_("Ok"),I_("Unknown Data Type.")));
    }
}

sub togglePackage {
    my ($ipkgsel)=@_;
    my $pkgselection=$pHlist->info(selection);
    if ((length($ipkgsel)) && (!($ipkgsel =~ m/Tk\:\:HList/))) {
	if ($DEBUG) {
	    foreach $parg (@_) {
		print "TOGGLEPKG: ARG: '$parg'\n";
	    }
	    print "TOGGLEPKG: Using arg '$ipkgsel' for selection\n";
	}
	$pkgselection=$ipkgsel;
    }
    my $pkgstr=$pHlist->entrycget($pkgselection,-text);

    if ($DEBUG) {
	print "TOGGLEPKG: Original '$pkgstr'\n";
    }
    my $isselected=substr($pkgstr,1,1);
    if ($isselected eq "R") {
	&msgWin(I_("Error"),I_("Ok"),I_("This package is marked for removal.  Unmark it first, then you can select it for upgrading."));
	return;
    }
    my $s1=substr($pkgstr,4);
    my @ss=split(/\s+/,$s1);
    my $pkgname=$ss[0];
    my $newvalue=0;
    $newvalue=1 unless ($isselected eq "*");
    if ($DEBUG) {
	print "TOGGLEPKG: Signalling server...\n";
    }
    my $scmd="\@SETPACKAGESTATE:$pkgname:$newvalue\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if (($line =~ m/ERROR/) || ($line =~ m/SUCCESS/)) {
		$waiting=0;
		if ($DEBUG) {
		    print "TOGGLEPKG: Response: '$line'\n";
		}
	    }
	    if ($line =~ m/ERROR/) {
		$err=1;
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    }
	}
    }
    if ($err) {
	&msgWin(I_("Error"),I_("Error changing state of package."));
	return;
    }
    my $skey="$pkgname" . "\\|";
    $skey =~ s/\+/\\\+/g;
    $skey =~ s/\-/\\\-/g;
    foreach $pk (keys(%PKGLINES)) {
	if ($pk =~ m/^$skey/) {
	    if ($isselected eq " ") {
		substr($PKGLINES{$pk},1,1)="*";
	    } else {
		substr($PKGLINES{$pk},1,1)=" ";
	    }
	    $pHlist->entryconfigure($pkgselection, -text => $PKGLINES{$pk});
	    if ($DEBUG) {
		print "TOGGLEPKG: '$PKGLINES{$pk}'\n";
	    }
	}
    }
    my $parentgroup=$pHlist->info(parent,$pkgselection);
    my $cc=getGroupStatus($parentgroup);
    if ($DEBUG) {
	print "GROUPSTATUS: [$cc]\n";
    }
    &toggleGroup($parentgroup,$cc);
    
}

sub setPackageByName {
    my ($pkgname,$setvalue,@pkglist)=@_;
    my $selfound=0;
    $pkgselection=0;
    if ($SHOWPACKAGES == 1) {
	foreach $possiblepkg (@pkglist) {
	    my $pkgstr=$pHlist->entrycget($possiblepkg, -text);
	    my $s1=substr($pkgstr,4);
	    my @ss=split(/\s+/,$s1);
	    my $xpkgname=$ss[0];
	    if ($xpkgname eq $pkgname) {
		$pkgselection=$possiblepkg;
		$selfound=1;
	    }
	}
	if ($selfound == 0) {
	    &msgWin(I_("Error"),I_("Ok"),I_("An unknown error has occurred changing the state of this package."));
	    return;
	}
    }
    my $newvalue=0;
    if ($setvalue eq "*") {
	$newvalue=1;
    }
    my $skey="$pkgname" . "\\|";
    $skey =~ s/\+/\\\+/g;
    $skey =~ s/\-/\\\-/g;
    foreach $pk (keys(%PKGLINES)) {
	if ($pk =~ m/^$skey/) {
	    substr($PKGLINES{$pk},1,1)="$setvalue";
	    if ($SHOWPACKAGES == 1) {
		$pHlist->entryconfigure($pkgselection, -text => $PKGLINES{$pk});
	    }
	    if ($DEBUG) {
		print "SETPKGBYNAME: '$PKGLINES{$pk}'\n";
	    }
	}
    }
}

sub toggleMulti {
    my ($pkgliststr,$newvalue)=@_;
    my @pkglist=split(/:/,$pkgliststr);
    my $numpkgs=($#pkglist + 1);
    if ($DEBUG) {
	print "MULTITOGGLEPKG: Signalling server to change state of $numpkgs packages...\n";
    }
    my $scmd="\@MULTISETPACKAGESTATE:$numpkgs:$newvalue:$pkgliststr\n";
    if ($DEBUG) {
	print "MULTITOGGLEPKG: '$scmd'\n";
    }
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if (($line =~ m/ERROR/) || ($line =~ m/SUCCESS/)) {
		$waiting=0;
		if ($DEBUG) {
		    print "MULTITOGGLEPKG: Response: '$line'\n";
		}
	    }
	    if ($line =~ m/ERROR/) {
		$err=1;
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    }
	}
    }
    if ($err) {
	&msgWin(I_("Error"),I_("Error changing state of package list."));
	return 0;
    } else {
	return 1;
    }
}

sub getGroupStatus {
    my $grpentry=shift;
    my $num_on=0;
    my $num_off=0;
    my @childlist=$pHlist->info(children,$grpentry);
    my $num_total=($#childlist + 1);
    foreach $child (@childlist) {
	my $childText=$pHlist->entrycget($child, -text);
	if (substr($childText,1,1) eq "*") {
	    $num_on++;
	} else {
	    $num_off++;
	}
    }
    if ($num_on == $num_total) {
	return '*';
    } elsif ($num_off == $num_total) {
	return ' ';
    } else {
	return '-';
    }
}
	

sub toggleGroup {
    &statusWin("none",I_("Please Wait..."),$BottomFrame);
    $MW->configure(-cursor => 'watch');
    $MW->update;
    my $ingrpselection=shift;
    my $multipkglist="";
    my $multipkgcount=0;
    my $grpselection=$pHlist->info(selection);
    my $expVal=shift;
    if ($expVal) {
	$grpselection=$ingrpselection;
    }    
    my $grpstr=$pHlist->entrycget($grpselection,-text);
    my $isselected=substr($grpstr,1,1);
    my $s1=substr($grpstr,4);
    my $grpname=$s1;
    if ($expVal) {
	if ($DEBUG) {
	    print "TOGGLEGRP: Explicit value specified '$expVal'\n";
	}
	substr($GROUPLINES{$grpname},1,1)=$expVal;
    } else {
	if ($isselected eq "-") {
	    $isselected=" ";
	}
	if ($isselected eq " ") {
	    substr($GROUPLINES{$grpname},1,1)="*";
	} else {
	    substr($GROUPLINES{$grpname},1,1)=" ";
	}
    }
    $pHlist->entryconfigure($grpselection, -text => $GROUPLINES{$grpname});
    if ($DEBUG) {
	print "TOGGLEGRP: '$GROUPLINES{$grpname}'\n";
    }
    if ($expVal) {
	$swMsg->DESTROY;
	$MW->configure(-cursor => 'left_ptr');
	$MW->update;
	return;
    }

    undef @grpmembers;
    if ($SHOWPACKAGES == 1) {
	@grpmembers=$pHlist->info(children,$grpname);
    } else {
	@grpmembers=split(/:/,$GPKGS{$grpname});
    }
    if (!($expVal)) {
	my $skey="\\|" . $grpname;
	foreach $pk (keys(%PKGLINES)) {
	    if ($pk =~ m/$skey$/) {
		my @pp=split(/\|/,$pk);
		$multipkglist .= $pp[0];
		$multipkglist .= ":";
		$multipkgcount++;
		if ($expVal) {
		    &setPackageByName($pp[0],$expVal,@grpmembers);
		} else {
		    if ($isselected eq " ") {
			&setPackageByName($pp[0],"*",@grpmembers);
		    } else {
			&setPackageByName($pp[0]," ",@grpmembers);
		    }
		}
	    }
	}
    }
    if (!($expVal)) {
	my $tglVal=0;
	if ($isselected eq " ") {
	    $tglVal=1;
	}
	
	chop $multipkglist;
	&toggleMulti($multipkglist,$tglVal);
    }
    $swMsg->DESTROY;
    $MW->configure(-cursor => 'left_ptr');
    $MW->update;
}

sub doQuery {
    my $selection=$pHlist->info(selection);
    my $selData=$pHlist->entrycget($selection, -data);
    if (substr($selData,0,1) eq "G") {
	&msgWin(I_("Error"),I_("Ok"),I_("Only individual packages may be queried."));
	return;
    }
    my $pkgname=substr($selData,2);
    if ($DEBUG) {
	print "QUERY: Querying '$pkgname'\n";
    }
    undef @QRLINES;
    my $qindex=0;
    my $scmd="\@QUERYPACKAGE:$pkgname\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    $line="";
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/SUCCESS\:/) {
		$waiting=0;
	    } elsif ($line =~ m/ERROR\:/) {
		$waiting=0;
		$err=1;
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    } else {
		if ($DEBUG) {
		    if ($line =~ m//) {
			print "QUERY: Line contains control-A subst char\n";
		    }
		}
		$line =~ s//\n/g;
		if ($DEBUG) {
		    print "QUERY: LINE: '$line'\n";
		}
		$QRLINES[$qindex]=$line;
		$qindex++;
	    }
	}
    }
    if ($err == 1) {
	&msgWin(I_("Error"),I_("Ok"),I_("Unable to query package."));
	return;
    }
    if ($DEBUG) {
	print "QUERY: Retrieved $qindex lines from query.\n";
    }
    if ($SHOWGRAPHICS) {
	$PICFN="";
	&statusWin("none",I_("Extracting Graphics"),$BottomFrame);
	$MW->update;
	my $gscmd="\@EXTRACTPICTURE:$pkgname:/var/tmp/" . "$pkgname" . "rpmpic" . "\n";
	$waiting=1;
	$err=0;
	send(SOCK,$gscmd,0);
	while ($waiting) {
	    if ($line = <SOCK>) {
		chop $line;
		if ($line =~ m/ERROR\:/) {
		    $waiting=0;
		    $err=1;
		    my @ee=split(/:/,$line);
		    if ($ee[1] =~ m/ServerTermBySignal/) {
			my @ee2=split(/\//,$ee[1]);
			$errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
			&msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
			exit(-1);
		    }
		} elsif ($line =~ m/PICTURE\:/) {
		    $waiting=0;
		    $PICFN=substr($line,8);
		    if ($PICFN eq "0") {
			$PICFN="";
		    } else {
			if ($DEBUG) {
			    print "PICTURE: Using picfile '$PICFN'\n";
			}
		    }
		}
	    }
	}
	if (substr($PICFN,-3,3) eq "xpm") {
	    $PICWIDGET=$MW->Pixmap(-file => $PICFN);
	} else {
	    $PICWIDGET=$MW->Photo(-file => $PICFN, -format => 'gif');
	}
	$swMsg->DESTROY;
    }
    $qrWin->DESTROY if Exists($qrWin);
    $qrWin=$MW->Toplevel;
    $qrWin->title(I_("Query Results"));
    $qrWin->configure(-bg => $FRAMEBGCOLOR);
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$qrWin->geometry($newGeometry);
    }    
    $qrWin->bind("<KeyPress-Return>" => sub {
	$qrOkButton->invoke;
    } );
    my $qrText="";
    for ($zz=0; $zz < $qindex; $zz++) {
	$qrText .= $QRLINES[$zz];
	$qrText .= "\n\n";
    }
    if (($SHOWGRAPHICS) && (length($PICFN))) {
	$qrGraphic=$qrWin->Label(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -border => 2,
				 -image => $PICWIDGET);
	$qrGraphic->pack(-side => 'top', -pady => '6');
    }
    $qrMsg=$qrWin->Message(-font => 'fixed', -width => '400', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => $qrText);
    $qrMsg->pack(-side => 'top', -pady => '10', -padx => '5');
    $qrButtons=$qrWin->Frame(-bg => $FRAMEBGCOLOR);
    $qrButtons->pack(-side => 'bottom', -pady => '6');
    $qrOkButton=$qrButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Ok'), -command => sub {
	$qrWin->DESTROY;
	if ($SHOWGRAPHICS) {
	    my $cleancmd="rm -f $PICFN";
	    system($cleancmd);
	}
    } );
    $qrDepButton=$qrButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -foreground => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_('Dependencies'), -command => sub {
	&showDependencies($pkgname);
    } );
    $qrOkButton->pack(-side => 'left', -padx => '4');
    $qrDepButton->pack(-side => 'left', -padx => '4');
}

sub showDependencies {
    my $pkgname=shift;
    if ($DEBUG) {
	print "DEPS: Getting dependencies for '$pkgname'\n";
    }
    undef @QREQS;
    undef @QPROVS;
    my $scmd="\@QUERYDEPS:$pkgname\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    while ($waiting) {
	if ($line = <SOCK>) {
	    if ($line =~ m/SUCCESS\:/) {
		$waiting=0;
		if ($DEBUG) {
		    print "DEPS: Received information successfully.\n";
		}
	    } elsif ($line =~ m/ERROR\:/) {
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    } elsif ($line =~ m/^REQS/) {
		chop $line;
		my @f=split(/:/,$line);
		@QREQS=split(//,$f[1]);

	    } elsif ($line =~ m/^PROVS/) {
		chop $line;
		my @f=split(/:/,$line);
		@QPROVS=split(//,$f[1]);
	    } else {
		if ($DEBUG) {
		    chop $line;
		    print "DEPS: Unknown line '$line'\n";
		}
	    }
	}
    }
    my $numreqs=$#QREQS + 1;
    my $numprovs=$#QPROVS + 1;
    if ($DEBUG) {	
	print "DEPS: $numreqs requirements, $numprovs provisions.\n";
    }
    if ($numreqs == 0) {
	$QREQS[0]=I_("(none)");
    }
    if ($numprovs == 0) {
	$QPROVS[0]=I_("(none)");
    }
    $depWin->DESTROY if Exists($depWin);
    $depWin=$MW->Toplevel;
    $depWin->configure(-bg => $FRAMEBGCOLOR);
    $depWin->title(I_("Package Dependencies"));
    if ($CASCADE) {
	my $mainGeometry=$qrWin->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$depWin->geometry($newGeometry);
    }
    $depTopText=sprintf(I_("Package Dependencies for '%s'"),$pkgname);
    $depTopMsg=$depWin->Message(-width => 300, -fg => $HEADINGCOLOR, -bg => $MAINBGCOLOR, -text => $depTopText);
    $depTopMsg->pack(-side => 'top', -pady => '10');
    $depReqFrame=$depWin->Frame(-bg => $MAINBGCOLOR);
    $depReqFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $depReqLabel=$depReqFrame->Label(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -font => 'fixed', -text => I_("Requires:"));
    $depReqLabel->pack(-side => 'left', -padx => '4');
    $depReqScroll=$depReqFrame->Scrollbar(-bg => $SECBGCOLOR, -highlightbackground => $MAINBGHILIGHT);
    $depReqScroll->pack(-side=> 'right', -fill => 'y');
    $depReqListbox=$depReqFrame->Listbox(-selectmode => 'multiple', -fg => $TEXTCOLOR, -bg => $SECBGCOLOR,-width => '30', -height => '5', -yscrollcommand => ['set', $depReqScroll]);

    $depReqListbox->pack(-side => 'right', -fill => 'y');
    $depReqScroll->configure(-command => ['yview', $depReqListbox]);
    $depReqListbox->insert(end,@QREQS);

    $depProvFrame=$depWin->Frame(-bg => $MAINBGCOLOR);
    $depProvFrame->pack(-side => 'top', -pady => '2', -fill => 'x');
    $depProvLabel=$depProvFrame->Label(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -font => 'fixed', -text => I_("Provides:"));
    $depProvLabel->pack(-side => 'left', -padx => '4');
    $depProvScroll=$depProvFrame->Scrollbar(-bg => $SECBGCOLOR, -highlightbackground => $MAINBGHILIGHT);
    $depProvScroll->pack(-side=> 'right', -fill => 'y');
    $depProvListbox=$depProvFrame->Listbox(-selectmode => 'multiple', -fg => $TEXTCOLOR, -bg => $SECBGCOLOR,-width => '30', -height => '5', -yscrollcommand => ['set', $depProvScroll]);

    $depProvListbox->pack(-side => 'right', -fill => 'y');
    $depProvScroll->configure(-command => ['yview', $depProvListbox]);
    $depProvListbox->insert(end,@QPROVS);

    $depButtons=$depWin->Frame(-bg => $FRAMEBGCOLOR);
    $depButtons->pack(-side => 'bottom', -pady => '6');
    $depOkButton=$depButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_("Ok"), -command => sub {
	$depWin->DESTROY;
    } );
    $depOkButton->pack(-side => 'left', -padx => '4');

    $depWin->bind("<KeyPress-Return>" => sub {
	$depOkButton->invoke;
    } );
}

sub servinfo_window {
    # my $scmd="\@REWELCOME\n";
#     send(SOCK,$scmd,0);
#     my $waiting=1;
#     while ($waiting) {
# 	if ($line = <SOCK>) {
# 	    $waiting=0;
# 	    if ($line =~ m/^HELLO/) {
# 		$waiting=0;
# 		$SERVERINFO=substr($line,6);
# 	    } else {
# 		if ($DEBUG) {
# 		    print "SERVINFO: Unexpected response from server (?)\n";
# 		}
# 	    }
# 	}
#     }
    chop $SERVERINFO unless (substr($SERVERINFO,-1,1) ne "\n");
    my ($pname,$pver,$pdate,$ptime)=split(/\//,$SERVERINFO);
    my $servtext=sprintf(I_("Server is '%s', version '%s'.\nBuilt on %s, %s.\n"),$pname,$pver,$pdate,$ptime);
    &msgWin(I_("Information"),I_("Ok"),$servtext);
}

sub isPackageInstalled {
    my $rc="";
    my $pkgname=shift;
    my $scmd="\@ISPACKAGEINSTALLED:$pkgname\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    if ($line =~ m/ERROR\:/) {
		$waiting=0;
		$err=1;
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    } elsif ($line =~ m/PACKAGEINSTALLED\:/) {
		$waiting=0;
		chop $line;
		my @f=split(/:/,$line);
		if ($f[1] eq $pkgname) {
		    $rc=$f[2];
		} else {
		    $err=1;		    
		}
	    }
	}
    }
    if ($err > 0) {
	&msgWin(I_("Error"),I_("Ok"),I_("Error determining state of package."));
	return "";
    }
    return $rc;
}

sub isPackageSelected {
    my $rc="";
    my $pkgname=shift;
    my $scmd="\@ISPACKAGESELECTED:$pkgname\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    if ($line =~ m/ERROR\:/) {
		$waiting=0;
		$err=1;
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    } elsif ($line =~ m/PACKAGESELECTED\:/) {
		$waiting=0;
		chop $line;
		my @f=split(/:/,$line);
		if ($f[1] eq $pkgname) {
		    $rc=$f[2];
		} else {
		    $err=1;
		}
	    }
	}
    }
    if ($err > 0) {
	&msgWin(I_("Error"),I_("Ok"),I_("Error determining state of package."));
	return "";
    }
    if ($rc == 1) {
	return "*";
    } else {
	return " ";
    }
}
	
	    

sub removePackage {
    &statusWin("remove",I_("Updating..."),$BottomFrame);
    $MW->update;
    my $selection=$pHlist->info(selection);
    my $pkgData=$pHlist->entrycget($selection, -data);
    my $pkgText=$pHlist->entrycget($selection, -text);
    if (substr($pkgData,0,1) ne "P") {
	$swMsg->DESTROY;
	&msgWin(I_("Error"),I_("Ok"),I_("You can only remove individual packages."));
	return;
    }
    my ($foo,$pkgname)=split(/:/,$pkgData);
    if ($DEBUG) {
	print "REMOVE: Marking '$pkgname' for removal.\n";
    }
    my $xc=&isPackageInstalled($pkgname);
    if (($xc =~ m/\s/) || (!(length($xc)))) {
	$swMsg->DESTROY;
	&msgWin(I_("Error"),I_("Ok"),I_("This package is not installed."));
	return;
    }
    my $scmd="\@REMOVEPACKAGE:$pkgname\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/SUCCESS\:/) {
		$waiting=0;
	    }
	    if ($line =~ m/ERROR\:/) {
		$waiting=0;
		$err=1;
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    }
	}
    }
    if ($err > 0) {
	$swMsg->DESTROY;
	&msgWin(I_("Error"),I_("Ok"),I_("Error removing package."));
	return;
    }
    if ($DEBUG) {
	print "REMOVE: Server reports remove toggle successful.\n";
    }

    my $isselected=&isPackageSelected($pkgname);
    my $isinstalled=&isPackageInstalled($pkgname);
    if ($isinstalled eq "R") {
	substr($pkgText,1,1)="R";
	substr($pkgText,-1,1)="R";
    } else {
	substr($pkgText,1,1)=$isselected;
	substr($pkgText,-1,1)=$isinstalled;
    }
    foreach $pkgline (keys(%PKGLINES)) {
	my $skey="$pkgname" . "\\|";
	$skey =~ s/\+/\\\+/g;
	$skey =~ s/\-/\\\-/g;
	if ($pkgline =~ m/^$skey/) {
	    $PKGLINES{$pkgline}=$pkgText;
	}
    }
    $pHlist->entryconfigure($selection, -text => $pkgText);
    $swMsg->DESTROY;
    $MW->update;
    &drawPackageList;
}

sub markAndRedraw {
    &toggleItem;
    &drawPackageList;
}

sub rightClickMenu {
    undef %RCM;
    $RCM{"doQuery"}=I_("Query Package Information");
    $RCM{"doSearch"}=I_("Search");
    $RCM{"removePackage"}=I_("Toggle Remove Flag");
    $RCM{"markAndRedraw"}=I_("Toggle Install/Upgrade Flag");
    $RCM{"treeExpand"}=I_("Expand/Collapse Tree");
    $RCM{"clearSelection"}=I_("Clear Selection");
    $RCM{"selectUpgrades"}=I_("Select All Upgrades");
    undef @RCMORDER;
    @RCMORDER=("doQuery",
	       "markAndRedraw",
	       "removePackage",
	       "selectUpgrades",
	       "clearSelection",
	       "doSearch",
	       "treeExpand");
    my $ycoord=(($pHlist->pointery) - ($pHlist->rooty));
    my $listid=$pHlist->nearest($ycoord);
    $pHlist->select('clear');
    $pHlist->select('set',$listid);
    $pHlist->see($listid);
    $MW->update;
    
    $rcmWin->DESTROY if Exists($rcmWin);
    $rcmWin=$MW->Toplevel;
    $rcmWin->title(I_("Options"));
    $rcmWin->configure(-bg => $FRAMEBGCOLOR);
    $rcmWin->transient($MW);
    my $wGeometry=$rcmWin->geometry;
    my @g=split(/\+/,$wGeometry);
    my $newX=$rcmWin->pointerx;
    my $newY=$rcmWin->pointery;
    my $newGeometry="=" . "+" . "$newX" . "+" . "$newY";
    if ($DEBUG) {
	print "[$newX,$newY] [$newGeometry]\n";
    }
    $rcmWin->geometry($newGeometry);
    my @buttonLabels=values(%RCM);
    my $maxWidth=0;
    for ($zz=0; $zz < $#buttonLabels;$zz++) {
	my $xx=length($buttonLabels[$zz]);
	if ($xx > $maxWidth) {
	    $maxWidth=$xx;
	}
    }
    foreach $rcmButton (@RCMORDER) {
	my $buttonName=$rcmButton . "_button";
	my $buttonAction=$rcmButton;
	my $buttonText=$RCM{$rcmButton};
	$WIDGETS{$buttonName}=$rcmWin->Button(-width => $maxWidth, -font => 'fixed', -background => $MAINBGHILIGHT, -foreground => $TEXTCOLOR, -highlightbackground => $MAINBGCOLOR, -activeforeground => $TEXTCOLOR, -activebackground => $SECBGCOLOR, -text => $buttonText, -relief => 'ridge', -command => sub {
	    &$buttonAction;
	    $rcmWin->DESTROY;
	} );
	$WIDGETS{$buttonName}->pack(-side => 'top');
    }
    $WIDGETS{"rcmCancel_button"}=$rcmWin->Button(-width => $maxWidth, -font => 'fixed', -background => $MAINBGHILIGHT, -foreground => $TEXTCOLOR, -highlightbackground => $MAINBGCOLOR, -activeforeground => $TEXTCOLOR, -activebackground => $SECBGCOLOR, -text => I_("Cancel"), -relief => 'ridge', -command => sub {
	$rcmWin->DESTROY;
    } );
    $WIDGETS{"rcmCancel_button"}->pack(-side => 'top');
}

sub treeExpand {
    if ($PROGRAM_STATE == 1) {
	$SHOWPACKAGES=abs(1 - $SHOWPACKAGES);
	my $curSel=$pHlist->info(selection);
	if ($curSel) {
	    my $curData=$pHlist->entrycget($curSel, -data);
	    if (substr($curData,0,1) eq "P") {
		$curSel=$pHlist->info(parent,$curSel);
	    }
	    if ($DEBUG) {
		print "TREE: Saving position at '$curSel'\n";
	    }
	}
	&drawPackageList(1);
	if ($curSel) {
	    $pHlist->selection(set,$curSel);
	    $pHlist->see($curSel);
	}
    }
}

sub doSearch {
    my $found=0;
    my $finished=0;
    my $prevkey=$searchkey;
    $searchkey=&entryWin(I_("Search"),$searchkey,I_("Enter the string to search for (case-insensitive"));
    if ($DEBUG) {
	print "SEARCH: Search key is '$searchkey'\n";
    }
    my $selection=$pHlist->info(selection);
    while (!($finished)) {
	if (!($selection)) {
	    $selection=$FIRSTENTRY;
	}
	if ($prevkey eq $searchkey) {
	    $nselection=$pHlist->info('next',$selection);
	    if (!(length($nselection))) {
		$selection=$FIRSTENTRY;
	    } else {
		$selection=$nselection;
	    }
	}
	while ((length($selection)) && (!($found))) {
	    my $selText=$pHlist->entrycget($selection, -text);
	    if ($selText =~ m/$searchkey/i) {
		$found=1;
		$pHlist->selection('clear');
		$pHlist->selection('set',$selection);
		$pHlist->see($selection);
	    }
	    $nselection=$pHlist->info('next',$selection);
	    $selection=$nselection;
	    
	}
	if (!($found)) {
	    my $yn=&choiceWin(I_("Not Found"),I_("Yes"),I_("No"),0,0,I_("Search term not found - retry from beginning of list?"));
	    if ($yn == 1) {
		$selection=$FIRSTENTRY;
		$finished=0;
	    } else {
		$finished=1;
	    }
	} else {
	    $finished=1;
	}
    }
}

sub checkDependencies {
    my $vDone=0;
    &statusWin("none",I_("Verifying Dependencies..."),$BottomFrame);
    my $scmd="\@VERIFYDEPS\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    my $udcount=0;
    undef %UDEPS;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/ERROR/) {
		$waiting=0;
		$err=1;
		if ($DEBUG) {
		    print "VERIFYDEPS: Received ERROR\n";
		}
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    } elsif ($line =~ m/SUCCESS\:/) {
		$waiting=0;
		if ($DEBUG) {
		    print "VERIFYDEPS: Server returned SUCCESS\n";
		}
	    } elsif ($line =~ m/UDEP\:/) {
		if ($DEBUG) {
		    print "VERIFYDEPS: '$line'\n";
		}
		$udcount++;
		my $udLine=substr($line,5);
		$UDEPS{$udLine}=1;
	    }
	}
    }
    if ($udcount == 0) {
	$swMsg->DESTROY;
	return 1;
    }
    undef %ADDPKGLIST;
    undef %URLIST;
    undef %WHATREQ;
    my $numUnresolved=0;
    foreach $ud (keys(%UDEPS)) {
	my ($pkg,$reqName,$reqVer,$suggestion)=split(/:/,$ud);
	my $xkey=$pkg . "";
	if (!($WHATREQ{$suggestion} =~ m/$xkey/)) {
	    $WHATREQ{$suggestion} .= "$pkg";
	    $WHATREQ{$suggestion} .= "";
	}
	if ($suggestion eq "0") {
	    $urString=$reqName;
	    if (length($reqVer)) {
		$urString .= " $reqVer";
	    }
	    $urString .= " (required by/conflicts with $pkg)";
	    $URLIST{$urString}=1;
	    $numUnresolved++;
	} else {
	    $ADDPKGLIST{$suggestion}=1;
	    if ($DEBUG) {
		print "VERIFYDEPS: Suggest install of '$suggestion'\n";
	    }
	}
    }
    if ($AUTODEPS == 1) {
	my @addlist=keys(%ADDPKGLIST);
	my $addlistStr=join(':',@addlist);
	my $trc=&toggleMulti($addlistStr,1);
	if ($trc == 0) {
	    &msgWin(I_("Error"),I_("Ok"),I_("Error selecting needed packages.  Synchronization reset is required."));
	    &restart_program;
	} else {
	    my $selection=$FIRSTENTRY;
	    while (length($selection)) {
		my $csData=$pHlist->entrycget($selection, -data);
		my ($csType,$csName)=split(/:/,$csData);
		if (($csType eq "P") && (&memberOf($csName,@addlist))) {
		    my $csText=$pHlist->entrycget($selection, -text);
		    my $plKey=$csName . "|" . $ALLPKGS{$csName};
		    substr($PKGLINES{$plKey},1,1)="*";
		    $pHlist->entryconfigure($selection,-text => $PKGLINES{$plKey});
		}
		my $nselection=$pHlist->info('next',$selection);
		$selection=$nselection;
	    }
	}
	# else {
# 	    foreach $addpkg (@addlist) {
# 		foreach $pkgent (keys(%PKGLINES)) {
# 		    my ($myName,$myGroup)=split(/\|/,$pkgent);
# 		    if ($myName eq $addpkg) {
# 			substr($PKGLINES{$pkgent},1,1)="*";
# 		    }
# 		}
# 	    }
# 	}
	if ($numUnresolved == 0) {
	    $swMsg->DESTROY;
	    &drawPackageList;
	    return 1;
	}
    }

    $vdepWin->DESTROY if Exists($vdepWin);
    $vdepWin=$MW->Toplevel;
    $vdepWin->configure(-bg => $FRAMEBGCOLOR);
    $vdepWin->title(I_("Dependencies"));
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$vdepWin->geometry($newGeometry);
    }
    $vdepTopMsg=$vdepWin->Label(-fg => $HEADINGCOLOR, -bg => $MAINBGCOLOR, -text => I_("Dependency Resolution"));
    $vdepTopMsg->pack(-side => 'top', -pady => '10');
    if ($AUTODEPS == 0) {
	$ADD_DEPS=0;
	my @addlist=keys(%ADDPKGLIST);
	$vdepAddFrame=$vdepWin->Frame(-bg => $MAINBGCOLOR);
	$vdepAddFrame->pack(-side => 'top', -pady => '4', -fill => 'x');
	$vdepAddLabel=$vdepAddFrame->Label(-font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("To Be Added:"));
	$vdepAddLabel->pack(-side => 'left', -padx => '4');
	undef %FADDLIST;
	foreach $xa (@addlist) {
	    my $plist=$WHATREQ{$xa};
	    chop $plist;
	    $plist =~ s//\, /g;
	    my $dStr="$xa (required by $plist)";
	    if ($DEBUG) {
		print "VERIFYDEPS: Displaying add of '$dStr'\n";
	    }
	    $FADDLIST{$dStr}=1;
	}
	&createScrolledListbox($vdepAddFrame,
			       "vdepaddlist",
			       50,
			       5,
			       'fixed',
			       'single',
			       (sort(keys(%FADDLIST))));
	$WIDGETS{vdepaddlist}->pack(-side => 'right', -padx => '4');
	$vdepAddCheckFrame=$vdepWin->Frame(-bg => $MAINBGCOLOR);
	$vdepAddCheckFrame->pack(-side => 'top', -fill => 'x');
	$vdepAddCheck=$vdepAddCheckFrame->Checkbutton(-font => 'fixed', -selectcolor => $SPECIALTEXTCOLOR, -text => I_("Install Needed Packages"), -variable => \$ADD_DEPS, -underline => -1, -background => $SECBGCOLOR, -activebackground => $MAINBGHILIGHT, -activeforeground => $TEXTCOLOR, -foreground => $TEXTCOLOR, -highlightbackground => $MAINBGCOLOR);
	$vdepAddCheck->pack(-side => 'left');
	$vdepAddCheck->invoke;
    }
    if ($numUnresolved > 0) {
	$vdepUFrame=$vdepWin->Frame(-bg => $MAINBGCOLOR);
	$vdepUFrame->pack(-side => 'top', -pady => '4', -fill => 'x');
	$vdepULabel=$vdepUFrame->Label(-font => 'fixed', -fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -text => I_("Unresolvable:"));
	$vdepULabel->pack(-side => 'left', -padx => '4');
	&createScrolledListbox($vdepUFrame,
			       "vdepUlist",
			       50,
			       5,
			       'fixed',
			       'single',
			       (sort(keys(%URLIST))));
	$WIDGETS{vdepUlist}->pack(-side => 'right', -padx => '4');
    }
    $vdepButtons=$vdepWin->Frame(-bg => $FRAMEBGCOLOR);
    $vdepButtons->pack(-side => 'bottom', -pady => '6');
    $vdepProceedButton=$vdepButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_("Proceed Anyway"), -command => sub {
	$vDone=1;
    } );
    $vdepBackButton=$vdepButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_("Go Back"), -command => sub {
	$vDone=2;
    } );
    $vdepProceedButton->pack(-side => 'left', -padx => '4');
    $vdepBackButton->pack(-side => 'left', -padx => '4');

    while ($vDone == 0) {
	$MW->update;
    }
    $vdepWin->DESTROY;

    if (($AUTO_DEPS == 0) && ($ADD_DEPS == 1)) {
	my @addlist=keys(%ADDPKGLIST);
	my $addlistStr=join(':',@addlist);
	my $trc=&toggleMulti($addlistStr,1);
	if ($trc == 0) {
	    &msgWin(I_("Error"),I_("Ok"),I_("Error selecting needed packages.  Synchronization reset is required."));
	    &restart_program;	    
	} else {	    
	    my $selection=$FIRSTENTRY;
	    while (length($selection)) {
		my $csData=$pHlist->entrycget($selection, -data);
		my ($csType,$csName)=split(/:/,$csData);
		if (($csType eq "P") && (&memberOf($csName,@addlist))) {
		    my $csText=$pHlist->entrycget($selection, -text);
		    my $plKey=$csName . "|" . $ALLPKGS{$csName};
		    substr($PKGLINES{$plKey},1,1)="*";
		    $pHlist->entryconfigure($selection,-text => $PKGLINES{$plKey});
		}
		my $nselection=$pHlist->info('next',$selection);
		$selection=$nselection;
	    }
	    # foreach $addpkg (@addlist) {
# 		foreach $pkgent (keys(%PKGLINES)) {
# 		    my ($myName,$myGroup)=split(/\|/,$pkgent);
# 		    if ($myName eq $addpkg) {
# 			substr($PKGLINES{$pkgent},1,1)="*";
# 		    }
# 		}
# 	    }
	}
    }
    $swMsg->DESTROY;
    &drawPackageList;
    return $vDone;
}

sub checkFreeSpace {
    my $scmd="\@CHECKFREESPACE\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    my $err=0;
    my $fsInfo="";
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/^FREESPACE\:/) {
		$fsInfo=substr($line,10);
		$waiting=0;
	    }
	    if ($line =~ m/ERROR/) {
		$err=1;
		if ($DEBUG) {
		    print "FREESPACE: Received ERROR\n";
		}
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    }
	}
    }
    if (!(length($fsInfo))) {
	&msgWin(I_("Error"),I_("Ok"),I_("Could not get free disk space information."));
	return 2;
    }
    my ($fsOk,$fsReq,$fsAvail)=split(/:/,$fsInfo);
    if ($fsOk) {
	my $fsText=sprintf(I_("Your package selections total %d megabytes (%d megabytes free).  Would you like to proceed or go back to the package selection window?"),$fsReq,$fsAvail);
	my $rc=&choiceWin(I_("Confirm Selections"),
			  I_("Proceed"),
			  I_("Go Back"),
			  0,
			  0,
			  $fsText);
	return $rc;
    } else {
	my $fsText=sprintf(I_("Your package selections total %d megabytes, but you only have %d megabytes available.  Make some more room on your hard disk and/or deselect some packages."),$fsReq,$fsAvail);
	&msgWin(I_("Error"),I_("Ok"),$fsText);
	return 2;
    }
    return 1;
}
			  
	    

sub doInstall {
    my $rc=-1;
    my $rPkgs=0;
    my $pkgsDone=0;
    my $bytesDone=0;

    if ($IGNOREDEPS) {
	my $drc=&choiceWin(I_("Warning"),I_("Ok"),I_("Go Back"),0,0,I_("You have chosen to ignore package dependencies.  This may break things on your system.  Select 'Go Back' to back out.  You can turn dependency checking back on in the Preferences window."));
	if ($drc == 2) {
	    return;
	}
    } else {
	$rc=&checkDependencies;
	if ($rc == 2) {
	    &drawPackageList;  # just in case ?
	    return;
	}
    }
    $rc=&checkFreeSpace;
    if ($rc == 2) {
	&drawPackageList;  # just in case ?
	return;
    }
    &remapMainFrame;
    $ButtonFrame->DESTROY;
    my $rccmd="\@REMOVECOUNT\n";
    send(SOCK,$rccmd,0);
    my $waiting=1;
    my $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/^REMOVECOUNT/) {
		$waiting=0;
		my @rf=split(/:/,$line);
		$rPkgs=$rf[1];
	    }
	    if ($line =~ m/ERROR/) {
		$err=1;
		if ($DEBUG) {
		    print "DOREMOVES: Received ERROR\n";
		}
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		}
	    }
	}
    }
    if ($rPkgs > 0) {
	my $rText=sprintf(I_("You have marked %d packages for removal.  Are you sure you want to remove them from your system?  Select 'Go Back' if you want to return to the package selection window."),$rPkgs);
	my $rrc=&choiceWin(I_("Remove Packages"),I_("Yes"),I_("No"),I_("Go Back"),0,$rText);
	if ($rrc == 3) {
	    &drawPackageList(1);
	    &remapMainButtons;
	    return;
	}
	if ($rrc == 1) {			   
	    my $rscmd="\@DOREMOVES\n";
	    send(SOCK,$rscmd,0);
	    $waiting=1;
	    $err=0;
	    while ($waiting) {
		if ($line = <SOCK>) {
		    chop $line;
		    if ($line =~ m/^SUCCESS/) {
			$waiting=0;
		    }
		    if ($line =~ m/ERROR/) {
			$err=1;
			if ($DEBUG) {
			    print "DOREMOVES: Received ERROR\n";
			}
			my @ee=split(/:/,$line);
			if ($ee[1] =~ m/ServerTermBySignal/) {
			    my @ee2=split(/\//,$ee[1]);
			    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),$ee2[1],$ee2[2]);
			    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
			    exit(-1);
			}
		    }
		}
	    }
	}
    }
    $CURRENTSIZE=0;
    $PKGPROGRESS=0;
    $INSTPROGRESS=0;
    &installWindow($MainFrame);
    my $scmd="\@STARTINSTALL\n";
    send(SOCK,$scmd,0);
    $waiting=1;
    $err=0;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/^SUCCESS\:/) {
		$waiting=0;
		if ($DEBUG) {
		    print "INSTALL: Server returned SUCCESS.\n";
		}
	    }
	    elsif (($line =~ m/^ERROR/) || ($line =~ m/^FATALERROR/)) {
		$err=1;
		if ($DEBUG) {
		    print "INSTALL: Received ERROR\n";
		}
		my @ee=split(/:/,$line);
		if ($ee[1] =~ m/ServerTermBySignal/) {
		    my @ee2=split(/\//,$ee[1]);
		    $errstr=sprintf(I_("FATAL ERROR: Server Terminated on Signal %s (%s)"),
				    $ee2[1],$ee2[2]);
		    &msgWin(I_("Fatal Error"),I_("Ok"),$errstr);
		    exit(-1);
		} else {
		    if ($DEBUG) {
			print "INSTALL: Received FATALERROR\n";
		    }
		    &msgWin(I_("Fatal Error"),I_("Ok"),I_("A fatal error occurred while installing packages.  Installation/Upgrade incomplete."));
		}
	    }
	    elsif ($line =~ m/^STATUS\:/) {
		my ($foo,$bar,$msgText)=split(/:/,$line);
		&statusWin("none",$msgText,$BottomFrame);
		$MW->update;
	    }
	    elsif ($line =~ m/^STATUSEND\:/) {
		$swMsg->DESTROY;
		$BottomFrame->configure(-height => 1);
		$MW->update;
	    }
	    elsif ($line =~ m/^INSTINFO/) {
		($pFoo,$pCount,$pBytes,$pLang)=split(/:/,$line);
		if ($DEBUG) {
		    print "INSTALL: Installation Set Info - $pCount packages, $pBytes bytes.\n";
		}
		my $pkvText=sprintf(I_("%d of %d Complete"),$pkgsDone,$pCount);
		$WIDGETS{Packages_value}->configure(-text => $pkvText);
		my $mbvText=sprintf(I_("%d of %d MB Complete"),($bytesDone / (1024 * 1024)),
				    ($pBytes / (1024 * 1024)));
		$WIDGETS{Megabytes_value}->configure(-text => $mbvText);
	    }
	    elsif ($line =~ m/^FTPSTART\:/) {
		my ($ftpFoo,$ftpFn,$ftpSize)=split(/:/,$line);
		if ($DEBUG) {
		    print "INSTALL: Getting package by FTP '$ftpFn'\n";
		}
		$PKGPROGRESS=0;
		my $sizeText=sprintf("%d kb",($ftpSize / 1024));
		$WIDGETS{Name_value}->configure(-text => $ftpFn);
		$WIDGETS{Size_value}->configure(-text => $sizeText);
		$WIDGETS{Summary_value}->configure(-text => I_("Receiving File by FTP"));
	    }
	    elsif ($line =~ m/^FTP\:/) {
		my ($ftpFoo,$ftpFn,$ftpSize,$ftpComplete)=split(/:/,$line);
		$PKGPROGRESS=$ftpComplete;
	    }
	    elsif ($line =~ m/^PKGSTART\:/) {
		my ($pkgFoo,$pkgName,$pkgSize,$pkgSumm)=split(/:/,$line);
		if ($DEBUG) {
		    print "INSTALL: Installing package '$pkgName'\n";
		}
		$CURRENTSIZE=$pkgSize;
		$PKGPROGRESS=0;
		my $sizeText=sprintf("%d kb",($pkgSize / 1024));
		$WIDGETS{Name_value}->configure(-text => $pkgName);
		$WIDGETS{Size_value}->configure(-text => $sizeText);
		$WIDGETS{Summary_value}->configure(-text => $pkgSumm);
	    }
	    elsif ($line =~ m/^PKG\:/) {
		my ($pkgFoo,$pkgProgress)=split(/:/,$line);
		$PKGPROGRESS=$pkgProgress;
	    }
	    elsif ($line =~ m/^PKGCOMPLETE\:/) {
		$pkgsDone++;
		$bytesDone += $CURRENTSIZE;
		my $pkvText=sprintf(I_("%d of %d Complete"),$pkgsDone,$pCount);
		$WIDGETS{Packages_value}->configure(-text => $pkvText);
		my $mbvText=sprintf(I_("%d of %d Complete"),($bytesDone / (1024 * 1024)),
				    ($pBytes / (1024 * 1024)));
		my $instprog=(($bytesDone * 100) / $pBytes);
		$INSTPROGRESS=$instprog;
		$WIDGETS{Megabytes_value}->configure(-text => $mbvText);
	    }	    
	}
	$MW->update;
    }
    &msgWin(I_("Finished"),I_("Ok"),I_("XTurboPkg has finished installing/upgrading/removing the selected packages."));
    &exit_program(1);
}

sub memberOf {
    my ($element,@array)=@_;
    foreach $e (@array) {
	if ($e eq $element) {
	    return 1;
	}
    }
    return 0;
}


sub clearEntries {
    &statusWin("none",I_("Clearing Selection"),$BottomFrame);
    $MW->update;
    my $scmd="\@CLEARSELECTION\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/SUCCESS\:/) {
		$waiting=0;
	    }
	}
    }
    my $selection=$FIRSTENTRY;
    while (length($selection)) {
 	my $eText=$pHlist->entrycget($selection, -text);
 	my $eData=$pHlist->entrycget($selection, -data);
 	if ((substr($eData,0,1) eq "P") && (substr($eText,1,1) eq "*")) {
	    my $pKey=(substr($eData,2)) . "|" . $ALLPKGS{(substr($eData,2))};
	    substr($PKGLINES{$pKey},1,1) = ' ';
	    $pHlist->entryconfigure($selection, -text => $PKGLINES{$pKey});
 	}
	elsif ((substr($eData,0,1) eq "G") && (substr($eText,1,1) eq "*")) {
	    my $gKey=(substr($eData,2));
	    substr($GROUPLINES{$gKey},1,1) = ' ';
	    $pHlist->entryconfigure($selection, -text => $GROUPLINES{$gKey});
 	}
 	my $nselection=$pHlist->info('next',$selection);
 	$selection=$nselection;
    }
    $swMsg->DESTROY;
    &drawPackageList;
}

sub selectUpgrades {
    my $numScanned=0;
    &statusWin("none",I_("Selecting Upgrades..."),$BottomFrame);
    $MW->configure(-cursor => 'watch');
    $MW->update;
    my $scmd="\@ALLUPGRADES:$DODOWNGRADES\n";
    send(SOCK,$scmd,0);
    my $waiting=1;
    while ($waiting) {
	if ($line = <SOCK>) {
	    chop $line;
	    if ($line =~ m/SUCCESS\:/) {
		if ($DEBUG) {
		    print "SELECTUPG: Server returned SUCCESS.\n";
		}
		$waiting=0;
	    }
	}
    }
    my $selection=$FIRSTENTRY;
    while (length($selection)) {
 	my $eText=$pHlist->entrycget($selection, -text);
 	my $eData=$pHlist->entrycget($selection, -data);
 	if (substr($eData,0,1) eq "P") {
	    my $pKey=(substr($eData,2)) . "|" . $ALLPKGS{(substr($eData,2))};
	    my $ugc=substr($PKGLINES{$pKey},-1,1);
#	    if ($DEBUG) {
#		print "SELECTUPG: Package '$pKey' last char is '$ugc'\n";
#	    }
	    if (($ugc eq "+") || ($ugc =~ m/\s/)) {
		substr($PKGLINES{$pKey},1,1) = '*';
	    } elsif (($ugc eq "-") && ($DODOWNGRADES == 1)) {
		substr($PKGLINES{$pKey},1,1) = '*';
	    }
	    $pHlist->entryconfigure($selection, -text => $PKGLINES{$pKey});
 	}
	elsif ((substr($eData,0,1) eq "G") && (substr($eText,1,1) eq "*")) {
	    my $gKey=(substr($eData,2));
	    substr($GROUPLINES{$gKey},1,1) = ' ';
	    $pHlist->entryconfigure($selection, -text => $GROUPLINES{$gKey});
 	}
 	my $nselection=$pHlist->info('next',$selection);
 	$selection=$nselection;
    }
    $swMsg->DESTROY;
    $MW->configure(-cursor => 'left_ptr');
    $MW->update;
    &drawPackageList;
}

sub installWindow {
    my ($parent) = @_;

    $iFrame=$parent->Frame(-bg => $MAINBGCOLOR);
    $iFrame->pack(-side => 'top', -fill => 'both');
    foreach $lw ("Name","Size","Summary") {
	my $labelText=sprintf("%-10s :",$lw);
	my $frameName=$lw . "_frame";
	my $keyName=$lw . "_key";
	my $valName=$lw . "_value";
	$WIDGETS{$frameName}=$iFrame->Frame(-bg => $MAINBGCOLOR);
	$WIDGETS{$frameName}->pack(-side => 'top', -pady => '2', -fill => 'x');
	$WIDGETS{$keyName}=$WIDGETS{$frameName}->Label(-fg => $HEADINGCOLOR,
						       -bg => $MAINBGCOLOR,
						       -font => 'fixed',
						       -text => $labelText);
	$WIDGETS{$valName}=$WIDGETS{$frameName}->Label(-fg => $TEXTCOLOR,
						       -bg => $MAINBGCOLOR,
						       -font => 'fixed',
						       -text => "N/A");
	$WIDGETS{$keyName}->pack(-side => 'left', -padx => '4');
	$WIDGETS{$valName}->pack(-side => 'left', -padx => '4');
    }
    $iPkgScale=$iFrame->Scale(-fg => $TEXTCOLOR,
			    -bg => $MAINBGCOLOR,
			    -troughcolor => $TROUGHCOLOR,
			    -highlightbackground => $MAINBGCOLOR,
			    -from => 0,
			    -to => 100,
			    -length => 300,
			    -resolution => 1,
			    -sliderlength => 10,
			    -tickinterval => 10,
			    -showvalue => 1,
			    -variable => \$PKGPROGRESS,
			    -font => 'fixed',
			    -orient => 'horizontal');
    $iPkgScale->pack(-side => 'top');
    
    $iStatFrame=$iFrame->Frame(-bg => $MAINBGCOLOR, -relief => 'ridge',
			       -highlightcolor => $MAINBGHILIGHT,
			       -highlightthickness => 2,
			       -highlightbackground => $MAINBGHILIGHT);
    $iStatFrame->pack(-side => 'top', -pady => '2', -fill => 'x');

    foreach $isw ("Packages","Megabytes") {
	my $labelText=sprintf("%-10s :",$isw);
	my $frameName=$isw . "_frame";
	my $keyName=$isw . "_key";
	my $valName=$isw . "_value";
	$WIDGETS{$frameName}=$iStatFrame->Frame(-bg => $MAINBGCOLOR);
	$WIDGETS{$frameName}->pack(-side => 'top', -pady => '2', -fill => 'x');
	$WIDGETS{$keyName}=$WIDGETS{$frameName}->Label(-fg => $HEADINGCOLOR,
						       -bg => $MAINBGCOLOR,
						       -font => 'fixed',
						       -text => $labelText);
	$WIDGETS{$valName}=$WIDGETS{$frameName}->Label(-fg => $TEXTCOLOR,
						       -bg => $MAINBGCOLOR,
						       -font => 'fixed',
						       -text => "N/A");
	$WIDGETS{$keyName}->pack(-side => 'left', -padx => '4');
	$WIDGETS{$valName}->pack(-side => 'left', -padx => '4');
    }

    $iInstScale=$iFrame->Scale(-fg => $TEXTCOLOR,
			    -bg => $MAINBGCOLOR,
			    -troughcolor => $TROUGHCOLOR,
			    -highlightbackground => $MAINBGCOLOR,
			    -from => 0,
			    -to => 100,
			    -length => 400,
			    -resolution => 1,
			    -sliderlength => 10,
			    -tickinterval => 10,
			    -showvalue => 1,
			    -variable => \$INSTPROGRESS,
			    -font => 'fixed',
			    -orient => 'horizontal');
    $iInstScale->pack(-side => 'top');
}


## Generic windowing routines

sub choiceWin {
    my ($title,$b1,$b2,$b3,$b4,$topmsg)=@_;
    my $cwDone=0;

    $cw->DESTROY if Exists($cw);
    $cw=$MW->Toplevel;
    $cw->configure(-bg => $FRAMEBGCOLOR);
    $cw->title($title);
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$cw->geometry($newGeometry);
    } else {
	my $newX=$cw->pointerx;
	my $newY=$cw->pointery;
	my $newGeometry="=" . "+" . "$newX" . "+" . "$newY";
	if ($DEBUG) {
	    print "[$newX,$newY] [$newGeometry]\n";
	}
	$cw->geometry($newGeometry);
    }
    if (($b2 == 0) && ($b3 == 0) && ($b4 == 0)) {
	$cw->bind("<KeyPress-Return>" => sub {
	    $cwbtns[1]->invoke;
	} );
    }
    $cwMsg=$cw->Message(-fg => $TEXTCOLOR, -bg => $MAINBGCOLOR, -width => 300, -text => $topmsg);
    $cwMsg->pack(-side => 'top', -pady => '10');
    $cwButtons=$cw->Frame(-bg => $FRAMEBGCOLOR);
    $cwButtons->pack(-side => 'bottom', -pady => '6');
    undef @cwbtns;
    for ($zz=1;$zz <= 4;$zz++) {
	my $ecmd="if (\$b$zz) { return 1; } else { return 0; }";
	if (eval($ecmd)) {
	    my $varname="\$b$zz";
	    my $returnID=$zz;
	    my $btext=eval($varname);
	    $cwbtns[$zz]=$cwButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => $btext, -command => sub {
		$cwDone=$returnID;
	    } );
	    $cwbtns[$zz]->pack(-side => 'left', -padx => '4');
	}
    }
    while ($cwDone == 0) {
	$MW->update;
    }
    $cw->DESTROY;
    if ($DEBUG) {
	print "Returning [$cwDone]\n";
    }
    return $cwDone;
}

sub msgWin {
    my ($title,$b1,$topmsg)=@_;
    return (&choiceWin($title,$b1,0,0,0,$topmsg));
}

sub entryWin {
    my ($title,$defaultentrytext,$topmsg)=@_;
    my $ewDone=0;
    my $rvalstr=$defaultentrytext;

    $ew->DESTROY if Exists($ew);
    $ew=$MW->Toplevel;
    $ew->configure(-bg => $FRAMEBGCOLOR);
    $ew->title($title);
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$ew->geometry($newGeometry);
    } else {
	my $newX=$ew->pointerx;
	my $newY=$ew->pointery;
	my $newGeometry="=" . "+" . "$newX" . "+" . "$newY";
	if ($DEBUG) {
	    print "[$newX,$newY] [$newGeometry]\n";
	}
	$ew->geometry($newGeometry);
    }
    $ewMsg=$ew->Message(-fg => $HEADINGCOLOR, -bg => $MAINBGCOLOR, -width => 300, -text => $topmsg);
    $ewMsg->pack(-side => 'top', -pady => '10');
    $ewEntry=$ew->Entry(-width => '40', -relief => 'sunken', -fg => $TEXTCOLOR, -bg => $SECBGCOLOR, -textvariable => \$rvalstr);
    $ewEntry->pack(-side => 'top', -pady => '2');
    $ew->bind("<KeyPress-Return>" => sub {
	$ewOkButton->invoke;
    } );
    $ewButtons=$ew->Frame(-bg => $FRAMEBGCOLOR);
    $ewButtons->pack(-side => 'bottom', -pady => '6');
    $ewOkButton=$ewButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_("Ok"), -command => sub {
		$ewDone=1;
	    } );
    $ewCancelButton=$ewButtons->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => I_("Cancel"), -command => sub {
		$ewDone=1;
		$rvalstr="";
	    } );
    $ewOkButton->pack(-side => 'left', -padx => '4');
    $ewCancelButton->pack(-side => 'left', -padx => '4');
    
    while ($ewDone == 0) {
	$MW->update;
    }
    $ew->DESTROY;
    if ($DEBUG) {
	print "EW: Returning '$rvalstr'\n";
    }
    return $rvalstr;
}

sub statusWin {
    if ($DEBUG) {
	foreach $arg (@_) {
	    print "SWARG: '$arg'\n";
	}
    }
    my ($title,$topmsg,$parent,$cancelbuttonText)=@_;
    $parent=$MainFrame unless $parent;
    if ($DEBUG) {
	if ($parent == $MainFrame) {
	    print "SW: Parent is MainFrame\n";
	}
    }
    $swMsg->DESTROY if Exists($swMsg);
    $swMsg=$parent->Frame(-bg => $MAINBGCOLOR);
    $swMsg->pack(-side => 'bottom', -pady => '10');
    $swMsgMsg=$swMsg->Message(-fg => $HEADINGCOLOR, -bg => $MAINBGCOLOR, -width => 400, -text => $topmsg);
    $swMsgMsg->pack(-side => 'left', -padx => '4');
    if (length($cancelbuttonText)) {
	$swMsgCancel=$swMsg->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => $cancelbuttonText, -command => sub {
	    $CANCELLED=1;
	} );
	$swMsgCancel->pack(-side => 'left', -padx => '4');
    }
}

sub statusWinSeparate {
    my ($title,$topmsg)=@_;
    $sw->DESTROY if Exists($sw);
    $sw=$MW->Toplevel;
    $sw->configure(-bg => $FRAMEBGCOLOR);
    $sw->title($title);
    if ($CASCADE) {
	my $mainGeometry=$MW->geometry;
	if ($DEBUG) {
	    print "maingeometry: '$mainGeometry'\n";
	}
	my @gg=split(/\+/,$mainGeometry);
	$gg[1] += $CASCADE_DISTANCE;
	$gg[2] += $CASCADE_DISTANCE;
	my $newGeometry="=" . "+" . "$gg[1]" . "+" . "$gg[2]";
	if ($DEBUG) {
	    print "newgeometry: '$newGeometry'\n";
	}
	$sw->geometry($newGeometry);
    } else {
	my $newX=$sw->pointerx;
	my $newY=$sw->pointery;
	my $newGeometry="=" . "+" . "$newX" . "+" . "$newY";
	if ($DEBUG) {
	    print "[$newX,$newY] [$newGeometry]\n";
	}
	$sw->geometry($newGeometry);
    }
    $swMsg=$sw->Message(-fg => $HEADINGCOLOR, -bg => $MAINBGCOLOR, -width => 300, -text => $topmsg);
    $swMsg->pack(-side => 'top', -pady => '10');
}

# This function performs no packing.
sub createScrolledListbox {
    my ($parent,$name,$width,$height,$font,$selectmode,@elements)=@_;

    my $scrollName=$name . "_scroll";
    my $listboxName=$name . "_listbox";
    $WIDGETS{$name}=$parent->Frame(-bg => $MAINBGCOLOR);
    $WIDGETS{$scrollName}=$WIDGETS{$name}->Scrollbar(-bg => $SECBGCOLOR, -highlightbackground => $MAINBGCOLOR, -troughcolor => $TROUGHCOLOR);
    $WIDGETS{$scrollName}->pack(-side=> 'right', -fill => 'y');
    $WIDGETS{$listboxName}=$WIDGETS{$name}->Listbox(-highlightbackground => $MAINBGCOLOR, -font => $font, -selectmode => $selectmode, -fg => $TEXTCOLOR, -bg => $SECBGCOLOR, -width => $width, -height => $height, -yscrollcommand => ['set', $WIDGETS{$scrollName}]);
    
    $WIDGETS{$listboxName}->pack(-side => 'right', -fill => 'y');
    $WIDGETS{$scrollName}->configure(-command => ['yview', $WIDGETS{$listboxName}]);
    $WIDGETS{$listboxName}->insert(end,@elements);
}

sub progressWin {
    if ($DEBUG) {
	foreach $arg (@_) {
	    print "PWARG: '$arg'\n";
	}
    }
    my ($title,$topmsg,$parent,$bindvar,$cancelbuttonText)=@_;
    $parent=$MainFrame unless $parent;
    if ($DEBUG) {
	if ($parent == $MainFrame) {
	    print "PW: Parent is MainFrame\n";
	}
    }
    $pwMsg=$parent->Frame(-bg => $MAINBGCOLOR);
    $pwMsg->pack(-side => 'bottom', -pady => '10');
    $pwMsgMsg=$pwMsg->Message(-font => 'fixed', -fg => $HEADINGCOLOR, -bg => $MAINBGCOLOR, -width => 400, -text => $topmsg);
    $pwMsgMsg->pack(-side => 'top', -pady => '4');
    if ($FTPSERVNOSIZE == 1) {
	$pwMsgScale=$pwMsg->Label(-fg => $TEXTCOLOR,
				  -bg => $MAINBGCOLOR,
				  -text => 'FTP Server does not support SIZE, progress N/A'
				  );
    } else {
	$pwMsgScale=$pwMsg->Scale(-fg => $TEXTCOLOR,
				  -bg => $MAINBGCOLOR,
				  -troughcolor => $TROUGHCOLOR,
				  -highlightbackground => $MAINBGCOLOR,
				  -from => 0,
				  -to => 100,
				  -length => 300,
				  -resolution => 1,
				  -sliderlength => 10,
				  -tickinterval => 10,
				  -showvalue => 1,
				  -variable => \$FTPPROGRESS,
				  -font => 'fixed',
				  -orient => 'horizontal');
    }
    $pwMsgScale->pack(-side => 'top', -pady => '2');
    if (length($cancelbuttonText)) {
	$pwMsgCancel=$pwMsg->Button(-bg => $BUTTONCOLOR, -highlightbackground => $TERBGCOLOR, -fg => $BUTTONTEXTCOLOR, -activebackground => $ACTIVEBUTTONCOLOR, -activeforeground => $BUTTONTEXTCOLOR, -text => $cancelbuttonText, -command => sub {
	    $CANCELLED=1;
	} );
	$pwMsgCancel->pack(-side => 'top', -pady => '2');
    }    
}
